//******************************************************************************
//  app-timers.c
//
//  Timer module useful for an application
//  This module provides a means for the application to perform actions at a
//  certain time, or continuously at a certain period. This functionality is
//  normally referred to as an application timer module.
//
//  DESCRIPTION:
//  The application timer module can handle a predetermined number of timers
//  (see app-timers.h)  Each timer has associated with it an expiration time, a
//  timer ID (the index into the timer structure array), a function pointer, the
//  parameter to give to the function when it's called, a flag indicating if
//  the timer is recurring, and the number of system ticks the timer is
//  supposed to recur.  The application adds a timer by making a call to the
//  addTimer function.
//
//  The application timer module has a core function (handleTimers) that needs
//  to be  called on a periodic basis (e.g. with the calls to emberTick in the
//  application). It  checks the processor clock to determine if a
//  predetermined amount of time has elapsed (currently set to 62.5 mS, or 16
//  times per second). If it has, the function will increment its system tick
//  and check if any timers have expired. If  a timer has expired, then the
//  function associated with the timer will be called.
//
//
//  Copyright 2004 by Ember Corporation. All rights reserved.              *80*
//******************************************************************************
#ifdef PLATFORM_HOST
#include <stdio.h>
#include "app/host-framework/c/host-api.h"
#else
#ifndef EMBER_TEST
#include <ioavr.h>
#endif
#include PLATFORM_HEADER //compiler/micro specifics, types
#include "stack/include/ember.h"
#include "stack/include/hal.h" // microGetSystemTick()
#endif


#include "stack/include/error.h"
#include "app/sample/app-timers.h"


//******************************************************************************

timer Timers[NUM_TIMERS];
int16u sysTicks = 0;
int16u previousTime = 0;

//******************************************************************************
// Timer utilities
//******************************************************************************

void resetTimerClock(void)
{
  sysTicks = 0;
  previousTime = 0;
}

//******************************************************************************

void initTimers()
{
  int8u i;
  for (i = 0; i < NUM_TIMERS; i++)
    Timers[i].timer_function = NULL;

  resetTimerClock();
}

//******************************************************************************

// Add a timer
// To keep things simple, just overwrite previous timers, w/o checking.
void addTimer(int8u timerId,
              TimerFunctionPointer function,
              int32u param,
              int16u ticks,
              boolean recurring)
{
  // Exit gracefully if bad timer index.
  if (timerId >= NUM_TIMERS) return;

  Timers[timerId].timerId = timerId;
  Timers[timerId].timer_function = function;
  Timers[timerId].expiration = sysTicks + ticks;

  Timers[timerId].param = param;
  if (recurring) {
    Timers[timerId].recurring = TRUE;
    Timers[timerId].recurValue = ticks;
  } else {
    Timers[timerId].recurring = FALSE;
    Timers[timerId].recurValue = 0;
  }
}

//******************************************************************************

// This function will increment the system tick every SYSTEM_CLOCK_PERIOD clocks
// (about SYSTEM_CLOCK_PERIOD milliseconds).
// At that time, it will check if any timers expired.

int8u count;
void handleTimers(void)
{
  TimerFunctionPointer function;
  int16u time;
  int16u timePassed;
  int8u i;

  time = microGetSystemTick();

  if (!previousTime) 
    previousTime = time;

  //check for a rollover
  if (time < previousTime) {
    timePassed  = (0xffff - previousTime) + time;    
  } 
  else {
    timePassed = time - previousTime;
  }

  if (timePassed >= SYSTEM_CLOCK_PERIOD) {
    
    sysTicks += (timePassed / SYSTEM_CLOCK_PERIOD);

    //fprintf(stdout, "time=%d, pretime=%d, sysTicks=%d\n", time, previousTime, sysTicks);

    // Check expirations of timers
    for (i = 0; i < NUM_TIMERS; i++) {
      if ((Timers[i].timer_function != NULL) &&
          (Timers[i].expiration <= sysTicks))
      { // Timer expired

        //fprintf(stdout, "timer expire = %d\n", Timers[i].expiration);

        function = Timers[i].timer_function; // save its function
        if (Timers[i].recurring)
          Timers[i].expiration = Timers[i].recurValue + sysTicks;
        else
          Timers[i].timer_function = NULL;
        // Run the timer function
        function(i, Timers[i].param);
      }
    }
    previousTime = time;
  }
}


//get time to the next event
int32u msToNextEvent(void) {
  int8u i;
  int16u leastTicks = 0xffff;
  int16u curTicksRemaining = 0;
  int32u msToNext = 0;
  for (i = 0; i < NUM_TIMERS; i++) {
    if (Timers[i].timer_function != NULL) {
      curTicksRemaining = (Timers[i].expiration - sysTicks);
       if (curTicksRemaining < leastTicks)
          leastTicks = curTicksRemaining;
    }
  }
  msToNext = (((leastTicks * 125) / 2) + 1 );  


  //debug
  //fprintf(stdout, "msToNext = %d\n", msToNext);

  return msToNext; //always overestimate by .5 - 1 ms.
}


//******************************************************************************


