/**
   @File host-api.c
   @brief Implementation of the host API 
   @see host-api.h, rnap.h, ember.h
   @author Guohui Liu (guohui@ember.com)

   This Ember Host API library is subject to the Ember corporation
   Development Kit Software License Agreement delivered with this file to
   licensee.
   * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
   */
#include <stdlib.h>             // for malloc(), exit()
#include <stdio.h> 
#include <string.h>             // for memcpy
#include <sys/time.h>           // for clock
#include <pthread.h>            // for thread
#include <signal.h>             // for signal
#include <sys/select.h>         // for select
#include "app/host-framework/c/host-api.h"
#include "app/host-framework/c/packet-buffer.h"
#include "app/host-framework/c/binding-table.h"
#include "app/host-framework/c/rnap-master.h"
#include "app/host-framework/c/rnap-message-queue.h"
#include "stack/include/error.h"
#include "app/util/rnap/rnap.h"

#ifdef EMBER_TEST
#include "tool/simulator/child/host-child.h"
#endif

//------------------------------------------------------------------------------
// Variables used in other files
//------------------------------------------------------------------------------
// Host binding table
Binding emHostBindingTable[EMBER_HOST_BINDING_TABLE_SIZE];
// Connection table
Connection emHostConnectionTable[EMBER_HOST_CONNECTION_TABLE_SIZE];
// The last rnap message that is sent 
EmberMessageBuffer emLastRnapMessageSent;
// The EUI64 for the node that the host talks to.
EmberEUI64 emLocalEui64;
//The Gateway Node's Pan Id
EmberPanId emLocalPanId;
//routing
boolean emEnableRelay;

//------------------------------------------------------------------------------
// Forward declarations
//------------------------------------------------------------------------------
static void clearMessages(void);
static EmberStatus clearTables(void);
static EmberStatus okToSend(EmberMessageBuffer message);
static int8u getNewCacheBindingIndex(int8u bindingIndex);
static void updateCacheBindingIndex(int8u bindingIndex, int8u cacheIndex);
static int8u getNewConnectionId(Connection *connection);
static EmberStatus sendSetBindingMessage(int8u bindingIndex,
                                         EmberBindingTableEntry *value);
static EmberStatus sendUnicast(boolean isDatagram, int8u bindingIndex,
                               EmberMessageBuffer message);
static Connection *updateRouteCost(int8u bindingIndex, boolean costOnly);
static void clearBinding(int8u bindingIndex);
static EmberStatus getBindingConnection(int8u bindingIndex);
static EmberStatus setBindingRoute(int8u bindingIndex);
static EmberInterfaceId *findInterface(EmberTerminal terminal);
static EmberStatus acknowledgeNodeReset(void);
static EmberStatus getOrSetValue(int8u type, int8u parameter, int8u cacheIndex,
                                 int8u *parameterData, int8u length);
static boolean sameSequenceNumber(EmberMessageBuffer buf, 
                                  int8u *sentNum,
                                  int8u ackedNum);
static boolean isUnicastMessage(EmberMessageBuffer buf,
                                int8u *bindingIndex,
                                int8u *messageId);
static void *timer_thread (void *arg);
static void *signal_thread (void *arg);
static boolean isSetBindingMessage(EmberMessageBuffer buf,
                                   int8u *hostBindingIndex);
static EmberStatus setCacheBinding(int8u bindingTableIndex);
static void processMessageSent(int8u index,
                               int8u messageId,
                               EmberStatus ackStatus);
static EmberStatus checkRnapMessageQueue(void);
static void retryLastRnapMessage(void);

//------------------------------------------------------------------------------
// Local variables
//------------------------------------------------------------------------------
// Preallocated for rnap command
static int8u commandBuffer[MAX_RNAP_PACKET_SIZE]; 
// Radio type, channel, and power of the node that the host talks to
static int8u radioType = RADIO_NULL;
static int8u radioChannel;
static int8s radioPower;
// Binding table size of the node binding table
static int8u nodeBindingTableSize;
// Time structs for microGetSystemTick()
static struct timeval tv;
static struct timezone tz;
// Connection id used by the next connection
static int8u nextConnectionId;
// Host side rnap version
static int8u version = 0;
// Host side sequence number
static int8u seqNum;      
// Next message id
static int8u messageToken;
// Outgoing messages that are not acked
static EmberMessageBuffer messageSent[256];
// Incoming out of sequenced messages
static EmberMessageBuffer messageReceived[256];
// Information extracted from the incoming message
static int8u currentMessageId, currentSendingTerminal;
int8u emCurrentRouteLength;
int8u emCurrentRoute[MAX_RNAP_PACKET_SIZE];
int16u emCurrentSender;
static EmberCost currentCost;
static EmberIncomingMessageType currentMessageType = -1;
// Datagram reply status
static boolean replySent = TRUE;
// Retry times for RNAP message
static int8u retry;
// Stack status
static EmberStatus stackStatus;
// Node (cache) binding index to host binding index mapping
static int8u nodeHostBindingIndexTable[EMBER_HOST_BINDING_TABLE_SIZE];
// Last picked cache binding index
static int8u lastReusedCacheIndex, lastCacheIndex;
// Thread related variables
static pthread_t  timer_thr_id, sig_thr_id; // thread id
static sigset_t  signal_mask;

/* NEW TO HOST API FOR RELEASE 3.3 */
/* The initialization routine on the gateway node has changed
   somewhat, the host should first call */
static EmberNetworkParameters emberNetworkParameters;
static boolean associateOnNodeStart;
static boolean nodeStarted;

//------------------------------------------------------------------------------
// Micros 
//------------------------------------------------------------------------------
// Invalid connection id
#define NULL_CONNECTION_ID 0
// Gets the valid nextConnectionId.
// If nextConnectionId is invalid (NULL_CONNECTION_ID), increases 1.
#define normalizeNextConnectionId()                     \
 do { if (nextConnectionId == NULL_CONNECTION_ID)       \
         nextConnectionId = NULL_CONNECTION_ID + 1;     \
 } while (0)
// Time period for checking the signal, in millisecond
#define TIMER_PERIOD 2000
// Maximum waiting period for the ack before retry
#define MAX_WAIT_TIME TIMER_PERIOD * 2 
// Maximum retry times
#define MAX_RETRY 3
// The binding is in the transition of being cleared.
// The cache index can not be used until the node informs the host about the
// clearing of the binding.
#define BINDING_IS_UNUSABLE 0xFE

//------------------------------------------------------------------------------
// RNAP Host APIs
//------------------------------------------------------------------------------
// Initializes the rnap communication.
// Clears all tables and variables on the host.
// Sends HostHasReset RNAP message to the node.
EmberStatus emberHostInit(int8u *name, int16u port, GatewayConnectionMode mode)
{
  
  // According to the man page, the signals need to be blocked
  // for ALL threads, or no guarantees on where the signal
  // actually gets handled. (fb:1592)
  // init signal mask
  // any newly created threads inherit the signal mask
    sigemptyset (&signal_mask);
    sigaddset (&signal_mask, SIGUSR1);
    if (pthread_sigmask (SIG_BLOCK, &signal_mask, NULL) != 0)
      return EMBER_ERR_FATAL;
  
  // create thread for reading the rnap port
  switch (mode) {
  case EMBER_HOST_MODE_SERIAL:
    if (emMasterConnectSerial(name) != EMBER_SUCCESS)
      return EMBER_RNAP_COMMUNICATION_FAILED;
    break;
  case EMBER_HOST_MODE_SOCKET:
    if (emMasterConnectSocket(name, port) != EMBER_SUCCESS)
      return EMBER_RNAP_COMMUNICATION_FAILED;
    break;
  case EMBER_HOST_MODE_STDIO:
    if (emMasterConnectStdIO() != EMBER_SUCCESS)
      return EMBER_RNAP_COMMUNICATION_FAILED;
    break;
  default:
    fprintf(stdout, "connection mode not recognized.\n");
    return EMBER_ERR_FATAL;
  } 

  // create the timer thread
  if (pthread_create (&timer_thr_id, NULL, timer_thread, NULL) != 0)
    return EMBER_ERR_FATAL;

  // create a signal thread
  if (pthread_create(&sig_thr_id, NULL, signal_thread, NULL) != 0)
    return EMBER_ERR_FATAL;

  // clear tables and messages
  if (clearTables() != EMBER_SUCCESS)
    return EMBER_ERR_FATAL;
  clearMessages();

  // initialize variables
  nextConnectionId = 1;
  stackStatus = EMBER_NETWORK_DOWN;
  retry = 0;
  replySent = TRUE;
  currentMessageType = -1;

  // send host has reset rnap message
  return emberHostHasReset();
}

// Indicates that this, the host processor, has reset
EmberStatus emberHostHasReset(void)
{
  // clear the RNAP message queue
  emRnapQInit();

  // compose and send the rnap message
  commandBuffer[0] = H_TO_N_HOST_HAS_RESET;
  commandBuffer[1] = seqNum++;
  commandBuffer[2] = version;
  return emMasterWrite(NULL, 0, commandBuffer, 3);
}

// Instructs the node to reset.
EmberStatus emberResetNode(void)
{
  commandBuffer[0] = H_TO_N_RESET_NODE;
  commandBuffer[1] = seqNum++;
  return emMasterWrite(NULL, 0, commandBuffer, 2);
}

//Check the rnap message queue to see if the 
//last function call has completed it's interaction
EmberStatus checkRnapMessageQueue(void) {
  if (emGetRnapQSize() > 0)
  {
    return EMBER_MAX_MESSAGE_LIMIT_REACHED;
  } else
    return EMBER_SUCCESS;
}


// Enable/disable the node to route packets on behalf of others.
EmberStatus emberEnableRouting(boolean enable)
{
  EmberStatus s = checkRnapMessageQueue();
  if (s != EMBER_SUCCESS)
    return s;

  return getOrSetValue(H_TO_N_SET_VALUE, NODE_ENABLE_ROUTING, 
                       NULL_BYTE, &enable, 1);
}

// Sets the radio channel.
// Returns error when the channel is out of range.
EmberStatus emberSetRadioChannel(int8u channel)
{
  EmberStatus s = checkRnapMessageQueue();
  if (s != EMBER_SUCCESS)
    return s;

  
  DEBUG_OUT(fprintf(stdout, "radio type = %d\n", radioType));
  switch (radioType) {
#if defined(EMBER_TEST)
  case 0:
    radioChannel = channel;
    return EMBER_SUCCESS;
#endif
  case RADIO_868:
    if (channel < 50 || channel > 52)
      return EMBER_PHY_INVALID_CHANNEL;
    break;
  case RADIO_915:
    if (channel > 48)
      return EMBER_PHY_INVALID_CHANNEL;
    break;
  case RADIO_2420:
    if (channel > 15)
      return EMBER_PHY_INVALID_CHANNEL;
    break;
  default:
    return EMBER_PHY_UNKNOWN_RADIO_TYPE;  
  }
  return getOrSetValue(H_TO_N_SET_VALUE, NODE_RADIO_CHANNEL, 
                       NULL_BYTE, &channel, 1);
}

// Sets the radio transmit power.
// Returns error when the power is out of range.
EmberStatus emberSetRadioPower(int8s power)
{
  EmberStatus s = checkRnapMessageQueue();
  if (s != EMBER_SUCCESS)
    return s;

  DEBUG_OUT(fprintf(stdout, "radio type = %d\n", radioType));
  switch (radioType) {
#if defined(EMBER_TEST)
  case 0:
    radioPower = power;
    return EMBER_SUCCESS;
#endif
  case RADIO_868:
  case RADIO_915:
    // -25 to -6
    if (power < -25 || power > -6)
      return EMBER_PHY_INVALID_POWER;
    break;
  case RADIO_2420:
    // -32 to -1
    if (power < -32 || power > -1)
      return EMBER_PHY_INVALID_POWER;
    break;
  default:
    return EMBER_PHY_UNKNOWN_RADIO_TYPE;      
  }
  return getOrSetValue(H_TO_N_SET_VALUE, NODE_RADIO_POWER, 
                       NULL_BYTE, &power, 1);
}

//Associates the gateway node with an Ember network
EmberStatus emberAssociate(EmberNetworkParameters *parameters)
{
  EmberStatus s = checkRnapMessageQueue();
  if (s != EMBER_SUCCESS)
    return s;

  //copy the params into our local memory location to save
  emberNetworkParameters.panId = parameters->panId;
  emberNetworkParameters.enableRelay = parameters->enableRelay;
  emberNetworkParameters.radioTxPower = parameters->radioTxPower;
  emberNetworkParameters.radioChannel = parameters->radioChannel;

  //if the node is started up, then we can send them the associate
  //call, otherwise we need to wait until the node has received our
  //NODE_START_ACK call.
  if (nodeStarted)
  {
    int8u buff[6];
    buff[0] = TRUE;
    buff[1] = HIGH_BYTE(parameters->panId);
    buff[2] = LOW_BYTE(parameters->panId);
    buff[3] = parameters->enableRelay;
    buff[4] = parameters->radioTxPower;
    buff[5] = parameters->radioChannel;
    return getOrSetValue(H_TO_N_SET_VALUE, NODE_NETWORK_ASSOCIATION,
                       NULL_BYTE, buff, 6);
  } else {
    //if the node hasn't started up yet then we need to remember to associate
    //when it does start
    associateOnNodeStart = TRUE;
  }
  return EMBER_SUCCESS;
}

//Disassociates the gateway node with an Ember network
EmberStatus emberDisassociate()
{
  EmberStatus s = checkRnapMessageQueue();
  if (s != EMBER_SUCCESS)
    return s;

  boolean assocValue = FALSE;
  return getOrSetValue(H_TO_N_SET_VALUE, NODE_NETWORK_ASSOCIATION,
                       NULL_BYTE, &assocValue, 1);
}
// Returns the radio channel.
int8u emberGetRadioChannel(void)
{
  return radioChannel;
}

// Returns the radio power.
int8s emberGetRadioPower(void)
{
  return radioPower;
}

//------------------------------------------------------------------------------
// The following functions are related to sending message.

// Sends multicast with limited hops.
EmberStatus emberSendLimitedMulticast(int8u bindingTableIndex,
                                      EmberMessageBuffer message,
                                      int8u hops)
{
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;
  
  EmberBindingTableEntry *entry =
    emHostBindingTable[bindingTableIndex].bindingEntry;
  int8u length = message->length;

  status = okToSend(message);
  if (status != EMBER_SUCCESS) 
    return status;

  if (bindingTableIndex >= EMBER_HOST_BINDING_TABLE_SIZE
      || entry == NULL
      || entry->type != EMBER_MULTICAST_BINDING)
    return EMBER_INVALID_BINDING_INDEX;

  commandBuffer[0] = H_TO_N_TX_MULTICAST;
  commandBuffer[1] = seqNum++;
  commandBuffer[2] = entry->identifier[0];
  commandBuffer[3] = entry->identifier[1];
  commandBuffer[4] = hops;
  commandBuffer[5] = length;

  return emMasterWrite(message->contents, length, commandBuffer, 6);
}

// Sends squenced message.
EmberStatus emberSendSequenced(int8u bindingTableIndex,
                               EmberMessageBuffer message)
{
  Binding *entry = &emHostBindingTable[bindingTableIndex];
  Connection *connection;
  
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  status = okToSend(message);
  if (status != EMBER_SUCCESS)
    return status;

  if (bindingTableIndex >= EMBER_HOST_BINDING_TABLE_SIZE
      || entry->bindingEntry == NULL
      || (entry->bindingEntry)->type != EMBER_UNICAST_BINDING)
    return EMBER_INVALID_BINDING_INDEX;

  connection = emHostBindingTable[bindingTableIndex].remoteNode;

  if (connection == NULL || connection->state != CON_ESTABLISHED)
    return EMBER_CONNECTION_CLOSED;

  // set binding on node if it is necessary
  status = setCacheBinding(bindingTableIndex);
  if (status != EMBER_SUCCESS)
    return status;

  // set the binding route if route is new
  if (entry->remoteNode != NULL && (entry->remoteNode)->routeNew) {
    status = setBindingRoute(bindingTableIndex);
    if (status != EMBER_SUCCESS)
      return status;
  }

  // send the sequenced message
  return sendUnicast(FALSE, bindingTableIndex, message);
}

// Sends datagram.
EmberStatus emberSendDatagram(int8u bindingTableIndex, 
                              EmberMessageBuffer message)
{
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  Binding *entry = &emHostBindingTable[bindingTableIndex];


  // check the ember stack status and message length
  status = okToSend(message);
  if (status != EMBER_SUCCESS)
    return status;

  // check the binding information
  if (bindingTableIndex >= EMBER_HOST_BINDING_TABLE_SIZE
      || entry->bindingEntry == NULL  
      || (entry->bindingEntry)->type != EMBER_UNICAST_BINDING)
    return EMBER_INVALID_BINDING_INDEX;

  // set binding on node if it is necessary
  status = setCacheBinding(bindingTableIndex);
  if (status != EMBER_SUCCESS)
    return status;
  
  // set the binding route if route is new
  if (entry->remoteNode != NULL && (entry->remoteNode)->routeNew) {
    status = setBindingRoute(bindingTableIndex);
    if (status != EMBER_SUCCESS)
      return status;
  }

  // send the datagram message
  return sendUnicast(TRUE, bindingTableIndex, message);
}

// Same again with destination given by the sending (local) terminal. 
EmberStatus emberSendDatagramFromTerminal(EmberTerminal sender,
                                          EmberMessageBuffer message)
{
  int8u index;
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  status = emFindUniqueTerminalBindingIndex(sender, &index);
  if (status == EMBER_SUCCESS)
    return emberSendDatagram(index, message);
  else
    return status;
}

EmberStatus emberSendMulticastFromTerminal(EmberTerminal sender,
                                           EmberMessageBuffer message)
{
  int8u index;
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  status = emFindUniqueTerminalBindingIndex(sender, &index);
  if (status == EMBER_SUCCESS)
    return emberSendMulticast(index, message);
  else
    return status;
}

EmberStatus emberSendSequencedFromTerminal(EmberTerminal sender,
                                           EmberMessageBuffer message)
{
  int8u index;
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  status = emFindUniqueTerminalBindingIndex(sender, &index);
  if (status == EMBER_SUCCESS)
    return emberSendSequenced(index, message);
  else
    return status;
}

// Sends reply to the incoming datagram message.
EmberStatus emberSendReply(EmberMessageBuffer reply)
{
  // cmd (1) | seqNumber (1) | destination (2) 
  // | messageId (1) | sendingTerminal (1) | routeLength(1) | route (n) 
  // | payloadLength (1) | payload (n)
 
  // NOTE: notice that we don't check the rnap message queue here because these 
  // reply's are time sensitive so we want to make sure they get out regardless
  // of what other messages might be on the message queue.

  
  // check to see if the incoming message is datagram and the reply is not sent
  if (currentMessageType != EMBER_INCOMING_DATAGRAM
      || replySent)
    return EMBER_INVALID_CALL;

  replySent = TRUE;
  commandBuffer[0] = H_TO_N_TX_DATAGRAM_REPLY;
  commandBuffer[1] = seqNum++;
  commandBuffer[2] = HIGH_BYTE(emCurrentSender);
  commandBuffer[3] = LOW_BYTE(emCurrentSender);
  commandBuffer[4] = currentMessageId;
  commandBuffer[5] = currentSendingTerminal;
  commandBuffer[6] = emCurrentRouteLength;
  memcpy(commandBuffer + 7, emCurrentRoute, emCurrentRouteLength);

  if (reply == EMBER_NULL_MESSAGE_BUFFER) {
    commandBuffer[7 + emCurrentRouteLength] = 0;
    return emMasterWrite(NULL, 0, commandBuffer, 8 + emCurrentRouteLength);
  }
  else {
    commandBuffer[7 + emCurrentRouteLength] = reply->length;
    return emMasterWrite(reply->contents, reply->length, 
                         commandBuffer, 8 + emCurrentRouteLength);
  }
}

//------------------------------------------------------------------------------
// The following functions are related to cost.

// Gets the cost to sender.
// The incoming message shouldn't be multicast.
EmberStatus emberGetCostToSender(EmberCost *returnLoc)
{
  if (currentMessageType == EMBER_INCOMING_MULTICAST)
    return EMBER_INVALID_CALL;
  
  *returnLoc = currentCost;
  return EMBER_SUCCESS;
}

// Returns the cost in returnLoc
EmberStatus emberGetCost(int8u index, EmberCost *returnLoc)
{
  Connection *node;
  EmberBindingTableEntry *entry = emHostBindingTable[index].bindingEntry;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= index)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;
  
  if (entry == NULL
      || entry->type == EMBER_MULTICAST_BINDING
      || entry->type == EMBER_UNUSED_BINDING)
    return EMBER_INVALID_BINDING_INDEX;
  
  node = emHostBindingTable[index].remoteNode;
  if ( node == NULL
       || node->nodeId == NULL_NODE_ID
       || node->cost == 0)
    return EMBER_COST_NOT_KNOWN;

  *returnLoc = node->cost;
  return EMBER_SUCCESS;
}

//------------------------------------------------------------------------------
// The following functions are related to connection.

// Opens a connection.
EmberStatus emberOpenConnection(int8u bindingTableIndex)
{
  Connection *connection;
  Binding *entry;
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= bindingTableIndex)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  // allocate connection if it doesn't exist 
  status = getBindingConnection(bindingTableIndex);
  if (status != EMBER_SUCCESS)
    return status;

  entry = &emHostBindingTable[bindingTableIndex];
  connection = entry->remoteNode;

  switch (connection->state) {
  case CON_LISTEN: {
    int8u status;
    int8u conId;

    // set binding on node if it is necessary
    status = setCacheBinding(bindingTableIndex);
    if (status != EMBER_SUCCESS)
      return status;
  
    commandBuffer[0] = H_TO_N_OPEN_CONNECTION;
    commandBuffer[1] = seqNum++;
    commandBuffer[2] = emHostBindingTable[bindingTableIndex].cacheIndex;
    conId = getNewConnectionId(connection); // connection id
    connection->localId = conId;
    commandBuffer[3] = conId;

    status = emMasterWrite(NULL, 0, commandBuffer, 4);

    if (status != EMBER_SUCCESS)
      return status;
    else {
      connection->state = CON_WAIT;    
      return EMBER_SUCCESS;
    }
  }
  case CON_ESTABLISHED:
    return EMBER_CONNECTION_OPEN;
  case CON_WAIT:
    return EMBER_SUCCESS;       // an open is already in progress
  default:
    return EMBER_CONNECTION_CLOSING;
  }
}

// Return the connection status
EmberStatus emberConnectionStatus(int8u bindingTableIndex)
{
  Connection *connection;
  
  if (EMBER_HOST_BINDING_TABLE_SIZE <= bindingTableIndex)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  connection = emHostBindingTable[bindingTableIndex].remoteNode;
  if (connection == NULL)
    return EMBER_CONNECTION_CLOSED;
  else
    switch (connection->state) {
    case CON_UNUSED:              // this should not happen
    case CON_LISTEN:
      return EMBER_CONNECTION_CLOSED;
    case CON_WAIT:
      return EMBER_CONNECTION_NOT_YET_OPEN;
    case CON_ESTABLISHED:
      return EMBER_CONNECTION_OPEN;
    default:
      return EMBER_CONNECTION_CLOSING;
    }
}

// Closes a connection.
EmberStatus emberCloseConnection(int8u bindingTableIndex)
{
  Connection *connection;
  EmberStatus status = checkRnapMessageQueue();
  if (status != EMBER_SUCCESS)
    return status;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= bindingTableIndex)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  connection = emHostBindingTable[bindingTableIndex].remoteNode;
  if (connection == NULL)
    return EMBER_CONNECTION_CLOSED;

  switch (connection->state) {
  case CON_LISTEN:
    return EMBER_CONNECTION_CLOSED;
  case CON_WAIT:
    return EMBER_CONNECTION_NOT_YET_OPEN;
  case CON_ESTABLISHED: {
    // change connection state    
    if (emHostBindingTable[bindingTableIndex].cacheIndex 
        == EMBER_NULL_BINDING) {
      connection->state = CON_LISTEN;
      return EMBER_CONNECTION_CLOSED;
    }
    else
      connection->state = CON_WAIT;    

    // cmd, seq, cache binding index
    commandBuffer[0] = H_TO_N_CLOSE_CONNECTION;
    commandBuffer[1] = seqNum++;
    commandBuffer[2] = emHostBindingTable[bindingTableIndex].cacheIndex;
    
    return emMasterWrite(NULL, 0, commandBuffer, 3);
  }
  default:
    return EMBER_CONNECTION_CLOSING;
  }
}

//------------------------------------------------------------------------------
// The following functions are related to binding.

// Returns the binding information in result.
EmberStatus emberGetBinding(int8u index, EmberBindingTableEntry *result)
{
  EmberBindingTableEntry entry;  

  if (EMBER_HOST_BINDING_TABLE_SIZE <= index)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  if (emHostBindingTable[index].bindingEntry == NULL) {
    int8u i;
    entry.type = 0;
    for (i = 0; i < EUI64_SIZE; i++)
      entry.identifier[i] = 0;
    entry.local = 0;
    entry.remote = 0;
  }
  else  
    entry = *emHostBindingTable[index].bindingEntry;

  memcpy(result, &entry, sizeof(EmberBindingTableEntry));
  
  return EMBER_SUCCESS;
} 

// Sets the binding to the sender of the incoming message and update the 
// sender information.
// The incoming message shouldn't be multicast.
EmberStatus emberSetReplyBinding(int8u index, EmberBindingTableEntry *entry)
{
  EmberStatus status;
  Binding *binding = &emHostBindingTable[index];

  if (entry == NULL)
    return EMBER_ERR_FATAL;

  if (currentMessageType == EMBER_INCOMING_MULTICAST
      || entry->type != EMBER_UNICAST_BINDING)
    return EMBER_INVALID_CALL;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= index)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  if (emberBindingIsActive(index))
    return EMBER_BINDING_IS_ACTIVE;

  if (emberTerminalEndpoint(entry->local) != 0
      && findInterface(entry->local) == NULL)
    return EMBER_INVALID_BINDING_TERMINAL;

  if (binding->bindingEntry == NULL)
    binding->bindingEntry = 
      (EmberBindingTableEntry *)malloc(sizeof(EmberBindingTableEntry));
  
  if (binding->bindingEntry == NULL)
    return EMBER_ERR_FATAL;

  memcpy(binding->bindingEntry, entry, sizeof(EmberBindingTableEntry));
  
  clearBinding(index);

  status = emberNoteSendersBinding(index);
  if (status != EMBER_SUCCESS)
    return status;

  return EMBER_SUCCESS;
}

// Updates the sender information.
EmberStatus emberNoteSendersBinding(int8u index)
{
  Connection *connection;

  if (currentMessageType == EMBER_INCOMING_MULTICAST)
    return EMBER_INVALID_CALL;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= index)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  connection = updateRouteCost(index, FALSE);
  
  if (connection != NULL) {
    emHostBindingTable[index].remoteNode = connection;
    return EMBER_SUCCESS;
  }
  else
    return EMBER_TOO_MANY_CONNECTIONS;
}

// Returns TRUE if the binding has no outgoing message.
boolean emberBindingIsActive(int8u bindingIndex)
{
  Connection *connection = emHostBindingTable[bindingIndex].remoteNode;

  return (connection != NULL) 
    && (connection->numOfOutStandingMessage != 0
        || connection->state == CON_WAIT);
}

// Sets the binding entry.
EmberStatus emberSetBinding(int8u index, EmberBindingTableEntry *value)
{
  Binding *binding = &emHostBindingTable[index];

  if (value == NULL
      || value->type == EMBER_UNICAST_BINDING)
    return EMBER_ERR_FATAL;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= index)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  if (emberTerminalEndpoint(value->local) != 0
      && findInterface(value->local) == NULL)
    return EMBER_INVALID_BINDING_TERMINAL;

  if (emberBindingIsActive(index))
    return EMBER_BINDING_IS_ACTIVE;

  if (binding->bindingEntry == NULL)
    binding->bindingEntry = 
      (EmberBindingTableEntry *)malloc(sizeof(EmberBindingTableEntry));

  if (binding->bindingEntry == NULL)
    return EMBER_ERR_FATAL;

  memcpy(binding->bindingEntry, value, sizeof(EmberBindingTableEntry));
  
  clearBinding(index);

  return EMBER_SUCCESS;
}

// Clears binding table.
EmberStatus emberClearBindingTable(void)
{
  int16u index;
  EmberStatus status = EMBER_SUCCESS;  

  for (index = 0; index < EMBER_HOST_BINDING_TABLE_SIZE; index++) {
    EmberStatus temp = emberDeleteBinding(index);
    if (temp != EMBER_SUCCESS)
      status = temp;
  }
  return status;
}

// Deletes one binding.
EmberStatus emberDeleteBinding(int8u index)
{
  EmberBindingTableEntry *entry = emHostBindingTable[index].bindingEntry;
  EmberStatus status;

  if (EMBER_HOST_BINDING_TABLE_SIZE <= index)
    return EMBER_BINDING_INDEX_OUT_OF_RANGE;

  if (emberBindingIsActive(index))
    return EMBER_BINDING_IS_ACTIVE;

  if (entry != NULL) {
    entry->type = EMBER_UNUSED_BINDING;
    status = emberSetBinding(index, entry);
    free(emHostBindingTable[index].bindingEntry);
    emHostBindingTable[index].bindingEntry = NULL;
    return status;
  }

  return EMBER_SUCCESS;
}

//------------------------------------------------------------------------------
// Other utility functions

// Gets the incoming datagram meesage's sending terminal.
EmberStatus emberGetSendingTerminal(EmberTerminal *returnLoc)
{
  if (currentMessageType != EMBER_INCOMING_DATAGRAM)
    return EMBER_INVALID_CALL;
  else {
    *returnLoc = currentSendingTerminal;
    return EMBER_SUCCESS;
  }
}

// Gets the size of an interface.  
// The first two bytes are the domain and type.
int8u emberInterfaceIdSize(EmberInterfaceId *interfaceId)
{
  switch (interfaceId->domain) {
  case ENDPOINT_ZERO_INTERFACE:         // spec is two byte identifer
    return 2 + 2;
  case AD_HOC_INTERFACE:                // first byte gives number of remaining
    return 2 + 1 + interfaceId->specification[0];
  case ZIGBEE_INTERFACE:                // two bytes of profile ID, two bytes
    return 2 + 2 + 2;                   // of device ID
  case EUI64_INTERFACE:                 // eight bytes of IEEE ID
    return 2 + 8;
  default:
    return 255;                         // seems safer than zero
  }
}

// Puts the gateway node into bootload mode.
// No ack for this RNAP message.
EmberStatus emberBootload(void)
{
  // compose and send the rnap message
  commandBuffer[0] = H_TO_N_BOOTLOAD;
  commandBuffer[1] = seqNum++;
  return emMasterWrite(NULL, 0, commandBuffer, 2);
}

//------------------------------------------------------------------------------
// Local functions

// Searching the nodeHostBindingIndexTable to get the first unused
// cache index or reuse one that has no outstanding message.
static int8u getNewCacheBindingIndex(int8u bindingTableIndex)
{
  int8u i;
  int8u index;
  int8u reusedIndex;

  for (i = 0; i < nodeBindingTableSize; i++)
    index = (i + lastCacheIndex + 1) % nodeBindingTableSize;
  if (nodeHostBindingIndexTable[index] == EMBER_NULL_BINDING) {
    nodeHostBindingIndexTable[index] = bindingTableIndex;
    return index;
  }

  for (i = 0; i < nodeBindingTableSize; i++) {
    reusedIndex = (i + lastReusedCacheIndex + 1) % nodeBindingTableSize;
    if (!emberBindingIsActive(nodeHostBindingIndexTable[reusedIndex])
        && nodeHostBindingIndexTable[reusedIndex] != BINDING_IS_UNUSABLE) {
      emHostBindingTable[nodeHostBindingIndexTable[reusedIndex]].cacheIndex = BINDING_IS_UNUSABLE;
      nodeHostBindingIndexTable[reusedIndex] = bindingTableIndex;
      lastReusedCacheIndex = reusedIndex;
      
      DEBUG_OUT(fprintf(stdout, "======reuse %d\n", reusedIndex));
      return reusedIndex;
    }
  }
  return EMBER_NULL_BINDING;
}

// Updates the nodeHostBindingIndexTable and the host binding table cache index.
static void updateCacheBindingIndex(int8u bindingIndex, int8u cacheIndex)
{
  int8u oldCacheIndex;

  oldCacheIndex = emHostBindingTable[bindingIndex].cacheIndex;
  if (oldCacheIndex != EMBER_NULL_BINDING)
    nodeHostBindingIndexTable[oldCacheIndex] = BINDING_IS_UNUSABLE;
  nodeHostBindingIndexTable[cacheIndex] = bindingIndex;
  emHostBindingTable[bindingIndex].cacheIndex = cacheIndex;
}

// Returns the next connection id.
// If the new id is the same as the one used last time, pick the next
// connection id.
static int8u getNewConnectionId(Connection *connection)
{
  int8u result;

  if (connection->localId == nextConnectionId) {
    nextConnectionId += 1;
    normalizeNextConnectionId();
  }
  
  result = nextConnectionId;
  nextConnectionId += 1;
  normalizeNextConnectionId();
  return result;
}

// Clears the entry of binding table.
static void clearBindingTableEntry(int8u bindingTableIndex)
{  
  emHostBindingTable[bindingTableIndex].cacheIndex = EMBER_NULL_BINDING;
  if (emHostBindingTable[bindingTableIndex].bindingEntry != NULL) {
    free(emHostBindingTable[bindingTableIndex].bindingEntry);
    emHostBindingTable[bindingTableIndex].bindingEntry = NULL;
  }
  emHostBindingTable[bindingTableIndex].remoteNode = NULL;
  emHostBindingTable[bindingTableIndex].remoteIndex = EMBER_NULL_BINDING;
  emHostBindingTable[bindingTableIndex].failedTimes = 0;
}

// Clears one entry of connetion table.
static EmberStatus clearConnectionTableEntry(Connection *remoteNode, 
                                             int8u state)
{
  if (remoteNode == NULL)
    return EMBER_ERR_FATAL;
    
  remoteNode->nodeId = NULL_NODE_ID;
  remoteNode->routeLength = 0;
  remoteNode->routeNew = FALSE;
  remoteNode->cost = 0xff;
  remoteNode->numOfOutStandingMessage = 0;
  remoteNode->state = state;
  remoteNode->localId = NULL_CONNECTION_ID;
  remoteNode->remoteId = NULL_CONNECTION_ID;
  remoteNode->inRetryId = 0;
  remoteNode->inRetryMask = 0;
  remoteNode->outRetryId = 0;

  return EMBER_SUCCESS;
}

// Clears the sent and received messages.
static void clearMessages(void)
{
  int16u i;

  for (i = 0; i <= 255; i++) {
    // clear sent message
    if (messageSent[i] != NULL) {
      free(messageSent[i]);
      messageSent[i] = NULL;
    }
    // clear received message
    if (messageReceived[i] != NULL) {
      free(messageSent[i]);
      messageReceived[i] = NULL;
    }
  }
  // clear last sent rnap message
  if (emLastRnapMessageSent != NULL) {
    free(emLastRnapMessageSent);
    emLastRnapMessageSent = NULL;
  }
}

// Clears tables: binding, connection and nodeHostBindingIndexTable.
static EmberStatus clearTables(void)
{
  int8u i;

  for (i = 0; i < EMBER_HOST_BINDING_TABLE_SIZE; i++) {
    clearBindingTableEntry(i);
    nodeHostBindingIndexTable[i] = EMBER_NULL_BINDING;
  }

  for (i = 0; i < EMBER_HOST_CONNECTION_TABLE_SIZE; i++)
    clearConnectionTableEntry(&emHostConnectionTable[i], CON_UNUSED); 

  return EMBER_SUCCESS;
}

// We look for an unused one.  Failing that we look for one which has no
// extant packets, in which case we reuse it.
static int8u findFreeConnection(void)
{
  int8u i;

  for (i = 0; i < EMBER_HOST_CONNECTION_TABLE_SIZE; i++)
    if (emHostConnectionTable[i].state == CON_UNUSED) {
      DEBUG_OUT(fprintf(stdout, "find free connection %d\n", i));
      return i;
    }
  for (i = 0; i < EMBER_HOST_CONNECTION_TABLE_SIZE; i++)
    if (emHostConnectionTable[i].state == CON_LISTEN) {
      DEBUG_OUT(fprintf(stdout, "reuse connection %d\n", i));
      return i;
    }

  return NULL_BYTE;
}

// Intializes a connection and assign it to the binding table's 
// corresponding remote node.
// Returns a pointer to the connection.
static Connection *initializeConnection(int8u connectionIndex,
                                        int8u bindingIndex)
{
  Connection *connection = &(emHostConnectionTable[connectionIndex]);

  DEBUG_OUT(fprintf(stdout, "[connection %d for binding %d]\n", 
                    connectionIndex, bindingIndex));

  clearConnectionTableEntry(connection, CON_LISTEN);
  emHostBindingTable[bindingIndex].remoteNode = connection;

  return connection;
}

// Returns the connection pointer if we can find a free entry.
// Otherwise returns NULL.
static Connection *allocateConnection(int8u bindingIndex)
{
  int8u connectionIndex = findFreeConnection();
  if (connectionIndex == NULL_BYTE)
    return NULL;
  else
    return initializeConnection(connectionIndex, bindingIndex);
}

// Finds or allocates a connection entry with the same nodeId as 
// emCurrentSender and updates the route and cost, or cost only.
static Connection *updateRouteCost(int8u bindingIndex, boolean costOnly)
{
  Connection *connection;
  int8u connectionIndex;

  connectionIndex = emFindNodeConnectionIndex(emCurrentSender);
  if (connectionIndex == NULL_BYTE) {
    if ((connection = allocateConnection(bindingIndex))  == NULL)
      return NULL;
    else {
      connection->nodeId = emCurrentSender;
      emHostBindingTable[bindingIndex].remoteNode = connection;
    }
  }
  else
    connection = &emHostConnectionTable[connectionIndex];

  if (connection->cost > currentCost && currentCost != 0) {
    if (costOnly) {
      connection->cost = currentCost;
      return connection;
    }
    else {
      connection->cost = currentCost;
      connection->routeLength = emCurrentRouteLength;
      memcpy(connection->route, emCurrentRoute, emCurrentRouteLength);
      connection->routeNew = TRUE;
      return connection; 
    }
  }
  else
    return connection; 
}

// Returns error code if the ember stack is not up or the message is too long.
static EmberStatus okToSend(EmberMessageBuffer message)
{
  int8u maxLength = MAX_APP_PAYLOAD_SIZE;

  if (stackStatus != EMBER_NETWORK_UP)
    return EMBER_NETWORK_DOWN;
    
  maxLength -= AGGREGATION_PAYLOAD_REDUCTION;

  return ((emberMessageBufferLength(message) > maxLength)
          ? EMBER_MESSAGE_TOO_LONG : EMBER_SUCCESS);
}

// Acknowledges a node reset.
static EmberStatus acknowledgeNodeReset(void)
{
  int8u i;

  //the node is in a limbo state until we ack their reset
  nodeStarted = FALSE;
  
  // inform user about the node reset
  stackStatus = EMBER_NETWORK_DOWN;
  emberStackStatusHandler(EMBER_NETWORK_DOWN);

  for (i = 0; i < EMBER_HOST_BINDING_TABLE_SIZE; i++) {
    // clear all cache index of the binding table
    emHostBindingTable[i].cacheIndex = EMBER_NULL_BINDING;
    // clear nodeHostBindingIndexTable
    nodeHostBindingIndexTable[i] = EMBER_NULL_BINDING;
  }

  for (i = 0; i < EMBER_HOST_CONNECTION_TABLE_SIZE; i++) {
    // clear numOfOutStandingMessage and connection data in connection table
    emHostConnectionTable[i].numOfOutStandingMessage = 0;
    if (emHostConnectionTable[i].state != CON_UNUSED)
      emHostConnectionTable[i].state = CON_LISTEN;
    emHostConnectionTable[i].localId = NULL_CONNECTION_ID;
    emHostConnectionTable[i].remoteId = NULL_CONNECTION_ID;
    emHostConnectionTable[i].inRetryId = 0;
    emHostConnectionTable[i].inRetryMask = 0;
    emHostConnectionTable[i].outRetryId = 0;
  }
  // clear message
  clearMessages();
  // clear rnap queue
  emRnapQInit();
  // clear variables
  retry = 0;

  // send ack message
  DEBUG_OUT(fprintf(stdout, "ack reset\n"));
  commandBuffer[0] = H_TO_N_NODE_START_ACK;
  commandBuffer[1] = seqNum++;
  commandBuffer[2] = version;
  return emMasterWrite(NULL, 0, commandBuffer, 3);
}

// General way to get/set node values
static EmberStatus getOrSetValue(int8u type, int8u parameter, int8u index,
                                 int8u *parameterData, int8u length) 
{
  commandBuffer[0] = type; // H_TO_N_SET/GET_VALUE;
  commandBuffer[1] = seqNum++;
  commandBuffer[2] = parameter;
  commandBuffer[3] = index;

  fprintf(stdout, "in set value\n");

  if (type == H_TO_N_SET_VALUE) {
    commandBuffer[4] = length;
    return emMasterWrite(parameterData, length, commandBuffer, 5);
  }
  else {
    return emMasterWrite(NULL, 0, commandBuffer, 4);
  }
}

// Composes and sends set binding message.
static EmberStatus sendSetBindingMessage(int8u bindingTableIndex, 
                                         EmberBindingTableEntry *value)

{
  Binding binding = emHostBindingTable[bindingTableIndex];
  int8u cacheIndex = binding.cacheIndex;
  int8u data[19]; // rnap message
  EmberStatus status;

  if (value == NULL)
    return EMBER_ERR_FATAL;

  data[0] = bindingTableIndex; // host index
  data[1] = value->type;
  memcpy(data + 2, &(value->identifier), 8);
  data[10] = value->local;
  data[11] = value->remote;
  data[13] = binding.remoteIndex;

  // connection data
  if (binding.remoteNode == NULL) {
    data[12] = FALSE; // isConnected
    data[14] = NULL_BYTE;
    data[15] = NULL_BYTE;
    data[16] = NULL_BYTE;
    data[17] = NULL_BYTE;
    data[18] = NULL_BYTE;
  }
  else {
    if ((binding.remoteNode)->state == CON_ESTABLISHED)
      data[12] = TRUE; // isConnected
    else
      data[12] = FALSE; // isConnected
    data[14] = (binding.remoteNode)->localId;
    data[15] = (binding.remoteNode)->remoteId;
    data[16] = (binding.remoteNode)->inRetryId;
    data[17] = (binding.remoteNode)->inRetryMask;
    data[18] = (binding.remoteNode)->outRetryId;
  }

  status = getOrSetValue(H_TO_N_SET_VALUE, NODE_BINDING, cacheIndex, data, 19);
  if (status == EMBER_SUCCESS)
    emHostBindingTable[bindingTableIndex].bindingStatus = BINDING_SENT_TO_NODE;

  return status;
}

// Sends datagram or sequenced message.
static EmberStatus sendUnicast(boolean isDatagram, 
                               int8u bindingIndex,
                               EmberMessageBuffer message)
{
  int8u status;
  int8u length = message->length;

  if (emHostBindingTable[bindingIndex].remoteNode == NULL)
    return EMBER_ERR_FATAL;

  // remember the message 
  messageSent[messageToken] = message;
  emberHoldMessageBuffer(message);

  commandBuffer[0] = H_TO_N_TX_UNICAST;
  commandBuffer[1] = seqNum++;
  commandBuffer[2] = messageToken++;
  commandBuffer[3] = emHostBindingTable[bindingIndex].cacheIndex;
  commandBuffer[4] = isDatagram;
  commandBuffer[5] = length;

  status = emMasterWrite(message->contents, length, commandBuffer, 6);

  if (status != EMBER_SUCCESS)
    return status;
  else {
    // increase number of outgoing message
    emHostBindingTable[bindingIndex].remoteNode->numOfOutStandingMessage++; 
    return EMBER_SUCCESS;
  }
}

// Allocates a connection for the binding if the connection doesn't exist.
static EmberStatus getBindingConnection(int8u bindingIndex)
{
  Connection *connection = emHostBindingTable[bindingIndex].remoteNode;

  if (connection == NULL) {
    connection = allocateConnection(bindingIndex);
    if (connection == NULL)
      return EMBER_TOO_MANY_CONNECTIONS;
    else
      return EMBER_SUCCESS;
  }
  else
    return EMBER_SUCCESS;
}

// Clears binding and maybe the related remote node infornation.
static void clearBinding(int8u index)
{
  Connection *remoteNode = emHostBindingTable[index].remoteNode;
  boolean lastOwner = TRUE;
  int8u i;
  
  if (remoteNode == NULL)
    return;

  for (i = 0; i < EMBER_HOST_BINDING_TABLE_SIZE; i++) {
    if (i != index && emHostBindingTable[i].remoteNode == remoteNode)
      lastOwner = FALSE;
  }
  if (lastOwner)
    clearConnectionTableEntry(remoteNode, CON_UNUSED);

  emHostBindingTable[index].remoteNode = NULL;
  emHostBindingTable[index].bindingStatus = BINDING_HOST_ONLY;
}

// Finds the interface for the provided terminal.
static EmberInterfaceId *findInterface(EmberTerminal terminal)
{
  int8u endpoint = emberTerminalEndpoint(terminal);
  int8u interfaceNumber = emberTerminalInterface(terminal);
  int8u i;
  EmberEndpoint *endpoints;

  for (endpoints = emberEndpoints, i = 0;
       i < emberEndpointCount;
       i++, endpoints++)
    if (endpoints->number == endpoint) {
      EmberEndpointDescription *description = endpoints->description;
      int8u j;
      for (j = 0; j < EMBER_MAXIMUM_ENDPOINT_INTERFACES; j++) {
        if ((*description)[j].number == interfaceNumber)
          return (*description)[j].interface;
      }
      return NULL;      // endpoint has no such interface
    }
  return NULL;          // no such endpoint
}

// Sends set route message.
static EmberStatus setBindingRoute(int8u bindingTableIndex)
{
  Binding binding = emHostBindingTable[bindingTableIndex];
  Connection *remoteNode = binding.remoteNode;
  int8u data[128]; // rnap message
  int8u routeLength;

  if (remoteNode == NULL)
    return EMBER_INVALID_BINDING_INDEX;
  
  routeLength = remoteNode->routeLength;
  // remote node ID (2) | route length (1) | route (n)
  data[0] = HIGH_BYTE(remoteNode->nodeId);
  data[1] = LOW_BYTE(remoteNode->nodeId);
  data[2] = routeLength;
  memcpy(data + 3, remoteNode->route, remoteNode->routeLength);

  // update routeNew
  emHostBindingTable[bindingTableIndex].remoteNode->routeNew = FALSE;

  return getOrSetValue(H_TO_N_SET_VALUE, 
                       NODE_BINDING_ROUTE, 
                       binding.cacheIndex, 
                       data, 
                       3 + routeLength);
}

// Processes the N_TO_H_COMMAND_ACK message.
static void processN_TO_H_COMMAND_ACK(int8u *message, int8u length)
{
  // cmd (1) | sequenceNumber (1) | originalCmd(1) | status (1) | param (n)
  int8u request = message[2]; // original command
  int8u status = message[3];
  int8u expectedSeqNum;

  DEBUG_OUT(fprintf(stdout, "ack status %d qsize %d retry %d\n", 
                    status, emGetRnapQSize(), retry));
  if (status != EMBER_SUCCESS)
    emberRnapStatusHandler(status);
  
  // check sequence number
  if (sameSequenceNumber(emLastRnapMessageSent, &expectedSeqNum, message[1])) {
    if (status != EMBER_SUCCESS) {
      int8u hostBindingIndex;
      int8u msgId;

      if (isUnicastMessage(emLastRnapMessageSent, &hostBindingIndex, &msgId))
        processMessageSent(hostBindingIndex, msgId, EMBER_RNAP_ACK_WITH_ERR);
      else if (isSetBindingMessage(emLastRnapMessageSent, &hostBindingIndex))
        // set the binding status to host only because the node failed to update
        emHostBindingTable[hostBindingIndex].bindingStatus = BINDING_HOST_ONLY;
    }
    else {
      //if we're getting an ack to our start ack, that means the node is up
      if (request == H_TO_N_NODE_START_ACK)
      {
        nodeStarted = TRUE;
        if (associateOnNodeStart)
        {
          //if the application wants us to now associate with the network we can call
          //ember associate with our saved networkParameters struct
          emberAssociate(&emberNetworkParameters);
        }
      }
      else if (length > 4) { // have payload
        if (request == H_TO_N_GET_VALUE || request == H_TO_N_SET_VALUE) {
          int8u registerId = message[4];
          switch (registerId) {
          case NODE_ENABLE_ROUTING:
            break;
          case NODE_RADIO_CHANNEL:
            radioChannel = message[7];
            break;
          case NODE_RADIO_POWER:
            radioPower = message[7];
            break;
          case NODE_NETWORK_ASSOCIATION:
            if (message[7])
            {
              emLocalPanId = HIGH_LOW_TO_INT(message[8], message[9]);
              emEnableRelay = message[10];
              radioPower = message[11];
              radioChannel = message[12];
              //debug
              fprintf(stdout, "Network Association Successful!\n");
              fprintf(stdout, "emPanId = %d, emRelay = %d, radioPower = %d, radioChannel=%d\n", 
                emLocalPanId, emEnableRelay, 
                radioPower, radioChannel);
            }
          case NODE_BINDING:
          case NODE_CLEAR_BINDING:  {
            // valueKey (1) | bindingTableCacheIndex (1) |length (1)
            //  | host binding index (1) | bindingType (1) | remoteEui64 (8)
            //  | localTerminal (1) | remoteTerminal (1) | isConnected (1)
            //  | connectionData (8)
            EmberStatus status = message[3];
            int8u cacheIndex = message[5];
            int8u length = message[6];
    
            if (length == 19) {
              int8u index = message[7]; // host index
        
              if (registerId == NODE_BINDING) {
                // should be the same
                if (cacheIndex != emHostBindingTable[index].cacheIndex)
                  DEBUG_OUT(fprintf(stdout, "assert: cache %x table cache %x host %x\n", 
                                    cacheIndex,
                                    emHostBindingTable[index].cacheIndex, 
                                    index));
                if (cacheIndex != emHostBindingTable[index].cacheIndex
                    || emHostBindingTable[index].remoteNode == NULL)
                  emberRnapStatusHandler(EMBER_RNAP_MISMATCH);
    
                if (message[19] == TRUE) // isConnected
                  emHostBindingTable[index].remoteNode->state = CON_ESTABLISHED;
                else
                  emHostBindingTable[index].remoteNode->state = CON_LISTEN;
    
                emHostBindingTable[index].remoteIndex = message[20];
                emHostBindingTable[index].remoteNode->localId = message[21];
                emHostBindingTable[index].remoteNode->remoteId = message[22];
                emHostBindingTable[index].remoteNode->inRetryId = message[23];
                emHostBindingTable[index].remoteNode->inRetryMask = message[24];
                emHostBindingTable[index].remoteNode->outRetryId = message[25];
                nodeHostBindingIndexTable[cacheIndex] = index;    
                emHostBindingTable[index].bindingStatus = BINDING_NODE_ACKED;
              }
              else if (registerId == NODE_CLEAR_BINDING
                       && status == EMBER_SUCCESS) {
                nodeHostBindingIndexTable[cacheIndex] = EMBER_NULL_BINDING;    
              }
            }
            break;
          }
          case NODE_BINDING_ROUTE: {
            //  valueKey (1) | bindingTableCacheIndex (1) | length (1)
            //  | remoteNode (2) | routeLength (1) | route (n) 
            int16u senderId = HIGH_LOW_TO_INT(message[7], message[8]);
            int8u connectionIndex = emFindNodeConnectionIndex(senderId);
            if (connectionIndex != NULL_BYTE) {
              emHostConnectionTable[connectionIndex].routeLength = message[9];
              memcpy(emHostConnectionTable[connectionIndex].route, 
                     message + 10, 
                     message[9]);
            }
            break;
          }    
          } // register id
        } // check if request is get or set value
      } // check length > 4
    } // check status

    // set need ack
    emSetNeedAck(FALSE);
    // clear retry
    retry = 0;
    // send one more message from the queue 
    if (emGetRnapQSize() > 0) {
      if (emLastRnapMessageSent != NULL)
        free(emLastRnapMessageSent);
      emLastRnapMessageSent = emRemoveRnapQMessage();  
      status = emMasterReallyWrite(emLastRnapMessageSent);
      if (status != EMBER_SUCCESS) {
        emberRnapStatusHandler(status);
        return;
      }
    }
  } // check for sequence number
}

// Processes N_TO_H_RX_UNICAST message.
static void processN_TO_H_RX_UNICAST(int8u *message)
{
  // cmd (1) | sequenceNumber (1) | messageType (1) | destination (1) | 
  // | sendingTerminal (1) | messageId (1) | sender (2) |routeLength (1) 
  // | route (n) | routeCost (1) | payloadLength (1) | payload (n)
  EmberMessageBuffer buf; // payload
  int8u bindingIndex; // binding table index
  int8u messageId = message[5];
  int8u type = message[2];

  DEBUG_OUT(fprintf(stdout, "in host-api rx unicast, sender %x %x\n", 
                    message[6], message[7]));
  currentSendingTerminal = message[4];
  emCurrentSender = HIGH_LOW_TO_INT(message[6], message[7]);
  emCurrentRouteLength = message[8];
  memcpy(emCurrentRoute, message + 9, emCurrentRouteLength);
  currentCost = message[emCurrentRouteLength + 9];
  currentMessageId = message[5];

  // fill the buffer with payload
  buf = emberFillLinkedBuffers(message + emCurrentRouteLength + 11,
                               message[emCurrentRouteLength + 10]);
  DEBUG_OUT(fprintf(stdout, "fillLinkedBuffers Success\n"));

  // get binding index
  // destination: For datagrams this is the destination terminal.
  //              For all other messages it is the destination binding index.
  if (type == EMBER_INCOMING_DATAGRAM) {
    bindingIndex = emFindIncomingDatagramBinding(emCurrentSender, 
                                                 message[3], 
                                                 currentSendingTerminal);
    DEBUG_OUT(fprintf(stdout, "find binding index %d for incoming datagram\n",
                      bindingIndex));
  }
  else
    bindingIndex = message[3];

  // update connection route and cost
  if (bindingIndex < EMBER_HOST_BINDING_TABLE_SIZE) {
    updateRouteCost(bindingIndex, FALSE);
    DEBUG_OUT(fprintf(stdout, "updateRoutCost Success\n"));
  }

  currentMessageType = type;
  switch (type) {
  case EMBER_INCOMING_DATAGRAM:
    replySent = FALSE;
    emberIncomingMessageHandler(EMBER_INCOMING_DATAGRAM, 
                                message[3], // destination terminal
                                bindingIndex, 
                                buf); // message
    DEBUG_OUT(fprintf(stdout, "incomingMessageHandler\n"));
    if (!replySent)
      emberSendReply(EMBER_NULL_MESSAGE_BUFFER);
    currentMessageType = -1;
    emberReleaseMessageBuffer(buf);
    break;
   
  case EMBER_INCOMING_SEQUENCED: {
    Connection *connection = emHostBindingTable[bindingIndex].remoteNode;
    int8u base;
      
    if (connection == NULL
        || emHostBindingTable[bindingIndex].bindingEntry == NULL)
      emberRnapStatusHandler(EMBER_RNAP_MISMATCH);

    base = connection->inRetryId;

    if (messageId != base) {
      if (PACKET_ID_IN_WINDOW(base, messageId))
        messageReceived[messageId] = buf;
      return;
    }
    else {
      emberIncomingMessageHandler(type,
                                  (emHostBindingTable[bindingIndex].bindingEntry)->local, // local terminal
                                  bindingIndex,  // binding index
                                  buf);          // message  
      emberReleaseMessageBuffer(buf);

      // check the following packet
      while (PACKET_ID_IN_WINDOW(base, ++messageId) 
             && messageReceived[messageId] != NULL) {
        emberIncomingMessageHandler(type,
                                    (emHostBindingTable[bindingIndex].bindingEntry)->local, // local terminal
                                    bindingIndex,  // binding index
                                    messageReceived[messageId]); // message
        emberReleaseMessageBuffer(messageReceived[messageId]);
      }

      connection->inRetryId = messageId;
    }     
    break;
  }   
  case EMBER_INCOMING_DATAGRAM_REPLY:
    emberIncomingMessageHandler(type,
				  (emHostBindingTable[bindingIndex].bindingEntry)->local, // local terminal
				  bindingIndex,  // binding index
				  buf);          // message
    emberReleaseMessageBuffer(buf);
    break;
  default:
    fprintf(stdout, "Error: Host API RX unicast message type we don't recognize\n");
    emberReleaseMessageBuffer(buf);
    break;
  }
}

// Processes N_TO_H_RX_MULTICAST message.
static void processN_TO_H_RX_MULTICAST(int8u *message)
{
  // cmd (1) | sequenceNumber (1) | sender (2) | cost (1) | multicastId (2)
  // | payload_length (1) | payload (n)
  EmberMessageBuffer buf; // payload
  int8u bindingIndex; // binding table index
  int8u terminal;
  int16u multicastId =  HIGH_LOW_TO_INT(message[5], message[6]);

  currentMessageType = EMBER_INCOMING_MULTICAST;
  emCurrentSender = HIGH_LOW_TO_INT(message[2], message[3]);
  currentCost = message[4]; // cost

  // fill the buffer with payload
  buf = emberFillLinkedBuffers(message + 8, message[7]);
  // find binding index
  bindingIndex = emFindMulticastBinding(0, multicastId);
    
  if (bindingIndex < EMBER_HOST_BINDING_TABLE_SIZE) {
    updateRouteCost(bindingIndex, TRUE); // update cost only
    terminal = emHostBindingTable[bindingIndex].bindingEntry->local;
    while (bindingIndex < EMBER_HOST_BINDING_TABLE_SIZE) {
      int8u nextIndex = emFindMulticastBinding(bindingIndex + 1, multicastId);
      emberIncomingMessageHandler(nextIndex == EMBER_NULL_BINDING
                                  ? EMBER_INCOMING_MULTICAST
                                  : EMBER_INCOMING_SHARED_MULTICAST,
                                  terminal,
                                  bindingIndex,
                                  buf);
      bindingIndex = nextIndex;
    }
  }
  emberReleaseMessageBuffer(buf);
}

// Processes N_TO_H_CONNECTION_STATUS message.
static void processN_TO_H_CONNECTION_STATUS(int8u *message)
{
  // cmd (1) | sequenceNumber (1) | bindingIndex (1) 
  // | status (1) | messageId (1)
  Connection *connection = emHostBindingTable[message[2]].remoteNode;

  if (connection == NULL)
    emberRnapStatusHandler(EMBER_RNAP_MISMATCH);

  switch (message[3]) {
  case EMBER_CONNECTION_OPEN:
    connection->inRetryId = message[4];
    connection->state = CON_ESTABLISHED;
    break;    
  case EMBER_CONNECTION_OPEN_FAILED:
  case EMBER_CONNECTION_FAILED:
  case EMBER_CONNECTION_CLOSED:
    connection->state = CON_LISTEN;
    break;
  case EMBER_CONNECTION_CLOSING:
    connection->state = CON_WAIT;
    break;
  }
  emberConnectionStatusHandler(message[2], message[3]);
}

// Checks the index range, updates the outstanding message number, and passes 
// message to user.
static void processMessageSent(int8u index, 
                               int8u messageId, 
                               EmberStatus ackStatus)
{
  if (index < EMBER_HOST_BINDING_TABLE_SIZE 
      && emHostBindingTable[index].remoteNode != NULL) {
    if (emHostBindingTable[index].remoteNode->numOfOutStandingMessage > 0)
      emHostBindingTable[index].remoteNode->numOfOutStandingMessage--;
  }
  else
    emberRnapStatusHandler(EMBER_RNAP_MISMATCH);

  if (index < EMBER_HOST_BINDING_TABLE_SIZE
      && messageSent[messageId] != EMBER_NULL_MESSAGE_BUFFER) {
    // pass the message to user
    emberMessageSent(index, messageSent[messageId], ackStatus); 
    // Logic to make sure it is not locked into a route
    // if the status is not success, increase the failure counter
    // if failed more than 2 times, update the cost to the max value
    if (ackStatus != EMBER_SUCCESS) {
      emHostBindingTable[index].failedTimes++;
      if (emHostBindingTable[index].failedTimes > 2) {
        if (emHostBindingTable[index].remoteNode != NULL)
          (emHostBindingTable[index].remoteNode)->cost = 0xff;
        emHostBindingTable[index].failedTimes = 0;
      }
    }
    else
      emHostBindingTable[index].failedTimes = 0;      
  }
  // release the messageSent buffer
  if (messageSent[messageId] != NULL)
    emberReleaseMessageBuffer(messageSent[messageId]);
}

// Processes N_TO_H_MESSAGE_SENT_STATUS message
static void processN_TO_H_MESSAGE_SENT_STATUS(int8u *message)
{
  // cmd (1) | sequenceNumber (1) | bindingIndex (1) 
  // | message token (1) | status (1) 
  // decrease the outstanding message
  int8u messageId = message[3];
  int8u index = message[2];
  processMessageSent(index, messageId, message[4]);
}

// Processes incoming messagess from the node that the host talks to.
void emProcessIncomingMessage(int8u *message, int8u length)
{
  int8u command = message[0];
  int8u i;

  //debug
  DEBUG_OUT(fprintf(stdout, "length=%d ", length));
  DEBUG_OUT(fprintf(stdout, "read: "));
  for (i = 0; i < length; i++)
    DEBUG_OUT(fprintf(stdout, "%d ", message[i])); 
  DEBUG_OUT(fprintf(stdout, "\n"));
 
  switch(command) {
  case N_TO_H_COMMAND_ACK: {
    processN_TO_H_COMMAND_ACK(message, length);

    //debug
    //fprintf(stdout, "got cmd ack\n");

    break; // N_TO_H_COMMAND_ACK
  }
  case N_TO_H_READ_ERROR:
    // cmd | seq | status
    // report the serial error to user
    retryLastRnapMessage();
    emberRnapStatusHandler(EMBER_RNAP_SERIAL_ERROR);
    break;

  case N_TO_H_NODE_START:
    // get node variables
    memcpy(emLocalEui64, message + 3, 8);
    radioType = message[11];
    radioChannel = message[12];
    radioPower = message[13];
    nodeBindingTableSize = message[14];
    // update the tables, release sent and received messages
    // and send RNAP ack message.
    acknowledgeNodeReset();
    break;
  case N_TO_H_RX_UNICAST:
    processN_TO_H_RX_UNICAST(message);
    break; // N_TO_H_RX_UNICAST
  case N_TO_H_RX_MULTICAST:
    processN_TO_H_RX_MULTICAST(message);
    break;
  case N_TO_H_MESSAGE_SENT_STATUS:
    processN_TO_H_MESSAGE_SENT_STATUS(message);
    break;
  case N_TO_H_CONNECTION_STATUS:
    processN_TO_H_CONNECTION_STATUS(message);
    break;  
  case N_TO_H_STACK_STATUS:
    // cmd (1) | sequenceNumber (1) | bindingIndex (1) | status (1) 
    stackStatus = message[2];
    emberStackStatusHandler(message[2]);
    break;
  case N_TO_H_NEED_BINDING: {
    // cmd (1) | sequenceNumber (1) | bindingIndex (1) 
    int8u neededBindingIndex = message[2];
    EmberBindingTableEntry *value = 
      emHostBindingTable[neededBindingIndex].bindingEntry;
    emberSetBinding(neededBindingIndex, value);
    setBindingRoute(neededBindingIndex);
    break;
  }
  default:
    break;
  }
}

int16u microGetSystemTick() {
#ifdef EMBER_TEST
  return simMicroGetSystemTick();
#else
  gettimeofday(&tv, &tz);
  return (int16u)(tv.tv_sec*1000 + tv.tv_usec/1000);
#endif
}

// Sends SIGUSR1 to the signal thread every TIMER_PERIOD.
static void *timer_thread (void *arg)
{
  struct timeval tv;

  while (TRUE) {
    // sleep a while
    tv.tv_sec = 0;
    tv.tv_usec = TIMER_PERIOD * 1000; // microsecond
    if (select(0, NULL, NULL, NULL, &tv) == 0)
      // send signal to signal handling thread
      pthread_kill(sig_thr_id, SIGUSR1);
  }
}

static void retryLastRnapMessage(void) {
    EmberStatus status;

  // retry one more time
  retry++;
  if (emLastRnapMessageSent != NULL) {
    status = emMasterReallyWrite(emLastRnapMessageSent);
    if (status != EMBER_SUCCESS)
      emberRnapStatusHandler(status);
      DEBUG_OUT(fprintf(stdout, "Retry the same message, %d\n", retry));
  }        
}


// Resends the RNAP message if the ack is not received within MAX_WAIT_TIME
// and the retry times is less than MAX_RETRY.
// If MAX_RETRY is reached and the ack is not received, informs the user
// about the failure and moves on to the next RNAP message in the queue.
static void *signal_thread (void *arg)
{
  int       sig_caught;    // signal caught
  EmberStatus status;

  while (TRUE) {
    if (sigwait (&signal_mask, &sig_caught) != 0)
      exit(0);

    switch (sig_caught) {
    case SIGUSR1:     // timer signal
      // check the wait time for ack
      DEBUG_OUT(fprintf(stdout, "Timer Thread Interrupt Once\n"));
      if (emWaitAckTooLong(TIMER_PERIOD, MAX_WAIT_TIME)) {
        if (emNeedAck()) {
          if (retry > MAX_RETRY) {
            int8u hostBindingIndex, msgId;

            // pass the status to user
            if (emLastRnapMessageSent != NULL)
              emberRnapMessageSent(emLastRnapMessageSent, EMBER_DELIVERY_FAILED);

            // call emberMessageSent if it is unicast message
            if (isUnicastMessage(emLastRnapMessageSent, 
                                 &hostBindingIndex, 
                                 &msgId))
              processMessageSent(hostBindingIndex, 
                                 msgId, 
                                 EMBER_DELIVERY_FAILED);

            // clear retry
            retry = 0;

            // discard the sent message, move on to the next message
            // send one more message from the queue
            if (emGetRnapQSize() > 0) {
              DEBUG_OUT(fprintf(stdout, "Wait ACK too long, send another message in the queue\n"));
              if (emLastRnapMessageSent != NULL)
                free(emLastRnapMessageSent);
              emLastRnapMessageSent = emRemoveRnapQMessage();
              status = emMasterReallyWrite(emLastRnapMessageSent);
              if (status != EMBER_SUCCESS)
                emberRnapStatusHandler(status);
            }
            else {
              DEBUG_OUT(fprintf(stdout, "Wait ACK too long, and no message in the Queue\n"));
              emSetNeedAck(FALSE); // no need to wait
            }
          }
          else {
            retryLastRnapMessage();
          }
        }
        else
          DEBUG_OUT(fprintf(stdout, "No need to be acked\n"));
      }
      break;
    default:          // should normally not happen
      fprintf(stdout, "\nUnexpected signal %d\n", sig_caught);
      break;
    }
  }
}

// Checks the sequence number in the ack message against the expected 
// sequence number.
// Returns TRUE if they are the same, FALSE otherwise.
static boolean sameSequenceNumber(EmberMessageBuffer buf, 
                                  int8u *expectedNum, 
                                  int8u ackedNum)
{
  if (buf == NULL)
    return FALSE;  
  else {
    // H_TO_N_HOST_HAS_RESET and H_TO_N_BOOTLOAD has no associated ack message
    if (buf->contents[2] == H_TO_N_HOST_HAS_RESET
        || buf->contents[2] == H_TO_N_BOOTLOAD)
      *expectedNum = buf->contents[3] + 1;
    else
      *expectedNum = buf->contents[3];
    if (*expectedNum == ackedNum)
      return TRUE;
    else 
      return FALSE;
  }
}

// Sets binding on node.
static EmberStatus setCacheBinding(int8u bindingTableIndex)
{
  Binding *entry = &emHostBindingTable[bindingTableIndex];
  EmberStatus status;

  // if the cache index is invalid we need to set it
  if (entry->cacheIndex >= nodeBindingTableSize
      || entry->bindingStatus != BINDING_NODE_ACKED) {

    //we only want a new cache index if the old one is invalid
    if (entry->cacheIndex >= nodeBindingTableSize)
    {
      int8u newCacheIndex = getNewCacheBindingIndex(bindingTableIndex);
      if (newCacheIndex == EMBER_NULL_BINDING)
        return EMBER_MAX_MESSAGE_LIMIT_REACHED;
      updateCacheBindingIndex(bindingTableIndex, newCacheIndex);
    }

    // either way we want to send a new set binding message just in case
    //there is a real problem with the setbinding status that isn't just 
    //related to having a bunch of messages in the rnap message queue
    status = sendSetBindingMessage(bindingTableIndex, entry->bindingEntry);
    if (status != EMBER_SUCCESS)
      return status;

    // send set route message
    if (entry->remoteNode != NULL && 
        ((entry->remoteNode)->cost != 0 || (entry->remoteNode)->cost != 0xFF)) {
      status = setBindingRoute(bindingTableIndex);
      if (status != EMBER_SUCCESS)
        return status;
    }
  }
  return EMBER_SUCCESS;
}

// Checks to see if the rnap message is sending unicast message.
// Format of the message in the buffer is:
//   [ length cmd seq messageToken cacheIndex ...
static boolean isUnicastMessage(EmberMessageBuffer buf,
                                int8u *hostBindingIndex,
                                int8u *messageId)
{
  if (buf == NULL
      || buf->contents[2] != H_TO_N_TX_UNICAST
      || buf->contents[5] > nodeBindingTableSize)
    return FALSE;
  else {
    *messageId = buf->contents[4];
    *hostBindingIndex = nodeHostBindingIndexTable[buf->contents[5]];
    return TRUE;
  }
}

// Checka to see if the rnap message is for setting binding  
// Format of the message in the buffer is:
//   [ length cmd seq setKey cacheIndex ...
static boolean isSetBindingMessage(EmberMessageBuffer buf,
                                   int8u *hostBindingIndex)
{
  if (buf == NULL
      && !(buf->contents[2] == H_TO_N_SET_VALUE
           && (buf->contents[4] == NODE_BINDING)))
    return FALSE;
  else {
    *hostBindingIndex = nodeHostBindingIndexTable[buf->contents[5]];
    return TRUE;
  }
}

// Checka to see if the rnap message is for open connection.
// Format of the message in the buffer is:
//   [ length cmd seq cacheIndex...
static boolean isOpenConnectionMessage(EmberMessageBuffer buf,
                                       int8u *hostBindingIndex)
{
  if (buf == NULL
      || buf->contents[2] != H_TO_N_OPEN_CONNECTION
      || buf->contents[4] > nodeBindingTableSize)
    return FALSE;
  else {
    *hostBindingIndex = nodeHostBindingIndexTable[buf->contents[4]];
    return TRUE;
  }
}

// Checks to see if binding is needed to be set on node.
boolean emNeedSetBindingOnNode(EmberMessageBuffer buf)
{
  int8u hostIndex, messageId; 
   
  if (isUnicastMessage(buf, &hostIndex, &messageId)
      || isOpenConnectionMessage(buf, &hostIndex))
    return (emHostBindingTable[hostIndex].bindingStatus != BINDING_NODE_ACKED);  
  else 
    return FALSE;
}

