/** @file host-api.h
 * @brief  Host-side APIs for use in RNAPII applications.
 * 
 * Use these APIs in conjunction with the EmberNet stack APIs to write a
 * host program for an RNAP II application. See the <i> EmberNet Stack 
 * Interface Guide</i> for information on the stack APIs. In particular,
 * see stack/include/error-def.h in the EmberNet stack API for definitions
 * of all EmberStatus return values.
 * 
 * Many declarations in host-api.h have identical counterparts in 
 * stack/include/ember.h. For documentation of these declarations, see
 * the <i> EmberNet Stack Interface Guide</i>. If you need to determine which 
 * declarations are present in both files, the comments in host-api.h clearly 
 * identify the common declarations.
 *
 * Because the EmberNet stack API supports stack functionality that is not 
 * supported by the host API, some ember.h functionality is different in the
 * host API, and some is not supported at all. See below for descriptions of these
 * differences.
 *
 * <b>EmberNet Stack Functions That Have Different Behavior in the Host API</b>
 * - emberSetBinding()<br>
 *      In the host API, only \b multicast bindings may be set using 
 *      this function. Unicast bindings must be set using emberSetReplyBinding().
 *      Attempts to set a unicast binding with this function will return an error
 *      in the host API.
 *
 * <b>EmberNet Stack Functions Not Supported by the Host API</b>
 * - emberEndpointFoundHandler()<br>
 *      RNAP II does not support endpoint searching, so applications built
 *      with the host API do not need to implement it.
 * - typedef int16u EmberPanId<br>
 *      The host API does not use or support the EmberPanId type.
 * - typedef int8u ResetType<br>
 *      RNAP II does not need the reset type because initialization of the host API 
 *      indirectly causes a reset of the gateway node, which is responsible for 
 *      indicating the reset type to the stack.
 * - emberInit()<br>
 *      The host API is initialized via emberHostInit(), which sets up the connection 
 *      to the gateway node that in turn is responsible for initializing the stack 
 *      through emberInit(). 
 * - emberTick()<br>
 *      The host application never calls embertick() because it is not responsible 
 *      for driving cleanup on the stack.
 * - emberGetPanId(), emberSetPanId()<br>
 *      The host API does not provide access to the gateway node's PAN ID.
 * - emberCancelMessage()<br>
 *      The host API does not allow the application to cancel a message that has 
 *      been sent.
 * - emberSleep(), emberWakeUp()<br>
 *      The host API does not provide direct control of the gateway node's sleep 
 *      and wake functionality.
 * - emberIncomingBeaconHandler(), emberBeaconTrack(), 
 *   emberGetLastBeaconReceiveTime(), EmberStatus emberTransmitBeacons(), 
 *   emberSetOutgoingBeaconNotifyCounter()<BR>
 *      The host API does not support beacons.
 *
  * <!--This RNAP Host API library is subject to the Ember Corporation-->
 * <!--Development Kit Software License Agreement delivered with this-->
 * <!--file to licensee.-->
 * <!--author:  Guohui Liu (guohui@ember.com)-->
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.                *80*-->
 */
#ifndef LIB_RNAP_HOST_API_H
#define LIB_RNAP_HOST_API_H

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** These public Ember types are documented in ember.h. */
typedef unsigned char  boolean;
typedef unsigned char  int8u;
typedef signed   char  int8s;
typedef unsigned short int16u;
typedef signed   short int16s;
typedef unsigned int   int32u;
typedef signed   int   int32s;

#define PGM  
#define PGM_P   const char *
#define PGM_PU  const unsigned char *
#define PGM_NO_CONST

#include <string.h>
#define MEMSET(d,v,l)  memset(d,v,l)
#define MEMCOPY(d,s,l) memcpy(d,s,l)
#define MEMCOMPARE(s0,s1,l) memcmp(s0, s1, l)
#define MEMPGMCOMPARE(s0,s1,l) memcmp(s0, s1, l)

/** Documented in ember.h. */
#define TRUE  1
#define FALSE 0

/** Documented in ember.h. */
#define EUI64_SIZE 8

#define NULL_BYTE 0xFF 

#ifdef DEBUG_HOST_API
#define DEBUG_OUT(x)  x;
#else
#define DEBUG_OUT(x)
#endif
#endif //DOXYGEN_SHOULD_SKIP_THIS

//sizes for binding and connection tables in the host api
#define EMBER_HOST_BINDING_TABLE_SIZE 254  // 0xFF and 0xFE is reserved
#define EMBER_HOST_CONNECTION_TABLE_SIZE EMBER_HOST_BINDING_TABLE_SIZE

//debug reporting stubs for ported code from the node
#define emberDebugReportOff() (FALSE)
#define emberDebugReportRestore(state) do {} while(FALSE)

/* @description The maximum size of an RNAP packet.
 */
#define MAX_RNAP_PACKET_SIZE 128

/* @description The NULL message buffer.
 */
#define EMBER_NULL_MESSAGE_BUFFER NULL

/* @brief A message buffer.
 */
typedef struct {
  /** Number of references. */
  int8u count;
  /**  Message length. */
  int8u length;
  /**  Message contents. */  
  int8u contents[MAX_RNAP_PACKET_SIZE];
} EmberMessageBufferS;

typedef EmberMessageBufferS *EmberMessageBuffer;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
//Return types for Ember functions 
/** Documented in ember.h. */
typedef int8u EmberStatus;

/** Documented in ember.h. */
typedef int8u EmberCost;

/** Documented in ember.h. */
typedef int8u EmberEUI64[EUI64_SIZE];

/** Documented in ember.h. */
extern EmberEUI64 emLocalEui64;

/** Documented in ember.h. */
#define emberGetEui64() (emLocalEui64)

/** Documented in ember.h. */
#define emberIsLocalEui64(eui64)                           \
(MEMCOMPARE(eui64, emLocalEui64, EUI64_SIZE) == 0)

/** Documented in ember.h. */
enum {
  ENDPOINT_ZERO_INTERFACE = 0,
  AD_HOC_INTERFACE        = 1,             // for use in closed networks
  ZIGBEE_INTERFACE        = 2,
  EUI64_INTERFACE         = 3
  // more as needed
};

/** Documented in ember.h. */
typedef int8u EmberInterfaceDomain;

/** Documented in ember.h. */
enum {
  SYMMETRIC_ENDPOINT          = 0,
  SERVER_ENDPOINT             = 1,
  CLIENT_ENDPOINT             = 2,
  AGGREGATION_SERVER_ENDPOINT = 3,
  AGGREGATION_CLIENT_ENDPOINT = 4
};

/** Documented in ember.h. */
typedef int8u EmberEndpointType;

/** Documented in ember.h. */
#define EMBER_MAXIMUM_INTERFACE_ID_SPECIFICATION 9

/** Documented in ember.h. */
#define EMBER_MAXIMUM_INTERFACE_ID_SIZE                 \
  (2 + EMBER_MAXIMUM_INTERFACE_ID_SPECIFICATION)

/** Documented in ember.h. */
typedef struct {
  EmberInterfaceDomain domain;
  EmberEndpointType type;
  int8u specification[EMBER_MAXIMUM_INTERFACE_ID_SPECIFICATION];
} EmberInterfaceId;

/** Documented in ember.h. */
#define EMBER_MAXIMUM_ENDPOINT_INTERFACES 8

/** Documented in ember.h. */
typedef struct {
  int8u number;
  EmberInterfaceId *interface;
} EmberEndpointDescription[EMBER_MAXIMUM_ENDPOINT_INTERFACES];

/** Documented in ember.h. */
typedef struct {
  int8u number;
  EmberEndpointDescription *description;
} EmberEndpoint;

/** Used to indicate the host's method of connecting to the gateway. */
typedef enum {
  EMBER_HOST_MODE_SERIAL,
  EMBER_HOST_MODE_SOCKET,
  EMBER_HOST_MODE_STDIO
} GatewayConnectionMode;


//----------------------------------------------------------------
// Binding table 

/** Documented in ember.h. */
enum {
  EMBER_UNUSED_BINDING    = 0,
  EMBER_UNICAST_BINDING   = 1,
  EMBER_MULTICAST_BINDING = 2
};

/** Documented in ember.h. */
typedef int8u EmberBindingType;

#define EMBER_NULL_BINDING 0xFF

/** Documented in ember.h. */
typedef int8u EmberTerminal;

/** These terminal defines are documented in ember.h. */
#define emberTerminalEndpoint(terminal)  (((terminal) >> 3) & 0x1F)
#define emberTerminalInterface(terminal) ((terminal) & 0x7)
#define emberMakeTerminal(endpoint, interface)  \
 ((((endpoint) & 0x1F) << 3) | ((interface) & 0x7))

/** Documented in ember.h. */
typedef struct {
  EmberBindingType type;
  int8u identifier[8];
  EmberTerminal local;
  EmberTerminal remote;
} EmberBindingTableEntry;

//Ember IDs
typedef int16u EmberNodeId;
/** @description 802.15.4 PAN ID. 
 */
typedef int16u EmberPanId;

#define NULL_NODE_ID 0xFFFF

enum {
  CON_UNUSED,             // struct not in use
  CON_LISTEN,             // has an index but there has been no open connection
  CON_WAIT,               // wait for the open/close connection response
  CON_ESTABLISHED         // connected and ready to send messages
};

/** @description The maximum number of active packets 
 * in one direction over one connection. */
#define PACKET_WINDOW_SIZE 8
#define PACKET_ID_IN_WINDOW(base, id) \
((((id) - (base)) & 0xFF) <= PACKET_WINDOW_SIZE)
typedef int8u PacketId;

typedef struct {
  EmberNodeId nodeId;    // 16 bits node id
  int8u routeLength;     // route length
  int8u route[MAX_RNAP_PACKET_SIZE];
  boolean routeNew;
  int8u cost;
  int8u numOfOutStandingMessage;

  // For connections and sequenced messages
  int8u state;           // state of the connection
  int8u localId;         // local connection ID
  int8u remoteId;        // remote connection ID

  // Incoming sequenced messages
  PacketId inRetryId;    // lowest ID not yet delivered to application
  int8u inRetryMask;     // mask of messages that have been received

  // Outgoing sequenced messages
  PacketId outRetryId;   // next ID to use
} Connection;

enum {
  BINDING_HOST_ONLY,
  BINDING_SENT_TO_NODE,
  BINDING_NODE_ACKED
};

typedef struct {
  int8u cacheIndex;     // binding index on the node
  EmberBindingTableEntry *bindingEntry;
  Connection *remoteNode;
  int8u remoteIndex;    // remote node binding index
  int8u bindingStatus;
  int8u failedTimes;
} Binding;

/** @description Holds network parameters. Documented in 
 * stack/include/ember-common.h.
 */
typedef struct {
  
  /** The networks's PAN identifier.
   */
  int16u  panId;
  
  /** TRUE if this node will relay messages for other nodes.
   */
  boolean enableRelay;
  
  /** A power setting, in dBm. For the available settings,
   *  see the technical specification for the RF communication module 
   *  in your Developer Kit.
   */
  int8s   radioTxPower;

  /** A radio channel. Be sure to specify a channel supported by the radio.
   * For a list of available channels, see the technical specification for the
   * RF communication module in your Developer Kit.
   */
  int8u   radioChannel;

} EmberNetworkParameters;


/** Documented in ember.h. */
// The application must provide definitions for the following two variables:
extern int8u emberEndpointCount;
extern EmberEndpoint emberEndpoints[];
#endif //DOXYGEN_SHOULD_SKIP_THIS

/** @description The behavior of this function in the host API differs significantly
 * from its behavior in the EmberNet stack API. In the host API, only multicast 
 * bindings may be set using this function. Unicast bindings must be set using 
 * emberSetReplyBinding(). Attempts to set a unicast binding with this function 
 * will return an error in the host API. For more information on the differences 
 * between the EmberNet stack API and the host API, see the comments at the top 
 * of this file. See stack/include/ember.h for further documentation of this 
 * function's parameters.
 */ 
EmberStatus emberSetBinding(int8u index, EmberBindingTableEntry *value);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// The following binding functions are documented in ember.h */
EmberStatus emberGetBinding(int8u index, EmberBindingTableEntry *result);
EmberStatus emberDeleteBinding(int8u index);
boolean emberBindingIsActive(int8u index);
EmberStatus emberClearBindingTable(void);

/** Documented in ember.h. */
// Byte manipulation.  Similarly, everyone defines this.
#define LOW_BYTE(n)                     (int8u)((n) & 0xFF)
#define HIGH_BYTE(n)                    (int8u)(LOW_BYTE((n) >> 8))
#define HIGH_LOW_TO_INT(high, low)      (((high) << 8) + (low))

/** @description Radio type.*/
enum {  
  /** Bit 0. */
  RADIO_868 = 1, 
  /** Bit 1. */
  RADIO_915 = 2,
  /** Bit 3. */
  RADIO_2420 = 8, 
  RADIO_NULL = 0xFF,
};
#endif //DOXYGEN_SHOULD_SKIP_THIS

/** @description Initializes the host. This function can be used to initialize
 * the host API for interaction with either a socket or a serial port. When 
 * initializing the host API for a socket, an address or hostname and port 
 * should be passed to this function. When initializing the host API for
 * a serial connection, the file location of the serial port (that is,
 * /dev/ttyS0) should be passed in as the name, but the port value is ignored 
 * so the value can be 0.
 *
 * <b>Configuring the serial port:</b>
 * This code was tested on a Linux machine running RedHat Linux. Serial 
 * configurations may vary between machines and operating systems. As 
 * a guideline, the high-level serial port configuration used for testing was reported 
 * by stty (stty -F /dev/ttyS0) as follows:
 *
 * <code>
 * speed 19200 baud; line = 0;                                                  \n
 * intr = \<undef\>; quit = \<undef\>; erase = \<undef\>; kill = \<undef\>;     \n
 * eof = \<undef\>; start = \<undef\>; stop = \<undef\>; susp = \<undef\>;      \n
 * werase = \<undef\>; lnext = \<undef\>; flush = \<undef\>; min = 1; time = 0; \n
 * -brkint -icrnl -imaxbel                                                      \n
 * -opost                                                                       \n
 * -icanon -echo                                                                \n
 * </code>
 *
 * The low-level serial port configuration was reported by setserial 
 * (setserial -a /dev/ttyS0) as follows:
 *

 * <code>
 * /dev/ttyS0, Line 0, UART: 16550A, Port: 0x03f8, IRQ: 4                       \n
 *      Baud_base: 115200, close_delay: 256, divisor: 0                         \n
 *      closing_wait: 15360                                                     \n
 *      Flags: spd_normal skip_test                                             \n
 * </code>
 *
 * Unfortunately, these settings do not persist after reboot. In order to persist
 * the settings chosen for a com port, you need to add the configuration commands to 
 * the file /etc/rc.local
 *
 * On RedHat Linux, add the following if you are using the default com port ttyS0(COM1).
 * For other ports, adjust as you see fit:
 *
 *  <code>
 *  setserial /dev/ttyS0 baud_base 115200 close_delay 256 divisor 0 closing_wait 15360     \n
 *  setserial /dev/ttyS0 spd_normal skip_test                                              \n
 *  stty -F /dev/ttyS0 speed 19200 baud                                                    \n
 *  stty -F /dev/ttyS0 -brkint -icrnl -imaxbel -opost -icanon -echo                        \n
 *  stty -F /dev/ttyS0 intr '' quit '' erase '' kill '' eof '' start '' stop '' susp\      \n
 *  '' werase '' lnext '' flush ''                                                         \n
 *  </code>
 *
 * @param name: For a socket connection, a pointer to the name of the host 
 * that the node is on. For a serial port connection to the gateway node
 * (that is, /dev/ttyS0), the file location.
 *               
 * @param port: The port on the Developer Kit carrier board to which
 * the RNAP host is connected. This value is ignored in the case of a gateway node
 *  connected over a serial port and can be set to 0.
 *
 * @param mode: The method of connecting to the gateway, either socket or serial.
 *
 * @return An EmberStatus value indicating the success or failure of the command. 
 */
EmberStatus emberHostInit(int8u *name, int16u port, GatewayConnectionMode mode);

/** @description Indicates that the host processor has reset.
 *
 * @return An EmberStatus value indicating the success or failure of the command. 
 */
EmberStatus emberHostHasReset(void);

/** @description Instructs a node to reset.
 *
 * @return An EmberStatus value indicating the success or failure of the command. 
 */
EmberStatus emberResetNode(void);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** Documented in ember.h. */
EmberStatus emberEnableRouting(boolean enable);

/** Documented in ember.h. */
EmberStatus emberSetRadioChannel(int8u channel);

/** Documented in ember.h. */
int8u emberGetRadioChannel(void);

/** Documented in ember.h. */
EmberStatus emberSetRadioPower(int8s power);

/** Documented in ember.h. */
int8s emberGetRadioPower(void);

/** Documented in ember.h. */
EmberStatus emberGetCost(int8u index, EmberCost *returnLoc);

/** Documented in ember.h. */
int8u emberInterfaceIdSize(EmberInterfaceId *interfaceId);

/** @description Puts the node into bootloader mode so that it can 
 * receive a new software image over the wired connection.
 *
 * @return An EmberStatus value indicating the success or
 * failure of the command. 
*/
EmberStatus emberBootload(void);

/** @description A general-purpose handler routine that parses 
 * messages passed to it by the host and  calls the appropriate
 * application-specific handler function.
 *
 * @param message: A pointer to message contents.
 *
 * @param length: Message length.
*/
void emProcessIncomingMessage(int8u *message, int8u length);

/** Documented in ember.h. */
EmberStatus emberSendDatagram(int8u bindingTableIndex,
			      EmberMessageBuffer message);

/** Documented in ember.h. */
EmberStatus emberSendLimitedMulticast(int8u bindingTableIndex,
				      EmberMessageBuffer message,
				      int8u hops);

// By default multicasts only go out a limited distance.
#ifndef EMBER_MULTICAST_HOPS
#define EMBER_MULTICAST_HOPS 12
#endif

/** Documented in ember.h. */
#define emberSendMulticast(bindingTableIndex, message) \
(emberSendLimitedMulticast(bindingTableIndex, message, EMBER_MULTICAST_HOPS))

/** Documented in ember.h. */
EmberStatus emberSendSequenced(int8u bindingTableIndex, 
			       EmberMessageBuffer message);

/** Documented in ember.h. */
EmberStatus emberSendDatagramFromTerminal(EmberTerminal sender,
					  EmberMessageBuffer message); 

/** Documented in ember.h. */
EmberStatus emberSendMulticastFromTerminal(EmberTerminal sender,
                                           EmberMessageBuffer message); 
/** Documented in ember.h. */
EmberStatus emberSendSequencedFromTerminal(EmberTerminal sender,
                                           EmberMessageBuffer message);

// Calls that can be made from within the emberIncomingMessageHandler() 
/** Documented in ember.h. */
EmberStatus emberSendReply(EmberMessageBuffer reply);

/** Documented in ember.h. */
EmberStatus emberGetCostToSender(EmberCost *returnLoc); 

/** Documented in ember.h. */
EmberStatus emberSetReplyBinding(int8u index, EmberBindingTableEntry *entry); 

/** Documented in ember.h. */
EmberStatus emberNoteSendersBinding(int8u index); 

/** Documented in ember.h. */
EmberStatus emberGetSendingTerminal(EmberTerminal *returnLoc);

// Opening and closing connections.
/** Documented in ember.h. */
EmberStatus emberOpenConnection(int8u bindingTableIndex); 

/** Documented in ember.h. */
EmberStatus emberConnectionStatus(int8u bindingTableIndex); 

/** Documented in ember.h. */
EmberStatus emberCloseConnection(int8u bindingTableIndex);

/** Documented in ember.h */
EmberStatus emberAssociate(EmberNetworkParameters *parameters);

/** Documented in ember.h */
EmberStatus emberDisassociate(void);


// -------------------------------------------------------------------
// The handlers below must be implemented (or at least stubbed out) by
// application developers.
// -------------------------------------------------------------------
/** Documented in ember.h. */
void emberMessageSent(int8u bindingTableIndex,
                      EmberMessageBuffer message,
                      EmberStatus status);

/** Documented in ember.h. */
enum {
  EMBER_INCOMING_DATAGRAM,
  EMBER_INCOMING_DATAGRAM_REPLY,
  EMBER_INCOMING_SEQUENCED,
  EMBER_INCOMING_MULTICAST,
  EMBER_INCOMING_SHARED_MULTICAST,
  EMBER_INCOMING_MULTICAST_LOOPBACK
};

typedef int8u EmberIncomingMessageType;

/** Documented in ember.h. */
void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 EmberTerminal terminal,
                                 int8u index,
                                 EmberMessageBuffer message);

/** Documented in ember.h. */
void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status);

/** Documented in ember.h. */
void emberStackStatusHandler(EmberStatus status);

/** @description A callback invoked when the status of the RNAP changes.
 *
 * @param status: RNAP status. One of the following:
 * - EMBER_ERR_FATAL
 * - EMBER_RNAP_MISMATCH
 * - EMBER_RNAP_ACK_WITH_ERR
 * - EMBER_RNAP_COMMUNICATION_FAILED
 * - EMBER_RNAP_DELIVERY_FAILED
 * - EMBER_RNAP_SERIAL_ERROR
 */
void emberRnapStatusHandler(EmberStatus status);

/** @description A callback invoked when the RNAP message delivery failed.
 *
 * @param status: rnap status. One of the following:
 * - EMBER_RNAP_DELIVERY_FAILED
 */
void emberRnapMessageSent(EmberMessageBuffer message, EmberStatus status);

/** Documented in hal.h  except here we return an int32u that
represents milleseconds of current system time.*/
int16u microGetSystemTick(void);

/** Documented in ember.h. */
#define BUILD_DEF(name, value) \
    name = value,
enum {
    #include "config/config.h"
};
#undef BUILD_DEF

#ifdef EMBER_TEST
	#include "tool/simulator/child/host-child.h"
#endif

#endif //DOXYGEN_SHOULD_SKIP_THIS

#endif // LIB_RNAP_HOST_API_H













