/**
   @File: packet-buffer.c
   @brief: Packet buffer allocation and management routines.
   @Author(s): Guohui Liu, guohui@ember.com

   These functions have the same signature as the packet-buffer functions
   defined on the ember stack. They are providing an easy transition from node
   to host application.

   This Ember Host API library is subject to the Ember corporation
   Development Kit Software License Agreement delivered with this file to
   licensee.
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */
#include <stdlib.h>             // for malloc(), exit()
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#include "app/host-framework/c/host-api.h"
#include "app/host-framework/c/packet-buffer.h"
#include "stack/include/error.h"

// Gets the pointer to the bufffer contents.
int8u *emberMessageBufferContents(EmberMessageBuffer buffer)
{
  return emberLinkedBufferContents(buffer);
}

// Sets the buffer contents length.
void emberSetMessageBufferLength(EmberMessageBuffer buffer,
                                 int8u newLength)
{
  if (buffer != NULL)
    buffer->length = newLength;
}

// Allocates a buffer,
EmberMessageBuffer emberAllocateLinkedBuffers(int8u unused)
{
  EmberMessageBuffer buffer;
  
  buffer = malloc(sizeof(EmberMessageBufferS));
  if (buffer != NULL) {
    buffer->count = 1;
    buffer->length = 0;
    return buffer;
  }
  else 
    return EMBER_NULL_MESSAGE_BUFFER;
}

// Allocates a buffer and fill it with the passed contents.
EmberMessageBuffer emberFillStackBuffer(int8u count, ...)
{
  va_list argPointer;
  EmberMessageBuffer buffer = emberAllocateStackBuffer();
  int8u i;
  
  emberSetMessageBufferLength(buffer, count);

  va_start(argPointer, count);
  for (i = 0; i < count; i++)
    emberLinkedBufferContents(buffer)[i] = LOW_BYTE(va_arg(argPointer, int));
  va_end(argPointer);
  
  return buffer;
}

// Allocates a buffer to hold 'length' bytes and initializes 
// them from 'contents'.
EmberMessageBuffer emberFillLinkedBuffers(int8u *contents, int8u length)
{
  EmberMessageBuffer buffer = emberAllocateStackBuffer();

//  fprintf(stdout, "[allocating buffer for length %d]\n", length);

  emberSetMessageBufferLength(buffer, length);
  if (contents != NULL)
    memcpy(buffer->contents, contents, length);

  return buffer;
}

// Almost the same thing again, except that the copying goes the other way.
void emberCopyFromLinkedBuffers(EmberMessageBuffer buffer,
				int8u startIndex,
				int8u *contents,
				int8u length)
{
  assert(startIndex + length <= emberMessageBufferLength(buffer));

  if (length == 0)
    return;

  memcpy(contents, emberLinkedBufferContents(buffer) + startIndex, length);
}

// Appends the contents to the buffer.
EmberStatus emberAppendToLinkedBuffers(EmberMessageBuffer buffer,
				       int8u *contents,
				       int8u length)
{
  int8u oldLength = emberMessageBufferLength(buffer);

  assert(oldLength + length <= MAX_RNAP_PACKET_SIZE);

  if (length == 0)
    return EMBER_SUCCESS;
  else if (emberSetLinkedBuffersLength(buffer, oldLength+length) 
	   != EMBER_SUCCESS)
    return EMBER_NO_BUFFERS;
  else {
    emberCopyToLinkedBuffers(contents, buffer, oldLength, length);
    return EMBER_SUCCESS;
  }
}

// Appends the contents to the buffer.
EmberStatus emberAppendPgmToLinkedBuffers(EmberMessageBuffer buffer,
                                          PGM_P contents,
                                          int8u length)
{
  return emberAppendToLinkedBuffers(buffer, (int8u *)contents, length);
}

// Appends a string to the buffer.
EmberStatus emberAppendPgmStringToLinkedBuffers(EmberMessageBuffer buffer,
						PGM_P suffix)
{
  int8u length;
  PGM_P temp;

  for (temp = suffix; *temp; temp++);
  length = temp - suffix;
  
  return (emberAppendPgmToLinkedBuffers(buffer, suffix, length));
}

// Sets the buffer contents length.
EmberStatus emberSetLinkedBuffersLength(EmberMessageBuffer buffer,
					int8u newLength)
{
  if (newLength < MAX_RNAP_PACKET_SIZE) {
    emberSetMessageBufferLength(buffer, newLength);
    return EMBER_SUCCESS;
  }
  else
    return EMBER_NO_BUFFERS;
}

// Gets the pointer to the buffer contents at the specified offset.
int8u *emberGetLinkedBuffersPointer(EmberMessageBuffer buffer, int8u index)
{
  assert(index < emberMessageBufferLength(buffer));
  return emberLinkedBufferContents(buffer) + index;
}

// Gets one byte of the buffer contens at the specified offset.
int8u emberGetLinkedBuffersByte(EmberMessageBuffer buffer, int8u index)
{
  return *emberGetLinkedBuffersPointer(buffer, index);
}

// Gets one byte of the buffer contens at the specified offset.
void emberSetLinkedBuffersByte(EmberMessageBuffer buffer, 
			       int8u index, 
			       int8u byte)
{
  *emberGetLinkedBuffersPointer(buffer, index) = byte;
}

// Allocates one buffer, then copy the data over.
EmberMessageBuffer emberCopyLinkedBuffers(EmberMessageBuffer buffer)
{     
  EmberMessageBuffer head;
  int8u length;
  
  assert(buffer != NULL);

  length = buffer->length;
  head = emberAllocateLinkedBuffers(1);
  emberSetMessageBufferLength(head, length);
  memcpy(head->contents, emberLinkedBufferContents(buffer), length);
  return head;
}

// Creates a new, unshared copy of 'buffer' if it is shared.  If it isn't
// shared we increment the reference count by one so that the user of the
// returned buffer can release it in either case.
EmberMessageBuffer emberMakeUnsharedLinkedBuffer(EmberMessageBuffer buffer,
						 boolean isShared)
{
  if (isShared)
    return emberCopyLinkedBuffers(buffer);
  else {
    emberHoldMessageBuffer(buffer);
    return buffer; 
  }
}
