/**
   @ File: packet-buffer.h
   @brief: Packet buffer allocation and management routines.
   @Author(s): Guohui Liu, guohui@ember.com
 
   These functions have the same signature as the packet-buffer functions
   defined on the ember stack. They are providing an easy way to write 
   application that works on both node and host.
  
   The linked, stack and message buffer are the same.
 
   Freeing a buffer automatically decrements the reference count of
   any following buffer, possibly freeing it as well.
 
   This Ember Host API library is subject to the Ember corporation
   Development Kit Software License Agreement delivered with this file to
   licensee.
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#ifndef __PACKET_BUFFER_H__
#define __PACKET_BUFFER_H__


#define PGM_P   const char *

/** Documented in ember.h. */
int8u *emberMessageBufferContents(EmberMessageBuffer buffer);

/** Documented in ember.h. */
#define emberMessageBufferLength(buffer)   (buffer->length)

/** Documented in ember.h. */
void emberSetMessageBufferLength(EmberMessageBuffer buffer, int8u newLength);

/** Documented in ember.h. */
#define emberHoldMessageBuffer(buffer)                                  \
do {                                                                    \
  buffer->count++;                                                      \
} while (FALSE)

/** Documented in ember.h. */
#define emberReleaseMessageBuffer(buffer)                               \
do {                                                                    \
  if ((--(buffer->count)) == 0) {                                       \
    free(buffer);                                                       \
    buffer = NULL;                                                      \
  }                                                                     \
} while (FALSE)

// Gets a pointer to a buffer's contents.
#define emberLinkedBufferContents(buffer) (buffer->contents)

// Returns the buffer. 
#define emberStackBufferLink(buffer) (buffer)

// Assigns the buffer a new value. 
#define emberSetStackBufferLink(buffer, newLink) ((buffer) = (newLink))

// Allocates one buffer.  
#define emberAllocateStackBuffer() (emberAllocateLinkedBuffers(1))

// Allocates one buffer.
EmberMessageBuffer emberAllocateLinkedBuffers(int8u unused);

// Allocates a buffer and fills in its contents.
EmberMessageBuffer emberFillStackBuffer(int8u count, ...);

// Allocates a buffer to hold 'length' bytes and initializes
// them from 'contents'.
EmberMessageBuffer emberFillLinkedBuffers(int8u *contents, int8u length);

// Copies 'length' bytes between 'buffer' and 'message', starting from
// 'startIndex' in the buffer.
#define emberCopyToLinkedBuffers(message, buffer, startIndex, length) \
memcpy(buffer->contents + startIndex, message, length);

// Copies 'length' bytes of data from a buffer to 'contents', starting at 
// a specified index in the buffer array. 
void emberCopyFromLinkedBuffers(EmberMessageBuffer buffer,
				int8u startIndex,
				int8u *contents,
				int8u length);

// Appends 'length' bytes from 'contents' onto buffer. 
EmberStatus emberAppendToLinkedBuffers(EmberMessageBuffer buffer,
				       int8u *contents,
				       int8u length);

// Appends 'length' bytes from 'contents' onto buffer. 
EmberStatus emberAppendPgmToLinkedBuffers(EmberMessageBuffer buffer,
					  PGM_P contents,
					  int8u length);

// Appends 'length' bytes from 'contents' onto buffer. 
EmberStatus emberAppendPgmStringToLinkedBuffers(EmberMessageBuffer buffer,
						PGM_P suffix);
// Sets the buffer length.
// Return EMBER_NO_BUFFERS if the length exceeds the limit
EmberStatus emberSetLinkedBuffersLength(EmberMessageBuffer buffer,
					int8u length);

// Gets a pointer to the index'th byte of a buffer.
int8u *emberGetLinkedBuffersPointer(EmberMessageBuffer buffer, int8u index);

// Returns the index'th byte from a buffer.
int8u emberGetLinkedBuffersByte(EmberMessageBuffer buffer, int8u index);

// Sets the index'th byte of a buffer.
void emberSetLinkedBuffersByte(EmberMessageBuffer buffer, 
			       int8u index, 
			       int8u byte);

// Makes a copy of the given buffer.
// Returns the buffer.
EmberMessageBuffer emberCopyLinkedBuffers(EmberMessageBuffer buffer);

// Creates a new, unshared copy of 'buffer' if it is shared.  If it isn't
// shared we increment the reference count by one so that the user of the
// returned buffer can release it in either case.
EmberMessageBuffer emberMakeUnsharedLinkedBuffer(EmberMessageBuffer buffer,
						 boolean isShared);

#ifdef SIMULATOR
void printPacketBuffers(EmberMessageBuffer buffer);
#endif 

#endif  // __PACKET_BUFFER_H__

