/**
   @file rnap-master-linux.c
   @brief Implementation of read/write for linux host.
   @see rnap-master.h
   @author Guohui Liu (guohui@ember.com)

   This Ember Host API library is subject to the Ember corporation
   Development Kit Software License Agreement delivered with this file to
   licensee.
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h> // inet_aton
#include <netdb.h> // gethostbyaddr
#include <string.h> // for memcpy
#include <pthread.h> // thread
#include <termios.h> //for serial port interaction
#include <time.h>


#include "app/host-framework/c/host-api.h"
#include "app/host-framework/c/packet-buffer.h"
#include "app/host-framework/c/rnap-message-queue.h"
#include "app/util/rnap/rnap.h"
#include "stack/include/error.h"
#include "app/host-framework/c/rnap-master.h"

static int fd;

boolean volatile rnapMasterConnected = FALSE;

extern boolean emNeedSetBindingOnNode(EmberMessageBuffer buf);

//rx rnap msg state
enum {
  READ_START_FRAME,
  READ_LENGTH,
  READ_MESSAGE,
  READ_END_FRAME
};

static int8u rnapRxState = READ_START_FRAME;
static int8u rnapMessageLength;
static int8u rnapMessageIndex;

// Recevied message
static int8u recvline[MAX_RNAP_PACKET_SIZE];

#ifdef DEBUG_HOST_TRANSACT_ON
  static int8u* 
  host2nodeMsgs[] = {"START_ACK", "TX_UNICAST", "TX_DATAGRAM_REPLY", \
	              	   "TX_MULTICAST", "RESET_NODE", "HOST_HAS_RESET", \
		                 "SET_VALUE", "GET_VALUE", "BOOTLOAD", "OPEN_CONNECTION", \
		                 "CLOSE_CONNECTION", "MAX_COMMAND"};
  static int8u* 
  node2hostMsgs[] = {"COMMAND_ACK", "READ_ERROR", "NODE_START", "RX_UNICAST", \
	              	   "RX_MULTICAST", "MESSAGE_SENT_STATUS", "CONNECTION_STATUS", \
		                 "STACK_STATUS", "NEED_BINDING", "MAX_COMMAND"};
#endif

// Reads in one byte.
static int8u readByte(void)
{
  int8u temp;
  int8s length;
  length = read(fd, &temp, 1);
  if (length <= 0) {
    close(fd);
    exit(0);
  }
  else {
    return temp;
  }
}

void buildRnapRxLine(int8u b) {
  switch (rnapRxState) {
    case READ_START_FRAME:
      if (b == FRAME_START) {
        rnapRxState = READ_LENGTH;
      }
      break;
    case READ_LENGTH:
      rnapMessageIndex = 0;
      rnapMessageLength = b;
      rnapRxState = READ_MESSAGE;
      break;
    case READ_MESSAGE:
      recvline[rnapMessageIndex] = b;
      rnapMessageIndex++;
      if (rnapMessageIndex == rnapMessageLength)
        rnapRxState = READ_END_FRAME;
      break;
    case READ_END_FRAME:
      if (b == FRAME_END) {
        #ifdef DEBUG_HOST_TRANSACT_ON
        {
          int8u i;
          time_t t1;
          time(&t1);
          fprintf(stdout, "<rx> %s %02x %02x ",
                  asctime(localtime(&t1)), FRAME_START, rnapMessageLength);
          for (i = 0; i < rnapMessageLength; i++) {
            fprintf(stdout, "%02x ", recvline[i]);
            if (i == 0)
            fprintf(stdout, "%s%s ", "N_TO_H_", 
            node2hostMsgs[recvline[i]]);
          }
          fprintf(stdout, "%02x </rx>\n", FRAME_END);
        }
        #endif
        emProcessIncomingMessage(recvline, rnapMessageLength);
      }
      rnapRxState = READ_START_FRAME;
      break;
    default:
      break;
  }   
}


static void *_emberbusReceiver(void *arg)
{
  rnapMasterConnected = TRUE;
  while (rnapMasterConnected) {
    buildRnapRxLine(readByte());
  }
  exit(0);
}

//initializes a connection to stdio
int8u emMasterConnectStdIO(void) {
  rnapMasterConnected = TRUE;
  return EMBER_SUCCESS;
}

// Initializes the serial connection.
int8u emMasterConnectSerial(int8u *name) 
{
  pthread_t tid;
  rnapMasterConnected = FALSE;

  // open the serial port
  fd = open(name, O_RDWR);  
  if (fd < 0) {
    fprintf(stdout, "Error opening Serial Port\n");
    return 1;
  }
  else {
    // default behavior
    return pthread_create(&tid, NULL, _emberbusReceiver, NULL);   
  }
}

// Initializes the socket connection.
int8u emMasterConnectSocket(int8u *hostName, int16u port)
{
  struct sockaddr_in address;
  struct in_addr inaddr;
  struct hostent *host;
  pthread_t tid;

  rnapMasterConnected = FALSE; 

  if(inet_aton(hostName, &inaddr))
    host = gethostbyaddr((char *)&inaddr, sizeof(inaddr), AF_INET);
  else
    host = gethostbyname(hostName);
  
  if (!host) {
    fprintf(stdout, "Error looking up host\n");
    return 1;
  }
  
  if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    fprintf(stdout, "Error open socket\n");
    return 2;
  }

  address.sin_family = AF_INET;
  address.sin_port = htons(port);
    memcpy(&address.sin_addr, host->h_addr_list[0], sizeof(address.sin_addr));
  if (connect(fd, (struct sockaddr *)&address, sizeof(address))) {
    fprintf(stdout, "Error connect\n");
    return 3;
  }

  // default behavior
  return pthread_create(&tid, NULL, _emberbusReceiver, NULL);
}

// Writes the buffer contents to the descriptor opened for 
// host node communication.
// Return code: EMBER_ERR_FATAL, EMBER_RNAP_COMMUNICATION_FAILED, EMBER_SUCCESS 
EmberStatus emMasterReallyWrite(EmberMessageBuffer buf)
{
  int length;

  if (buf == EMBER_NULL_MESSAGE_BUFFER)
    return EMBER_ERR_FATAL;

  if (rnapMasterConnected) {

   /*
    * Check to see if this message can be sent to the node for instance
    * if it is a unicast and there is no binding for it on the node it wouldn't
    * make sense to send it down and we should return an error. This won't block
    * the queue though, it will just error and dump the message on the floor
    */
    if (emNeedSetBindingOnNode(buf)) {
      fprintf(stdout, "Attempt to send rnap message failed due to EMBER_RNAP_MISMATCH\n");
      return EMBER_RNAP_MISMATCH;
    }

    #ifdef DEBUG_HOST_TRANSACT_ON
    {
      int8u i;
      time_t t1;
      time(&t1);
      fprintf(stdout, "<tx> %s ",  asctime(localtime(&t1)));
      for (i = 0; i < (buf->length); i++) {
        fprintf(stdout, "%02x ", buf->contents[i]);
        if (i == 2)
	      fprintf(stdout, "%s%s ", "H_TO_N_", 
		    host2nodeMsgs[buf->contents[i]]);
      }
      fprintf(stdout, "</tx>\n");
    }
    #endif

    
    //write the rnap packet
    length = writeIO(fd, buf->contents, buf->length);
    
    /*  
      If we don't set need ack to true we'll never use the
      rnap message queue and messages will always be written 
      directly out to the gateway which is appropriate for RNAP 1
    */
    emSetNeedAck(TRUE);
  
    if (length < 0) {
      emberRnapStatusHandler(EMBER_RNAP_COMMUNICATION_FAILED);
      return EMBER_RNAP_COMMUNICATION_FAILED;
    }
  }
  else {
    fprintf(stdout, "[attempt to write to node before connecting]\n");
    emberRnapStatusHandler(EMBER_RNAP_COMMUNICATION_FAILED);
    return EMBER_RNAP_COMMUNICATION_FAILED;
  }
  return EMBER_SUCCESS;
}

// Adds the framing bytes, prefix '[ length' and suffix ']'.
// Calls emMasterReallyWrite if there is no message in the queue and 
// no need to wait for an ack,
// Otherwise add the message to the end of the queue.
EmberStatus emMasterWrite(int8u *userData, int8u userLength,
			  int8u *systemData, int8u systemLength)
{
  int8u index = 0;
  int8u temp[MAX_RNAP_PACKET_SIZE];
  int8u length = systemLength + userLength;
  EmberMessageBuffer buf;

  if ((length + RNAP_PACKET_OVERHEAD) > MAX_RNAP_PACKET_SIZE)
    return EMBER_MESSAGE_TOO_LONG;

  temp[index++] = FRAME_START;
  temp[index++] = length;
  memcpy(temp + index, systemData, systemLength);
  memcpy(temp + index + systemLength, userData, userLength);
  temp[length + index++] = FRAME_END;
  
  buf = emberFillLinkedBuffers(temp, length + index);

  //debug
  //fprintf(stdout, "queue size %d, needAck %d \n",
	//	    emGetRnapQSize(), emNeedAck());
  
  if (emGetRnapQSize() == 0 && !emNeedAck()) {

    if (emLastRnapMessageSent != NULL)
      free(emLastRnapMessageSent);

    emLastRnapMessageSent = buf;
    return emMasterReallyWrite(buf);
  }
  else if (!emAddRnapQMessage(buf)) {
    return EMBER_MAX_MESSAGE_LIMIT_REACHED;
  }
    else {
    if (!emNeedAck()) {
      buf = emRemoveRnapQMessage();
      
      //if we have a pointer to the last message hanging around, free it
      if (emLastRnapMessageSent != NULL)
        free(emLastRnapMessageSent);
      
      emLastRnapMessageSent = buf;
      return emMasterReallyWrite(buf);
    }
    else
      return EMBER_SUCCESS;
  }
  return EMBER_SUCCESS;
}
