/**
   @File rnap-message-queue.c
   @brief Manage the messages in the queue.
   @see rnap-message-queue.h
   @author Guohui Liu (guohui@ember.com)

   The operation on the queue is thread safe.

   This Ember Host API library is subject to the Ember corporation
   Development Kit Software License Agreement delivered with this file to
   licensee.
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#include "host-api.h"
#include "rnap-message-queue.h"
#include "packet-buffer.h"

// mutex for queue access
pthread_mutex_t queue_mutex = PTHREAD_MUTEX_INITIALIZER; 

static RnapMessageQueue rnapMessageQueue;

// Initializes the message queue
void emRnapQInit()
{
  int8u i;

  // init mutex
  while (pthread_mutex_init(&queue_mutex, NULL) != 0)
    ;
  
  // lock the queue
  pthread_mutex_lock(&queue_mutex); 
  // clear the queue
  rnapMessageQueue.head = 0;
  rnapMessageQueue.tail = 0;
  rnapMessageQueue.length = 0;
  rnapMessageQueue.needAck = FALSE;
  rnapMessageQueue.waitTime = 0;
  for (i = 0; i < MAX_RNAP_Q_SIZE; i++) {
    if (rnapMessageQueue.rnapMessages[i] != EMBER_NULL_MESSAGE_BUFFER) {
      free(rnapMessageQueue.rnapMessages[i]);
      rnapMessageQueue.rnapMessages[i] = EMBER_NULL_MESSAGE_BUFFER;
    }
  }
  // unlock the queue
  pthread_mutex_unlock(&queue_mutex); 
}

// Adds one message to the end of the queue
boolean emAddRnapQMessage(EmberMessageBuffer buf)
{
  if (buf == EMBER_NULL_MESSAGE_BUFFER)
      return FALSE;
            
  // lock the queue
  pthread_mutex_lock(&queue_mutex); 
  // add the message to the end of the queue
  if (rnapMessageQueue.length < MAX_RNAP_Q_SIZE) {
      Q_ENQUEUE(rnapMessageQueue, buf);
      // unlock the queue
      pthread_mutex_unlock(&queue_mutex); 
      DEBUG_OUT(fprintf(stdout, "Add one message to queue, length %d head %d tail %d wait %d\n", 
			rnapMessageQueue.length, 
			rnapMessageQueue.head,
			rnapMessageQueue.tail,
			rnapMessageQueue.waitTime));
      return TRUE;
  }
  else {
    // unlock the queue
    pthread_mutex_unlock(&queue_mutex); 
    DEBUG_OUT(fprintf(stdout, "Queue Full \n"));
    return FALSE;
  }
}

// Removes the message from the head of the queue.
EmberMessageBuffer emRemoveRnapQMessage()
{
  EmberMessageBuffer buf = EMBER_NULL_MESSAGE_BUFFER;

  pthread_mutex_lock(&queue_mutex); 

  if (rnapMessageQueue.length > 0) {
    buf = Q_DEQUEUE(rnapMessageQueue);
    
    DEBUG_OUT(fprintf(stdout, "Remove one rnap message from queue, length %d head %d tail %d wait %d\n", 
		      rnapMessageQueue.length, 
		      rnapMessageQueue.head, 
		      rnapMessageQueue.tail,
		      rnapMessageQueue.waitTime));
  }
  pthread_mutex_unlock(&queue_mutex); 

  return buf;
}

// Gets the number of message in the queue.
int8u emGetRnapQSize()
{
  int8u length;

  pthread_mutex_lock(&queue_mutex); 
  length = rnapMessageQueue.length;
  pthread_mutex_unlock(&queue_mutex); 

  return length;
}

// Sets the needAck value.
void emSetNeedAck(boolean needAck)
{
  pthread_mutex_lock(&queue_mutex); 
  rnapMessageQueue.needAck = needAck;
  pthread_mutex_unlock(&queue_mutex); 
}

// Gets the needAck value
boolean emNeedAck()
{
  boolean needAck;

  pthread_mutex_lock(&queue_mutex); 
  needAck = rnapMessageQueue.needAck;
  pthread_mutex_unlock(&queue_mutex);
  
  return needAck;
}

// Increases the wait time and check if max wait reached
boolean emWaitAckTooLong(int32u timePassed, int32u maxWait)
{
  rnapMessageQueue.waitTime += timePassed;
  if (rnapMessageQueue.waitTime >= maxWait)
    return TRUE;
  else
    return FALSE;
}
