// File: command-interpreter.h
// Description: processes commands incoming over the serial port.
// Originally written for the eval-kit application.
//
// Culprit(s): Richard Kelsey (kelsey@ember.com)
//
// Copyright 2004 by Ember Corporation.  All rights reserved.               *80*

// Applications using this module must define a commands table

// A comment consists of a '#' plus any following characters up to the next
// end-of-line.
#ifndef EMBER_COMMENT_CHARACTER
#define EMBER_COMMENT_CHARACTER	   '#'
#endif

// The maximum number of characters in a command.
#ifndef EMBER_MAX_COMMAND_LENGTH
#define EMBER_MAX_COMMAND_LENGTH    12
#endif

// A command may have an optional one-character short name, with zero indicating
// that the command has no short name.
//
// The argumentTypes string specifies the number and types of arguments the
// command accepts.  The argument specifiers are:
//   b   buffer
//   u1  one-byte unsigned 
//   u2  two-byte unsigned 
//   s1  one-byte signed 
//   s2  two-byte signed
//   *   zero or more of the previous type; must be the last specifier
// [A possible extension would be:
//   ?   zero or one of the previous type; must be the last specifier ]

struct EmberCommandStateS;      // To allow forward use of this type.

// Commands are case-insensitive.  The names should be given in lowercase.

typedef PGM struct {
  int8u shortName;              // lower case letters and '_'
  PGM_P longName;               // lower case letters and '_'
  void (*action)(struct EmberCommandStateS *state);
  PGM_P argumentTypes;
  PGM_P description;
} EmberCommandEntry;

//----------------
// Errors.
//
// emberCommandErrorNames[] in command-interpreter.c has strings describing
// these, so if you change this list you should change that one as well.

enum {
  EMBER_CMD_SUCCESS,
  EMBER_CMD_ERR_PORT_PROBLEM,
  EMBER_CMD_ERR_NO_BUFFER_AVAILABLE,
  EMBER_CMD_ERR_NO_SUCH_COMMAND,
  EMBER_CMD_ERR_WRONG_COMMAND_ARGUMENTS,
  EMBER_CMD_ERR_TOO_MANY_COMMAND_ARGUMENTS,
  EMBER_CMD_ERR_ARGUMENT_OUT_OF_RANGE,
  EMBER_CMD_ERR_ARGUMENT_SYNTAX_ERROR,
  EMBER_CMD_ERR_STRING_TOO_LONG,
  EMBER_CMD_ERR_NO_BINARY_STRING_TERMINATOR
};
typedef int8u EmberCommandStatus;

extern PGM_NO_CONST PGM_P emberCommandErrorNames[];

//----------------
// In order to allow commands to be read from multiple sources the state of
// the command interpreter is bundled up into a struct.  The user provides
// an instance of the struct to the interpreter.  The following sample code
// is from command-interpreter-test.c.
//
// static void commandErrorHandler(int8u errorCode)
// {
//   // ...
// }
// 
// static EmberCommandEntry commands[] = {
// //       |----------| MAX_COMMAND_LENGTH is currently 12
//   { 'i', "integers",   integersCommand,   "u1u2s1s2", "integers"},
//   { 'j', "intlist",    intListCommand,    "u1*",      "intlist"},
//   { 0,   "stringlist", stringListCommand, "bb*",      "stringlist"},
//   { 0, NULL, NULL, NULL, NULL}}; // NULL action makes this a terminator
// 
// int16u arguments[4];
// 
// EmberCommandState testState =
//   {
//     commands,
//     commandErrorHandler,
//     arguments,
//     4,                    // maximum command arguments
//     10                    // integer base
//   };
//
// The integer base determines whether integers are parsed using base 10
// or base 16.

typedef struct EmberCommandStateS {
  // Stuff provided by the user.
  EmberCommandEntry *commands;
  void (* errorHandler)(EmberCommandStatus);
  EmberMessageBuffer *arguments;                    // filled in by the interpreter
  int8u maximumArguments;
  int8u integerBase;                    // must be 10 or 16
  
  // Set by the interpreter for use by the command function.
  int8u argumentCount;

  //----------------
  // The rest is internal gubbish.

  int8u state;                  // finite-state machine's current state
  int8u error;                  // first error found in this command

  // For reading in and remembering the current command
  int8u command[EMBER_MAX_COMMAND_LENGTH + 1];  // + 1 for trailing nul
  int8u commandIndex;
  PGM_P argumentSpecs;          // stepped forward as we read arguments

  int8u stringLength;           // bytes read in so far
  int8u binaryLength;           // remaining bytes in binary block
  EmberMessageBuffer stringBuffer;

  int16u integerValue;
  int8u integerSign;            // '+' or '-'
  
} EmberCommandState;

//----------------
// Three macros to help command actions retrieve their arguments from the
// command interpreter state object.  They pull out unsigned integers,
// signed integers, and buffers.

#define emberUnsignedCommandArgument(state, index) \
((int32u)((state)->arguments[index]))

#define emberSignedCommandArgument(state, index) \
((int32s) ((state)->arguments[index]))

#define emberBufferCommandArgument(state, index) \
((state)->arguments[index])

//----------------
// Initialize a command state.
void emberCommandReaderInit(EmberCommandState *state);

// Parses 'size' characters from 'input'.
void emberProcessCommandString(EmberCommandState *state,
                               int8u *input,
                               int8u size);

// Parses any input coming in on the serial port, returning when no more
// is available.  Just a front for emberProcessCommandString() with a null
// string and the port passed in as the string size.
// void emberProcessCommandInput(EmberCommandState *state, int8u port);
#define emberProcessCommandInput(state, port) \
emberProcessCommandString(state, NULL, port)


