/**
 * File: cipher-meter.c
 * Description: meter code for AMR application. This application waits
 * to hear from an aggregator. When it does, it begins sending periodic
 * datagrams to the aggregator representing meter data. The meter can also
 * respond to datagram requests from the aggregator for report data as well
 * as history data. Report data is meant to represent the latest report. History
 * data is meant to represent a history of the meter's reports and is transmitted
 * via sequenced message. For a deeper look at datagrams and sequenced messages
 * @see stack/include/ember.h
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */
#include <stdlib.h>
#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "stack/include/packet-buffer.h"
#include "stack/include/error.h"
#include "stack/include/hal.h"
#include "app/util/serial/serial.h"
#include "app/util/serial/command-interpreter.h"
#include "hal/board/board.h"
#include "hal/micro/micro.h"
#include "app/Cipher/rnap/cipher-meter.h"
#include "app/Cipher/rnap/cipher-rf.h"
#include "app/Cipher/rnap/cipher.h"
#include "app/Cipher/rnap/create-packets.h"
#include "app/Cipher/rnap/cipher-meter-cmds.c"

// definitions for application timers
#include "app/Cipher/app-timers.h"

//------------------------------------------------------------------------------
//for a more comprehensive description of these see amr-common.c
#ifdef EMBER_TEST
  extern int16u simulatorId;
#endif

//a flag to track if the network is up or down
boolean networkUp = FALSE;
boolean NetworkUPdate;

int8u PacketFailure;

//------------------------------------------------------------------------------
#define MULTICAST_BINDING_INDEX MAXIMUM_AGGREGATORS

#define EMBER_TRANSPORT_CONNECTION_COUNT        1
#define EMBER_BINDING_TABLE_SIZE                (MAXIMUM_AGGREGATORS + 1)
#define EMBER_TEMPORARY_BINDING_ENTRIES         EMBER_BINDING_TABLE_SIZE
#define EMBER_MAXIMUM_AGGREGATION_ROUTES        0

#include "stack/include/ember-static-configuration.h"


//------------------------------------------------------------------------------

static EmberInterfaceId PGM clientInterface = {
  AD_HOC_INTERFACE,
  AGGREGATION_CLIENT_ENDPOINT,
  4, 'A', 'M', 'R', 0
};

static EmberEndpointDescription PGM endpointDescription = {
  { 0, &clientInterface }
};

int8u emberEndpointCount = 1;
EmberEndpoint emberEndpoints[1] = {
  { 1, &endpointDescription }
};

PGM_NO_CONST PGM_P applicationString = "Meter";

//------------------------------------------------------------------------------

void RespondCUadvertisment(int8u *contents);
void SendAdvertismentToCU(int8u Dev_type);
void RespondNetworkUpdate(void);
void TimedEvents(void);
void Senddata(void);
void SendSearchCU(void);

void initializeEmberStack(void)
{
  EmberStatus status;
  EmberNetworkParameters networkParams;
  EmberResetType resetInfo = microGetResetInfo();
  boardBoot();
  microBoot();
  microStartSystemTick();
  INTERRUPTS_ON();
  emberDebugInit(0);
  emberSerialInit(SERIAL_PORT, BAUD_19200, PARITY_NONE, 1);
    // Initialize the Ember Stack.
  status = emberInit(resetInfo);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(SERIAL_PORT, 
      "error returned from emberInit: 0x%x\r\n",
      status);
    assert(FALSE);
  }

  // Set up our default network parameters.
  networkParams.panId = 0;
  networkParams.enableRelay = TRUE;   //
  networkParams.radioTxPower = CIPHER_SAMPLE_POWER;
  networkParams.radioChannel = CIPHER_SAMPLE_CHANNEL;
  
  // associate with the network
  status = emberAssociate(&networkParams);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(SERIAL_PORT, 
      "error returned from emberAssociate: 0x%x\r\n",
      status);
    assert(FALSE);
  }
   adcSleep();  //turn off adc
    
}

//commands used by the command interpreter
extern EmberCommandEntry commands[];
extern EmberCommandState commandState;

//------------------------------------------------------------------------------
/* FORWARD DECLARATIONS OF METER SPECIFIC HANDLERS                            */
static void incomingDatagramHandler(int8u localTerminal, int8u bindingTableIndex,
                                    EmberMessageBuffer message);
static void incomingMulticastHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message);



//------------------------------------------------------------------------------
//if we are in test mode, nodeMain replaces main()
#ifdef EMBER_TEST
int nodeMain(void)
#else
int main(void)
#endif
{
  
  InitStatusSruct();
  initializeEmberStack(); 
  emberCommandReaderInit(&commandState);
  createMulticastBinding(MULTICAST_BINDING_INDEX);

  //emberSerialGuaranteedPrintf(SERIAL_PORT,"RESET\r\n");  //debug only
  //blink a heartbeat
  #ifndef EMBER_TEST
    addTimer(BLINK_LED_TIMER_INDEX, toggleBlinker, 0, 8, 1);
  #endif

  while(TRUE) {
    
    /*** NODE / HOST DIFFERENCE ***/
    //On a node we need to service the watchdog 
    //and emberTick. so the node doesn't reset.
    __watchdog_reset();
    emberTick();//lets try to do more in a loop
    emberSerialBufferTick();

    //need to service the timer module
    handleTimers();

    //need to service the command processor module
    emberProcessCommandInput(&commandState, SERIAL_PORT);
    
    emberTick();//lets try to do more in a loop
    emberSerialBufferTick();
    
    TimedEvents();
    
    emberTick();//lets try to do more in a loop
    
    if(TXnow) Senddata();
    
    emberTick();//lets try to do more in a loop
    emberSerialBufferTick();
    
    if(GotDataPacketACK)PacketsStatus();  //packet retrying etc only activated by a data packet (Ember) ack/nack
    
    emberTick();//lets try to do more in a loop
    
    if((FindCU) && (!RStatus.bMesher) && (!RStatus.bConnected)) SendSearchCU(); //leaf node only
    
    emberTick();  //lets try to do more in a loop
    emberSerialBufferTick();
       
    #ifdef EMBER_TEST
      simulatedTimePassesMs(msToNextEvent());
    #endif
    if(Sleep) {
        if(emberSleep()){
              GoToSleep();
              emberWakeUp();
              Sleep = FALSE;
        }
    }/**/
    if(PacketFailure > ACCEPTABLEFAILURES){
        RStatus.bConnected = FALSE;           //tell nigel i have lost CU
        emberDeleteBinding(CU_BINDING_INDEX); //delete CU binding
        PacketFailure = 0;                    //Reset dropped packet counter
    }
    emberTick();      //lets try to do more in a loop
    emberSerialBufferTick();
  }
  //return 0;  
}



//------------------------------------------------------------------------------
// METER SPECIFIC MESSAGE HANDLERS

// Handle datagrams coming from the aggregator
static void incomingDatagramHandler(int8u localTerminal,
                                    int8u bindingTableIndex,
                                    EmberMessageBuffer message)
{
  if (message != EMBER_NULL_MESSAGE_BUFFER)
  {
    //int8u *contents = emberMessageBufferContents(message);
    int8u *contents = emberGetLinkedBuffersPointer(message,0);
  
    switch (contents[0]) {
      case CU_OKBIND: 
          RStatus.bConnected = TRUE;
          RStatus.bDeliveryFail = FALSE;
          break;
   
      case NODE_INCOMMING_CMD:
          RStatus.bShortIdle = FALSE;
          //store the cmd in a buffer
          //CmdBuff contents[1]+1 = SIZE BYTES TO FOLLOW PLUS 1 FOR THE SIZE
          emberCopyFromLinkedBuffers(message, 1, CmdBuff, contents[1]+1);
          RStatus.bCuRecord = TRUE;                         //
          break;
       
      case CU_REMOVE:
          RStatus.bConnected = FALSE;           //tell nigel i have lost CU
          RStatus.bBusy = FALSE;
          RStatus.bDelivered = FALSE;
          RStatus.bDeliveryFail = FALSE;
          emberDeleteBinding(CU_BINDING_INDEX); //delete CU binding
          break;
      
      case NODE_CMD_WAITING:
          RStatus.bShortIdle = TRUE;                         //
          break;
      
      
      default:
        //emberSerialPrintf(SERIAL_PORT, "Datagram msg error\n");
          break;
    }  // end switch
  }
}

// Called by emberIncomingMessageHandler
static void incomingMulticastHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message)
{
  //int8u *contents = emberMessageBufferContents(message);
  int8u *contents = emberGetLinkedBuffersPointer(message,0);
  EmberBindingTableEntry MyCu;

  switch (contents[0]) {
  case CU_ADVERTISEMENT:
    if(!RStatus.bConnected && RStatus.bCommissioned &&!WaitingToRespond)RespondCUadvertisment(contents);
    break;
    
  case NETWORK_UPDATE:
    if(RStatus.bConnected){
      emberGetBinding(CU_BINDING_INDEX, &MyCu);
      if(!MEMCOMPARE(contents+1,MyCu.identifier,EUI_LEN)) //if its our CU will respond otherwise will ignor
      {
        UpdateRespondTimer = ((firmGetRandom()) % 20) + 5;    //1 to 5 sec's
        NetworkUPdate = TRUE;
        //emberSerialGuaranteedPrintf(SERIAL_PORT,"Recieved Update from CU\r\n");  //debug only
      }
    }
    break;
      
    default:
    //emberSerialPrintf(SERIAL_PORT, "Multicast msg error\n");  //debug only
    break;
  }  // end switch
}



//------------------------------------------------------------------------------
// Stack Callbacks
// @see stack/include/ember.h for a complete description
void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 int8u localTerminal,
                                 int8u bindingTableIndex,
                                 EmberMessageBuffer message)
{
  //int8u *data = emberMessageBufferContents(message);
  int8u *data = emberGetLinkedBuffersPointer(message,0);
  switch (type) {
  case EMBER_INCOMING_DATAGRAM_REPLY:
  case EMBER_INCOMING_DATAGRAM:
    incomingDatagramHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_MULTICAST:
    incomingMulticastHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_SEQUENCED:
  case EMBER_INCOMING_SHARED_MULTICAST:
  case EMBER_INCOMING_MULTICAST_LOOPBACK:
  default:
    break;
  }
}

//@see stack/include/ember.h
void emberMessageSent(int8u bindingTableIndex, EmberMessageBuffer message, EmberStatus status)
{
  int8u PacketNumber;
  int8u *data = emberMessageBufferContents(message);
  
  switch(data[0])          
  {
      case SENSOR_DATAPACKET_START:
      case MOBILE_DATAPACKET_START:
      case HANDHELD_DATAPACKET_START:
        PacketNumber = 1;
        GotDataPacketACK = TRUE;
        break;
      
      case SENSOR_DATAPACKET_2:
      case MOBILE_DATAPACKET_2:
      case HANDHELD_DATAPACKET_2:
        PacketNumber = 2;
        GotDataPacketACK = TRUE;
        break;
        
      case SENSOR_DATAPACKET_3:
      case MOBILE_DATAPACKET_3:
      case HANDHELD_DATAPACKET_3:
        PacketNumber = 3;
        GotDataPacketACK = TRUE;
        break;
        
      case SENSOR_DATAPACKET_4:
      case MOBILE_DATAPACKET_4:
      case HANDHELD_DATAPACKET_4:
        PacketNumber = 4;
        GotDataPacketACK = TRUE;
        break;
  }
  if(status == EMBER_SUCCESS){    //packet was delivered
         if(PacketNumber == 1)MessageResults.Packet1Rx = TRUE;
         if(PacketNumber == 2)MessageResults.Packet2Rx = TRUE;
         if(PacketNumber == 3)MessageResults.Packet3Rx = TRUE;
         if(PacketNumber == 4)MessageResults.Packet4Rx = TRUE;
         PacketFailure = 0;
  }
  else{       //packet failed
         if(PacketNumber == 1)MessageResults.Packet1Fail = TRUE;
         if(PacketNumber == 2)MessageResults.Packet2Fail = TRUE;
         if(PacketNumber == 3)MessageResults.Packet3Fail = TRUE;
         if(PacketNumber == 4)MessageResults.Packet4Fail = TRUE;
         PacketFailure++;
  }
}


//@see stack/include/ember.h
void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status)
{
  //emberSerialPrintf(SERIAL_PORT,
  //  "Connection to binding %d: %x\r\n", bindingTableIndex, status);

  //connectedToAggregator = (status == EMBER_CONNECTION_OPEN);
}

//@see stack/include/ember.h
void emberStackStatusHandler(EmberStatus status)
{
  if (status == EMBER_NETWORK_UP) {
    networkUp = TRUE;
  }  
   
}

// @see stack/include/ember.h
EmberStatus emberRemoteSetBindingHandler(int8u index,
                                         EmberBindingTableEntry *entry) 
                                         {
  //we're not allowing external control of our binding table since this
  //isn't really a provisonable application.
  return EMBER_BINDING_TABLE_WRITE_FAILED;
}

//The meter doesn't search for endpoints, but if it did, they would 
//come back here
void emberEndpointFoundHandler(EmberInterfaceId *desiredInterface,
                               EmberEUI64 node,
                               EmberNetworkDistance distance)
{
}

void emberBeaconLossCallback(void) {

}

void temporaryIncomingBeaconHandler(void) {

}

//-----------------------------------------------------------------
//board related functions, have to implement these for the dev board
void boardButtonCallback(int8u button, int8u state)
{
}

void toggleBlinker(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED3);
}


void TimedEvents(void)
{
  //in16u LastEventTime = 0;
  int16u Time;
    
  Time = microGetSystemTick();
  if((Time - LastEventTime) > 200 /*ms*/) //this is now a 200ms timer
  {
        LastEventTime = Time; //update the 200ms timer
       
        if(WaitingToRespond)
        {
             CURespondTimer--;
      
             if(CURespondTimer <1){
                SendAdvertismentToCU(Cipher_Params.DeviceType);
                WaitingToRespond = FALSE;
                return;
             }
        }
        
        if(NetworkUPdate)   //responce to CU network up-date
        {
            UpdateRespondTimer--;
            if(UpdateRespondTimer < 1){
               RespondNetworkUpdate();
               NetworkUPdate = FALSE;
               return;
            }
        }
  }

}

void Senddata(void)
{
  int8u size = 0; 
  int8u BlockLength = 0;
  int LeftToSend = 0;

  BlockLength = MonsterBlock[0];
  InitPacketRetryStuct(BlockLength);
  LeftToSend = BlockLength - PACKET1SIZE;   //Pack1 payload       PACKET1
  if(LeftToSend<=0){     //indicates only needs 1 packet (if less than zero)
      size = LeftToSend + PACKET1SIZE;
      MessageResults.Packet1size = size;
      CreatePacket1(size, BlockLength, Cipher_Params.DeviceType);  //will create and send a packet
      TXnow = FALSE;
      return;
  }
  else{ //indicates we can fill up packet 1 completely
      size = PACKET1SIZE;
      MessageResults.Packet1size = size;
      CreatePacket1(size, BlockLength, Cipher_Params.DeviceType);  //will create and send a packet
  }
  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------
  LeftToSend = LeftToSend - PACKET2SIZE;   //Pack2 payload        PACKET2
   if(LeftToSend<=0){        //indicates only needs 2 packets  (if less than zero)
      size = LeftToSend + PACKET2SIZE;
      MessageResults.Packet2size = size;
      CreatePacket(size, PACKET2, Cipher_Params.DeviceType);  //will create and send a packet
      TXnow = FALSE;
      return;
  }
  else{ //indicates we can fill up packet 2 completely
      size = PACKET2SIZE;
      MessageResults.Packet2size = size;
      CreatePacket(size, PACKET2, Cipher_Params.DeviceType);  //will create and send a packet
      
  }
  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------
   LeftToSend = LeftToSend - PACKET3SIZE;   //Pack3 payload       PACKET3
   if(LeftToSend<=0){        //indicates only needs 3 packets  (if less than zero)
      size = LeftToSend + PACKET3SIZE;
      MessageResults.Packet3size = size;
      CreatePacket(size, PACKET3, Cipher_Params.DeviceType);  //will create and send a packet
      TXnow = FALSE;
      return;
  }
  else{ //indicates we can fill up packet 3 completely
      size = PACKET3SIZE;
      MessageResults.Packet3size = size;
      CreatePacket(size, PACKET3, Cipher_Params.DeviceType);  //will create and send a packet
      
  }
  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------
   LeftToSend = LeftToSend - PACKET4SIZE;   //Pack4 payload       PACKET4
   if(LeftToSend<=0){        //indicates only needs 4 packets  (if less than zero)
      size = LeftToSend + PACKET4SIZE;
      MessageResults.Packet4size = size;
      CreatePacket(size, PACKET4, Cipher_Params.DeviceType);  //will create and send a packet
      TXnow = FALSE;
      return;
  }
  else{ //indicates we can fill up packet 4 completely
      size = PACKET4SIZE;
      MessageResults.Packet4size = size;
      CreatePacket(size, PACKET4, Cipher_Params.DeviceType);  //will create and send a packet
  }
  
  TXnow = FALSE;
}

void SendSearchCU(void)
{
  EmberStatus status;
  EmberMessageBuffer buffer = 0; 

  globalBuffer[0] = LEAFNODE_ADVERTISEMENT;
  MEMCOPY(&(globalBuffer[1]), emberGetEui64(), EUI_LEN);
  
  //Need to add Serial Number
  
  MEMCOPY(&globalBuffer[9], &Cipher_Params.SerialNum, 10);
  
  buffer = emberFillLinkedBuffers(globalBuffer, 9+10); //9 = msg type + eui64 + comark s/n
  // check to make sure a buffer is available
  if (buffer == EMBER_NULL_MESSAGE_BUFFER) 
  {
    emberSerialGuaranteedPrintf(SERIAL_PORT,"No buffer");
    RStatus.bConnected = FALSE;
    FindCU = TRUE;
    return;
  }
  // send the message
  status = emberSendMulticast(MULTICAST_BINDING_INDEX, buffer);

  // done with the packet buffer
  emberReleaseMessageBuffer(buffer);
  FindCU = FALSE;
}


//creates and sets a binding to send multicasts to
//or in the case of a meter, receive multicasts from related
//to the amr application
void createMulticastBinding(int8u index)
{
  EmberBindingTableEntry entry;

  entry.type = EMBER_MULTICAST_BINDING;
  entry.identifier[0] = HIGH_BYTE(MULTICAST_GROUP);
  entry.identifier[1] = LOW_BYTE(MULTICAST_GROUP);
  entry.local = ourTerminal;

  assert(emberSetBinding(index, &entry) == EMBER_SUCCESS);
}

//Application Functions

void RespondCUadvertisment(int8u *contents) //when node recieves a CU advertisment (multicast) it creates a binding to it
{                                           //and then asks the CU to confirm its binding, i do this so i can make this responce
  EmberBindingTableEntry entry;             //a datagram message and not a multicast for better network performance.  The responce
                                            //is sent after a random backoff time of 1 to 5 seconds to help any data contention
  entry.type = EMBER_UNICAST_BINDING;
  entry.local = emberMakeTerminal(1, 0);//ourTerminal;  (1, 0)
  entry.remote = emberMakeTerminal(1, 0);// ourTerminal; (1, 0)
  MEMCOPY(entry.identifier, &(contents[1]), EUI_LEN);
  
  assert(emberSetReplyBinding(CU_BINDING_INDEX, &entry) == EMBER_SUCCESS);   //refer to matts email in log book
  //wait 1 to 5 sec's before responding
  if(RStatus.bMesher)  CURespondTimer = ((firmGetRandom()) % 20) + 5;
  if(!RStatus.bMesher) CURespondTimer = 1;
  WaitingToRespond = TRUE;
  
  //emberSerialGuaranteedPrintf(SERIAL_PORT,"Recieved Advertisment from CU\r\n");  //debug only
}

void SendAdvertismentToCU(int8u Dev_type)
{
  EmberStatus status;
  EmberMessageBuffer buffer = 0; 
  
  //emberSerialGuaranteedPrintf(SERIAL_PORT,"Sending Advertisment to CU\r\n");  //debug only
  
  switch(Dev_type)
  {
    case SENSOR:    globalBuffer[0] = SENSOR_ADVERTISEMENT;   break;
    case MOBILE:    globalBuffer[0] = MOBILE_ADVERTISEMENT;   break;
    case HANDHELD:  globalBuffer[0] = HANDHELD_ADVERTISEMENT; break;
  }
  
  
  MEMCOPY(&(globalBuffer[1]), emberGetEui64(), EUI_LEN);
  //add the node type information mesher or leaf-node
  //String_Copy(node_type, 0, globalBuffer, DATA_START_INDEX+SER_NUM_LEN, 1);
  MEMCOPY(&globalBuffer[9], &Cipher_Params, sizeof(commission_t));
  
  buffer = emberFillLinkedBuffers(globalBuffer, DATA_START_INDEX + sizeof(commission_t)); //9 = msg type + eui64, +7
  
  // check to make sure a buffer is available
  if (buffer == EMBER_NULL_MESSAGE_BUFFER) 
  {
    //emberSerialGuaranteedPrintf(SERIAL_PORT,"No buffer");
    RStatus.bConnected = FALSE;
    return;
  }
    
  // send the message
  status = emberSendDatagram(CU_BINDING_INDEX, buffer);
 //emberSerialGuaranteedPrintf(SERIAL_PORT,"status is %x\r\n",status);  //debug only
  // done with the packet buffer
  emberReleaseMessageBuffer(buffer);
}

void RespondNetworkUpdate(void)
{
  EmberStatus status;
  EmberMessageBuffer buffer = 0; 
  
  //build a responce to send to the CU  NODE_CAN_I_BIND
  // the data
  globalBuffer[0] = NETWORK_UPDATE_OK;
  MEMCOPY(&(globalBuffer[1]), emberGetEui64(), EUI_LEN);
  MEMCOPY(&globalBuffer[9], &Cipher_Params, sizeof(commission_t));
    
  /// copy the data into a packet buffer    1+11+12+7+2
  buffer = emberFillLinkedBuffers(globalBuffer, DATA_START_INDEX + sizeof(commission_t)); //9 = msg type + eui64, +7
 
  // check to make sure a buffer is available
  if (buffer == EMBER_NULL_MESSAGE_BUFFER) 
  {
    //emberSerialGuaranteedPrintf(SERIAL_PORT,"No buffer");
    return;
  }
    
  // send the message
  status = emberSendDatagram(CU_BINDING_INDEX, buffer);
  // done with the packet buffer
  emberReleaseMessageBuffer(buffer);
  //emberSerialGuaranteedPrintf(SERIAL_PORT,"Sent update signal\n\r");  //debug only
}






















