/* File: cipher-rf.h
 * 
 * DESCRIPTION:
 * This is a sample AMR (Automated Meter Reading) application
 * That is written in such a way as to highlight the ease of porting
 * a node application to a host processor using (RNAP) Remote Network Access
 * Protocol and the supporting Host API. 
 * 
 * RUNNING THE AGGREGATOR APPLICATION:
 * This application consists of two components, The AMR Aggregator and the
 * AMR Meter. The Aggregator represents a host application which collects
 * data from a network of meters. The Meter represents a node connected to
 * some type of data generating meter device. Obviously since a meter application
 * deployed on a test network is not expected to actually be connected to a real
 * meter, false data is generated by the meter application to represent
 * meter data.
 *
 * A make file, MakeHostSample is provided to compile the AMR Aggregator. To
 * compile the aggregator, copy the MakeHostSample file to the root of the 
 * ember directory and use "make -f MakeHostSample". This will create a host
 * aggregator application called "amr-host." The amr-host application takes
 * two arguments, address and port of the rnap gateway it will use to connect
 * to the Ember Network.
 * 
 * When the host application runs, it sends out periodic multicasts onto the network
 * advertising itself as an amr aggregator. Any meters that hear this multicast will
 * begin to send periodic datagrams to the aggregator with meter data. Once meters begin
 * connecting to the aggregator, a user can pull data at will from each of them using
 * the data request command at the command line.
 *
 * USING THE COMMAND INTERPRETER IN THE AGGREGATOR APPLICATION:
 * The amr-aggregator includes a command interpreter. To see all possible commands
 * type 'help' or simply 'h' at the command line. 
 *
 * Commands listed below are in the form, (long command) (short command) (parameters) : (Explanation)
 * 
 * status () (): 
 *      Prints out the EUI64 ID, radio channel, and radio power of the connected
 *      gateway node.
 *
 * set-channel () (int channel): 
 *      Changes the current radio channel of the connected gateway node.
 *
 * set-power () (int power): 
 *      Changes the current radio power of the connected gateway node.
 *
 * set-period (p) (int seconds):
 *      Set the number of seconds between multicast advertisements
 *      from the aggregator to pick up new meters. You may wish to 
 *      set this to a higher value once the aggregator has picked up
 *      all the meters so as to reduce network traffic.
 * 
 * set-heartbeat (l) (int boolean): 
 *      Turns the aggregator's heartbeat on and off. On the host
 *      this simply prints a '*' once a second. This command takes a single
 *      argument either 1 or 0 indicating on or off respectively.
 *                    
 * print-meters (m) (): 
 *      Prints out a list of all the meters that the aggregator has found
 *      along with their EUI64 IDs.
 *
 * print-bindings (b) (): 
 *      Prints out the entire binding table. On a host aggregator this list
 *      will be quite large since the binding table is itself very large.
 *
 * request-data (d) (int meter id): 
 *      Requests an immediate report from the meter indicated 
 *      The meter id is not the same as the EUI64 Id, it is related to the 
 *      order in which that meter was discovered. All of the meter ids can be seen
 *      using the print-meters command above. This command highlights the ability 
 *      to send a datagram to any number of meters from the host aggregator.
 *
 * request-history (s) (int meter id): 
 *      Requests an immediate report of historical meter data from the meter
 *      This is intended to show how larger message types called sequenced
 *      messages can be sent to and from the host aggregator from any one
 *      of the discovered meters.
 *
 * help (h) ():
 *      The help command, prints out a list of all the available commands
 *      along with a brief discription of how and why they are used.
 *
 * quit (q) ():
 *      Causes the host aggregator to shut down.
 * 
 * UNDERSTANDING THE AGGREGATOR APPLICATION:
 * The AMR Aggregator consists of a single core file "amr-aggregator.c" and
 * also shares functionality with the amr-meter defined in this file, amr.h and 
 * also amr-common.c.
 *
 * The AMR Aggregator is written to compile either for the node or the host
 * depending on the existence of a single definition. If PLATFORM_HOST
 * is defined, the amr-aggregator expects that it is being compiled for a
 * host processor and will include all of the appropriate files related to 
 * the host-api. If PLATFORM_HOST is not defined, the amr-aggregator expects that
 * it is being compiled to run on a node and will include the appropriate files from
 * the EmberAPI.
 *
 * This build process highlights the degree to which the host-api abstracts the differences between
 * the node and the host making it very easy for a developer to port a node application
 * over to a host processor connected to an RNAP Gateway.
 *
 * The few differences which for various reasons could not be abstracted have been
 * highlighted in the code by the words "***NODE / HOST VERSION DIFFERENCE***". Each
 * instance of this comment includes an explanation of why the differntiated code
 * was necessary. 
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#ifndef AMR_H
#define AMR_H

#define EUI_LEN 8
#define DATA_START_INDEX 9
#define SER_NUM_LEN 8 //8 for ascii 3 for binary


//the serial port used for interaction with the host and the gateway-node
#define SERIAL_PORT     1

//the multicast group for this specific application. The meters will listen
//to multicasts of 
#define MULTICAST_GROUP          0xA555

//the terminal shared by the aggregator and the meter for aggregation interaction
#define ourTerminal          emberMakeTerminal(1, 0)

//the default radio channel and power used by this application
#define CIPHER_SAMPLE_CHANNEL  0
#define CIPHER_SAMPLE_POWER -1
#define CU_BINDING_INDEX (0)

// For identifying ourselves over the serial port.
extern PGM_NO_CONST PGM_P applicationString;

#define MAXBLOCK_SIZE   255
#define MAXCMD_SIZE     90
int8u globalBuffer[90];
int8u MonsterBlock[MAXBLOCK_SIZE+1];
int8u CmdBuff[MAXCMD_SIZE+1];

//typedef struct {		//this will not be used for the CU look in Device-management.h
//        int8u         Model;          //for nick
//        int8u         DeviceType;     //SENSOR,MOBILE,HANDHELD
//		int8u         bMesher;	      //Whether a constant mesher
//		int8u         SerialNum[11];
//		int8u         ModelStr[12];
//		int8u         FirmwareStr[7]; //Nigels
//		int8u         RFFirmwareStr[7];//Mine
//	} commission_t;

//commission_t Cipher_Params;				moved to cipher-meter.h

//These are application level definitions used
//in interaction within the amr application 
enum {
  //CU sends this message to indicate it is avaliable to any "un-bound" sensors
  //Multicast Message
  CU_ADVERTISEMENT,
  //CU sends this message to sensor after validating App Serial number indicating the i should be part of our Network
  //Datagram message
  CU_OKBIND,
  //Sensor sends this message which will have serial number etc asking to be validated
  //Datagram Message
  SENSOR_ADVERTISEMENT,
  //Multicast message sent from CU to initiate a responce from all nodes that hear it to send information back to 
  //update the network
  NETWORK_UPDATE,
  //Unicast message from device to CU
  //
  NETWORK_UPDATE_OK,
  //Mobile device sends this message which will have serial number etc asking to be validated
  //Datagram Message
  MOBILE_ADVERTISEMENT,
  //Handheld device sends this message which will have serial number etc asking to be validated
  //Datagram Message
  HANDHELD_ADVERTISEMENT,
  //--------------------------------------------------------------------------------------------------------
  //The following are datagram messages to the CU  Data packets!!!!
  //
  SENSOR_DATAPACKET_START,
  //
  //
  SENSOR_DATAPACKET_2,
  //
  //
  SENSOR_DATAPACKET_3,
  //
  //
  SENSOR_DATAPACKET_4,
  //
  //
  MOBILE_DATAPACKET_START,
  //
  //
  MOBILE_DATAPACKET_2,
  //
  //
  MOBILE_DATAPACKET_3,
  //
  //
  MOBILE_DATAPACKET_4,
  //
  //
  HANDHELD_DATAPACKET_START,
  //
  //
  HANDHELD_DATAPACKET_2,
  //
  //
  HANDHELD_DATAPACKET_3,
  //
  //
  HANDHELD_DATAPACKET_4,
  //
  //
  NODE_INCOMMING_CMD,
  //
  //Message from CU to individual transmitter telling it to remove itself from the system
  CU_REMOVE,
  //
  //Messgae telling node there are cmd's waiting for it when it is its turn to go one on one with the CU
  NODE_CMD_WAITING,
  //
  //
  LEAFNODE_ADVERTISEMENT,
};


#endif //AMR_H

