//******************************************************************************
//  connection-app.c
//
//  Sample application to show operation of connections and sequenced messages.
//
//  This sample application will show some basic usage of connection
//  functionality.
//  Given a group of nodes, one node will have a serial port connection for I/O.
//  This node will be referred to as a gateway.
//  On startup, the gateway will send a multcast datagram SEARCH message to all
//  nodes (remote nodes).  Each of the remote nodes will respond with a
//  multicast RESPONDER message announcing its presence.  The gateway will
//  create binding entries to the remote nodes, and each remote node will create
//  a binding entry to the gateway.  The multicast messaging is being used for
//  simplicity in this application.  Normally the node would reply on remote
//  provisioning being set up by a specialized provisioning node.
//  The gateway will accept serial commands to query a remote node for data.
//  The remote node will send random data to the gateway until the gateway tells
//  it to stop sending data.
//  The gateway will also accept a command to tell all the nodes to delete their
//  binding entry, and it will delete its binding entries to the nodes.
//
//  KEY INGREDIENTS:
//  - Definitions of interfaces and endpoints:
//      A symmetric Multicast endpoint used by all nodes
//      A Server endpoint - accepts messages from gateway
//      A Client endpoint - accepts messages from remote nodes
//  - Multicast bindings
//  - Finding remote nodes via multicast query (datagram message)
//  - Creating unicast bindings
//  - Opening connections
//  - Sending sequenced messages
//
//  This example also provides some simple commands that can be sent to the
//  node via the serial port. A brief explanation of the commands is
//  provided below:
//  ('h') prints the help menu
//  ('b') Put node in bootload mode
//  ('i') Print some node info out to the serial port
//  ('r') Reset the node
//  ('c channel') Set the radio channel
//  ('s') Search the network for nodes
//  ('p') Print this node's binding table
//  ('q binding') Query remote node to send data
//  ('e binding') End data from remote node
//  ('d') Delete bindings on all nodes
//
//  Copyright 2004 by Ember Corporation. All rights reserved.               *80*
//******************************************************************************

#include <ioavr.h>
#include PLATFORM_HEADER //compiler/micro specifics, types
#include "stack/include/ember.h"
#include "stack/include/error.h"
#include "stack/include/hal.h" // firmGetRandom()
#include "hal/board/board.h"   // LED definitions
#include "hal/micro/micro.h"   // microBoot, memcopy, sleep, launch bootloader
#include "stack/include/packet-buffer.h"
#include "app/util/serial/serial.h"

// ************************************
// Ember static memory requirements
//
// There are constants that define the amount of static memory
// required by the stack. If the application does not set them then
// the default values from ember-static-configuration.h are used.
//
// We need up to 12 bindings:
// 1 for multicast messaging
// 11 bindings on the gateway to other nodes
#define EMBER_BINDING_TABLE_SIZE        12
// all entries in binding table are temporary (in RAM, not flash)
#define EMBER_TEMPORARY_BINDING_ENTRIES 12
#include "stack/include/ember-static-configuration.h"

// ************************************
// Include definitions for application timers
#include "app-timers.h"

// ************************************
// Serial port definitions.  We define the debug serial port because it needs
// to be initialized in case the debug stack library is being used.
#define DEBUG_SERIAL 0
#define IO_SERIAL 1

//******************************************************************************
// Define endpoints and interfaces for the Ember Stack
// Design:
// We will use one endpoint.
// This endpoint will have the following interfaces:
// 1) A fixed multicast interface for general messages
// 2) A server interface (for a controlled device)
// 3) A client interface (for the gateway)

// Define interfaces
// (indices into the list of interfaces in the endpoint structure)
#define MULTICAST_INTERFACE 0
#define SERVER_INTERFACE    1
#define CLIENT_INTERFACE    2

// Define endpoints.
#define DEVICE_CONTROL_ENDPOINT   1

// Specification IDs for the ad-hoc interfaces
#define MULTICAST_SPEC              0x56
#define DEVICE_CONTROL_INTERFACE_ID 0x71

// Define the interface structures
// Ad-hoc interfaces contain: interface domain, endpoint type,
// #-specification-bytes and specification (unique identifier)

// Multicast interface, so all devices can talk to each other.
static PGM EmberInterfaceId mcastInterfaceId = {
  AD_HOC_INTERFACE,
  SYMMETRIC_ENDPOINT,
  1, MULTICAST_SPEC };

// Server interface, for controlled devices (many)
static PGM EmberInterfaceId deviceControlServerInterfaceId = {
  AD_HOC_INTERFACE,
  SERVER_ENDPOINT,
  1, DEVICE_CONTROL_INTERFACE_ID };

// Client interface, for the gateway
static PGM EmberInterfaceId deviceControlClientInterfaceId = {
  AD_HOC_INTERFACE,
  CLIENT_ENDPOINT,
  1, DEVICE_CONTROL_INTERFACE_ID };


static PGM EmberEndpointDescription deviceControlEndpoint = {
  { MULTICAST_INTERFACE, &mcastInterfaceId },
  { SERVER_INTERFACE,    &deviceControlServerInterfaceId },
  { CLIENT_INTERFACE,    &deviceControlClientInterfaceId }
};


// Global storage required by the Ember stack
// This tells the stack we're using only one endpoint:
int8u emberEndpointCount = 1;

// This defines the endpoints we're using:
EmberEndpoint emberEndpoints[] = {
  { DEVICE_CONTROL_ENDPOINT, &deviceControlEndpoint}
};


//-------------------------------------
// Definitions for binding table entries
#define MAX_BINDING_INDEX   (EMBER_BINDING_TABLE_SIZE - 2)
#define MCAST_BINDING_INDEX (EMBER_BINDING_TABLE_SIZE - 1)
#define GATEWAY_BINDING_INDEX 0  // index of binding to gateway

// Multicast address used for binding entry
#define MCAST_ADDRESS 0x0384


//-------------------------------------
// Message types
typedef enum {
  // Multicast messages
  SEARCHER     = 0x10, // Gateway sends to find devices in network
  NET_DOWN     = 0x11, // Gateway sends to tell all nodes to delete bindings
                       //  to gateway

  // Unicast messages from remote nodes to gateway
  RESPONDER    = 0x18, // Remote nodes send to identify themselves to gateway

  // Unicast datagram messages from gateway to remote nodes
  QUERY_DEVICE = 0x20, // Gateway asking for data
  STOP_REPORT  = 0x21, // Gateway asking node to stop sending data

  // Unicast sequenced messages from node to gateway
  DATA_REPORT  = 0x30  // Data messages from node to gateway
} DeviceControlMessageType;


//******************************************************************************
// Application-specific constants and globals

// Current status of the stack
static EmberStatus stackStatus = EMBER_NETWORK_DOWN;

static int8u myChannel = 0;
static int8s myPower = -1;

// Holds count of nodes that respond to node search (gateway)
static int8u nodesFound = 0;

// Flag indicating a gateway exists on the network
static boolean gatewayExists = FALSE;

// Flag indicating this node is the gateway
static boolean amGateway = FALSE;

// Flag indicating we have a connection open to the gateway
static boolean connectedToGateway = FALSE;

// structure to store necessary network parameters of the node
// (which are panId, enableRelay, radioTxPower, and radioChannel)
EmberNetworkParameters networkParams;

//******************************************************************************

// Forward declarations
static void processSerialInput(void);
static boolean isDuplicateBinding(EmberBindingTableEntry* newEntry);
static void blinkLEDremote(int8u timerId, int32u param);
static void blinkLEDGateway(int8u timerId, int32u param);
static void endNodeSearch(int8u timerId, int32u param);
static void sendStreamingData(int8u timerId, int32u param);

//******************************************************************************

// Actions grouped together because they need to be done with emberTick
static void runSystem(void)
{
  __watchdog_reset();
  emberTick();
  emberSerialBufferTick();
  handleTimers();
}

//******************************************************************************
// Main application loop

int main(void)
{
  EmberResetType reset;
  EmberBindingTableEntry entry;
  EmberStatus status;
  
  // Determine the cause of the reset (powerup, etc).
  reset = microGetResetInfo();
  microBoot();

  // to init the leds and such..
  boardBoot();

  // reset the system tick
  // enable the system tick interrupt
  microStartSystemTick();
  // allow interrupts
  INTERRUPTS_ON();

  // inititialize the serial port
  emberSerialInit(IO_SERIAL, BAUD_19200, PARITY_NONE, 1);
  emberDebugInit(DEBUG_SERIAL);

  // Initialize the Ember Stack.
  status = emberInit(reset);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, 
      "error returned from emberInit: 0x%x\r\n",
      status);
    if (status == EMBER_UNINITIALIZED_EUI64) {
      // Run to allow over-the-air configuration.
      while (TRUE) {
        __watchdog_reset();
        emberTick();
      }
    } else {
      assert(FALSE);
    }
  }

  initTimers();

  // Bring the network up.
  networkParams.panId = 0;
  networkParams.enableRelay = TRUE;
  networkParams.radioTxPower = myPower;
  networkParams.radioChannel = myChannel;
  
  status = emberAssociate(&networkParams);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, 
      "error returned from emberAssociate: 0x%x\r\n",
      status);
    assert(FALSE);
  }
  
  //-------------------------------------------------------

  // Wait for stack to come up before doing anything
  while (stackStatus != EMBER_NETWORK_UP)
    runSystem();

  //-------------------------------------------------------

  boardLEDClear(BOARDLED0);
  boardLEDClear(BOARDLED1);
  boardLEDClear(BOARDLED2);
  boardLEDSet(BOARDLED3);
  // Create timer to blink LEDs for heartbeat
  // Blink every 1/2 second (8 clock ticks)
  addTimer(0, blinkLEDremote, 0, 8, 1);

  emberSerialPrintf(IO_SERIAL,
              "\r\nConnection sample app running on node %X%X, channel %d\r\n",
              emberGetEui64()[1], emberGetEui64()[0], myChannel);
  emberSerialPrintf(IO_SERIAL, "Press 'h' for menu of commands\r\n");

  //-------------------------------------------------------

  // make the binding table entry
  entry.type = EMBER_MULTICAST_BINDING;
  entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT, MULTICAST_INTERFACE);
  *(int16u *)entry.identifier = MCAST_ADDRESS;
  if (emberSetBinding(MCAST_BINDING_INDEX, &entry) != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL, "Failed to make binding\r\n");

  //-------------------------------------------------------

  while (TRUE) {
    runSystem();
    // Don't run anything else unless the stack is running
    if (stackStatus == EMBER_NETWORK_UP) {
      processSerialInput();
    }
  }
}

//******************************************************************************
// Application Timer functions
// These functions get set up by code below to execute after a certain delay.
//******************************************************************************

// This will blink a red LED to indicate this is not the gateway node.
static void blinkLEDremote(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED3);
}

//---------------------------------------------------------

// This will blink a green LED to indicate this node is the gateway.
static void blinkLEDGateway(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED2);
}

//---------------------------------------------------------

// This timer function is how the gateway ends its search for remote nodes.
// It is set up when the search is started.
static void endNodeSearch(int8u timerId, int32u param)
{
  emberSerialGuaranteedPrintf(
    IO_SERIAL, "ending search,%d nodes found\r\n", nodesFound);

  if (nodesFound) {
    // Guard against multiple gateways
    gatewayExists = TRUE;
    amGateway = TRUE;
  
    // Visually denote the gateway
    if (nodesFound) {
      boardLEDClear(BOARDLED0);
      boardLEDClear(BOARDLED1);
      boardLEDClear(BOARDLED2);
      boardLEDClear(BOARDLED3);
      addTimer(0, blinkLEDGateway, 0, 8, 1);
    }
  }
}

//---------------------------------------------------------

// This timer function is set up to run by a remote node when the gateway tells
// it to start sending data.  The point is to keep sending data messages to the
// gateway until we're told to stop.
static void sendStreamingData(int8u timerId, int32u param)
{
  int8u message[64];
  EmberMessageBuffer pbuff;
  int8u i;

  if (connectedToGateway) {
    // Fill data message
    message[0] = DATA_REPORT;
    for (i = 1; i < sizeof(message); i++) {
      message[i] = i;
    }
    pbuff = emberFillLinkedBuffers(message, sizeof(message));

    emberSendSequenced(GATEWAY_BINDING_INDEX, pbuff);
    emberReleaseMessageBuffer(pbuff);
  }
  else {
    emberSerialGuaranteedPrintf(IO_SERIAL, "Not connected\r\n");
    // Retry opening connection to gateway
    emberOpenConnection(GATEWAY_BINDING_INDEX);
  }

  // Reschedule next set of data
  addTimer(timerId, sendStreamingData, param, param, 0);
}

//---------------------------------------------------------

// Send search response back to gateway, with our EUI64 ID.
// This is a timer function because we set a random timeout to perform this
// action, so that remote nodes respond at different times.
static void respondToGateway(int8u timerId, int32u param)
{
  EmberMessageBuffer responseMessage;
  EmberStatus status;
  int8u messageBuffer[9];

  messageBuffer[0] = RESPONDER;
  MEMCOPY(messageBuffer + 1, emberGetEui64(), 8);

  responseMessage =
    emberFillLinkedBuffers(messageBuffer, sizeof(messageBuffer));

  status = emberSendDatagram(GATEWAY_BINDING_INDEX, responseMessage);
  if (status != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "SendDatagram failed: %x\r\n", status);

  emberReleaseMessageBuffer(responseMessage);
}

//******************************************************************************
// Utility functions
//******************************************************************************

// The user will want to know which binding entries to use.  This is most
// useful on the gateway, or for debugging a remote node (to make sure it
// has opened a connection to the gateway).
void printBindingTable(void)
{
  int8u i;
  EmberStatus status;
  EmberBindingTableEntry result;
  int8u* id;
  int8u indexLow;
  int8u indexHigh;
  emberSerialPrintf(IO_SERIAL, "status index type local remote id\r\n");

  for (i=0; i<EMBER_BINDING_TABLE_SIZE; i++) {
    emberSerialWaitSend(IO_SERIAL);

    status = emberGetBinding(i, &result);
    
    indexLow = (i % 10) + 48;
    indexHigh = ((i - (i % 10))/10) + 48;

    id = result.identifier;

    emberSerialPrintf(IO_SERIAL, "0x%x  | %c%c | ", 
                      status, indexHigh, indexLow);

    if (EMBER_SUCCESS == status) {
      switch(result.type) {
      case EMBER_UNUSED_BINDING:
        emberSerialPrintf(IO_SERIAL, "unused....|");
        break;
      case EMBER_UNICAST_BINDING:
        emberSerialPrintf(IO_SERIAL, "unicast...|");
        break;
      case EMBER_MULTICAST_BINDING:
        emberSerialPrintf(IO_SERIAL, "multicast.|");
        break;
      default:
        emberSerialPrintf(IO_SERIAL, "?????????.|");
      }

      emberSerialPrintf(IO_SERIAL, "0x%x  0x%x  ",
                        result.local, result.remote);

      emberSerialPrintf(IO_SERIAL, "%x %x %x %x %x %x %x %x\r\n",
                        id[0],id[1],id[2],id[3],id[4],id[5],id[6],id[7]);
      
    } else {
      emberSerialPrintf(IO_SERIAL, " --- ERROR ---\r\n ");
    }
  }
  emberSerialWaitSend(IO_SERIAL);
}

//---------------------------------------------------------

// This function is run by the gateway to send a multicast message to nodes in
// the area to get them to respond, so that the gateway can set bindings to them.
static void startNodeSearch()
{
  EmberMessageBuffer searchMessage;
  EmberStatus status;
  int8u messageBuffer[9];

  nodesFound = 0;

  messageBuffer[0] = SEARCHER;
  MEMCOPY(messageBuffer + 1, emberGetEui64(), 8);
  searchMessage = emberFillLinkedBuffers(messageBuffer, sizeof(messageBuffer));

  status = emberSendMulticast(MCAST_BINDING_INDEX, searchMessage);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "SendMulticast failed: %x", status);
  }

  // we are done with the buffer
  emberReleaseMessageBuffer(searchMessage);

  // Create search timeout - 56 => 3.5 seconds
  addTimer(1, endNodeSearch, 1, 56, 0);
}

//---------------------------------------------------------

static boolean isDuplicateBinding(EmberBindingTableEntry* newEntry)
{
  int8u i;
  EmberBindingTableEntry existingEntry;
  for (i = 0; i < EMBER_BINDING_TABLE_SIZE; i++) {
    // If error, just return that we found a duplicate entry,
    // so we don't try to add a binding entry with an invalid index.
    if (emberGetBinding(i, &existingEntry) == EMBER_BINDING_INDEX_OUT_OF_RANGE)
      return TRUE;
    if (MEMCOMPARE(&existingEntry,
                   newEntry,
                   sizeof(EmberBindingTableEntry)) == 0) {
      return TRUE;
    }
  }
  return FALSE;
}

//---------------------------------------------------------

// Gateway sends a datagram to a remote node, to get it to start sending data.
void gatewayQueryRemoteDevice(int8u bindingIndex)
{
  EmberMessageBuffer pbuff;
  int8u message[] = {QUERY_DEVICE};
  EmberStatus status;

  pbuff = emberFillLinkedBuffers(message, sizeof(message));

  status = emberSendDatagram(bindingIndex, pbuff);
  if (status != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "Failed to send query msg: %X\r\n", status);
  emberReleaseMessageBuffer(pbuff);
}

//---------------------------------------------------------

// Sent by the gateway to a remote node, to have it stop sending data.
void gatewayEndQueryRemoteDevice(int8u bindingIndex)
{
  EmberMessageBuffer pbuff;
  int8u message[] = {STOP_REPORT};
  EmberStatus status;

  pbuff = emberFillLinkedBuffers(message, sizeof(message));

  status = emberSendDatagram(bindingIndex, pbuff);
  if (status != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "Failed to send end query msg, %X\r\n", status);
  emberReleaseMessageBuffer(pbuff);
}

//---------------------------------------------------------

// Executed by a remote node when it's told by the gateway to stop sending data,
// or when it's told to stop network interactions and delete all bindings.
static void stopStreamingData(void)
{
  // End streaming data to gateway
  addTimer(2, NULL, 0, 0, 0);

  if (connectedToGateway) {
    // Delete connection to gateway
    emberCloseConnection(GATEWAY_BINDING_INDEX);
    connectedToGateway = FALSE;
  }
}

//---------------------------------------------------------

// The gateway sends a multicast message telling all nodes to delete their
// binding table entries (and also stop sending data).
void gatewayShutdownRemoteDevices(void)
{
  EmberMessageBuffer pbuff;
  int8u message[] = {NET_DOWN};
  EmberStatus status;

  pbuff = emberFillLinkedBuffers(message, sizeof(message));

  status = emberSendMulticast(MCAST_BINDING_INDEX, pbuff);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "SendMulticast failed: %x", status);
  }

  emberReleaseMessageBuffer(pbuff);
}

//---------------------------------------------------------

// Run by the gateway to delete all binding entries.
void deleteAllLocalBindings(void)
{
  int8u i;

  // Delete bindings
  for (i = 0; i < nodesFound; i++) {
    emberDeleteBinding(i);
//    MEMSET(&devicesFound[i], 0, sizeof(EmberEUI64));
    emberSerialGuaranteedPrintf(IO_SERIAL, "local binding %d deleted\r\n", i);
  }
  // Last: clear binding counter
  nodesFound = 0;
}

//******************************************************************************
// Interface handlers (called from emberIncomingMessageHandler)
//******************************************************************************

// Called by emberIncomingMessageHandler to handle EMBER_INCOMING_DATAGRAM
static void incomingDatagramHandler(int8u localTerminal,
                                    int8u bindingTableIndex,
                                    EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  emberSerialGuaranteedPrintf(IO_SERIAL, "Datagram: ");
  if (emberTerminalInterface(localTerminal) == SERVER_INTERFACE) {
    //-----------------------------------------------------
    // Unicast messages received by a remote node
    //-----------------------------------------------------
    if (data[0] == QUERY_DEVICE) {
      emberSerialGuaranteedPrintf(
        IO_SERIAL, "query command from gateway.\r\n");

      // Open connection to gateway
      emberOpenConnection(GATEWAY_BINDING_INDEX);

      // Start streaming data to gateway
      // Set it up as a self-rescheduling timer application
      // Timer 2, parameter = 16 (# ticks, or 1 second)
      // Start process after 2 seconds
      addTimer(2, sendStreamingData, 16, 32, 0);
    //-----------------------------------------------------
    } else if (data[0] == STOP_REPORT) {
      emberSerialGuaranteedPrintf(
        IO_SERIAL, "stop report command from gateway.\r\n");

      stopStreamingData();
    //-----------------------------------------------------
    }
    else  // First byte of message unknown
      emberSerialGuaranteedPrintf(
        IO_SERIAL,
        "unknown message [%x] from binding %d\r\n",
        data[0], bindingTableIndex);
    //-----------------------------------------------------
  } else if (emberTerminalInterface(localTerminal) == CLIENT_INTERFACE) {
    //-----------------------------------------------------
    // Unicast messages received by the gateway
    //-----------------------------------------------------
    if (data[0] == RESPONDER) {
      EmberBindingTableEntry entry;
      EmberStatus status;

      emberSerialGuaranteedPrintf(
        IO_SERIAL, "response from %x%x\r\n", data[2], data[1]);

      // Set bindings to remote nodes as they respond.
      entry.type = EMBER_UNICAST_BINDING;
      MEMCOPY(entry.identifier, &data[1], 8);
      entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                      CLIENT_INTERFACE);
      entry.remote = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                       SERVER_INTERFACE);

      // Guard against duplicate binding table entries.
      if (isDuplicateBinding(&entry)) {
        emberSerialGuaranteedPrintf( IO_SERIAL, "already have binding set\r\n");
      } else {
        // Set a binding for ourselves to that controller
        status = emberSetReplyBinding(nodesFound, &entry);
        if (status == EMBER_SUCCESS) {
          emberSerialGuaranteedPrintf(IO_SERIAL,
            "binding %d set\r\n", nodesFound);
          nodesFound += 1;
        }
      }
    }
    //-----------------------------------------------------
  } else {
    emberSerialGuaranteedPrintf(
      IO_SERIAL,
      "message [%x] from binding %d, bad terminal\r\n",
      data[0], bindingTableIndex);
  }
}

//---------------------------------------------------------

// Called by emberIncomingMessageHandler to handle EMBER_INCOMING_MULTICAST
static void incomingMulticastHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  if (emberTerminalInterface(localTerminal) == MULTICAST_INTERFACE) {
    //-----------------------------------------------------
    if (data[0] == SEARCHER) {
      EmberBindingTableEntry entry;
      EmberStatus status;

      emberSerialGuaranteedPrintf(
        IO_SERIAL, "search from %x%x\r\n", data[2], data[1]);

      // Guard against multiple gateways
      gatewayExists = TRUE;
      amGateway = FALSE;

      // Create binding to gateway
      // We're using emberSetReplyBinding because at this point, the
      // stack doesn't have to rediscover the original sender.
      entry.type = EMBER_UNICAST_BINDING;
      MEMCOPY(entry.identifier, &data[1], 8);
      entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                      SERVER_INTERFACE);
      entry.remote = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                       CLIENT_INTERFACE);
      status = emberSetReplyBinding(GATEWAY_BINDING_INDEX, &entry);
      if (status == EMBER_SUCCESS) {
        emberSerialGuaranteedPrintf(IO_SERIAL, "binding set\r\n");

        // Schedule a response back to the gateway
        // The timeout will be random: 0 thru 3.1 seconds
        addTimer(1, respondToGateway, 0, (firmGetRandom() & 0x1F), 0);
      } else {
        emberSerialGuaranteedPrintf(IO_SERIAL, "binding failed(%x)\r\n", status);
      }
    //-----------------------------------------------------
    } else if (data[0] == NET_DOWN) {
      // Make sure streaming data to gateway is stopped
      stopStreamingData();

      // Delete binding to gateway
      emberDeleteBinding(GATEWAY_BINDING_INDEX);
      emberSerialGuaranteedPrintf(IO_SERIAL, "binding deleted\r\n");

      // Now there's no gateway.
      gatewayExists = FALSE;
      amGateway = FALSE;
      connectedToGateway = FALSE;
    }
    //-----------------------------------------------------
    else  // First byte of message unknown
      emberSerialGuaranteedPrintf(IO_SERIAL,
                        "Unknown mcast message from binding %x\r\n",
                        bindingTableIndex);
    //-----------------------------------------------------
  }
  else { // Not multicast
    emberSerialGuaranteedPrintf(IO_SERIAL,
                      "Unknown message (not mcast) from binding %x\r\n",
                      bindingTableIndex);
  } // end if(MULTICAST_INTERFACE)
}

//******************************************************************************
// Handlers required to use the Ember Stack
// Populate as needed
//******************************************************************************

void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 int8u localTerminal,
                                 int8u bindingTableIndex,
                                 EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  if (!emberMessageBufferLength(message))
    return;

  switch (type) {
  case EMBER_INCOMING_DATAGRAM:
    incomingDatagramHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_MULTICAST:
    incomingMulticastHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_SEQUENCED:
    emberSerialGuaranteedPrintf(IO_SERIAL,
                      "Received data from binding %x\r\n",
                      bindingTableIndex);
    break;

  case EMBER_INCOMING_DATAGRAM_REPLY:
  case EMBER_INCOMING_SHARED_MULTICAST:
  case EMBER_INCOMING_MULTICAST_LOOPBACK:
  default:
    emberSerialGuaranteedPrintf(
      IO_SERIAL,
      "Unexpected message (type %d) [ %x] binding %d, for terminal %d\r\n",
      type, data[0], bindingTableIndex, localTerminal);
    break;
  }
}

//---------------------------------------------------------

void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status)
{
  emberSerialGuaranteedPrintf(
    IO_SERIAL, "Connection to binding %d: %x\r\n", bindingTableIndex, status);

  connectedToGateway = (status == EMBER_CONNECTION_OPEN);
}

//---------------------------------------------------------

void emberMessageSent(int8u bindingTableIndex,
                      EmberMessageBuffer message,
                      EmberStatus status)
{
  int8u *data = emberMessageBufferContents(message);
  emberSerialGuaranteedPrintf(
    IO_SERIAL,
    "Sent message [%x] from binding %d, status = 0x%x\r\n",
    data[0], bindingTableIndex, status);
}

//---------------------------------------------------------

EmberStatus emberRemoteSetBindingHandler(int8u index,
                                         EmberBindingTableEntry *entry)
{
  // Don't let anyone change our bindings.
  return EMBER_INVALID_BINDING_INDEX;
}

//---------------------------------------------------------

void emberEndpointFoundHandler(EmberInterfaceId *desiredInterface,
                               EmberEUI64 node,
                               EmberNetworkDistance distance)
{
}

//---------------------------------------------------------

void boardButtonCallback(int8u button, int8u state)
{
}

//---------------------------------------------------------

void emberStackStatusHandler(EmberStatus status)
{
  stackStatus = status;
}

//******************************************************************************
// End of Handlers
//******************************************************************************


//******************************************************************************
// Serial command interface functions
//******************************************************************************

static void getParams(char *line, int8u length, int32u *argv, int8u *argc)
{
  int8u i;
  int32u temp = 0;
  int8u count = 0;
  for (i = 0; i < length; i++) {
    if ((line[i] >= '0') && (line[i] <= '9')) {
      temp = (temp * 10) + (line[i] - '0');
    } else if ((line[i] == 0x20) || (line[i] == 0x09)) {
      argv[count++] = temp;
      temp = 0;
    } else {
      count--;
      break; // die on first invalid character
    }
  }
  argv[count++] = temp;
  *argc = count;
}

#define LINE_BUFFER_MAX 30
static int8u lineLen = 0;
static char lineBuffer[LINE_BUFFER_MAX];

static void processSerialInput(void) {
  int8u ch;
  int8u cmd = '\0';
  int32u argv[2] = {0, 0};  // Handle up to 2 numeric arguments
  int8u argc = 0;
  EmberStatus status;

  if (emberSerialReadByte(IO_SERIAL, &ch) != EMBER_SUCCESS) return;

  if (ch == 0x0A)
    return;
  else if ((ch == 0x0D) || (lineLen >= LINE_BUFFER_MAX)) {
    cmd = lineBuffer[0];
  } else if (ch == 0x08) { // Backspace
    lineLen--;
    lineBuffer[lineLen] = '\0';
    emberSerialPrintf(IO_SERIAL," \010");
    return;
  } else {
    lineBuffer[lineLen++] = ch;
    return;
  }

  if (lineLen >= 2) {
    getParams(&lineBuffer[2], lineLen-2, argv, &argc);
    lineBuffer[0] = '\0';
  }

  emberSerialPrintf(IO_SERIAL,"\r\n");

  switch(cmd) {
  // bootload
  case 'b':
    emberSerialGuaranteedPrintf(IO_SERIAL, "starting bootloader...\r\n");
    microLaunchBootloader(MICRO_LAUNCH_BOOTLOADER_KEY);
    break;

  // info
  case 'i':
    myChannel = emberGetRadioChannel();
    myPower = emberGetRadioPower();
    emberSerialPrintf(IO_SERIAL,
                "\r\nConnection app running on node %X%X, channel %d, power %d\r\n",
                emberGetEui64()[1], emberGetEui64()[0], myChannel, emberGetRadioPower());
    break;

  // reset using watchdog timeout
  case 'r':
    emberSerialPrintf(IO_SERIAL, "\r\nresetting...\r\n");
    while (TRUE) ;  // wait for watchdog timer reset
//    break;

  case 'c':
   if (argc == 1) {
      if(emberDisassociate() != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"Cannot disassociate from network\r\n");  
        return;
      } 
      // waiting for the stack to come down after disassociate from network
      while(stackStatus != EMBER_NETWORK_DOWN){
       __watchdog_reset();
       emberTick();
      }
      networkParams.radioChannel = argv[0];
      status = emberAssociate(&networkParams);
      if(status != EMBER_SUCCESS){
        emberSerialPrintf(IO_SERIAL,"can't set radio channel, 0x%x\r\n", status);
        while(1) ;
      }
      else {
        emberSerialPrintf(IO_SERIAL,"Set channel successfully, waiting for stack to come up");
        while(stackStatus != EMBER_NETWORK_UP){
          __watchdog_reset();
          emberTick();
        }
        
      }
    } else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

      //change power
    case 'w':
    if (argc == 1) {
      if(emberDisassociate() != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"Cannot disassociate from network\r\n");  
        return;
      } 
      // waiting for the stack to come down after disassociate from network
      while(stackStatus != EMBER_NETWORK_DOWN){
       __watchdog_reset();
       emberTick();
      }
      networkParams.radioTxPower = argv[0];
      status = emberAssociate(&networkParams);
      if(status != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"can't set radio power, 0x%x\r\n", status);
        while(1) ;
      }
      else {
        emberSerialPrintf(IO_SERIAL,"Set power successfully, waiting for stack to come up");
        while(stackStatus != EMBER_NETWORK_UP){
          __watchdog_reset();
          emberTick();
        }
        
      }
    } else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

  // search
  case 's':
    if (!gatewayExists) {
      emberSerialPrintf(IO_SERIAL, "starting search\r\n");
      startNodeSearch();
    } else {
      emberSerialPrintf(IO_SERIAL, "A gateway already exists!\r\n");
    }
    break;

  case 'p':
    printBindingTable();
    break;

  // Query node to start sending data
  case 'q':
    if (argc == 1) // single argument: binding index
      gatewayQueryRemoteDevice(argv[0]);
    else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

  // Tell node to stop sending data
  case 'e':
    if (argc == 1) // single argument: binding index
      gatewayEndQueryRemoteDevice(argv[0]);
    else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

  case 'd':
    if (amGateway) {
      gatewayShutdownRemoteDevices();
  
      deleteAllLocalBindings();
  
      boardLEDClear(BOARDLED0);
      boardLEDClear(BOARDLED1);
      boardLEDClear(BOARDLED2);
      boardLEDSet(BOARDLED3);
      addTimer(0, blinkLEDremote, 0, 8, 1);
  
      // Now there's no gateway
      gatewayExists = FALSE;
      amGateway = FALSE;
      connectedToGateway = FALSE;
    }
    break;

  //-----------------------------------------------------

  case 'h':
    emberSerialGuaranteedPrintf(IO_SERIAL, "\r\nNode %X%X, channel %d\r\n",
                emberGetEui64()[1], emberGetEui64()[0], myChannel);
    emberSerialGuaranteedPrintf(IO_SERIAL, "b: bootload\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "i: info\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "r: reset\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "c chan: set chan\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "w power: set power\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "s: node search\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "p: print binding table\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "q binding: Query remote node\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "e binding: End data from remote node\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "d: del bindings\r\n");
    break;

  //-----------------------------------------------------

  case '\0':
  case '\n':
  case '\r':
    break;

  default:
    emberSerialPrintf(IO_SERIAL, "unknown cmd\r\n");
    break;
  } // end switch

  cmd = 0;
  lineLen = 0;
  lineBuffer[0] = '\0';
  lineBuffer[1] = '\0';
}

//******************************************************************************

