//******************************************************************************
//  endpoint0-app.c
//
//  Sample application for endpoint 0 operation (DevKit)
//
//  This sample application will show some basic usage of Endpoint 0
//  functionality.
//  Given a group of nodes, when button #1 is pressed on one of them, it will
//  designate itself as a controlled device, and will set bindings on the other
//  nodes enabling them as controllers.  When button #1 is pushed on one of the
//  controller devices, it will cause the controlled device to beep and flash
//  an LED.  When button #2 is pushed on the controlled device, it will give up
//  being the controlled device, clearing the bindings on the other nodes.  The
//  process can be repeated with any node in the network.
//  
//  This example also provides some simple commands that can be sent to the
//  node via the serial port. A brief explanation of the commands is
//  provided below:
//  ('h') prints the help menu
//  ('b') Put node in bootload mode
//  ('i') Print some node info out to the serial port
//  ('r') Reset the node
//  ('c channel') Set the radio channel
//  ('p') Print binding table
//
//  KEY INGREDIENTS:
//  - Definitions of interfaces and endpoints:
//      A Server endpoint - this is used by the controlled device
//      A Client endpoint - this is used by the controller devices
//  - Search for interfaces:
//      When button #1 is pressed on a node, it searches the network for client
//      interfaces (emberStartEndpointSearch).  As responses come in, it will
//      store the EUI64 node IDs in an array to use later to set bindings.
//      After a certain amount of time it will end the search
//      (emberStopEndpointSearch)
//  - Finding remote interfaces:
//      After the interface search is started, emberEndpointFoundHandler will be
//      called by the Ember stack to notify the application that a node with the
//      client interface is responding with the search.  We store it's EUI64
//      address for later.
//  - On endpoint search completion, the node will use it's table of EUI64
//      addresses to set bindings to itself on all the remote nodes (controller
//      devices), as well as setting bindings to the remote nodes for itself.
//      This is done by sending a EMBER_SET_BINDING_REQUEST message directly to
//      endpoint 0 on the remote nodes.  Note that a temporary binding entry
//      must be made in order to send the message to the remote node, which then
//      gets removed once the message is sent.
//  - Remote nodes are notified of the binding intention by the stack calling
//      emberRemoteSetBindingHandler.  The application sets the binding itself,
//      returning the status to the stack.
//  - Once the distinction between controlled device and controlling devices is
//      made (binding entries created on all nodes), the controlled device waits
//      for an identify message from any of the controlling devices.  When
//      button #1 is pressed on any of the controlling devices, a datagram is
//      sent to the controlled device (via the binding entry set by the
//      controlled device). The controlled device receives the datagram and
//      performs the necessary action (beep and flash the LED).
//
//  Copyright 2004 by Ember Corporation. All rights reserved.              *80*
//******************************************************************************


#include <ioavr.h>
#include PLATFORM_HEADER //compiler/micro specifics, types
#include "stack/include/ember.h"
#include "stack/include/error.h"
#include "stack/include/hal.h" // microDelayMicroSeconds()
#include "hal/board/board.h"   // LED definitions
#include "hal/micro/micro.h"   // microBoot, memcopy, sleep, launch bootloader
#include "stack/include/packet-buffer.h"
#include "app/util/serial/serial.h"

// For Endpoint 0 communications
#include "stack/include/provisioning.h"

// ************************************
// Ember static memory requirements
//
// There are constants that define the amount of static memory 
// required by the stack. If the application does not set them then
// the default values from ember-static-configuration.h are used. 
// 
// Up to 12 nodes in the network, minus the controlled device,
// plus one temporary one.
#define EMBER_BINDING_TABLE_SIZE        12
// all entries in binding table are temporary (in RAM, not flash)
#define EMBER_TEMPORARY_BINDING_ENTRIES 12
#include "stack/include/ember-static-configuration.h"

// ************************************
// Include definitions for application timers
#include "app-timers.h"

// ************************************
// Serial port definitions.  We define the debug serial port because it needs
// to be initialized in case the debug stack library is being used.
#define DEBUG_SERIAL 0
#define IO_SERIAL 1

//******************************************************************************
// Define endpoints and interfaces for the Ember Stack
// Design:
// We will use one endpoint (endpoint #1)
// We can't use endpoint 0 because it belongs to the Transport layer.
// This endpoint will have the following interfaces:
// 1) A fixed multicast interface for general messages
// 2) A device control server interface (for a controlling device)
// 3) A device control client interface (for a device to be controlled)

// Define interfaces
// (indices into the list of interfaces in the endpoint structure)
#define DEVICE_CONTROL_SERVER_INTERFACE 0
#define DEVICE_CONTROL_CLIENT_INTERFACE 1

// Define endpoints.  Note we start at 1, not 0.
// (indices into the list of endpoints)
#define DEVICE_CONTROL_ENDPOINT   1

// Specification IDs for the ad-hoc interfaces
#define DEVICE_CONTROL_INTERFACE_ID 0x70

// Define the interface structures
// Ad-hoc interfaces contain: interface domain, endpoint type,
// #-specification-bytes and specification (unique identifier)

// Server interface, for controlling devices (many)
static PGM EmberInterfaceId deviceControlServerInterfaceId = {
  AD_HOC_INTERFACE,
  SERVER_ENDPOINT,
  1, DEVICE_CONTROL_INTERFACE_ID };

// Client interface, for controlled device (one)
static PGM EmberInterfaceId deviceControlClientInterfaceId = {
  AD_HOC_INTERFACE,
  CLIENT_ENDPOINT,
  1, DEVICE_CONTROL_INTERFACE_ID };


static PGM EmberEndpointDescription deviceControlEndpoint = {
  { DEVICE_CONTROL_SERVER_INTERFACE, &deviceControlServerInterfaceId },
  { DEVICE_CONTROL_CLIENT_INTERFACE, &deviceControlClientInterfaceId },
};


// Global storage required by the Ember stack
// This tells the stack we're using only one endpoint:
int8u emberEndpointCount = 1;

// This defines the endpoints we're using:
EmberEndpoint emberEndpoints[] = {
  { DEVICE_CONTROL_ENDPOINT, &deviceControlEndpoint}
};


//-------------------------------------
// Definitions for binding table entries
#define MAX_BINDING_INDEX          (EMBER_BINDING_TABLE_SIZE - 2)
#define PROVISIONING_BINDING_INDEX (EMBER_BINDING_TABLE_SIZE - 1)
#define CONTROLLER_BINDING_INDEX   0


//-------------------------------------
// Message types

// Messages used over the device control interface
typedef enum {
  // Messages from server to client
  // (none defined)

  // Messages from client to server
  IDENTIFY_DEVICE = 0xA0
} DeviceControlMessageType;

//******************************************************************************
// Application-specific constants and globals

// Current status of the stack
static EmberStatus stackStatus = EMBER_NETWORK_DOWN;

static int8u myChannel = 0;
static int8s myPower = -1;

static boolean button1_pressed = FALSE;
static boolean button2_pressed = FALSE;

// Flag indicating if this node has become the controlled device
static boolean amControlledDevice = FALSE;
// Flag indicating a controlled device exists somewhere on the network
// (could even be this node)
static boolean controlledDeviceExists = FALSE;

// Array of nodes found on endpoint search.
// Number of nodes on the network, minus the controlled device
static EmberEUI64 controllersFound[EMBER_BINDING_TABLE_SIZE-1];
// When endpoint search is done, this will hold the number of controllers.
static int8u nextBindingIndex = 0;

static int8u controllerBindingIndex = 0xFF;

static int8u PGM identifyTune[] = {
  NOTE_E4,  1,
  0,        0
};

// structure to store necessary network parameters of the node
// (which are panId, enableRelay, radioTxPower, and radioChannel)
EmberNetworkParameters networkParams;

#define REMOTE_BINDING_SET_DELAY  (4)

//******************************************************************************

// Forward declarations
static void processSerialInput(void);
static void microMemPGMCopy(void *dest, void PGM *src, int8u bytes);
static boolean isDuplicateBinding(EmberBindingTableEntry* newEntry);
EmberStatus setRemoteBinding(EmberEUI64 node, int8u type);
static void blinkLEDNotBound(int8u timerId, int32u param);
static void blinkLEDController(int8u timerId, int32u param);
static void blinkLEDControlled(int8u timerId, int32u param);
static void identify(int8u timerId, int32u param);
static void endClientSearch(int8u timerId, int32u param);
static void setNextRemoteBinding(int8u timerId, int32u param);
void startClientSearch();
EmberStatus setRemoteBinding(EmberEUI64 node, int8u type);
static void deleteNextRemoteBinding(int8u timerId, int32u param);
void deleteAllLocalBindings(void);
void controllerIdentifyControlledDevice(void);
void identifyControlledDevice(void);

//******************************************************************************

// Actions grouped together because they need to be done with emberTick
static void runSystem(void)
{
  __watchdog_reset();
  emberTick();
  emberSerialBufferTick();
  handleTimers();
}

//******************************************************************************
// Main application loop
int main(void)
{
  EmberResetType reset;
  EmberStatus status;

  // Determine the cause of the reset (powerup, etc).
  reset = microGetResetInfo();
  microBoot();

  // to init the leds and such..
  boardBoot();

  // reset the system tick
  // enable the system tick interrupt
  microStartSystemTick();
  // allow interrupts
  INTERRUPTS_ON();

  // inititialize the serial port (debugging only)
  emberSerialInit(IO_SERIAL, BAUD_19200, PARITY_NONE, 1);
  emberDebugInit(DEBUG_SERIAL);

  // Initialize the Ember Stack.
  status = emberInit(reset);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, 
      "error returned from emberInit: 0x%x\r\n",
      status);
    if (status == EMBER_UNINITIALIZED_EUI64) {
      // Run to allow over-the-air configuration.
      while (TRUE) {
        __watchdog_reset();
        emberTick();
      }
    } else {
      assert(FALSE);
    }
  }

  initTimers();

  // Bring the network up.
  networkParams.panId = 0;
  networkParams.enableRelay = TRUE;
  networkParams.radioTxPower = myPower;
  networkParams.radioChannel = myChannel;
  
  status = emberAssociate(&networkParams);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, 
      "error returned from emberAssociate: 0x%x\r\n",
      status);
    assert(FALSE);
  }
  
  //-------------------------------------------------------

  // Wait for stack to come up before doing anything
  while (stackStatus != EMBER_NETWORK_UP)
    runSystem();

  //-------------------------------------------------------

  boardLEDClear(BOARDLED0);
  boardLEDClear(BOARDLED1);
  boardLEDClear(BOARDLED2);
  boardLEDClear(BOARDLED3);
  // Create timer to blink LEDs for heartbeat
  // Blink every 1/4 second
  addTimer(0, blinkLEDNotBound, 0, 4, 1);

  // print a startup message
  emberSerialGuaranteedPrintf(IO_SERIAL,
              "\r\nEndpoint0 app running on node %X%X, channel %d\r\n",
              emberGetEui64()[1], emberGetEui64()[0], myChannel);

  //-------------------------------------------------------

  while (TRUE) {
    runSystem();

    // Don't run anything else unless the stack is running
    if (stackStatus == EMBER_NETWORK_UP) {
      processSerialInput();

      // Process button presses
      if (button1_pressed) {
        button1_pressed = FALSE;

        // Can only become the controlled device if we're not already
        // and one doesn't currently exist in the network.
        if (!amControlledDevice) {
          if (!controlledDeviceExists) {
            // This node now becomes the controlled device
            amControlledDevice = TRUE;
            controlledDeviceExists = TRUE;

            // Need to search for controlling devices
            startClientSearch();
          } else {
            // This is a controller device and a controlled device exists
            // Send control message to the controlled device
            controllerIdentifyControlledDevice();
          }
        } // end if (not controlled device)
      } // end button 1 handling

      if (button2_pressed) {
        button2_pressed = FALSE;

        if (amControlledDevice && controlledDeviceExists) {
          // Schedule deletion of remote nodes' bindings,
          // processing one at a time.
          addTimer(1, deleteNextRemoteBinding, 0, REMOTE_BINDING_SET_DELAY, 0);
        }
      } // end button 2 handling
    } // end if(EMBER_NETWORK_UP)
  }  // end while(TRUE)
}


//******************************************************************************
// Timer functions
//******************************************************************************

// Blink red LED to indicate node is not bound to a controlled device.
static void blinkLEDNotBound(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED1);
}

//---------------------------------------------------------

// Blink red LED to indicate node as a controller device in the network.
static void blinkLEDController(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED3);
}

//---------------------------------------------------------

// Blink green LED to indicate node as the controlled device in the network.
static void blinkLEDControlled(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED2);
}

//---------------------------------------------------------

// Identify function - when a controller device sends the IDENTIFY message
// to the controlled device.
static void identify(int8u timerId, int32u param)
{
  if (param) {
    boardLEDToggle(BOARDLED1);
    addTimer(2, identify, param-1, 4, 0);
  } else
    boardLEDClear(BOARDLED1);
}

//******************************************************************************
// Utility functions
//******************************************************************************

// Function used by a device that's trying to become the controlled device.
// Will only be called if there's not already a controlled device in the
// network.  This will start an endpoint search for controller devices.
void startClientSearch()
{
  EmberInterfaceId iface;
  EmberInterfaceId* ifacePtrs[1];

  // Look for nodes with client iface
  // Need to look for deviceControlClientInterfaceId, but that's in flash,
  // and we need to use a RAM address.

  // have to make non-PGM copy of interfaces
  microMemPGMCopy(&iface,
                  &deviceControlClientInterfaceId,
                  sizeof(EmberInterfaceId));

  ifacePtrs[0] = &iface;
  emberStartEndpointSearch(ifacePtrs, 1);

  // Add timeout for endpoint search (3 seconds)
  addTimer(1, endClientSearch, 0, 48, 0);
}

//---------------------------------------------------------

// Timer function used by the controlled device to end its search for
// controller devices.
static void endClientSearch(int8u timerId, int32u param)
{
  emberStopEndpointSearch();
  emberSerialGuaranteedPrintf(IO_SERIAL, "Found %d controller devices\r\n",
                              nextBindingIndex);

  if (nextBindingIndex) {
    // Need to set bindings on controllers found.  Unfortunately, we can't do
    // it all at once in a loop because we'd be trying to send multiple
    // messages without calling emberTick in between, and we'd run out
    // of buffers.
    // So what we'll do is schedule a task to set bindings one at a time.
    addTimer(1,setNextRemoteBinding, 0, REMOTE_BINDING_SET_DELAY, 0);
  }
}

//---------------------------------------------------------

// Timer function used by the controlled device to set remote bindings
// one at a time.
static void setNextRemoteBinding(int8u timerId, int32u param)
{
  EmberBindingTableEntry entry;
  EmberStatus status;

  if (param == nextBindingIndex) {
    // Change LED blinking to indicate controllers have been found
    if (nextBindingIndex) {
      boardLEDClear(BOARDLED0);
      boardLEDClear(BOARDLED1);
      boardLEDClear(BOARDLED2);
      boardLEDClear(BOARDLED3);
      addTimer(0, blinkLEDControlled, 0, 8, 1);
    }
    return;
  }

  entry.type = EMBER_UNICAST_BINDING;
  MEMCOPY(entry.identifier, controllersFound[param], 8);
  entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                  DEVICE_CONTROL_SERVER_INTERFACE);
  entry.remote = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                   DEVICE_CONTROL_CLIENT_INTERFACE);

  // Guard against duplicate binding table entries.
  if (isDuplicateBinding(&entry)) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "already have binding set\r\n");
    return;
  }

  // Set a binding for ourselves to that controller
  status = emberSetBinding(param, &entry);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "Failed to set local binding #%d\r\n", param);
    return;
  } else
    emberSerialGuaranteedPrintf(IO_SERIAL, "binding %d set\r\n", param);

  // Use remote provisioning to set bindings on the remote node
  setRemoteBinding(controllersFound[param], EMBER_UNICAST_BINDING);

  // Reschedule for next controller found
  addTimer(1, setNextRemoteBinding, param+1, REMOTE_BINDING_SET_DELAY, 0);
}

//---------------------------------------------------------

// Utility function used by the controlled device to set a binding on a
// controller device via remote provisioning in Endpoint 0.
EmberStatus setRemoteBinding(EmberEUI64 node, int8u type)
{
  EmberBindingTableEntry entry;
  EmberStatus status = EMBER_SUCCESS;
  EmberBindingTableEntry bindingEntryToMe;
  EmberMessageBuffer pbuff;
  int8u temp;

  entry.type = EMBER_UNICAST_BINDING;
  MEMCOPY(entry.identifier, node, 8);
  entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                  DEVICE_CONTROL_SERVER_INTERFACE);
  entry.remote = 0; // endpoint 0 on remote node

  // Set a temporary binding to send the remote provisioning message
  status = emberSetBinding(PROVISIONING_BINDING_INDEX, &entry);
  if (status != EMBER_SUCCESS)
    return status;

  // Set up binding to be set on remote node
  bindingEntryToMe.type = type;
  MEMCOPY(bindingEntryToMe.identifier, emberGetEui64(), 8);

  bindingEntryToMe.local =
    emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                      DEVICE_CONTROL_CLIENT_INTERFACE);
  bindingEntryToMe.remote =
    emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                      DEVICE_CONTROL_SERVER_INTERFACE);

  // Create datagram message to send to remote endpoint 0
  temp = EMBER_SET_BINDING_REQUEST;
  pbuff = emberFillLinkedBuffers(&temp, 1);
  temp = CONTROLLER_BINDING_INDEX;  // remote node's binding index to us
  emberAppendToLinkedBuffers(pbuff, &temp, 1);

  if (pbuff != EMBER_NULL_MESSAGE_BUFFER) {
    emberAppendToLinkedBuffers(pbuff,
                               (int8u *)&bindingEntryToMe,
                               sizeof(bindingEntryToMe));
    status = emberSendDatagram(PROVISIONING_BINDING_INDEX, pbuff);
    emberReleaseMessageBuffer(pbuff);
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "remote binding set, %x\r\n", status);
  }
  else {
    emberSerialGuaranteedPrintf(IO_SERIAL, "error setting remote binding\r\n");
  }

  // Delete temporary remote provisioning binding entry
  emberDeleteBinding(PROVISIONING_BINDING_INDEX);

  return status;
}

//---------------------------------------------------------

// Timer function used by the controlled device to use remote provisioning to
// delete bindings to itself on all controller devices, one at a time.
// This function will reschedule itself for each binding.
// When it finishes deleting bindings on remote nodes, it will
// delete all local bindings.
static void deleteNextRemoteBinding(int8u timerId, int32u param)
{
  EmberBindingTableEntry entry;

  if (param == nextBindingIndex) {
    deleteAllLocalBindings();
  } else {
    emberGetBinding(param, &entry);
    // "delete" = set to unused
    setRemoteBinding(entry.identifier, EMBER_UNUSED_BINDING);
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "remote binding %d deleted\r\n", param);
    // Schedule deletion of next remote binding
    addTimer(1, deleteNextRemoteBinding, param+1, REMOTE_BINDING_SET_DELAY, 0);
  }
}

//---------------------------------------------------------

// Used by the controlled device to delete bindings to all controller devices.,
// and clear all evidence of its being a controlled device.
void deleteAllLocalBindings(void)
{
  int8u i;

  for (i = 0; i < nextBindingIndex; i++) {
    emberDeleteBinding(i);
    MEMSET(&controllersFound[i], 0, sizeof(EmberEUI64));
    emberSerialGuaranteedPrintf(IO_SERIAL, "local binding %d deleted\r\n", i);
  }
  // Last: clear binding counter
  nextBindingIndex = 0;

  boardLEDClear(BOARDLED0);
  boardLEDClear(BOARDLED1);
  boardLEDClear(BOARDLED2);
  boardLEDClear(BOARDLED3);
  addTimer(0, blinkLEDNotBound, 0, 4, 1);

  // This node now gives up its status as the controlled device
  amControlledDevice = FALSE;
  controlledDeviceExists = FALSE;
}

//---------------------------------------------------------

// Used by a controller device to have the controlled device identify itself.
void controllerIdentifyControlledDevice(void)
{
  EmberMessageBuffer pbuff;
  int8u message[] = {IDENTIFY_DEVICE};

  pbuff = emberFillLinkedBuffers(message, sizeof(message));

  emberSendDatagram(controllerBindingIndex, pbuff);
  emberReleaseMessageBuffer(pbuff);
}

//---------------------------------------------------------

// Used by the controlled device to identify itself when told to by a
// controller device.
void identifyControlledDevice(void)
{
  boardPlayTune_P(identifyTune, 0);
  addTimer(2, identify, 16, 1, 0);
}

//---------------------------------------------------------

static void microMemPGMCopy(void *dest, void PGM *src, int8u bytes)
{
  int8u *d=dest;
  int8u PGM *s=src;

  while(bytes>=4) {
    bytes -= 4;
    *d++ = *s++;
    *d++ = *s++;
    *d++ = *s++;
    *d++ = *s++;
  }
  for(;bytes;bytes--) {
    *d++ = *s++;
  }
}

//---------------------------------------------------------

void printBindingTable(void)
{
  EmberBindingTableEntry entry;
  int8u i;

  emberSerialPrintf(IO_SERIAL, "\r\nI\tT\tL\tR\tM");
  for(i=0; i < EMBER_BINDING_TABLE_SIZE; i++) {
    if (emberGetBinding(i,&entry) == EMBER_SUCCESS) {
      emberSerialPrintf(IO_SERIAL, "\r\n%d\t%d\t%d\t%d\t%x%x",
                        i, entry.type, entry.local, entry.remote,
                        entry.identifier[1], entry.identifier[0]);
      microDelayMicroSeconds(5000);
    }
  }
  emberSerialPrintf(IO_SERIAL, "\r\n");
}

//---------------------------------------------------------

static boolean isDuplicateBinding(EmberBindingTableEntry* newEntry)
{
  int8u i;
  EmberBindingTableEntry existingEntry;
  for (i = 0; i < EMBER_BINDING_TABLE_SIZE; i++) {
    assert(emberGetBinding(i, &existingEntry) == EMBER_SUCCESS);
    if (MEMCOMPARE(&existingEntry,
                   newEntry,
                   sizeof(EmberBindingTableEntry)) == 0) {
      return TRUE;
    }
  }
  return FALSE;
}

//---------------------------------------------------------

//******************************************************************************
// Interface handlers (called from emberIncomingMessageHandler)
//******************************************************************************

// Called by emberIncomingMessageHandler to handle EMBER_INCOMING_DATAGRAM
static void incomingDatagramHandler(int8u localTerminal,
                                    int8u bindingTableIndex,
                                    EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  emberSerialGuaranteedPrintf(IO_SERIAL, "Datagram: ");
  if (emberTerminalInterface(localTerminal) ==
      DEVICE_CONTROL_SERVER_INTERFACE) {
    //---------------------------------
    if (data[0] == IDENTIFY_DEVICE) {
      // Got identify message from a controller device
      emberSerialGuaranteedPrintf(
        IO_SERIAL, "identify command from controller.\r\n");
      identifyControlledDevice();
    //---------------------------------
    }
    else  // First byte of message unknown
      emberSerialGuaranteedPrintf(
        IO_SERIAL,
        "unknown message [%x] from binding %d\r\n",
        data[0], bindingTableIndex);
    //---------------------------------
  }
  else { // Not for server
    emberSerialGuaranteedPrintf(
      IO_SERIAL,
      "message [%x] from binding %d, bad terminal\r\n",
      data[0], bindingTableIndex);
  }
}

//******************************************************************************
// Callback handlers required to use the Ember Stack
//******************************************************************************

void emberEndpointFoundHandler(EmberInterfaceId *desiredInterface,
                               EmberEUI64 node,
                               EmberNetworkDistance distance)
{
  emberSerialGuaranteedPrintf(
    IO_SERIAL,"Found node %X%X\r\n", node[1], node[0]);

  // Drop response if we've maxed out our connections.
  if (nextBindingIndex > MAX_BINDING_INDEX) {
    emberSerialGuaranteedPrintf(
      IO_SERIAL, "dropped - max bindings reached\r\n");
    return;
  }

  if ((*desiredInterface).type == CLIENT_ENDPOINT) {
    MEMCOPY(controllersFound[nextBindingIndex], node, 8);
    nextBindingIndex++;
  } // end interface type check
}

//---------------------------------------------------------

void emberMessageSent(int8u bindingTableIndex,
                      EmberMessageBuffer message,
                      EmberStatus status)
{
  int8u *data = emberMessageBufferContents(message);
  emberSerialGuaranteedPrintf(
    IO_SERIAL,
    "Sent message [%x] from binding %d, status = 0x%x\r\n",
    data[0], bindingTableIndex, status);
}

//---------------------------------------------------------

void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 int8u localTerminal,
                                 int8u bindingTableIndex,
                                 EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  if (!emberMessageBufferLength(message))
    return;

  switch (type) {
  case EMBER_INCOMING_DATAGRAM:
    incomingDatagramHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_DATAGRAM_REPLY:
    if (data[0] == EMBER_SET_BINDING_RESPONSE)
      emberSerialGuaranteedPrintf(IO_SERIAL,
        "SET_BINDING_RESPONSE message from %d\r\n", bindingTableIndex);
    break;

  case EMBER_INCOMING_MULTICAST:
  case EMBER_INCOMING_SEQUENCED:
  case EMBER_INCOMING_SHARED_MULTICAST:
  case EMBER_INCOMING_MULTICAST_LOOPBACK:
  default:
    emberSerialGuaranteedPrintf(
      IO_SERIAL,
      "Unexpected message (type %d) [ %x] binding %d, for terminal %d\r\n",
      type, data[0], bindingTableIndex, localTerminal);
    break;
  }
}

//---------------------------------------------------------

EmberStatus emberRemoteSetBindingHandler(int8u index,
                                         EmberBindingTableEntry *entry)
{
  emberSerialGuaranteedPrintf(
    IO_SERIAL, "Setting binding index %d, %x\r\n", index, entry->type);
  amControlledDevice = FALSE;

  if (entry->type == EMBER_UNUSED_BINDING) {
    controlledDeviceExists = FALSE;
    controllerBindingIndex = 0xFF;

    boardLEDClear(BOARDLED0);
    boardLEDClear(BOARDLED1);
    boardLEDClear(BOARDLED2);
    boardLEDClear(BOARDLED3);
    addTimer(0, blinkLEDNotBound, 0, 4, 1);
  } else {
    controlledDeviceExists = TRUE;
    controllerBindingIndex = index;

    boardLEDClear(BOARDLED0);
    boardLEDClear(BOARDLED1);
    boardLEDClear(BOARDLED2);
    boardLEDClear(BOARDLED3);
    addTimer(0, blinkLEDController, 0, 8, 1);
  }

  return emberSetBinding(index, entry);
}

//---------------------------------------------------------

void boardButtonCallback(int8u button, int8u state)
{
  if (state == BUTTON_PRESSED) {
    if ((button == BUTTON_1) &&
        (boardButtonState(BUTTON_2) == BUTTON_RELEASED) &&
        !button1_pressed) {
      button1_pressed = TRUE;
    } else if ((button == BUTTON_2) &&
               (boardButtonState(BUTTON_1) == BUTTON_RELEASED) &&
               !button2_pressed) {
      button2_pressed = TRUE;
    }
  }  // end if (state == pressed)
}

//---------------------------------------------------------

void emberStackStatusHandler(EmberStatus status)
{
  stackStatus = status;
}

//---------------------------------------------------------

void emberConnectedHandler(int8u bindingTableIndex)
{
}

//---------------------------------------------------------

void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status)
{
}

//******************************************************************************
// End of Handlers
//******************************************************************************


//******************************************************************************
// Serial command interface functions
//******************************************************************************

static void getParams(char *line, int8u length, int32u *argv, int8u *argc)
{
  int8u i;
  int32u temp = 0;
  int8u count = 0;
  for (i = 0; i < length; i++) {
    if ((line[i] >= '0') && (line[i] <= '9')) {
      temp = (temp * 10) + (line[i] - '0');
    } else if ((line[i] == 0x20) || (line[i] == 0x09)) {
      argv[count++] = temp;
      temp = 0;
    } else {
      count--;
      break; // die on first invalid character
    }
  }
  argv[count++] = temp;
  *argc = count;
}

#define LINE_BUFFER_MAX 30
static int8u lineLen = 0;
static char lineBuffer[LINE_BUFFER_MAX];

static void processSerialInput(void) {
  int8u ch;
  int8u cmd = '\0';
  int32u argv[2] = {0, 0};  // Handle up to 2 numeric arguments
  int8u argc = 0;
  EmberStatus status;

  if (emberSerialReadByte(IO_SERIAL, &ch) != EMBER_SUCCESS) return;

  if (ch == 0x0A)
    return;
  else if ((ch == 0x0D) || (lineLen >= LINE_BUFFER_MAX)) {
    cmd = lineBuffer[0];
  } else if (ch == 0x08) { // Backspace
    lineLen--;
    lineBuffer[lineLen] = '\0';
    emberSerialPrintf(IO_SERIAL," \010");
    return;
  } else {
    lineBuffer[lineLen++] = ch;
    return;
  }

  if (lineLen >= 2) {
    getParams(&lineBuffer[2], lineLen-2, argv, &argc);
    lineBuffer[0] = '\0';
  }

  emberSerialPrintf(IO_SERIAL,"\r\n");

  switch(cmd) {
  // bootload
  case 'b':
    emberSerialGuaranteedPrintf(IO_SERIAL, "starting bootloader...\r\n");
    microLaunchBootloader(MICRO_LAUNCH_BOOTLOADER_KEY);
    break;

  // info
  case 'i':
    myChannel = emberGetRadioChannel();
    myPower = emberGetRadioPower();
    emberSerialGuaranteedPrintf(IO_SERIAL,
                "\r\nEndpoint0 node %X%X, channel %d, power %d\r\n",
                emberGetEui64()[1], emberGetEui64()[0], myChannel, emberGetRadioPower());
    break;

  // reset using watchdog timeout
  case 'r':
    emberSerialPrintf(IO_SERIAL, "\r\nresetting...\r\n");
    while (TRUE) ;  // wait for watchdog timer reset
//    break;

  case 'c':
  if (argc == 1) {
      if(emberDisassociate() != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"Cannot disassociate from network\r\n");  
        return;
      } 
      // waiting for the stack to come down after disassociate from network
      while(stackStatus != EMBER_NETWORK_DOWN){
       __watchdog_reset();
       emberTick();
      }
      networkParams.radioChannel = argv[0];
      status = emberAssociate(&networkParams);
      if(status != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"can't set radio channel, 0x%x\r\n", status);
        while(1) ;
      }
      else {
        emberSerialPrintf(IO_SERIAL,"Set channel successfully, waiting for stack to come up");
        while(stackStatus != EMBER_NETWORK_UP){
          __watchdog_reset();
          emberTick();
        }
      }
    } else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

      //change power
    case 'w':
    if (argc == 1) {
      if(emberDisassociate() != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"Cannot disassociate from network\r\n");  
        return;
      } 
      // waiting for the stack to come down after disassociate from network
      while(stackStatus != EMBER_NETWORK_DOWN){
       __watchdog_reset();
       emberTick();
      }
      networkParams.radioTxPower = argv[0];
      status = emberAssociate(&networkParams);
      if(status != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"can't set radio power, 0x%x\r\n", status);
        while(1) ;
      }
      else {
        emberSerialPrintf(IO_SERIAL,"Set power successfully, waiting for stack to come up");
        while(stackStatus != EMBER_NETWORK_UP){
          __watchdog_reset();
          emberTick();
        }
      }
    } else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

  case 'p':
    printBindingTable();
    break;

  case 'h':
    emberSerialGuaranteedPrintf(IO_SERIAL, "b: bootload\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "i: info\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "r: reset\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "c chan: set chan\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "w power: set power\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "p: print binding table\r\n");
    break;

  //-----------------------------------------------------

  case '\0':
  case '\n':
  case '\r':
    break;

  default:
    emberSerialPrintf(IO_SERIAL, "unknown cmd\r\n");
    break;
  } // end switch

  cmd = 0;
  lineLen = 0;
  lineBuffer[0] = '\0';
  lineBuffer[1] = '\0';
}

//******************************************************************************
