/**
 * File: amr-aggregator-cmds.c
 * Description: aggregator commands and run function
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "app/host-framework/c/host-api.h"
#include "app/host-framework/c/packet-buffer.h"
#include "app/host-framework/c/util/command-interpreter.h"
#include "stack/include/error.h"
#include "app/sample/app-timers.h"
#include "app/sample/rnap/amr.h"
#include "app/sample/rnap/amr-aggregator.h"

//------------------------------------------------------------------------------
// Commands
// This is the command utility configuration. we're using the command-interpreter 
// utility located at app/util/serial/ for the node and 
// app/host-framework/c/util for the host.

//commands used by the command interpreter
EmberCommandEntry commands[] = {
  { 0,    "status",         statusCommand,                  "",   "status"},             
  { 0,    "setchannel",     setChannelCommand,              "u1", "set channel"},        
  { 0,    "setpower",       setPowerCommand,                "s1", "set power"},          
  {'h',   "help",           helpCommand,                    "", ""},
  {'p',   "setperiod",      setAdvertisementPeriodCommand,  "u2", "set advertisement period" },
  {'m',   "printmeters",    showMetersCommand,              "", "print a list of meters and their ids" },
  {'e',   "requestempty",    sendEmptyDatagram,         "u1", "send empty datagram to meter" },
  {'d',   "requestdata",    sendDataRequestCommand,         "u1", "request data from meter" },
  {'s',   "requesthistory", sendHistoryRequestCommand,      "u1", "request history data from meter" },
  {'b',   "printbindings",  printBindingTable,              "", "print binding table"},
  {'l',   "setheartbeat",   setLedBlink,                    "u1", "turn the heart beat on (1) and off (0)"},
  {'q',   "quit",           quit,                           "", "quit the application"},
  { 0, NULL, NULL, NULL, NULL } // NULL action makes this a terminator
};

EmberMessageBuffer arguments[4];

EmberCommandState commandState =
  {
    commands,
    commandErrorHandler,
    arguments,
    4,                    // maximum command arguments
    10                    // integer base
  };


extern int16u advertisementPeriod;  //needed by setAdvertisementPeriodCommand
extern int8u nextReportId;          //needed by sendDataRequestCommand


//----------------------------------------------------------------
void commandErrorHandler(int8u errorCode)
{
  fprintf(stderr, "Error: %p\r\n",
           emberCommandErrorNames[errorCode]);  
}

//----------------------------------------------------------------
// Standard commands provided by the command interpreter as defined in amr.h

//print the EUI64ID, channel and power of the node or connected gateway device
void statusCommand(EmberCommandState *state)
{
  fprintf(stderr, "%s ", applicationString);
  printEui64(emberGetEui64());
  fprintf(stderr, " channel %d power %d\r\n",
                    emberGetRadioChannel(),
                    emberGetRadioPower());
}

//print out a listing of all the entries in the binding table
void printBindingTable(EmberCommandState *state)
{
  int8u i;
  EmberStatus status;
  EmberBindingTableEntry result;
  int8u* id;

  fprintf(stderr, "status index type      local remote id\r\n");

  for (i=0; i<EMBER_HOST_BINDING_TABLE_SIZE; i++) {
    
    status = emberGetBinding(i, &result);
    
    id = result.identifier;

    fprintf(stderr, "0x%x  | %d | ", status, i);

    if (EMBER_SUCCESS == status) {
      switch(result.type) {
      case EMBER_UNUSED_BINDING:
        fprintf(stderr, "unused....|");
        break;
      case EMBER_UNICAST_BINDING:
        fprintf(stderr, "unicast...|");
        break;
      case EMBER_MULTICAST_BINDING:
        fprintf(stderr, "multicast.|");
        break;
      default:
        fprintf(stderr, "?????????.|");
      }

      fprintf(stderr, "0x%x  0x%x  ", result.local, result.remote);

      fprintf(stderr, "%x %x %x %x %x %x %x %x\r\n",
               id[0],id[1],id[2],id[3],id[4],id[5],id[6],id[7]);
      
    } else {
      fprintf(stderr, " --- ERROR ---\r\n ");
    }
  }
}


//a utility for printing the EUI64
void printEui64(EmberEUI64 eui64)
{
  fprintf(stderr, "%X%X%X%X%X%X%X%X",
           eui64[0], eui64[1], eui64[2], eui64[3],
           eui64[4], eui64[5], eui64[6], eui64[7]);
}

//handles set-channel from the command line
void setChannelCommand(EmberCommandState *state)
{
  int8u channel = emberUnsignedCommandArgument(state, 0);
  EmberStatus status = emberSetRadioChannel(channel);
  if (status == EMBER_SUCCESS)
    statusCommand(state);
  else
    fprintf(stderr, "emberSetChannel(%d) -> 0x%X\r\n",
                      channel, status);
}

//handles set-power from the command line
void setPowerCommand(EmberCommandState *state)
{
  int8s power = emberSignedCommandArgument(state, 0);
  EmberStatus status = emberSetRadioPower(power);
  if (status == EMBER_SUCCESS)
    statusCommand(state);
  else
    fprintf(stderr, "emberSetPower(%d) -> 0x%X\r\n",
                      power, status);
}

//print the list of commands
void helpCommand(EmberCommandState *state)
{
  int8u i;

  fprintf(stderr, "Command summary:\r\n");
  for (i = 0; commands[i].action != NULL; i++)
  {
    if (commands[i].description != NULL) {
      if (commands[i].shortName != 0) {      
        fprintf(stderr, "%c | ", commands[i].shortName);
      } else {
        fprintf(stderr, "  | ");
      }
      fprintf(stderr, commands[i].longName);
      fprintf(stderr, " ");
      fprintf(stderr, commands[i].description);
      fprintf(stderr, " ");
      fprintf(stderr, "\r\n");
    } 
  }
}

// Set the number of seconds between aggregator advertisements
void setAdvertisementPeriodCommand(EmberCommandState *state)
{
  int16u period = emberSignedCommandArgument(state, 0);
  advertisementPeriod = period << 4;

  // Overwrite timer to send aggregator advertisement
  addTimer(SEND_ADVERT_TIMER_INDEX, sendAdvertisement, 0, advertisementPeriod, 1);
}

//request data from a meter
void sendDataRequestCommand(EmberCommandState *state)
{
  int8u report[2];
  int8u bindingIndex = emberUnsignedCommandArgument(state, 0);

  report[0] = AGGREGATOR_DATA_REQUEST;
  report[1] = nextReportId;

  nextReportId += 1;

  fprintf(stderr, "sending data request for meter %d\n", bindingIndex);

  transmitAmrMessage(bindingIndex, report, 2);
}

void sendEmptyDatagram(EmberCommandState *state)
{
   int8u report[2];
  int8u bindingIndex = emberUnsignedCommandArgument(state, 0);
  fprintf(stderr, "sending empty datagram to meter %d\n", bindingIndex);
  transmitAmrMessage(bindingIndex, report, 0);
}


/*  ****NODE / HOST DIFFERENCE**** */
// On a normal node application, there would be no problem with having the 
// meter simply open a connection to the aggregator and send its history
// data back. The difference here is that the Host Api is not capable of handling
// incoming connection requests. Here we have implemented this function
// so that the Aggregator is responsible for opening the connection to the meter
// for pulling down history data.
void sendHistoryRequestCommand(EmberCommandState *state)
{
  //add a periodic call back to requestMeterHistory which will ensure that a connection gets
  //opened to the proper meter.
  addTimer(REQUEST_METER_HISTORY, requestMeterHistory, emberUnsignedCommandArgument(state, 0), 2, TRUE);
}


// display a list of the available meters
void showMetersCommand(EmberCommandState *state)
{
  int8u i;
  EmberStatus status;
  EmberBindingTableEntry result;
  fprintf(stderr, "Available meters:\n");
  for (i =0; i<MAX_METERS; i++ )
  {
    status = emberGetBinding(i, &result);
    if (result.type == EMBER_UNICAST_BINDING)
    {
      fprintf(stderr, "Meter %d | ", i);
      printEui64(result.identifier);
      fprintf(stderr, " |\n");
    }    
  }
}

// Turn the led blinking on and off
void setLedBlink(EmberCommandState *state)
{
  if (emberUnsignedCommandArgument(state, 0))
    addTimer(BLINK_LED_TIMER_INDEX, toggleBlinker, 0, 8, 1);    
  else
    addTimer(BLINK_LED_TIMER_INDEX, NULL, 0, 0, 0);
}

//exit the host application
void quit(EmberCommandState *state) 
{
  assert(FALSE);
}


