/**
   @file amr-aggregator.c
   @description This is the main code for the aggregator portion of 
   the amr sample application. This code is written with the intention
   that it can compiled for either a node or a host. For a more comprehensive
   description of the use of this application see amr.h. Comments about the 
   specific features of the aggregator are provided inline.

    @see amr.h

   Copyright 2004 by Ember Corporation. All rights reserved.              *80*
 */


/**** NODE / HOST DIFFERENCE ****
  Many of the differences between the node and the host are abstracted
  for the developer by Ember's host-framework module referred to as the 
  "host-api". The host-api is included below. For a more comprehensive 
  description of how it works see app/host-framework/c/host-api.h

 @see app/host-framework/c/host-api.h

*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include "app/host-framework/c/host-api.h"
#include "app/host-framework/c/packet-buffer.h"
#include "app/host-framework/c/util/command-interpreter.h"
#include "app/sample/app-timers.h"
#include "stack/include/error.h"
#include "app/sample/rnap/amr.h"
#include "app/sample/rnap/amr-aggregator.h"

//------------------------------------------------------------------------------
//a flag to track when the network is up or down
boolean networkUp = FALSE;

//Endpoints only need to be defined for the node, not for the host.
static EmberInterfaceId PGM serverInterface = {
  AD_HOC_INTERFACE,
  AGGREGATION_SERVER_ENDPOINT,
  4, 'A', 'M', 'R', 0
};

static EmberEndpointDescription PGM endpointDescription = {
  { 0, &serverInterface }
};

int8u emberEndpointCount = 1;
EmberEndpoint emberEndpoints[1] = {
  { 1, &endpointDescription }
};

PGM_NO_CONST PGM_P applicationString = "Aggregator";

// Times are in quarter seconds.
int16u advertisementPeriod =  7 * 16; //number of quarterSeconds between broadcasts from aggregator
#define advertisementRadius 10 //number of hops for the broadcast

//the id of the next meter we expect a report from so we'll know if
//their skipping a node or not.
int8u nextReportId = 0;

//number of meters we know about
static int8u meterCount = 0;

/*
  **** NEW WITH VERSION 3.3 ****
  Before the host application can do anything, it must first associate with
  a network. I have provided a default network parameters struct here which
  is passed into the new emberAssociate() call in the Host API.
*/
static EmberNetworkParameters networkParams =
  {
    0,                  // PAN ID
    TRUE,               // enable relaying
    AMR_SAMPLE_POWER,
    AMR_SAMPLE_CHANNEL
  };

//commands used by the command interpreter
extern EmberCommandEntry commands[];
extern EmberCommandState commandState;

//-----------------------------------------------------------------------------
/* FORWARD DECLARATIONS OF AGGREGATOR SPECIFIC MESSAGE HANDLERS */
static void incomingDatagramHandler(int8u localTerminal, int8u bindingTableIndex,
                                    EmberMessageBuffer message);
static void incomingSequencedHandler(int8u localTerminal, int8u bindingTableIndex,
                                     EmberMessageBuffer message);



//------------------------------------------------------------------------------
/*
  ****NODE / HOST DIFFERENCE****
  On the node, we have to initialize the Ember Stack, 
  On the host we need to handle arguments which are passed in
  so that we can initialize the host using the host-api's emberHostInit() which
  takes in two arguments "address and port" so that it can locate the RNAP gateway
  node which proxys its requests on the Ember network.
*/
int main(int argc, char **argv) //host main
{

    /* NECESSARY HOST API CALL:
     In order for the Host API to initialize properly the host application must first make a 
     call to emberHostInit informing it where to locate the gateway node. The Host API is 
     written to connect to a gateway connected to a serial port, a socket or a simulated node
     over stdio.
     
     1. If two arguments are given at run time, the code below assumes that the host intends to connect
     to an RNAP gateway on a socket where the two arguments are address and port.
     
     2. If only one argument is given, the Host API assumes that it is connecting to 
     a gateway on a serial port where the argument is port. 
     
     3. When no arguments are given, the Host API assumes that it is connecting to
     a simulated gateway over stdio. This connection mode is used internally by Ember
     for simulated testing and development of the Host API and RNAP protocol.
     
  */
  EmberStatus status;
  if (argc == 3)
    status = emberHostInit(argv[1], atoi(argv[2]), EMBER_HOST_MODE_SOCKET);
  else if (argc == 2)
    status = emberHostInit(argv[1], 0, EMBER_HOST_MODE_SERIAL);
  else {
    fprintf(stderr, "No Gateway location specified, connecting in test mode.\n");
    status = emberHostInit((int8u *)0, 0, EMBER_HOST_MODE_STDIO);
  }
  if (status != 0) {
    fprintf(stderr, "ERROR connecting to gateway, EmberStatus = %d.\n", status);
    return status;
  }  
  
  /* NEW TO VERSION 3.3 
    NECESSARY HOST API CALL.
    The application must tell the stack
    that it wants to associate with the network. On the host, this can be done
    at any time, as the host-api will just save the parameter information until the 
    gateway node is ready to receive it. The Host API waits until the initial node start
    handshaking is done with the RNAP Gateway before sending the emberAssociate message
    to the gateway.
    
    NOTE: The networkParams struct is defined in this file above and draws it's default
    settings for channel and power from the definitions in amr.h. If you are using this
    application as a template you will want to be sure and change the default values for
    channel and power in amr.h to your desired settings.
    */
  emberAssociate(&networkParams);

  /* Non-RNAP related module used by this sample application:
     The Command Reader is used by the Host application to read key strokes coming in from 
     the user. The commands available are defined in amr-aggregator-cmds.c. You do not need 
     to use the command reader to run RNAP it is just a module we wrote for this sample application.
     You are free to use it or discard it if you wish.
  */
  emberCommandReaderInit(&commandState);  
  
  /* Necessary for this sample app: create multicast binding for meter interaction. Both the 
     amr-aggregator and the meters subscribe to a shared multicast binding so that they can 
     find eachother when the aggregator sends out multicast advertisements.
  */ 
  createMulticastBinding(MULTICAST_BINDING_INDEX);
  
  while(TRUE) {

    /*NOTE: This call is used internally by Ember for development. If you are using this 
      application as a template for application development you can remove all of the
      references to EMBER_TEST and included code.
    */
    #ifdef EMBER_TEST
      simulatedTimePassesMs(msToNextEvent());    
    #endif
    
    /* Non-RNAP related module used by this sample application:
       This amr sample app uses the timer module included with the sample apps to schedule
       timed events. This module is not part of, and is not used by the RNAP Host API. For
       the sake of this application, timers must be serviced but if you are using this
       app as template and don't want to use the timer module, you do not have
       to service the timers and can remove this call to handleTimers.
    */
    handleTimers();

    /* Non-RNAP related module usde by this sample application:
       As mentioned above with regard to emberCommandReaderInit, this call is part of 
       the command reader module included with this sample application. The command reader
       is not used by the Host API so the call below is only necessary if you are using
       the command module.
    */
    emberProcessCommandInput(&commandState, SERIAL_PORT);

    /*
      This main loop was chewing up 100% of the cpu so to make it relax a bit 
      i've added a sleep here.
    */
    usleep(100);  


  }
  return 0;
}


//------------------------------------------------------------------------------
// AGGREGATOR SPECIFIC MESSAGE HANDLERS

//handles messages coming in from meters
static void incomingDatagramHandler(int8u localTerminal,
                                    int8u bindingTableIndex,
                                    EmberMessageBuffer message)
{
  int8u *contents = emberMessageBufferContents(message);
    
  switch (contents[0]) {
  case METER_ANNOUNCE: {
    registerMeterWithAggregator(contents);
    break;
    }
  case METER_HISTORY_REPLY:
    fprintf(stderr, "error received history as datagram, should be sequenced\n");
    break;
  case METER_REPORT:
  case METER_DATA_REPLY: {
    if (bindingTableIndex == EMBER_NULL_BINDING) {
      //a meter we don't know (probably because we rebooted)
      fprintf(stderr, "rx report from meter we don't know sending announce request\n");
      EmberMessageBuffer datagramAnnounceRequest = emberFillStackBuffer(1, AGGREGATOR_ANNOUNCE_REQUEST);
      emberSendReply(datagramAnnounceRequest); //ask for an introduction
      emberReleaseMessageBuffer(datagramAnnounceRequest);
    } else {
      //a meter we know sending data
      fprintf(stderr, "RX data from meter %d\n", bindingTableIndex);
	  }
    break;
    }
  default:
    fprintf(stderr, "rx, datagram we don't understand.\n");
    break;
  } // end switch
}

//handles sequenced messages coming from the meters
static void incomingSequencedHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message)
{
  int8u msgType;
  emberCopyFromLinkedBuffers(message, 0, &msgType, 1);

  switch(msgType) {
  case METER_HISTORY_REPLY: {
    int8u index = 1;
    int8u contents[16];
    int8u length = emberMessageBufferLength(message) - 1;
    fprintf(stderr, "Received history data from meter %x\r\n", bindingTableIndex);
    while (length) {
      int8u remaining = (length >= 16) ? 16 : length;
      int8u i;
      emberCopyFromLinkedBuffers(message, index, contents, remaining);
      for (i = 0; i < remaining; i++) {
        fprintf(stderr, "%x ", contents[i]);
      } 
      fprintf(stderr, "\r\n");
      length -= remaining;
      index += remaining;
    }
    break;
  }

  default:
    fprintf(stderr, "rx sequenced message we don't understand.\n");
    break;
  } // end switch
}

//------------------------------------------------------------------------------
// Host Api Callbacks
// handle incoming messages from the host-api for a more complete explanation
// @see stack/include/ember.h
void emberBeaconLossCallback(void) {

}

void temporaryIncomingBeaconHandler(void) {

}

void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 int8u localTerminal,
                                 int8u bindingTableIndex,
                                 EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  if (!emberMessageBufferLength(message))
    return;

  switch (type) {
  case EMBER_INCOMING_DATAGRAM_REPLY:
    fprintf(stderr, "host rx incoming datagram reply\n");
  case EMBER_INCOMING_DATAGRAM:
    incomingDatagramHandler(localTerminal, bindingTableIndex, message);
    break;
    
  case EMBER_INCOMING_SEQUENCED:
    incomingSequencedHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_MULTICAST:
  case EMBER_INCOMING_SHARED_MULTICAST:
  case EMBER_INCOMING_MULTICAST_LOOPBACK:
  default:
    fprintf(stderr, "Unexpected message (type %d) [ %x] binding %d, for terminal %d\r\n",
      type, data[0], bindingTableIndex, localTerminal);
    break;
  }
}

// Called when a datagram has been transmitted. The status indicates
// the success or failure of the transmission
void emberMessageSent(int8u bindingTableIndex, 
                      EmberMessageBuffer message, 
                      EmberStatus status)
{
  fprintf(stderr, status == EMBER_SUCCESS
                    ? "msg succeeded\r\n"
                    : "msg failed\r\n");
}


// Indicates the success or failure of a connection.
void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status)
{
  fprintf(stderr, "Connection to binding %d: %x\r\n", bindingTableIndex, status);
}

/* 
   Indicates the state of the stack. When the stack is not able to communicate
   with the network or is down for some other reason, this function will be called
   by the stack or the gateway-node in the case of a host application. The stack is not
   able to transmit messages unless it reports a status of EMBER_NETWORK_UP.
*/
void emberStackStatusHandler(EmberStatus status)
{
  fprintf(stderr, "agg: rx stackStatus=%d UP=%d DN=%d\n", status, EMBER_NETWORK_UP, EMBER_NETWORK_DOWN);
  if (status == EMBER_NETWORK_UP) {
    networkUp = TRUE;
    
    //I moved chanel set to part of initialization cuz it doesnt' make sense
    //to do it here, we were just told we were good on whatever the default channel was, 
    //then we're going to switch channels? that doesn't make sense.
    //set our channel
    //emberSetRadioChannel(AMR_SAMPLE_CHANNEL);

    addTimer(SEND_ADVERT_TIMER_INDEX, sendAdvertisement, 0, advertisementPeriod, 1);

  } else if (status == EMBER_NETWORK_DOWN) {
    networkUp = FALSE;
    // Delete timer to send aggregator advertisement
    addTimer(SEND_ADVERT_TIMER_INDEX, NULL, 0, 0, 0);
  }
}


//---------------------------------------------------------------------
/* RNAP SPECIFIC HANDLERS                                            */

// Indicates the rnap status, report the following code
void emberRnapStatusHandler(EmberStatus status)
{
  switch (status) { 
  case EMBER_RNAP_MISMATCH:
    fprintf(stderr, "Host and Node binding information mismatched\n");
    break;
  case EMBER_RNAP_ACK_WITH_ERR:
    fprintf(stderr, "Node acked with error code\n");
    break;
  case EMBER_RNAP_COMMUNICATION_FAILED:
    fprintf(stderr, "Rnap communication with the Node failed\n");
    break;
  default:
    fprintf(stderr, "Error Rnap status\n");
    break;
    
  }
}

// Indicates the failure of sending rnap message.
void emberRnapMessageSent(EmberMessageBuffer rnapMessage, EmberStatus status)
{
  if (status == EMBER_RNAP_DELIVERY_FAILED) {
    fprintf(stderr, "Rnap message failed to be delivered\n");
  }
  else {
    fprintf(stderr, "Error Status for emberRnapMessageSent\n");
  }
}


//------------------------------------------------------------------------------
// Aggregator Application Functions

//adds a meter to the aggregators list of meters 
void registerMeterWithAggregator(int8u *contents) {
    EmberBindingTableEntry binding;
    int8u availBinding;
    int8u i;

    //first check to see if an entry has already been allocated
    //for this meter
    for (i = 0; i < MAX_METERS; i++) {
      emberGetBinding(i, &binding);
      if (MEMCOMPARE(binding.identifier, (contents+1), 8) == 0) {
	      //we should note the senders binding to update the cost and route
	      //to the meter.
	      emberNoteSendersBinding(i);
	      return;  
      }
    }

    binding.type = EMBER_UNICAST_BINDING;
    MEMCOPY(binding.identifier, contents + 1, 8);
    binding.local = ourTerminal;
    binding.remote = ourTerminal;
    
    availBinding = findAvailableBinding(contents+1);
    if ((availBinding != 0xFF) &&
        (emberSetReplyBinding(availBinding, &binding) == EMBER_SUCCESS))
    {
      fprintf(stderr, "have meter %d ( ",
                        availBinding);
      printEui64(binding.identifier);
      fprintf(stderr, ")\n");
    } else
      fprintf(stderr, "Could not set binding entry.\n");
}



void sendAdvertisement(int8u timerId, int32u param)
{
  EmberMessageBuffer advert = emberFillStackBuffer(1,
                           meterCount == MAX_METERS
                             ? AGGREGATOR_REMINDER
                             : AGGREGATOR_ADVERTISEMENT);

  if (advert == EMBER_NULL_MESSAGE_BUFFER)
    return;

  fprintf(stderr, "sending advert\n");

  emberAppendToLinkedBuffers(advert, emberGetEui64(), 8);
  emberSendLimitedMulticast(MULTICAST_BINDING_INDEX,
                            advert,
                            advertisementRadius);
  emberReleaseMessageBuffer(advert);
}


//find an available binding in the binding table
int8u findAvailableBinding(int8u *remoteEui64)
{
  EmberBindingTableEntry binding;
  int8u i;
  for (i = 0; i < MAX_METERS; i++) {
    emberGetBinding(i, &binding);
    if (binding.type == EMBER_UNUSED_BINDING) {
      return i;  
    }
  }
  return 0xFF;
}

void requestMeterHistory(int8u timerId, int32u param) {
  
  int8u report[2];
  int8u bindingIndex = (int8u)param;
  
  if (emberConnectionStatus(bindingIndex) != EMBER_CONNECTION_OPEN) {
    emberOpenConnection(bindingIndex); //if we're not connected, connect
    return;
  }

  report[0] = AGGREGATOR_HISTORY_REQUEST;
  report[1] = nextReportId;

  nextReportId += 1;

  transmitAmrMessage(bindingIndex, report, 2);
  addTimer(REQUEST_METER_HISTORY, NULL, 0, 0, 0);

}

//creates and sets a binding to send multicasts to
//or in the case of a meter, receive multicasts from related
//to the amr application
void createMulticastBinding(int8u index)
{
  EmberBindingTableEntry entry;

  entry.type = EMBER_MULTICAST_BINDING;
  entry.identifier[0] = HIGH_BYTE(MULTICAST_GROUP);
  entry.identifier[1] = LOW_BYTE(MULTICAST_GROUP);
  entry.local = ourTerminal;

  assert(emberSetBinding(index, &entry) == EMBER_SUCCESS);
}

//A utility function to simpify the code involved in sending a datagram
//within the amr application.
EmberStatus transmitAmrMessage(
  int8u bindingIndex,
  int8u *message,
  int8u length)
{
  EmberStatus status = EMBER_SUCCESS;
  EmberMessageBuffer buffer = emberFillLinkedBuffers(message, length);

  if (buffer == EMBER_NULL_MESSAGE_BUFFER)
    return EMBER_NO_BUFFERS;

  status = emberSendDatagram(bindingIndex, buffer);

  emberReleaseMessageBuffer(buffer);
  
  return status;
}

void toggleBlinker(int8u timerId, int32u param)
{
  fprintf(stderr, "*");
}


