/**
 * File: amr-meter-cmds.c
 * Description: meter's run and commands
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

//Node includes
#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "stack/include/packet-buffer.h"
#include "stack/include/hal.h"
#include "app/util/serial/serial.h"
#include "app/util/serial/command-interpreter.h"
#include "hal/board/board.h"
#include "hal/micro/micro.h"
#include "stack/include/error.h"
#include "app/sample/app-timers.h"
#include "app/sample/rnap/amr-meter.h"
#include "app/sample/rnap/amr.h"

extern int16u reportPeriod;         //needed by setReportPeriodCommand
extern int8u reportSize;            //needed by reportSizeCommand
extern EmberCost *aggregatorCosts;   //needed by printAggregatorsCommand
extern int8u *aggregatorAge;         //needed by printAggregatorCommand



//------------------------------------------------------------------------------
// Commands
EmberCommandEntry commands[] = {
  { 0,  "status",         statusCommand,            "",     "status"},             
  { 0,  "setchannel",     setChannelCommand,        "u1",   "set channel"},        
  { 0,  "setpower",       setPowerCommand,          "s1",   "set power"},          
  {'h', "help",           helpCommand,              "",     ""},
  { 0,  "setperiod",      setReportPeriodCommand,   "u2",   "set report period" },
  { 0,  "setsize",        setReportSizeCommand,     "u1",   "set report size" },
  {'i', "info",           infoCommand,              "",     "get report info" },
  {'b', "bootload",       bootloadCommand,          "",     "bootloader"},
  {'d', "data",           sendReportCommand,        "",     "send data to aggregator" },
  {'p', "printbinding",   printBindingTable,        "",     "print binding table"},
  {'g', "aggregators",    printAggregatorsCommand,  "",     "print aggregators" },
  { 0, NULL, NULL, NULL, NULL } // NULL action makes this a terminator
};

int16u arguments[4];

EmberCommandState commandState =
  {
    commands,
    commandErrorHandler,
    arguments,
    4,                    // maximum command arguments
    10                    // integer base
  };


//----------------------------------------------------------------
// Standard commands provided by the command interpreter as defined in amr.h

//print the EUI64ID, channel and power of the node or connected gateway device
void statusCommand(EmberCommandState *state)
{
  emberSerialPrintf(SERIAL_PORT, "%p ", applicationString);
  printEui64(emberGetEui64());
  emberSerialPrintf(SERIAL_PORT, " channel %d power %d\r\n",
                    emberGetRadioChannel(),
                    emberGetRadioPower());
}

//print out a listing of all the entries in the binding table
void printBindingTable(EmberCommandState *state)
{
  int8u i;
  EmberStatus status;
  EmberBindingTableEntry result;
  int8u* id;
  //  int8u indexLow;
  //int8u indexHigh;
  emberSerialPrintf(SERIAL_PORT, "status index type      local remote id\r\n");

  for (i=0; i<getBindingTableSize(); i++) {
    
    status = emberGetBinding(i, &result);
    
    id = result.identifier;

    emberSerialPrintf(SERIAL_PORT, "0x%x  | %d | ", status, i);

    if (EMBER_SUCCESS == status) {
      switch(result.type) {
      case EMBER_UNUSED_BINDING:
        emberSerialPrintf(SERIAL_PORT, "unused....|");
        break;
      case EMBER_UNICAST_BINDING:
        emberSerialPrintf(SERIAL_PORT, "unicast...|");
        break;
      case EMBER_MULTICAST_BINDING:
        emberSerialPrintf(SERIAL_PORT, "multicast.|");
        break;
      default:
        emberSerialPrintf(SERIAL_PORT, "?????????.|");
      }

      emberSerialPrintf(SERIAL_PORT, "0x%x  0x%x  ", result.local, result.remote);

      emberSerialPrintf(SERIAL_PORT, "%x %x %x %x %x %x %x %x\r\n",
                        id[0],id[1],id[2],id[3],id[4],id[5],id[6],id[7]);
      
    } else {
      emberSerialPrintf(SERIAL_PORT, " --- ERROR ---\r\n ");
    }
  }
}


//a utility for printing the EUI64
void printEui64(EmberEUI64 eui64)
{
  emberSerialPrintf(SERIAL_PORT, "%X%X%X%X%X%X%X%X",
                    eui64[0], eui64[1], eui64[2], eui64[3],
                    eui64[4], eui64[5], eui64[6], eui64[7]);
}

//handles set-channel from the command line
void setChannelCommand(EmberCommandState *state)
{
  int8u channel = emberUnsignedCommandArgument(state, 0);
  EmberStatus status = emberSetRadioChannel(channel);
  if (status == EMBER_SUCCESS)
    statusCommand(state);
  else
    emberSerialPrintf(SERIAL_PORT, "emberSetChannel(%d) -> 0x%X\r\n",
                      channel, status);
}

//handles set-power from the command line
void setPowerCommand(EmberCommandState *state)
{
  int8s power = emberSignedCommandArgument(state, 0);
  EmberStatus status = emberSetRadioPower(power);
  if (status == EMBER_SUCCESS)
    statusCommand(state);
  else
    emberSerialPrintf(SERIAL_PORT, "emberSetPower(%d) -> 0x%X\r\n",
                      power, status);
}

//print the list of commands
void helpCommand(EmberCommandState *state)
{
  int8u i;

  emberSerialPrintf(SERIAL_PORT, "Command summary:\r\n");
  for (i = 0; commands[i].action != NULL; i++)
  {

    __watchdog_reset(); 

    if (commands[i].description != NULL) {
      if (commands[i].shortName != 0) {      
        emberSerialPrintf(SERIAL_PORT, "%c | ", commands[i].shortName);
      } else {
        emberSerialPrintf(SERIAL_PORT,"  | ");
      }
      emberSerialPrintf(SERIAL_PORT,commands[i].longName);
      emberSerialPrintf(SERIAL_PORT," ");
      emberSerialPrintf(SERIAL_PORT,commands[i].description);
      emberSerialPrintf(SERIAL_PORT," ");
      emberSerialPrintf(SERIAL_PORT, "\r\n");
    } 
  }
}

//------------------------------------------------------------------------------
// handle a command to start the bootloader
void bootloadCommand(EmberCommandState *state)
{
  emberSerialGuaranteedPrintf(SERIAL_PORT, "starting bootloader...\r\n");
  microLaunchBootloader(MICRO_LAUNCH_BOOTLOADER_KEY);
}

// set the number of seconds between sending reports to the aggregator
void setReportPeriodCommand(EmberCommandState *state)
{
  int16u period = emberSignedCommandArgument(state, 0);
  reportPeriod = period << 4;
  addTimer(SEND_REPORT_TIMER_INDEX, sendReport, 0, reportPeriod, 1);
}

// set the ammount of data reported in a datagram
void setReportSizeCommand(EmberCommandState *state)
{
  int8u newSize = emberUnsignedCommandArgument(state, 0);
  if ((newSize >= minimumReportSize) && (newSize <= maximumReportSize))
    reportSize = newSize;
  else
    emberSerialPrintf(SERIAL_PORT,
                      "Command error, report size: min = %u, max = %u\r\n",
                      minimumReportSize, maximumReportSize);
}

void sendReportCommand(EmberCommandState *state)
{
  sendReport(0, 0);
}

// print out the current report settings
void infoCommand(EmberCommandState *state)
{
  emberSerialPrintf(SERIAL_PORT,
                    "Report %u bytes every %u seconds\r\n",
                    reportSize,
                    reportPeriod >> 4);
}

// print out the current aggregators that this meter knows about
void printAggregatorsCommand(EmberCommandState *state)
{
  int8u i;
  emberSerialPrintf(SERIAL_PORT,
                    "Index Cost Age\r\n");
  for (i = 0; i < MAXIMUM_AGGREGATORS; i++) {
    emberSerialPrintf(SERIAL_PORT,
                    "%d     %x   %x\r\n", i, aggregatorCosts[i], aggregatorAge[i]);
  }
}


//----------------------------------------------------------------
void commandErrorHandler(int8u errorCode)
{
  emberSerialPrintf(SERIAL_PORT, "Error: %p\r\n",
                    emberCommandErrorNames[errorCode]);  
}

