/**
 * File: meter.c
 * Description: meter code for AMR application. This application waits
 * to hear from an aggregator. When it does, it begins sending periodic
 * datagrams to the aggregator representing meter data. The meter can also
 * respond to datagram requests from the aggregator for report data as well
 * as history data. Report data is meant to represent the latest report. History
 * data is meant to represent a history of the meter's reports and is transmitted
 * via sequenced message. For a deeper look at datagrams and sequenced messages
 * @see stack/include/ember.h
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "stack/include/packet-buffer.h"
#include "stack/include/error.h"
#include "stack/include/hal.h"
#include "app/util/serial/serial.h"
#include "app/util/serial/command-interpreter.h"
#include "hal/board/board.h"
#include "hal/micro/micro.h"
#include "app/sample/rnap/amr-meter.h"
#include "app/sample/rnap/amr.h"

// definitions for application timers
#include "app/sample/app-timers.h"

//------------------------------------------------------------------------------
//for a more comprehensive description of these see amr-common.c
#ifdef EMBER_TEST
  extern int16u simulatorId;
#endif

//a flag to track if the network is up or down
boolean networkUp = FALSE;

//------------------------------------------------------------------------------
#define MULTICAST_BINDING_INDEX MAXIMUM_AGGREGATORS

#define EMBER_TRANSPORT_CONNECTION_COUNT        1
#define EMBER_BINDING_TABLE_SIZE                (MAXIMUM_AGGREGATORS + 1)
#define EMBER_TEMPORARY_BINDING_ENTRIES         EMBER_BINDING_TABLE_SIZE
#define EMBER_MAXIMUM_AGGREGATION_ROUTES        0

#include "stack/include/ember-static-configuration.h"


//------------------------------------------------------------------------------

static EmberInterfaceId PGM clientInterface = {
  AD_HOC_INTERFACE,
  AGGREGATION_CLIENT_ENDPOINT,
  4, 'A', 'M', 'R', 0
};

static EmberEndpointDescription PGM endpointDescription = {
  { 0, &clientInterface }
};

int8u emberEndpointCount = 1;
EmberEndpoint emberEndpoints[1] = {
  { 1, &endpointDescription }
};

PGM_NO_CONST PGM_P applicationString = "Meter";

//------------------------------------------------------------------------------

void initializeEmberStack(void)
{
  EmberStatus status;
  EmberNetworkParameters networkParams;
  EmberResetType resetInfo = microGetResetInfo();
  boardBoot();
  microBoot();
  microStartSystemTick();
  INTERRUPTS_ON();
  emberDebugInit(0);
  emberSerialInit(SERIAL_PORT, BAUD_19200, PARITY_NONE, 1);
    // Initialize the Ember Stack.
  status = emberInit(resetInfo);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(SERIAL_PORT, 
      "error returned from emberInit: 0x%x\r\n",
      status);
    assert(FALSE);
  }

  // Set up our default network parameters.
  networkParams.panId = 0;
  networkParams.enableRelay = TRUE;
  networkParams.radioTxPower = -1;
  networkParams.radioChannel = AMR_SAMPLE_CHANNEL;
  
  // associate with the network
  status = emberAssociate(&networkParams);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(SERIAL_PORT, 
      "error returned from emberAssociate: 0x%x\r\n",
      status);
    assert(FALSE);
  }
}

// Times are in quarter seconds.
int16u reportPeriod = 15 * 16;   // fifteen seconds

// This is the number of bytes in the reports sent back to the aggregator.
int8u reportSize = 32;  // number of data bytes in a report

//the aggregator we're currently talking to
static int8u currentAggregator = AGGREGATOR_DEFAULT;
EmberCost aggregatorCosts[MAXIMUM_AGGREGATORS];
int8u aggregatorAge[MAXIMUM_AGGREGATORS];

//what meter reports we've supplied
static int8u historyReportCount;

// Flag indicating we have a connection open to the aggregator
static boolean connectedToAggregator = FALSE;

#define MAXIMUM_AGGREGATOR_AGE 6
static int8u nextReportId = 0;


// We ignore cost differences less than this.
#define costHysteresis 2

//commands used by the command interpreter
extern EmberCommandEntry commands[];
extern EmberCommandState commandState;

//------------------------------------------------------------------------------
/* FORWARD DECLARATIONS OF METER SPECIFIC HANDLERS                            */
static void incomingDatagramHandler(int8u localTerminal, int8u bindingTableIndex,
                                    EmberMessageBuffer message);
static void incomingMulticastHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message);



//------------------------------------------------------------------------------
//if we are in test mode, nodeMain replaces main()
#ifdef EMBER_TEST
int nodeMain(void)
#else
int main(void)
#endif
{
  
  initializeEmberStack(); 
  emberCommandReaderInit(&commandState);
  createMulticastBinding(MULTICAST_BINDING_INDEX);

  {
    int8u i;
    for (i = 0; i < MAXIMUM_AGGREGATORS; i++)
      aggregatorAge[i] = 0xFF;
  }

  //blink a heartbeat
  #ifndef EMBER_TEST
    addTimer(BLINK_LED_TIMER_INDEX, toggleBlinker, 0, 8, 1);
  #endif

  while(TRUE) {
    
    /*** NODE / HOST DIFFERENCE ***/
    //On a node we need to service the watchdog 
    //and emberTick. so the node doesn't reset.
    __watchdog_reset();
    emberTick();
    emberSerialBufferTick();

    //need to service the timer module
    handleTimers();

    //need to service the command processor module
    emberProcessCommandInput(&commandState, SERIAL_PORT);

    #ifdef EMBER_TEST
      simulatedTimePassesMs(msToNextEvent());
    #endif
    
  }
  //return 0;
}



//------------------------------------------------------------------------------
// METER SPECIFIC MESSAGE HANDLERS

// Handle datagrams coming from the aggregator
static void incomingDatagramHandler(int8u localTerminal,
                                    int8u bindingTableIndex,
                                    EmberMessageBuffer message)
{
  if (message != EMBER_NULL_MESSAGE_BUFFER)
  {
    int8u *contents = emberMessageBufferContents(message);
  
    switch (contents[0]) {
    case AGGREGATOR_REJECTION: {
      int8u i;
      if (findEui64Binding(contents + 1, &i)) {
        if (i == currentAggregator)
          currentAggregator = 0xFF;
        aggregatorAge[i] = MAXIMUM_AGGREGATOR_AGE;
      }
      break;
    }
    case AGGREGATOR_DATA_REQUEST:
      if (bindingTableIndex == currentAggregator)
        sendDataReply();
      break;
    case AGGREGATOR_HISTORY_REQUEST:
      // Don't send history data if it wasn't our current aggregator that requested.
      if (bindingTableIndex == currentAggregator) {
        historyReportCount = 4;
        // Schedule history report because it will be long.
        addTimer(SEND_HISTORY_TIMER_INDEX, sendHistory, 0, 2, TRUE);
      }
      break;
    case AGGREGATOR_ANNOUNCE_REQUEST:
      emberSerialPrintf(SERIAL_PORT, "got aggregator announce request\n");
      announcePresence();
      break;
    default:
      emberSerialPrintf(SERIAL_PORT, "msg error\n");
      break;
    }  // end switch
  } else {
    emberSerialPrintf(SERIAL_PORT, "rx empty message\n");
  }
}

// Called by emberIncomingMessageHandler
static void incomingMulticastHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message)
{
  int8u *contents = emberMessageBufferContents(message);

  switch (contents[0]) {
  case AGGREGATOR_ADVERTISEMENT:
  case AGGREGATOR_REMINDER:
    considerAggregatorAdvertisement(contents);
    break;
  default:
    emberSerialPrintf(SERIAL_PORT, "msg error\n");
    break;
  }  // end switch
}



//------------------------------------------------------------------------------
// Stack Callbacks
// @see stack/include/ember.h for a complete description
void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 int8u localTerminal,
                                 int8u bindingTableIndex,
                                 EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);
  switch (type) {
  case EMBER_INCOMING_DATAGRAM_REPLY:
  case EMBER_INCOMING_DATAGRAM:
    incomingDatagramHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_MULTICAST:
    incomingMulticastHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_SEQUENCED:
  case EMBER_INCOMING_SHARED_MULTICAST:
  case EMBER_INCOMING_MULTICAST_LOOPBACK:
  default:
    emberSerialPrintf(
      SERIAL_PORT,
      "Unexpected message (type %d) [ %x] binding %d, for terminal %d\r\n",
      type, data[0], bindingTableIndex, localTerminal);
    break;
  }
}

//@see stack/include/ember.h
void emberMessageSent(int8u bindingTableIndex,
                      EmberMessageBuffer message,
                      EmberStatus status)
{
  emberSerialPrintf(SERIAL_PORT,
                    status == EMBER_SUCCESS
                    ? "msg succeeded\r\n"
                    : "msg failed\r\n");
  emberDebugPrintf(status == EMBER_SUCCESS
                    ? "msg succeeded"
                    : "msg failed");
}

//@see stack/include/ember.h
void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status)
{
  emberSerialPrintf(SERIAL_PORT,
    "Connection to binding %d: %x\r\n", bindingTableIndex, status);

  connectedToAggregator = (status == EMBER_CONNECTION_OPEN);
}

//@see stack/include/ember.h
void emberStackStatusHandler(EmberStatus status)
{
  if (status == EMBER_NETWORK_UP) {
    networkUp = TRUE;
    
    //set a delay to start reporting
    if (currentAggregator != 0xFF) {
      int16u rand = (firmGetRandom() & 0xFF) + 1;  // 1...256
      addTimer(START_REPORTING_TIMER_INDEX, startReporting, 0, ((rand * reportPeriod) >> 8), 0);
    }
  } else if (status == EMBER_NETWORK_DOWN) {
    networkUp = FALSE;
    addTimer(SEND_REPORT_TIMER_INDEX, NULL, 0, 0, 0); //stop reporting
  }
}

// @see stack/include/ember.h
EmberStatus emberRemoteSetBindingHandler(int8u index,
                                         EmberBindingTableEntry *entry) 
                                         {
  //we're not allowing external control of our binding table since this
  //isn't really a provisonable application.
  return EMBER_BINDING_TABLE_WRITE_FAILED;
}

//The meter doesn't search for endpoints, but if it did, they would 
//come back here
void emberEndpointFoundHandler(EmberInterfaceId *desiredInterface,
                               EmberEUI64 node,
                               EmberNetworkDistance distance)
{
}

void emberBeaconLossCallback(void) {

}

void temporaryIncomingBeaconHandler(void) {

}

//-----------------------------------------------------------------
//board related functions, have to implement these for the dev board
void boardButtonCallback(int8u button, int8u state)
{
}

void toggleBlinker(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED3);
}


//------------------------------------------------------------------------------
// Meter Application Functions

// Whenever we hear an advertisement from a aggregator that is in our table
// we set its age to zero and increment the age of any other aggregators.
// If a aggregator's age value reaches the age limit we remove it from the
// table as being unreliable.
void ageAggregators(int8u current)
{
  int8u i;
  for (i = 0; i < MAXIMUM_AGGREGATORS; i++) {
    if (i == current)
      aggregatorAge[i] = 0;
    else if (aggregatorAge[i] < MAXIMUM_AGGREGATOR_AGE)
      aggregatorAge[i] += 1;
  }
}


// The preference is:
//   - entry already exists
//   - an unused entry
//   - a stale entry
//   - an entry with a higher cost
void considerAggregatorAdvertisement(int8u *contents)
{
  int8u i;
  EmberCost cost;
  int8u maxCost = 0;
  boolean isOld;
  int8u maxCostIndex = 0xFF;

  if (emberGetCostToSender(&cost) != EMBER_SUCCESS)
    return;

  if (findEui64Binding(contents + 1, &i)) {
    aggregatorCosts[i] = cost;
    emberNoteSendersBinding(i);
    if (i == currentAggregator
        || contents[0] == AGGREGATOR_ADVERTISEMENT)
      ageAggregators(i);
    else
      aggregatorAge[i] = MAXIMUM_AGGREGATOR_AGE;
    return;
  }

  if (contents[0] == AGGREGATOR_REMINDER)
    return;                             // it's full up

  for (i = 0; i < MAXIMUM_AGGREGATORS; i++) {
    int8u thisAge = aggregatorAge[i];
    EmberCost thisCost = aggregatorCosts[i];

    if (thisAge == 0xFF) {
      isOld = TRUE;
      maxCostIndex = i;
      break;
    } else if (thisAge == MAXIMUM_AGGREGATOR_AGE) {
      isOld = TRUE;
      maxCostIndex = i;
    } else if (! isOld && maxCost < thisCost) {
      maxCost = thisCost;
      maxCostIndex = i;
    }
  }

  if (maxCostIndex != 0xFF
      && (isOld || cost + costHysteresis < maxCost)) {
    EmberBindingTableEntry binding;

    binding.type = EMBER_UNICAST_BINDING;
    MEMCOPY(binding.identifier, contents + 1, 8);
    binding.local = ourTerminal;
    binding.remote = ourTerminal;

    assert(emberSetReplyBinding(maxCostIndex, &binding) == EMBER_SUCCESS);

    aggregatorCosts[maxCostIndex] = cost;
    ageAggregators(maxCostIndex);

    if (currentAggregator == 0xFF) {
      currentAggregator = maxCostIndex;
      announcePresence();
      emberSerialPrintf(SERIAL_PORT,
                        "have aggregator %d ",
                        currentAggregator);
      printEui64(binding.identifier);
    }
  }
}

//------------------------------------------------------------------------------

void announcePresence(void)
{
  int32u rand = ((firmGetRandom() & 0x3F) + 1);
  // Schedule meter announcement to aggregator
  addTimer(SEND_ANNOUNCE_TIMER_INDEX, sendAnnounce, 0, rand, 0);
  emberSerialPrintf(SERIAL_PORT, "announce delay %d\r\n", rand/16);
}

void startReporting(int8u timerId, int32u param)
{
  addTimer(SEND_REPORT_TIMER_INDEX, sendReport, 0, reportPeriod, 1);
}

//------------------------------------------------------------------------------

// Announce our presence to the aggregator
void sendAnnounce(int8u timerId, int32u param)
{
  EmberCost costToAggregator;
  EmberMessageBuffer report = emberFillLinkedBuffers(NULL, 11);

  if (report == EMBER_NULL_MESSAGE_BUFFER)
    return;

  if (emberGetCost(currentAggregator, &costToAggregator) != EMBER_SUCCESS)
    costToAggregator = 0;


  emberSetLinkedBuffersByte(report, 0, METER_ANNOUNCE);
  emberCopyToLinkedBuffers(emberGetEui64(), report, 1, 8);
  emberSetLinkedBuffersByte(report, 9, nextReportId);
  emberSetLinkedBuffersByte(report, 10, costToAggregator);

  if (emberSendDatagram(currentAggregator, report) == EMBER_SUCCESS)
    nextReportId += 1;
  emberReleaseMessageBuffer(report);

  addTimer(SEND_ANNOUNCE_TIMER_INDEX, NULL, 0, 0, 0);
  addTimer(SEND_REPORT_TIMER_INDEX, sendReport, 0, reportPeriod, 1);
}

//------------------------------------------------------------------------------

void sendReport(int8u timerId, int32u param)
{
  EmberCost costToAggregator;
  EmberMessageBuffer report;
  int8u i;

    report = emberFillLinkedBuffers(NULL, reportSize);
    if (report == EMBER_NULL_MESSAGE_BUFFER)
      return;

    if (emberGetCost(currentAggregator, &costToAggregator) != EMBER_SUCCESS)
      costToAggregator = 0;

    i = 0;
    emberSetLinkedBuffersByte(report, i++, METER_REPORT);
    emberSetLinkedBuffersByte(report, i++, nextReportId);
    for (; i < reportSize; i++)
      emberSetLinkedBuffersByte(report, i, i);

    if (emberSendDatagram(currentAggregator, report) == EMBER_SUCCESS)
      nextReportId += 1;
    emberReleaseMessageBuffer(report);
}

//------------------------------------------------------------------------------

// When the aggregator requests data, the meter needs to respond with meter data
// immediately as part of the datagram reply
void sendDataReply(void)
{
  EmberMessageBuffer report;
  int8u i;

  report = emberFillLinkedBuffers(NULL, reportSize);
  if (report == EMBER_NULL_MESSAGE_BUFFER)
    return;

  i = 0;
  emberSetLinkedBuffersByte(report, i++, METER_DATA_REPLY);
  emberSetLinkedBuffersByte(report, i++, nextReportId);
  for (; i < reportSize; i++)
    emberSetLinkedBuffersByte(report, i, i);

  if (emberSendReply(report) == EMBER_SUCCESS)
    nextReportId += 1;
  else
    emberSerialPrintf(SERIAL_PORT, "SendReply failed\r\n");
  emberReleaseMessageBuffer(report);
}

//------------------------------------------------------------------------------

//send historical data from the meter to the aggregator
void sendHistory(int8u timerId, int32u param)
{
  EmberStatus status;
  EmberMessageBuffer report;
  int8u message[maximumReportSize];
  int8u i;

  if (connectedToAggregator) {
    if (historyReportCount == 0) {
      // Close connection to aggregator when done
      emberCloseConnection(currentAggregator);
      connectedToAggregator = FALSE;
      addTimer(SEND_HISTORY_TIMER_INDEX, NULL, 0, 0, 0); //remove us from the timer module
      return;
    }

    // Fill data message with successive report.
    i = 0;
    message[i++] = METER_HISTORY_REPLY;
    message[i++] = nextReportId;
    for (; i < reportSize; i++) {
      message[i] = i + historyReportCount; // make up some data...
    }
    report = emberFillLinkedBuffers(message, reportSize);
    if (report == EMBER_NULL_MESSAGE_BUFFER)
      return;

    status = emberSendSequenced(currentAggregator, report);
    if (status == EMBER_SUCCESS) {
      nextReportId += 1;
      historyReportCount -= 1;
    } else {
      emberSerialPrintf(SERIAL_PORT, "SendSequenced failed (%x)\r\n", status);
      // stop sending right now
      historyReportCount = 0;
    }
    emberReleaseMessageBuffer(report);
  }
}

//creates and sets a binding to send multicasts to
//or in the case of a meter, receive multicasts from related
//to the amr application
void createMulticastBinding(int8u index)
{
  EmberBindingTableEntry entry;

  entry.type = EMBER_MULTICAST_BINDING;
  entry.identifier[0] = HIGH_BYTE(MULTICAST_GROUP);
  entry.identifier[1] = LOW_BYTE(MULTICAST_GROUP);
  entry.local = ourTerminal;

  assert(emberSetBinding(index, &entry) == EMBER_SUCCESS);
}

//A utility function to simpify the code involved in sending a datagram
//within the amr application.
EmberStatus transmitAmrMessage(
  int8u bindingIndex,
  int8u *message,
  int8u length)
{
  EmberStatus status = EMBER_SUCCESS;
  EmberMessageBuffer buffer = emberFillLinkedBuffers(message, length);

  if (buffer == EMBER_NULL_MESSAGE_BUFFER)
    return EMBER_NO_BUFFERS;

  status = emberSendDatagram(bindingIndex, buffer);

  emberReleaseMessageBuffer(buffer);
  
  return status;
}

//find a binding by the Eui64 Id
boolean findEui64Binding(EmberEUI64 key, int8u *index)
{
  int8u i;
  int8u unused = 0xFF;

  for (i = 0; i < emberBindingTableSize - 1; i ++) {
    EmberBindingTableEntry binding;
    if (emberGetBinding(i, &binding) == EMBER_SUCCESS) {
      if (binding.type == EMBER_UNICAST_BINDING
          && MEMCOMPARE(key, binding.identifier, 8) == 0) {
        *index = i;
        return TRUE;
      } else if (binding.type == EMBER_UNUSED_BINDING
                 && unused == 0xFF)
        unused = i;
    }
  }

  *index = unused;
  return FALSE;
}

//make binding table size available to other portions of the code
int8u getBindingTableSize() {
  return EMBER_BINDING_TABLE_SIZE;
}




