//******************************************************************************
//  sequenced-msg-app.c
//
//  Sample application to show transferring a large amount of data
//  from one node to another using sequenced messages.
//  This is done by fragmenting the data and sending each fragment as its
//  own sequenced message.
//  We also need to keep track of how many buffers we're using.  We can only
//  send so many messages before buffers are used up, so we set up a message
//  count limit, after which we wait for a message-sent confirmation from the
//  emberMessageSent callback.
//
//  Copyright 2004 by Ember Corporation. All rights reserved.               *80*
//******************************************************************************

#if !defined(SIMULATOR)
#include <ioavr.h>
#endif
#include PLATFORM_HEADER //compiler/micro specifics, types
#include "stack/include/ember.h"
#include "stack/include/error.h"
#include "stack/include/hal.h" // firmGetRandom()
#include "hal/board/board.h"   // LED definitions
#include "hal/micro/micro.h"   // microBoot, memcopy, sleep, launch bootloader
#include "stack/include/packet-buffer.h"
#include "app/util/serial/serial.h"

// ************************************
// Ember static memory requirements
//
// There are constants that define the amount of static memory
// required by the stack. If the application does not set them then
// the default values from ember-static-configuration.h are used.
//
// 2 bindings:
// 1 for multicast messaging
// 1 binding to another node
#define EMBER_BINDING_TABLE_SIZE        15
// all entries in binding table are temporary (in RAM, not flash)
#define EMBER_TEMPORARY_BINDING_ENTRIES 15
#include "stack/include/ember-static-configuration.h"

// ************************************
// Include definitions for application timers
#include "app/sample/app-timers.h"

// ************************************
// Serial port definitions.  We define the debug serial port because it needs
// to be initialized in case the debug stack library is being used.
#define DEBUG_SERIAL 0
#define IO_SERIAL 1

//******************************************************************************
// Define endpoints and interfaces for the Ember Stack
// Design:
// We will use one endpoint.
// This endpoint will have the following interfaces:
// 1) A fixed multicast interface for general messages
// 2) A server interface (for a controlled device)
// 3) A client interface (for the gateway)

// Define interfaces
// (indices into the list of interfaces in the endpoint structure)
#define MULTICAST_INTERFACE 0
#define SERVER_INTERFACE    1
#define CLIENT_INTERFACE    2

// Define endpoints.
#define DEVICE_CONTROL_ENDPOINT   1

// Specification IDs for the ad-hoc interfaces
#define MULTICAST_SPEC              0x56
#define DEVICE_CONTROL_INTERFACE_ID 0x71

// Define the interface structures
// Ad-hoc interfaces contain: interface domain, endpoint type,
// #-specification-bytes and specification (unique identifier)

// Multicast interface, so all devices can talk to each other.
static PGM EmberInterfaceId mcastInterfaceId = {
  AD_HOC_INTERFACE,
  SYMMETRIC_ENDPOINT,
  1, MULTICAST_SPEC };

// Server interface
static PGM EmberInterfaceId deviceServerInterfaceId = {
  AD_HOC_INTERFACE,
  SERVER_ENDPOINT,
  1, DEVICE_CONTROL_INTERFACE_ID };

// Client interface
static PGM EmberInterfaceId deviceClientInterfaceId = {
  AD_HOC_INTERFACE,
  CLIENT_ENDPOINT,
  1, DEVICE_CONTROL_INTERFACE_ID };


static PGM EmberEndpointDescription deviceControlEndpoint = {
  { MULTICAST_INTERFACE, &mcastInterfaceId },
  { SERVER_INTERFACE,    &deviceServerInterfaceId },
  { CLIENT_INTERFACE,    &deviceClientInterfaceId }
};


// Global storage required by the Ember stack
// This tells the stack we're using only one endpoint:
int8u emberEndpointCount = 1;
int8u index = 0;
// This defines the endpoints we're using:
EmberEndpoint emberEndpoints[] = {
  { DEVICE_CONTROL_ENDPOINT, &deviceControlEndpoint}
};


//-------------------------------------
// Definitions for binding table entries
#define REMOTE_BINDING_INDEX 0
#define MCAST_BINDING_INDEX  EMBER_BINDING_TABLE_SIZE - 1

// Multicast address used for binding entry
#define MCAST_ADDRESS 0x4A50


//-------------------------------------
// Message types
typedef enum {
  // Multicast messages
  SEARCHER     = 0x10, // Node sends to find another device in network

  // Unicast messages from remote node to searcher node
  RESPONDER    = 0x11, // Remote node sends to identify itself to searcher

  // Unicast datagram messages
  QUERY_DEVICE = 0x20, // Gateway asking for data
  STOP_REPORT  = 0x21, // Gateway asking node to stop sending data

  // Unicast sequenced messages
  DATA_REPORT  = 0x30  // Data messages from remote node
} DeviceControlMessageType;


//******************************************************************************
// Application-specific constants and globals

//#define DEBUG_DATA
#if defined(DEBUG_DATA)
  #define DBG(x) do { x; } while(0)
#else
  #define DBG(x)
#endif

// Number of bytes in data block to transfer
#define DATA_BLOCK_SIZE 1550
// Size of each fragment of data block sent as one message
#define MAX_FRAGMENT_SIZE (MAX_APP_PAYLOAD_SIZE-10) //64
#define BUFFERS_PER_FRAGMENT (MAX_FRAGMENT_SIZE / PACKET_BUFFER_SIZE + 1)

// Number of messages to have active at one time before waiting for the
// MessageSent callback.
// This is done because if, for example, EMBER_PACKET_BUFFER_COUNT
// (in ember-static-configuration.h) is 24, and we're using 3 buffers per message
// (MAX_FRAGMENT_SIZE / PACKET_BUFFER_SIZE), then if we send 8 messages, we've
// used up all the buffers in the system, leaving none for the stack to do
// negotiation of the messages being sent!
// This is being set such that there are about 8 buffers (that we know of)
// available to the rest of the system.
#define MAX_MESSAGES ((EMBER_PACKET_BUFFER_COUNT - 8) / BUFFERS_PER_FRAGMENT)

// Overall byte counter on sender and receiver.
static int16u dataBlockCount = 0;
// Receiver is told in the first fragment how many bytes to expect altogether.
// It stores it here, and uses this to tell when it's done receiving fragments.
static int16u receivedDataBlockSize = 0;

// Data block storage - dummy.  Should really be DATA_BLOCK_SIZE, but we don't
// have that much space left in RAM.
static int8u dataBlock[MAX_FRAGMENT_SIZE];
static int8u * dataBlockStartAddress;

// Outgoing message counter, for guarding against using up all the buffers
// in the system.
// Range is from 0 to MAX_MESSAGES-1
static int8u messageCount = 0;
// Flag indicating previous set of messages have been sent (emberMessageSent callback)
static boolean previousMessagesSent = FALSE;

// Current status of the stack
static EmberStatus stackStatus = EMBER_NETWORK_DOWN;

static int8u myChannel = 0;
static int8s myPower = -1;

// Holds count of nodes that respond to node search
static int8u remoteNodeFound = 0;

// Flag indicating a binding has been created to a remote node
static boolean bindingExists = FALSE;

// Flag indicating we have a connection open
static boolean connectedToRemote = FALSE;

// structure to store necessary network parameters of the node
// (which are panId, enableRelay, radioTxPower, and radioChannel)
EmberNetworkParameters networkParams;

//******************************************************************************

// Forward declarations
static void processSerialInput(void);
static boolean isDuplicateBinding(EmberBindingTableEntry* newEntry);
static void endNodeSearch(int8u timerId, int32u param);
static void sendStreamingData(int8u timerId, int32u param);
static void stopStreamingData(void);

static int8u * getDataHandler(int16u receivedDataBlockSize);
static void dataBlockDoneCallback(void);

//******************************************************************************

// This will blink a red LED to indicate this is not the gateway node.
static void blinkLED(int8u timerId, int32u param)
{
  boardLEDToggle(BOARDLED3);
}

// Actions grouped together because they need to be done with emberTick
static void runSystem(void)
{
  __watchdog_reset();
  emberTick();
  emberSerialBufferTick();
  handleTimers();
  simulatedTimePasses();
}

//******************************************************************************
// Main application loop

#if defined(SIMULATOR)
int nodeMain(void)
#else
int main(void)
#endif
{
  EmberResetType reset;
  EmberStatus status;
  EmberBindingTableEntry entry;

  // Determine the cause of the reset (powerup, etc).
  reset = microGetResetInfo();
  microBoot();

  // to init the leds and such..
  boardBoot();

  // reset the system tick
  // enable the system tick interrupt
  microStartSystemTick();
  // allow interrupts
  INTERRUPTS_ON();

  // inititialize the serial port
  emberSerialInit(IO_SERIAL, BAUD_19200, PARITY_NONE, 1);
  emberDebugInit(DEBUG_SERIAL);

  // Initialize the Ember Stack.
  status = emberInit(reset);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, 
      "error returned from emberInit: 0x%x\r\n",
      status);
    if (status == EMBER_UNINITIALIZED_EUI64) {
      // Run to allow over-the-air configuration.
      while (TRUE) {
        __watchdog_reset();
        emberTick();
      }
    } else {
      assert(FALSE);
    }
  }

  initTimers();

  // Bring the network up.
  networkParams.panId = 0;
  networkParams.enableRelay = TRUE;
  networkParams.radioTxPower = myPower;
  networkParams.radioChannel = myChannel;
  
  status = emberAssociate(&networkParams);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, 
      "error returned from emberAssociate: 0x%x\r\n",
      status);
    assert(FALSE);
  }

  //-------------------------------------------------------

  // Wait for stack to come up before doing anything
  while (stackStatus != EMBER_NETWORK_UP)
    runSystem();

  //-------------------------------------------------------

  boardLEDClear(BOARDLED0);
  boardLEDClear(BOARDLED1);
  boardLEDClear(BOARDLED2);
  boardLEDSet(BOARDLED3);
  // Create timer to blink LEDs for heartbeat
  // Blink every 1/2 second (8 clock ticks)
  addTimer(0, blinkLED, 0, 8, 1);

  emberSerialGuaranteedPrintf(IO_SERIAL,
              "\r\nSequenced Message sample app running on node %X%X, channel %d\r\n",
              emberGetEui64()[6], emberGetEui64()[7], myChannel);
  emberSerialGuaranteedPrintf(IO_SERIAL, "Press 'h' for menu of commands\r\n");

  DBG(
    emberSerialGuaranteedPrintf(IO_SERIAL, "\r\nMAX_FRAGMENT_SIZE=%d\r\n", MAX_FRAGMENT_SIZE);
    emberSerialGuaranteedPrintf(IO_SERIAL, "BUFFERS_PER_FRAGMENT=%d\r\n", BUFFERS_PER_FRAGMENT);
    emberSerialGuaranteedPrintf(IO_SERIAL, "MAX_MESSAGES=%d\r\n", MAX_MESSAGES);
  );

  //-------------------------------------------------------

  // make the multicast binding table entry
  entry.type = EMBER_MULTICAST_BINDING;
  entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT, MULTICAST_INTERFACE);
  *(int16u *)entry.identifier = MCAST_ADDRESS;
  if (emberSetBinding(MCAST_BINDING_INDEX, &entry) != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL, "Failed to make binding\r\n");

  //-------------------------------------------------------

  while (TRUE) {
    runSystem();
    // Don't run anything else unless the stack is running
    if (stackStatus == EMBER_NETWORK_UP) {
      processSerialInput();
    }
  }
}

//******************************************************************************
// Application Timer functions
// These functions get set up by code below to execute after a certain delay.
//******************************************************************************

// This timer function is how the gateway ends its search for remote nodes.
// It is set up when the search is started.
static void endNodeSearch(int8u timerId, int32u param)
{
  if (remoteNodeFound) {
    emberSerialGuaranteedPrintf(
      IO_SERIAL, "ending search,node found\r\n");
  } else
    emberSerialGuaranteedPrintf(
      IO_SERIAL, "ending search,NO NODE FOUND!\r\n");
}

//---------------------------------------------------------

// This timer function is set up to run by a remote node when the gateway tells
// it to send a large block of data.
static void sendStreamingData(int8u timerId, int32u param)
{
  int8u message[MAX_FRAGMENT_SIZE];
  EmberMessageBuffer pbuff;
  int8u i, currentSize;
  EmberStatus status;

  while ((dataBlockCount < DATA_BLOCK_SIZE) &&
         connectedToRemote &&
         previousMessagesSent &&
         (messageCount < MAX_MESSAGES)) {

    // Fill data message
    message[0] = DATA_REPORT;
    i = 1;
    // First message contains the total size of the data block being sent.
    if (dataBlockCount == 0) {
      *(int16u *)&message[1] = DATA_BLOCK_SIZE;
      i += 2;
    }
    for (currentSize = 0;
         (i < sizeof(message)) && (dataBlockCount < DATA_BLOCK_SIZE);
         i++) {
      message[i] = i;
      currentSize += 1;
      dataBlockCount += 1;
    }
    message[3] = emFreePacketBufferCount();
    message[4] = i;
    DBG(emberSerialGuaranteedPrintf(IO_SERIAL,
         "free buffers: %d, want %d bytes\r\n", message[3], i));
    pbuff = emberFillLinkedBuffers(message, i);
    DBG(emberSerialGuaranteedPrintf(IO_SERIAL,
          "got %d bytes\r\n", emberMessageBufferLength(pbuff)));

    // Didn't get as many bytes as we requested: fail
    if (emberMessageBufferLength(pbuff) != i) {
      DBG(emberSerialGuaranteedPrintf(IO_SERIAL, "wanted %d bytes, got %d\r\n",
        i, emberMessageBufferLength(pbuff)));
      emberReleaseMessageBuffer(pbuff);
      pbuff = EMBER_NULL_MESSAGE_BUFFER;
    }

    // Ran out of buffers: fail
    if (pbuff == EMBER_NULL_MESSAGE_BUFFER) {
      dataBlockCount -= currentSize;  // Current block failed: rewind
      DBG(emberSerialGuaranteedPrintf(IO_SERIAL,
            "Buffer alloc failed (%d)\r\n", dataBlockCount));
      // Reschedule
      addTimer(timerId, sendStreamingData, 0, 1, 0);
      return;
    }

    DBG(emberSerialGuaranteedPrintf(IO_SERIAL,
         "Calling SendSequenced (%d, %d, %d)\r\n",
         messageCount, currentSize, dataBlockCount));
    status = emberSendSequenced(REMOTE_BINDING_INDEX, pbuff);
    emberReleaseMessageBuffer(pbuff);
    if (status != EMBER_SUCCESS) {
      dataBlockCount -= currentSize;  // Current block failed: rewind
      DBG(emberSerialGuaranteedPrintf(IO_SERIAL,
            "SendSequenced failed %x (%d)\r\n", status, dataBlockCount));
      // Reschedule
      addTimer(timerId, sendStreamingData, 0, 4, 0);
      return;
    }
    // Update and check message counter.
    messageCount += 1;
    if (messageCount >= MAX_MESSAGES) {
      previousMessagesSent = FALSE;
      messageCount = 0;
    }

    // Need to run these to handle buffers and the Tx messages
    emberTick();
    emberSerialBufferTick();
  }  // end while

  // Check if we're done
  if (dataBlockCount >= DATA_BLOCK_SIZE) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "Data done (%d bytes).\r\n", dataBlockCount);
    stopStreamingData();
    dataBlockCount = 0;
  } else
    // Schedule next set of messages
    addTimer(timerId, sendStreamingData, 0, 1, 0);
}

//---------------------------------------------------------

// Send search response back to searching node, with our EUI64 ID.
// This is a timer function because we set a random timeout to perform this
// action, so that remote nodes respond at different times.
static void respondToSearch(int8u timerId, int32u param)
{
  EmberMessageBuffer responseMessage;
  EmberStatus status;
  int8u messageBuffer[9];

  messageBuffer[0] = RESPONDER;
  MEMCOPY(messageBuffer + 1, emberGetEui64(), 8);

  responseMessage =
    emberFillLinkedBuffers(messageBuffer, sizeof(messageBuffer));
  if (responseMessage == EMBER_NULL_MESSAGE_BUFFER) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "Alloc failed\r\n");
    return;
  }

  status = emberSendDatagram(REMOTE_BINDING_INDEX, responseMessage);
  if (status != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "SendDatagram failed: %x\r\n", status);

  emberReleaseMessageBuffer(responseMessage);
}

//******************************************************************************
// Utility functions
//******************************************************************************

// The user will want to know which binding entries to use.  This is most
// useful on the gateway, or for debugging a remote node (to make sure it
// has opened a connection to the gateway).
void printBindingTable(void)
{
  int8u i;
  EmberStatus status;
  EmberBindingTableEntry result;
  int8u* id;
  int8u indexLow;
  int8u indexHigh;
  emberSerialPrintf(IO_SERIAL, "status index type local remote id\r\n");

  for (i=0; i<EMBER_BINDING_TABLE_SIZE; i++) {
    emberSerialWaitSend(IO_SERIAL);

    status = emberGetBinding(i, &result);

    indexLow = (i % 10) + 48;
    indexHigh = ((i - (i % 10))/10) + 48;

    id = result.identifier;

    emberSerialPrintf(IO_SERIAL, "0x%x  | %c%c | ",
                      status, indexHigh, indexLow);

    if (EMBER_SUCCESS == status) {
      switch(result.type) {
      case EMBER_UNUSED_BINDING:
        emberSerialPrintf(IO_SERIAL, "unused....|");
        break;
      case EMBER_UNICAST_BINDING:
        emberSerialPrintf(IO_SERIAL, "unicast...|");
        break;
      case EMBER_MULTICAST_BINDING:
        emberSerialPrintf(IO_SERIAL, "multicast.|");
        break;
      default:
        emberSerialPrintf(IO_SERIAL, "?????????.|");
      }

      emberSerialPrintf(IO_SERIAL, "0x%x  0x%x  ",
                        result.local, result.remote);

      emberSerialPrintf(IO_SERIAL, "%x %x %x %x %x %x %x %x\r\n",
                        id[0],id[1],id[2],id[3],id[4],id[5],id[6],id[7]);

    } else {
      emberSerialPrintf(IO_SERIAL, " --- ERROR ---\r\n ");
    }
  }
  emberSerialWaitSend(IO_SERIAL);
}

//---------------------------------------------------------

// This function is run by a node to search for another node.
static void startNodeSearch()
{
  EmberMessageBuffer searchMessage;
  EmberStatus status;
  int8u messageBuffer[9];

  remoteNodeFound = 0;

  messageBuffer[0] = SEARCHER;
  MEMCOPY(messageBuffer + 1, emberGetEui64(), 8);
  searchMessage = emberFillLinkedBuffers(messageBuffer, sizeof(messageBuffer));
  if (searchMessage == EMBER_NULL_MESSAGE_BUFFER) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "Alloc failed\r\n");
    return;
  }

  status = emberSendMulticast(MCAST_BINDING_INDEX, searchMessage);
  if (status != EMBER_SUCCESS) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "SendMulticast failed: %x", status);
  }

  // we are done with the buffer
  emberReleaseMessageBuffer(searchMessage);

  // Create search timeout: 48 => 3 seconds
  addTimer(1, endNodeSearch, 1, 48, 0);
}

//---------------------------------------------------------

static boolean isDuplicateBinding(EmberBindingTableEntry* newEntry)
{
  int8u i;
  EmberBindingTableEntry existingEntry;
  for (i = 0; i < EMBER_BINDING_TABLE_SIZE; i++) {
    // If error, just return that we found a duplicate entry,
    // so we don't try to add a binding entry with an invalid index.
    if (emberGetBinding(i, &existingEntry) == EMBER_BINDING_INDEX_OUT_OF_RANGE)
      return TRUE;
    if (MEMCOMPARE(&existingEntry,
                   newEntry,
                   sizeof(EmberBindingTableEntry)) == 0) {
      return TRUE;
    }
  }
  return FALSE;
}

//---------------------------------------------------------

// Node sends a datagram to a remote node, to get it to send a large block
// of data as sequenced messages.
void queryRemoteDevice(int8u bindingIndex)
{
  EmberMessageBuffer pbuff;
  int8u message[] = {QUERY_DEVICE};
  EmberStatus status;

  pbuff = emberFillLinkedBuffers(message, sizeof(message));
  if (pbuff == EMBER_NULL_MESSAGE_BUFFER) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "Alloc failed\r\n");
    return;
  }

  status = emberSendDatagram(bindingIndex, pbuff);
  if (status != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "Failed to send query msg: %X\r\n", status);
  emberReleaseMessageBuffer(pbuff);
}

//---------------------------------------------------------
/*
// Sent to have remote node stop sending data.
// Used to cancel a block of data being sent.
void endQueryRemoteDevice(int8u bindingIndex)
{
  EmberMessageBuffer pbuff;
  int8u message[] = {STOP_REPORT};
  EmberStatus status;

  pbuff = emberFillLinkedBuffers(message, sizeof(message));
  if (pbuff == EMBER_NULL_MESSAGE_BUFFER) {
    emberSerialGuaranteedPrintf(IO_SERIAL, "Alloc failed\r\n");
    return;
  }

  status = emberSendDatagram(bindingIndex, pbuff);
  if (status != EMBER_SUCCESS)
    emberSerialGuaranteedPrintf(IO_SERIAL,
      "Failed to send end query msg, %X\r\n", status);
  emberReleaseMessageBuffer(pbuff);

  dataBlockCount = 0;
}
*/
//---------------------------------------------------------

// Executed by a node when it's told to stop sending data,
// or when it's told to stop network interactions and delete all bindings.
static void stopStreamingData(void)
{
  // End streaming data to gateway
  addTimer(2, NULL, 0, 0, 0);

  emberCloseConnection(REMOTE_BINDING_INDEX);

  dataBlockCount = 0;
}

//---------------------------------------------------------

// Run to delete all binding entries.
void deleteAllLocalBindings(void)
{
  EmberStatus status;
  // Delete bindings
  status = emberDeleteBinding(REMOTE_BINDING_INDEX);
//  MEMSET(&devicesFound[i], 0, sizeof(EmberEUI64));
  emberSerialGuaranteedPrintf(IO_SERIAL, "local binding %d deleted (%x)\r\n",
    REMOTE_BINDING_INDEX, status);

  // Last: clear binding counter
  remoteNodeFound = 0;
}

//******************************************************************************
// Interface handlers (called from emberIncomingMessageHandler)
//******************************************************************************

// Called by emberIncomingMessageHandler to handle EMBER_INCOMING_DATAGRAM
static void incomingDatagramHandler(int8u localTerminal,
                                    int8u bindingTableIndex,
                                    EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  emberSerialGuaranteedPrintf(IO_SERIAL, "Datagram: ");
  if (emberTerminalInterface(localTerminal) == SERVER_INTERFACE) {
    //-----------------------------------------------------
    // Unicast messages received from client
    //-----------------------------------------------------
    if (data[0] == QUERY_DEVICE) {
      emberSerialGuaranteedPrintf(
        IO_SERIAL, "query command.\r\n");

      // Open connection
      emberOpenConnection(REMOTE_BINDING_INDEX);

      // Schedule streaming data, to get out of message receive handlers quick
      dataBlockCount = 0;
      addTimer(2, sendStreamingData, 0, 1, 0);

      // Need to set this TRUE to get transmission started
      previousMessagesSent = TRUE;
    //-----------------------------------------------------
    } else if (data[0] == STOP_REPORT) {
      emberSerialGuaranteedPrintf(
        IO_SERIAL, "stop report command.\r\n");

      stopStreamingData();
    //-----------------------------------------------------
    }
    else  // First byte of message unknown
      emberSerialGuaranteedPrintf(
        IO_SERIAL,
        "unknown message [%x] from binding %d\r\n",
        data[0], bindingTableIndex);
    //-----------------------------------------------------
  } else if (emberTerminalInterface(localTerminal) == CLIENT_INTERFACE) {
    //-----------------------------------------------------
    // Unicast messages received from server
    //-----------------------------------------------------
    if (data[0] == RESPONDER) {
      EmberBindingTableEntry entry;
      EmberStatus status;

      emberSerialGuaranteedPrintf(
        IO_SERIAL, "response from %x%x\r\n", data[2], data[1]);

      // Set bindings to remote nodes as they respond.
      entry.type = EMBER_UNICAST_BINDING;
      MEMCOPY(entry.identifier, &data[1], 8);
      entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                      CLIENT_INTERFACE);
      entry.remote = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                       SERVER_INTERFACE);

      // Guard against duplicate binding table entries.
      if (isDuplicateBinding(&entry)) {
        emberSerialGuaranteedPrintf( IO_SERIAL, "already have binding set\r\n");
      } else {
        // Set a binding for ourselves to that node
        if(index < EMBER_BINDING_TABLE_SIZE) {
          status = emberSetReplyBinding(index, &entry);
          index += 1;
        }
        if (status == EMBER_SUCCESS) {
          emberSerialGuaranteedPrintf(IO_SERIAL,
            "binding %d set\r\n", remoteNodeFound);
          bindingExists = TRUE;
        }
      }
      remoteNodeFound += 1;
    }
    else  // First byte of message unknown
      emberSerialGuaranteedPrintf(
        IO_SERIAL,
        "unknown message [%x] from binding %d\r\n",
        data[0], bindingTableIndex);
    //-----------------------------------------------------
  } else {
    emberSerialGuaranteedPrintf(
      IO_SERIAL,
      "message [%x] from binding %d, bad terminal\r\n",
      data[0], bindingTableIndex);
  }
}

//---------------------------------------------------------

// Called by emberIncomingMessageHandler to handle EMBER_INCOMING_MULTICAST
static void incomingMulticastHandler(int8u localTerminal,
                                     int8u bindingTableIndex,
                                     EmberMessageBuffer message)
{
  int8u *data = emberMessageBufferContents(message);

  if (emberTerminalInterface(localTerminal) == MULTICAST_INTERFACE) {
    //-----------------------------------------------------
    if (data[0] == SEARCHER) {
      EmberBindingTableEntry entry;
      EmberStatus status;

      emberSerialGuaranteedPrintf(
        IO_SERIAL, "search from %x%x\r\n", data[2], data[1]);

      // Create binding to searcher node
      // We're using emberSetReplyBinding because at this point, the
      // stack doesn't have to rediscover the original sender.
      entry.type = EMBER_UNICAST_BINDING;
      MEMCOPY(entry.identifier, &data[1], 8);
      entry.local = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                      SERVER_INTERFACE);
      entry.remote = emberMakeTerminal(DEVICE_CONTROL_ENDPOINT,
                                       CLIENT_INTERFACE);
      status = emberSetReplyBinding(REMOTE_BINDING_INDEX, &entry);
      if (status == EMBER_SUCCESS) {
        emberSerialGuaranteedPrintf(IO_SERIAL, "binding set\r\n");
        bindingExists = TRUE;

        // Schedule a response back to the searcher node
        // The timeout will be random: 0 thru almost 2 seconds
        addTimer(1, respondToSearch, 0, (firmGetRandom() & 0x1F), 0);
      } else {
        emberSerialGuaranteedPrintf(IO_SERIAL, "binding failed(%x)\r\n", status);
      }
    //-----------------------------------------------------
    } else  // First byte of message unknown
      emberSerialGuaranteedPrintf(IO_SERIAL,
                        "Unknown mcast message from binding %x\r\n",
                        bindingTableIndex);
    //-----------------------------------------------------
  }
  else { // Not multicast
    emberSerialGuaranteedPrintf(IO_SERIAL,
                      "Unknown message (not mcast) from binding %x\r\n",
                      bindingTableIndex);
  } // end if(MULTICAST_INTERFACE)
}

//******************************************************************************
// Handlers required to use the Ember Stack
// Populate as needed
//******************************************************************************

void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 int8u localTerminal,
                                 int8u bindingTableIndex,
                                 EmberMessageBuffer message)
{
  int8u *data; // = emberMessageBufferContents(message);

  if (!emberMessageBufferLength(message))
    return;

  data = emberMessageBufferContents(message);

  switch (type) {
  case EMBER_INCOMING_DATAGRAM:
    incomingDatagramHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_MULTICAST:
    incomingMulticastHandler(localTerminal, bindingTableIndex, message);
    break;

  case EMBER_INCOMING_SEQUENCED:
  {
    if (data[0] == DATA_REPORT) {
      int8u size;
      int8u * msgAddr;
      if (dataBlockCount == 0) {
        // First message contains the total size of the data block being sent.
        receivedDataBlockSize = *(int16u *)&data[1];
        size = emberMessageBufferLength(message)-3;
        msgAddr = &data[3];
        dataBlockStartAddress = getDataHandler(receivedDataBlockSize);
      } else {
        size = emberMessageBufferLength(message)-1;
        msgAddr = &data[1];
      }

      // Copy the fragment into local RAM.
      // We only have room for one fragment.  This is OK because we just want to
      // take up time doing the copy for demonstration purposes.
      //MEMCOPY(dataBlockStartAddress+dataBlockCount, msgAddr, size);
      MEMCOPY(dataBlockStartAddress, msgAddr, size);
      dataBlockCount += size;

      DBG(
        emberSerialGuaranteedPrintf(IO_SERIAL,
          "Received data from binding %x (%d), %d free bfrs\r\n",
          bindingTableIndex, dataBlockCount, emFreePacketBufferCount());
        for (i = 1; i <= 4; i++) {
          emberSerialGuaranteedPrintf(IO_SERIAL, "%x ", data[i]);
        }
        emberSerialGuaranteedPrintf(IO_SERIAL, "\r\n");
      );
      if (dataBlockCount >= receivedDataBlockSize) {
        dataBlockDoneCallback();
        dataBlockCount = 0;
      }
    }
    break;
  }

  case EMBER_INCOMING_DATAGRAM_REPLY:
  case EMBER_INCOMING_SHARED_MULTICAST:
  case EMBER_INCOMING_MULTICAST_LOOPBACK:
  default:
    emberSerialGuaranteedPrintf(
      IO_SERIAL,
      "Unexpected message (type %d) [ %x] binding %d, for terminal %d\r\n",
      type, data[0], bindingTableIndex, localTerminal);
    break;
  }
}

//---------------------------------------------------------

void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status)
{
  emberSerialGuaranteedPrintf(
    IO_SERIAL, "Connection to binding %d: %x\r\n", bindingTableIndex, status);

  connectedToRemote = (status == EMBER_CONNECTION_OPEN);
}

//---------------------------------------------------------

void emberMessageSent(int8u bindingTableIndex,
                      EmberMessageBuffer message,
                      EmberStatus status)
{
  if (status == EMBER_SUCCESS)
    previousMessagesSent = TRUE;
}

//---------------------------------------------------------

EmberStatus emberRemoteSetBindingHandler(int8u index,
                                         EmberBindingTableEntry *entry)
{
  // Don't let anyone change our bindings.
  return EMBER_INVALID_BINDING_INDEX;
}

//---------------------------------------------------------

void emberEndpointFoundHandler(EmberInterfaceId *desiredInterface,
                               EmberEUI64 node,
                               EmberNetworkDistance distance)
{
}

//---------------------------------------------------------

void boardButtonCallback(int8u button, int8u state)
{
}

//---------------------------------------------------------

void emberStackStatusHandler(EmberStatus status)
{
  stackStatus = status;
}

//******************************************************************************
// End of Handlers
//******************************************************************************

//******************************************************************************
// Application handlers and callbacks
//******************************************************************************

static int8u * getDataHandler(int16u receivedDataBlockSize)
{
  emberSerialGuaranteedPrintf(IO_SERIAL,
    "Receiving %d bytes of data...\r\n", receivedDataBlockSize);
  return dataBlock;
}

static void dataBlockDoneCallback(void)
{
  emberSerialGuaranteedPrintf(IO_SERIAL,
    "Received %d bytes of data\r\n", dataBlockCount);
}

//******************************************************************************
// Serial command interface functions
//******************************************************************************

static void getParams(char *line, int8u length, int32u *argv, int8u *argc)
{
  int8u i;
  int32u temp = 0;
  int8u count = 0;
  for (i = 0; i < length; i++) {
    if ((line[i] >= '0') && (line[i] <= '9')) {
      temp = (temp * 10) + (line[i] - '0');
    } else if ((line[i] == 0x20) || (line[i] == 0x09)) {
      argv[count++] = temp;
      temp = 0;
    } else {
      count--;
      break; // die on first invalid character
    }
  }
  argv[count++] = temp;
  *argc = count;
}

#define LINE_BUFFER_MAX 30
static int8u lineLen = 0;
static char lineBuffer[LINE_BUFFER_MAX];

static void processSerialInput(void) {
  int8u ch;
  int8u cmd = '\0';
  int32u argv[2] = {0, 0};  // Handle up to 2 numeric arguments
  int8u argc = 0;
  EmberStatus status;

  if (emberSerialReadByte(IO_SERIAL, &ch) != EMBER_SUCCESS) return;

  if (ch == 0x0A)
    return;
  else if ((ch == 0x0D) || (lineLen >= LINE_BUFFER_MAX)) {
    cmd = lineBuffer[0];
  } else if (ch == 0x08) { // Backspace
    lineLen--;
    lineBuffer[lineLen] = '\0';
    emberSerialPrintf(IO_SERIAL," \010");
    return;
  } else {
    lineBuffer[lineLen++] = ch;
    return;
  }

  if (lineLen >= 2) {
    getParams(&lineBuffer[2], lineLen-2, argv, &argc);
    lineBuffer[0] = '\0';
  }

  emberSerialPrintf(IO_SERIAL,"\r\n");

  switch(cmd) {
  // bootload
  case 'b':
    emberSerialGuaranteedPrintf(IO_SERIAL, "starting bootloader...\r\n");
    microLaunchBootloader(MICRO_LAUNCH_BOOTLOADER_KEY);
    break;

  // info
  case 'i':
    myChannel = emberGetRadioChannel();
    emberSerialGuaranteedPrintf(IO_SERIAL,
                "\r\nnode %X%X, channel %d, %d free buffers\r\n",
                emberGetEui64()[1], emberGetEui64()[0], myChannel,
                emFreePacketBufferCount());
    break;

  // reset using watchdog timeout
  case 'r':
    emberSerialPrintf(IO_SERIAL, "\r\nresetting...\r\n");
    while (TRUE) ;  // wait for watchdog timer reset
//    break;

  case 'c':
    if (argc == 1) {
      if(emberDisassociate() != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"Cannot disassociate from network\r\n");  
        return;
      } 
      // waiting for the stack to come down after disassociate from network
      while(stackStatus != EMBER_NETWORK_DOWN){
       __watchdog_reset();
       emberTick();
      }
      networkParams.radioChannel = argv[0];
      status = emberAssociate(&networkParams);
      if(status != EMBER_SUCCESS) {
        emberSerialPrintf(IO_SERIAL,"can't set radio channel, 0x%x\r\n", status);
        while(1) ;
      }
      else {
        emberSerialPrintf(IO_SERIAL,"Set channel successfully, waiting for stack to come up");
        while(stackStatus != EMBER_NETWORK_UP){
          __watchdog_reset();
          emberTick();
        }
      }
    } else
      emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

  // search
  case 's':
    if (!bindingExists) {
      emberSerialPrintf(IO_SERIAL, "starting search\r\n");
      startNodeSearch();
    } else {
      emberSerialPrintf(IO_SERIAL, "A binding already exists!\r\n");
    }
    break;

  case 'p':
    printBindingTable();
    break;

  // Query node to start sending data
  case 'q':
    if (argc == 1) {
      queryRemoteDevice(argv[0]);
      } else
        emberSerialPrintf(IO_SERIAL,"Incorrect arguments\r\n");
    break;

  // Tell node to stop sending data
/*  case 'e':
    endQueryRemoteDevice(0);
    break;
*/
  case 'd':
    if (connectedToRemote)
      emberCloseConnection(REMOTE_BINDING_INDEX);
    deleteAllLocalBindings();
    // No binding, no connection
    bindingExists = FALSE;
    connectedToRemote = FALSE;
    break;

  //-----------------------------------------------------

  case 'h':
    emberSerialGuaranteedPrintf(IO_SERIAL, "\r\nNode %X%X, channel %d\r\n",
                emberGetEui64()[1], emberGetEui64()[0], myChannel);
    emberSerialGuaranteedPrintf(IO_SERIAL, "b: bootload\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "i: info\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "r: reset\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "c chan: set chan\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "s: node search\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "p: print binding table\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "q binding: Query remote node\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "e binding: End data from remote node\r\n");
    emberSerialGuaranteedPrintf(IO_SERIAL, "d: del bindings\r\n");
    break;

  //-----------------------------------------------------

  case '\0':
  case '\n':
  case '\r':
    break;

  default:
    emberSerialPrintf(IO_SERIAL, "unknown cmd\r\n");
    break;
  } // end switch

  cmd = 0;
  lineLen = 0;
  lineBuffer[0] = '\0';
  lineBuffer[1] = '\0';
}

//******************************************************************************

