/** @file app/util/rnap/rnap-app.h
 * @brief Declarations for an RNAP serial connection on the gateway.
 *
 * This file provides internal declarations for general RNAP serial connection
 * on the gateway node. If you want to modify the RNAP gateway application
 * so that it can talk over something other than the serial bus, you must
 * implement the functions represented by these macros, to interact with the bus
 * connection.
 *
 * The macros are set up to use serial functions included in app/util/serial.
 * This is because the sample RNAP gateway is set up
 * to use a predefined serial port (in this cas,e serial port 1) and the 
 * serial library shipped with the Ember Developer Kit. To redirect
 * the reading and writing of RNAP data to and from another bus, simply implement
 * versions of these functions, which talk over your specific bus, in place of
 * the functions provided in app/util/serial.
 *
 * The RNAP library uses these serial functions heavily. As a result, they
 * are necessary for the proper usage of the library.
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.-->
 */


/** @description Location of the serial port used to read and write
 * the RNAP protocol.
 */
#define SERIAL_PORT     1

/** @description A macro that may be defined to indicate how data
 * is read by the node's RNAP interpreter.
 *
 * @param data: The byte that the program wishes to read off the serial port.
 */
#define RNAP_READ_BYTE(data) emberSerialReadByte(SERIAL_PORT, data)

/** @description A macro that may be defined to indicate how 
 * the RNAP interpreter should write data out to the host.
 *
 * @param data: A pointer to the first byte in an array of bytes that the 
 * application wishes to write.
 *
 * @param length: The length of the total bytes that the application wishes
 * to write.
 */
#define RNAP_WRITE_DATA(data, length) emberSerialWriteData(SERIAL_PORT, data, length)

/** @description A macro that is used to wait until the port that is being used
 *   to send data is clear. The function should then return.
 */
#define RNAP_WRITE_WAIT_SEND emberSerialWaitSend(SERIAL_PORT)

/** @description A macro that the RNAP interpreter can use to write
 * remaining available bytes out to the host.
 */
#define RNAP_WRITE_AVAILABLE() emberSerialWriteAvailable(SERIAL_PORT)

/** @description A macro that the RNAP interpreter uses to write from a 
 * buffer out to the host.
 *
 * @param buffer: The EmberMessageBuffer to write out over the serial port. 
 *
 * @param start: The starting point within the buffer.
 *
 * @param length: The number of bytes to write out over the serial port.
 */
#define RNAP_WRITE_BUFFER(buffer, start, length) emberSerialWriteBuffer(SERIAL_PORT, buffer, start, length)

/** @description A function that contains all of the initialization steps
 * for setting up RNAP on the gateway node. This function is implemented in the 
 * RNAP library and must be called by the gateway node application at startup.
 */
void emRnapInit(void);

/** @description A function that must be called in a continuous loop by the 
 * gateway node so that the gateway node can process the host datastream.
 */
void emProcessRnapInput(void);

/** @description An internal array that needs to be visible for the RNAP library.*/
extern int8u emBindingHostIndex[];

/** @description An internal array that needs to be visible for the RNAP library.*/
extern int8u emMessageTokens[];

