/** @file app/util/rnap/rnap.h
 * @brief Internal declarations for RNAPII functionality.
 *
 * This file contains node-to-host commands, host-to-node commands, and node 
 * value keys. 
 *
 * The node value keys are used to set and read values stored on the node. 
 * Node values are set using the H_TO_N_SET_NODE_VALUE command and
 * read using H_TO_N_GET_NODE_VALUE.
 *
 * Most types of packets declared in this file include a \c status parameter
 * that is an EmberStatus value. See stack/include/error-def.h for definitions of  
 * all EmberStatus return values.
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.-->
 */
 
/** \mainpage RNAP II Interface Guide
 *
 * RNAP II stands for Remote Network Access Protocol type 2. This protocol 
 * allows any host processor to perform memory-related services for a 
 * gateway node. RNAP II is intended for aggregation applications in which the gateway
 * node periodically retrieves and forwards data from a large number of network nodes.
 * Such applications typically require more memory than is available on a node.
 *
 * An RNAP II application consists of the following:
 * - A host program that manages the binding table and other 
 *   memory-related business application services.
 * - An RNAP II interpreter that sits on the gateway node and handles communication
 *   between network nodes and the gateway node program.
 * - Business application programs that sit on network nodes and in some cases
 *   the gateway node also.
 *
 * The host program has the following responsibilities:
 * - Maintain an extended binding table, including the following:
 *   - Binding table index
 *   - Binding type
 *   - Binding identifier
 *   - Local terminal
 *   - Remote terminal
 *   - Node ID 
 *   - Route
 *   - Connection
 *   - Connection state
 * - Keep track of the nodes' current binding entries, so that the host 
 *   processor can provide a route if one does not currently exist on a node. 
 * - Inform the host application when the gateway node is running an RNAP II 
 *   interpreter.
 * - Create new custom routes on the network from routing data that is returned
 *   to the host application through the gateway.
 * - Respond to all datagrams and sequenced messages so that the gateway 
 *   node can provide an ACK. 
 *
 * The RNAP II interpreter has these responsibilities:
 * - Communicate node IDs and routes to the host program so that it can maintain 
 *   the binding table. 
 * - Inform the host program when messages have been delivered so that it can 
 *   maintain its knowledge of the network and network-related data.
 * - Expose most of the functionality of the EmberNet stack API to support a 
 *   complete version of the EmberNet stack API on the host processor.
 *
 * <b>RNAP II Sample Application</b>
 *
 * The RNAP sample application models an automated meter reading (AMR) application 
 * in which network nodes receive data from meters and funnel that data to a gateway
 * node termed an \b aggregator. This type of application typically involves very 
 * high memory requirements, making it ideal for demonstrating the advantages of RNAP II.
 * 
 * Three files model the sample application components:
 * - \b amr-meter.\b c models the behavior of a meter.
 * - \b node-gateway.\b c is a simple instance of an RNAP II interpreter built to 
        run on a node physically connected to an Ethernet network.
 * - \b amr-aggregator.\b c models either the RNAP II host or a non-RNAP II aggregator, 
 *      depending how defines in the file are set (for more information, see 
 *      comments in the file).
 *
 * When the aggregator receives data from the network, it interacts with the host 
 * application via the RNAP II API, to inform the host of network activity. The 
 * host application interacts with the node using the RNAP II API to send an 
 * acknowledgement for a message received or when originating a message.
 *
 * It is straightforward to convert a node application to use RNAP II. Apart from 
 * the normal porting code that would be expected when moving from a microprocessor 
 * to a host Linux or Windows platform (for example, emberSerialPrint() becomes 
 * fprintf), the application-related code is unchanged. The file amr-aggregator.c 
 * includes extensive comments pointing out where node applications should be changed.
 *
 * <b>Writing an RNAP II Application</b>
 *
 * Minimally, you must write a host program using the RNAP II API, host-api.h. 
 * This API includes a subset of the EmberNet stack calls plus a few 
 * calls specific to RNAP II. The shared EmberNet stack API calls are identified in
 * both host-api.h and stack/include/ember.h and documented in the <i>EmberNet Stack Interface 
 * Guide</i>. 
 *
 * It is important to note that RNAP II only supports applications that use aggregation
 * (client / server) endpoints. This is because the aggregation functionality 
 * of RNAP II relies on source routing that is only available to aggregation type 
 * endpoints. The different types of endpoints are both enumerated and
 * documented in stack/include/ember.h.
 * 
 * The host application must must call emberHostInit() and must include
 * host-api.h and the RNAP II version of stack/include/packet-buffer.h.
 *
 * Application developers must implement, or at least stub out, the following
 * EmberNet stack handler functions, because they are not implemented as part of
 * the host API.
 * - emberMessageSent()
 * - emberIncomingMessageHandler()
 * - emberConnectionStatusHandler()
 * - emberStackStatusHandler()
 *
 * <b>RNAP Protocol Packet Format</b>
 *
 * An RNAP message is structured as follows: 
 * <table border=1><tr>
 * <td>Num Bytes
 * </td><td>1
 * </td><td>1
 * </td><td>1
 * </td><td>1
 * </td><td>n EMBER_UNICAST_BINDING 
 * </td><td>1
 * </td></tr><tr>
 * <td>Values
 * </td><td>[
 * </td><td>length
 * </td><td>command
 * </td><td>sequenceNumber
 * </td><td>parameters
 * </td><td>]
 * </td></tr></table>
 * 
 * The \c length byte equals the number of parameter bytes plus 2 (for the command and
 * sequence number). The maximum length for an RNAP packet is 128 bytes and is set below.
 * Together \c length, \c command, and \c sequenceNumber
 * form the RNAP header.
 * 
 * The command byte of an RNAP message is followed by a 1-byte sequence 
 * number, which is incremented by the sending party (either host or node)
 * and is used to keep track of messages being sent by either party.
 * 
 * All multi-byte parameters, such as network node IDs, are sent
 * most-significant-byte first.
 *
 * The node must ACK every message sent by the host, but the host is not
 * required to ACK all messages sent by the node. The format of an 
 * ACK is similar to that of a generic RNAP message with one important 
 * exception: The sequence number in an ACK reflects the sequence number of 
 * the message that was recieved. 
 * 
 * An RNAP ACK is structured as follows:<br>
 * <table>
 * <tr><td>Num Bytes
 * </td><td>1
 * </td><td>1
 * </td><td>1
 * </td><td>1
 * </td><td>1
 * </td><td>1
 * </td><td>n
 * </td><td>1
 * </td></tr>
 * <tr><td>Values
 * </td><td>[ 
 * </td><td>length
 * </td><td>command (ack)
 * </td><td>sequenceNumber (being ACKed)
 * </td><td>command (being ACKed)
 * </td><td>status
 * </td><td>parameters
 * </td><td>]
 * </td></tr></table>
 *
 * The number and type of parameters in an ACK depend on the command being
 * ACKed.  Typically they include an EmberStatus value giving the success or
 * failure of the command.
 */

/** @description The RNAP version currently implemented. This value is sent by both 
 * the node and the host during startup.
 * @see N_TO_H_NODE_START
 * @see H_TO_N_NODE_START_ACK
 * @see H_TO_N_HOST_HAS_RESET
 */
#define RNAP_VERSION 0

/** @description The maximum size of an RNAP packet.
*/
#define MAX_RNAP_PACKET_SIZE 128

/** @description A character that signifies the start of a valid RNAP packet. 
 */
#define FRAME_START '['

/** @description A character that signifies the end of a valid RNAP packet. 
 */
#define FRAME_END ']'



enum {
  /** @description The ACK returned by the node for each command sent by 
   * the host. This message is special because the sequence number must be
   * the same as the sequence number of the message being ACKed.
   *       
   * @packetFormat
   * [ rnapHeader (3)
   * | command (1)
   * | status (1)
   * | parameters (n) ]
   *       
   * @param command: The command being ACKed.
   * @param status: An EmberStatus value indicating the success or
   *  failure of the command.
   * @param parameters: The size and format of the parameters depends
   *  on the command being ACKed. See the individual H_TO_N command
   *  descriptions for details.
   *
   * @reqResponse None   
   */
  N_TO_H_COMMAND_ACK = 0, 

  /** @description Sent by the node in response to an error in reading
   * a command.
   *
   * @packetFormat
   * [ rnapHeader (3)
   * | status (1) ]
   *       
   * @param status: The EmberStatus value corresponding to the serial error
   *  that occurred.
   *
   * @reqResponse None   
   */
  N_TO_H_READ_ERROR = 1,

  /** @description Sent by the node when it reboots.
   *                               
   * @packetFormat
   * [ rnapHeader (3)
   * | rnapProtocolVersion (1)
   * | eui64 (8)
   * | radioType (1)
   * | radioChannel (1)
   * | radioPower (1)
   * | bindingTableSize (1)]
   *          
   * @param rnapProtocolVersion: The version of the RNAP protocol that the
   *  node implements.
   * @param eui64: The node's 64-bit EUI identfier.
   * @param radioType: A bit mask indicating which bands the radio supports.
   *  <!--(the meanings of which are defined in stack/phy/phy.h)-->
   * @param radioChannel: The channel currently being used.
   * @param radioPower: The current radio transmit power.
   * @param bindingTableSize: The number of entries in the node's binding
   *  table.
   *          
   * @reqResponse The host must ACK this command by sending an
   * H_TO_N_NODE_START_ACK command so that the node is (1) aware that
   * that the host is aware of the reboot and (2) agrees to use the
   * specified version of the protocol.
   *          
   * @see H_TO_N_NODE_START_ACK
   */
  N_TO_H_NODE_START = 2, 

  /** @description Sent by the node when it receives an incoming unicast
   * message.
   *
   * @packetFormat       
   * [ rnapHeader(3) 
   * | messageType (1)
   * | destination (1)
   * | sendingTerminal (1)
   * | messageId (1)
   * | sender (2)
   * | routeLength (1)
   * | route (n)
   * | routeCost (1)
   * | payloadLength (1)
   * | payload (n) ]
   *        
   * @param messageType: The type of message, as an
   *   EmberIncomingMessageType.
   * @param destination: For datagrams, this is the destination
   *  terminal. For all other messages, it is the destination binding index.
   * @param sendingTerminal: The terminal from which the message originated.
   *  This parameter is valid only for datagrams.
   * @param messageId: The sequence number assigned by the sender. This
   *  parameter is not valid for datagram replies.
   * @param sender: The network ID of the sending node.
   * @param routeLength: The number of bytes in the route information.
   * @param route: The route taken by the message.
   * @param routeCost: The cost of the route taken by the message.
   * @param payloadLength: The length of the application payload.
   * @param payload: The application payload of the message.
   * 
   * @reqResponse The host must respond to datagrams with an
   *   H_TO_N_TX_DATAGRAM_REPLY command to acknowledge receipt of the
   *   message. 
   */
  N_TO_H_RX_UNICAST = 3,

  /** @description Sent by the node when it receives a multicast message.
   *       
   * @packetFormat
   * [ rnapHeader(3)
   * | sender (2)
   * | cost (1)
   * | multicastId (2)
   * | payloadLength (1)
   * | payload (n)]
   *       
   * @param sender: The network ID of the sending node.
   * @param cost: The cost to the sending node.
   * @param multicastId: The multicast identifier.
   * @param payloadLength: The length of the application payload.
   * @param payload: The application payload.
   *
   * @reqResponse None   
   */
  N_TO_H_RX_MULTICAST = 4,

  /** @description Sent by the node in response to a call to
   * emberMessageSentHandler(). The parameters correspond to the values
   * passed to emberMessageSent() with the EmberMessageBuffer value being
   * replaced by the message token passed in the RNAP command that sent the
   * message.
   *       
   * @packetFormat
   * [ rnapHeader(3)
   * | bindingIndex (1)
   * | messageToken (1)
   * | status (1) ]
   *       
   * @param bindingIndex: The binding index to which the message was sent.
   * @param messageToken: The message token sent.
   * @param status: Status of the message. 
   *
   * @reqResponse None   
   */
  N_TO_H_MESSAGE_SENT_STATUS = 5,
    
  /** @description Sent by the node in response to a call to
   * emberConnectionStatusHandler().
   *       
   * @packetFormat
   * [ rnapHeader(3)
   * | bindingIndex (1)
   * | status (1)
   * | messageId (1) ]
   *       
   * @param bindingIndex: The binding index whose connection status has changed.
   * @param status: The new connection status.
   * @param messageId: If \c status is EMBER_CONNECTION_OPEN, this is the
   *  message ID of the initial sequenced message. For all other values of
   *  \c status the value of \c messageId is unspecified. This value is
   *  used by the host in reordering incoming sequenced messages.
   *
   * @reqResponse None   
   */
  N_TO_H_CONNECTION_STATUS = 6,
    
  /** @description Sent by the node in response to a call to
   * emberStackStatusHandler().
   * 
   * @packetFormat
   * [rnapHeader(3)
   * | status (1) ]
   * 
   * @param status: The status passed to emberStackStatusHandler().
   *
   * @reqResponse None   
   */
  N_TO_H_STACK_STATUS = 7,

  /** @description Sent when the node receives a sequenced message from
   * a node whose binding and connection are not currently in the node's
   * cache.
   * 
   * @packetFormat
   * [rnapHeader(3)
   * | bindingIndex (1) ]
   * 
   * @param bindingIndex: The index of the binding needed by the node.
   *
   * @reqResponse The host must respond by sending the binding and
   *  binding route for the given binding index. See H_TO_N_SET_VALUE,
   *  NODE_BINDING, and NODE_BINDING_ROUTE.
   */
  N_TO_H_NEED_BINDING =  8,

#ifndef DOXYGEN_SHOULD_SKIP_THIS
  // Maximum node-to-host command.
  N_TO_H_MAX_COMMAND =  8
#endif
    
};

enum {
  /** @description Sent by the host in response to a H_TO_N_NODE_START_ACK
   * message from the node.
   *     
   * @packetFormat
   * [ rnapHeader (3)
   * | version (1)  ]
   *
   * @param version: The RNAP version that the host implements. This should
   * be the same as the version that the node implements.
   *
   * @reqResponse The status value in ACK for this command is 
   *  EMBER_SUCCESS if the node code is compatible with the host's
   *  RNAP version and EMBER_ERR_FATAL if not.
   */
  H_TO_N_NODE_START_ACK = 0, 

  /** @description Sent by the host to transmit a unicast message.
   *     
   * @packetFormat
   * [ rnapHeader(3)
   * | messageToken (1)
   * | bindingTableCacheIndex (1)
   * | isDatagram (1)
   * | payloadLength (1)
   * | payload (n) ]
   *    
   * @param messageToken: A unique identifier used by the node to identify the
   *  message in the subsequent N_TO_H_MESSAGE_SEND_STATUS sent by the
   *  node.
   * @param bindingTableCacheIndex: The index into the node's binding-table cache
   *  for the destination.
   * @param isDatagram: TRUE if the message is to be sent as a datagram and
   *  FALSE if it is to be sent as a sequenced message.
   * @param payloadLength: The length of the application payload.
   * @param payload: The application payload.
   * 
   * @reqResponse The ACK for this command contains the status value returned
   *  by the call to emberSendDatagram() or emberSendSequenced() that
   *  was used to send the message.
   *
   * @see N_TO_H_MESSAGE_SEND_STATUS
   */
  H_TO_N_TX_UNICAST = 1,  
  
  /** @description Used to respond to a previously received datagram.
   * The host must send a datagram reply for each datagram received, even
   * if there is no payload associated with the reply, so that the sending
   * node will know that the datagram was received.
   *   
   * @packetFormat
   * [ rnapHeader(3)
   * | destination (2)
   * | messageId (1)
   * | sendingTerminal (1)
   * | routeLength (1)
   * | route (n)
   * | payloadLength (1)
   * | payload (n) ]
   *     
   * @param destination: The network ID of the sender of the original datagram.
   * @param transportSegmentMessageId: The message ID for the datagram being 
   *  replied to.
   * @param sendingTerminal: The destination terminal of the original datagram.
   * @param routeLength: The length of the route being provided for this
   *  message.
   * @param route: The route to be used for sending this message.
   * @param payloadLength: The length of the application payload (may be zero).
   * @param payload: The application payload.
   *
   * @reqResponse None  
   *
   * @see N_TO_H_RX_UNICAST
   */
  H_TO_N_TX_DATAGRAM_REPLY = 2,

  /** @description Used to send a multicast message. Because multicasts are not 
   * ACKed within the transport layer, there is no corresponding 
   * N_TO_H_MESSAGE_SEND_STATUS and thus no need for a message token.
   *    
   * @packetFormat
   * [ rnapHeader (3)
   * | multicastId (2)
   * | numberOfHops (1)
   * | payloadLength (1)
   * | payload (n) ]
   *     
   * @param multicastId: A 2-byte multicast ID to which to send the message.
   * @param numberOfHops: The number of hops allowed for the sent multicast.
   * @param payloadLength: The length of the application.
   * @param payload: The application payload.
   *
   * @reqResponse None
   */
  H_TO_N_TX_MULTICAST = 3,  

  /** @description Used by the host to request that the node reset itself.  
   *    
   * @packetFormat
   [ rnapHeader(3) ] 
   *
   * @reqResponse None  
  */
  H_TO_N_RESET_NODE = 4,  
  
  /** @description Used to indicate that the host has reset, so that the
   * node may reset its knowledge of the host's state.
   *    
   * @packetFormat
   *  [ rnapHeader(3) | version (1) ]
   *    
   * @param version: The RNAP version that the host implements.
   *
   * @reqResponse None
   */
  H_TO_N_HOST_HAS_RESET = 5,  
  
  /** @description Used by the host to set a value on the node.
   *    
   * @packetFormat
   * [ rnapHeader(3)
   * | nodeValueKey (1)
   * | index (1)
   * | valueLength (1)
   * | value (n) ]
   *     
   * @param nodeValueKey: One of the NODE_... keys indicating which value
   *  is to be set. 
   * @param index: An index used to set a particular value in an array. 
   *  This is ignored for values that are not arrays. See the NODE_... 
   *  enumeration for more information.
   * @param valueLength: The length of the value. See the NODE_... enumeration
   *  for the size and format of the various values. The value may
   *  include an index for setting elements of an array.
   * @param value: The value to be set.  
   *
   * @reqResponse The ACK for this message includes the key and value from
   *  the original command.
   *     
   * @reqResponsePacketFormat
   * [ status (1)
   * | nodeValueKey (1)
   * | index (1)
   * | valueLength (1)
   * | value (n) ]
   *     
   */
  H_TO_N_SET_VALUE = 6,

  /** @description Used by the host to read a value from the node.
   * 
   * @packetFormat
   * [ rnapHeader(3)
   * | nodeValueKey (1)
   * | index (1) ]
   *     
   * @param nodeValueKey: One of the NODE_... keys indicating which value
   *  is to be set. 
   * @param index: An index used to select a particular value
   *  from an array. This is ignored for values that are not arrays.
   *  See the NODE_... enumeration for more information.
   *
   * @reqResponse The ACK for this message includes the key, the length of
   *  the value, and the value itself (including any indexes).  This is the
   *  same format as H_TO_N_SET_VALUE ACKs.
   *
   * @reqResponsePacketFormat 
   * [ status (1)
   * | nodeValueKey (1)
   * | valueLength (1)
   * | value (n) ]
   */
  H_TO_N_GET_VALUE = 7,  

  /** @description Puts the node into bootloader mode so that it can receive a new 
    software image over the wired connection.
  */
  H_TO_N_BOOTLOAD = 8,

  /** @description Used by the host to open a connection.
   * 
   * @packetFormat
   * [ rnapHeader(3)
   * | bindingTableCacheIndex (1)
   * | connectionId (1) ]
   *     
   * @param bindingTableCacheIndex: The index into the node's binding-table cache
   *  for the binding for which to open a connection.
   * @param connectionId: The ID to use for this connection. It should be
   *  distinct from the IDs used in other recent connections to the same
   *  node.
   *
   * @reqResponse In addition to the status, the ACK for this message includes
   *  \c bindingTableCacheIndex.
   *
   * @reqResponsePacketFormat
   * [ status (1)
   * | bindingTableCacheIndex (1) ]
   */
  H_TO_N_OPEN_CONNECTION = 9,

  /** @description Used by the host to close a connection.
   * 
   * @packetFormat
   * [ rnapHeader(3)
   * | bindingTableCacheIndex (1) ]
   *     
   * @param bindingTableCacheIndex: Index into the node's binding table cache
   *  for the binding whose connection should be closed.
   */
  H_TO_N_CLOSE_CONNECTION = 10,

#ifndef DOXYGEN_SHOULD_SKIP_THIS
  // Maximum host-to-node command.
  H_TO_N_MAX_COMMAND = 10  
#endif

};

enum {
  /** @description Enables or disables multihop routing on the node.
   * This value may only be set. It may not be read.
   *
   * @packetFormat
   * [ enableRouting (1) ]
   *     
   * @param value: A boolean value that is passed to emberEnableRouting().
   */  
  NODE_ENABLE_ROUTING = 0,  
  
  /** @description Sets or gets the node's radio channel.
   *
   * @packetFormat
   * [ channel (1) ]
   *     
   * @param value: The channel that the node is currently on. Valid values for this
   * parameter can be identified through the \c radioType field.
   *
   * @see N_TO_H_NODE_START
   */
  NODE_RADIO_CHANNEL = 1,  
  
  /** @description Sets or gets the radio's transmission power setting.
   *
   * @packetFormat
   * [ power (1) ]
   *     
   * @param value: Power at which the radio is transmitting.
   */
  NODE_RADIO_POWER = 2,

  /** @description Sets or gets bindings within the node's cached binding 
   * table. \b Note: Four of the parameters for this packet together constitute an 
   * EmberBindingTableEntry: \c type, \c remoteEui64, \c localTerminal, and
   * \c remoteTerminal. For information on binding table entries, see the 
   * <i>EmberNet Stack Interface Guide</i>.
   *
   * @packetFormat
   * [ hostBindingIndex 
   * | bindingType (1)
   * | remoteEui64 (8)
   * | localTerminal (1)
   * | remoteTerminal (1)
   * | isConnected (1)
   * | connectionData (6) ]
   *
   * @param bindingType: The type of binding. For writes this should be
   *  EMBER_UNICAST_BINDING. For reads this should be EMBER_UNICAST_BINDING
   *  for bindings that are in use and EMBER_UNUSED_BINDING for bindings
   *  that are not in use.
   * @param remoteEui64: The 64-bit EUI ID assigned to the binding.
   * @param localTerminal: The terminal that is the local side of this binding.
   * @param remoteTerminal: The terminal that is the remote side of this binding.
   * @param remoteBindingIndex: The index of the matching binding in the
   *  binding table on the remote node.
   * @param isConnected: TRUE if this binding has an open connection, and
   *  FALSE if it does not.
   * @param connectionData: The connection's internal state. It is
   *  ignored if \c isConnected is FALSE.
   */
  NODE_BINDING = 3,  

  /** @description Returns the contents of a binding to the host and then
   * clears it.
   *
   * The value format is identical to that of NODE_BINDING.
   */
  NODE_CLEAR_BINDING = 4,

  /** @description Sets or gets the routing information associated with a binding.
   *
   * @packetFormat
   * [ remoteNode (2)
   * | routeLength (1)
   * | route (n) ]
   *
   * @param remoteNode: The network ID of the remote node.
   * @param routeLength: The length of the route being provided for this
   *  binding.
   * @param route: The route to be used for this binding.
   */
  NODE_BINDING_ROUTE = 5,

  /** @description Tells the gateway node to associate and dissassociate with 
   *    the network.
   *
   * @packetFormat
   * [ enableAssociation(1)
   * | panId (2)
   * | enableRelay (1)
   * | radioTxPower (1)
   * | radioChannel (1) ]
   *
   * @param enableAssociation: A boolean value that tells the gateway node
   *  to either associate or dissassociate with the network.
   * @param panId: A settable parameter for the node's PAN ID.
   * @param enableRelay: A boolean value determining whether the gateway node
    * will relay messages to other nodes.
   * @param radioTxPower: The transmit power for the gateway node.
   * @param radioChannel: The radio channel on which to operate the gateway node.
   */
  NODE_NETWORK_ASSOCIATION = 6,
  
};
