/**
 * File: command-interpreter.c
 * Description: processes commands incoming over the serial port.
 *   Originally written for the eval-kit application.
 *
 * Culprit(s): Richard Kelsey (kelsey@ember.com)
 *
 * Copyright 2004 by Ember Corporation.  All rights reserved.               *80*
 */

#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "include/packet-buffer.h"
#include "stack/include/hal.h"
#include "stack/include/error.h"
#include "app/util/serial/serial.h"
#include "app/util/serial/command-interpreter.h"

//----------------------------------------------------------------
// Forward declarations.
static void readError(EmberCommandState *state, int8u error);
static boolean commandLookup(EmberCommandState *state);
static void callCommandAction(EmberCommandState *state);
static boolean addIntegerArgument(EmberCommandState *state);
static boolean addStringArgument(EmberCommandState *state, boolean isBinary);
static boolean argumentsDone(PGM_P argumentSpecs);
static void releaseArgumentBuffers(EmberCommandState *state);

//----------------------------------------------------------------
// Command parsing states.

enum {
  CMD_AWAITING_COMMAND,                // nothing on this line yet
  CMD_READING_COMMAND,
  CMD_READING_ARGUMENTS,
  CMD_READING_SIGNED_INTEGER,
  CMD_READING_INTEGER,
  CMD_READING_STRING,                  // have read opening " but not closing "
  CMD_READING_BAD_STRING,              // ran out of buffers in the middle
  CMD_READING_BINARY_STRING_LENGTH,    // have read opening [
  CMD_READING_BINARY_STRING,           // have read opening [ and length
  CMD_READING_BAD_BINARY_STRING,       // ran out of buffers in the middle
  CMD_READING_WHITESPACE_TO_EOL,       // have read message
  CMD_READING_TO_EOL                   // clean up after error
};

#ifdef __unix__
char *rxStateNames[] =
  {
    "awaiting command",
    "reading command",
    "reading arguments",
    "reading signed integer",
    "reading integer",
    "reading string",
    "reading bad string",
    "reading binary string length",
    "reading binary string",
    "reading bad binary string",
    "reading whitespace to eol",
    "reading to eol"
  };
#endif

//----------------------------------------------------------------
// Character types for parsing commands.
//
// Some characters can be of multiple types so we use masks.  We put the
// value of the digit characters in the low four bits.  It isn't clear that
// this hackery saves much time or flash, but it was fun to do.

enum {
  C_OTHER         = 0,      // As 0 in the array to make it easier to read.

  // There is a two-byte type field and bits for decimal and hexidecimal digits.
  // The value for digits is in the low four bits.
  C_TYPE_MASK     = 0x30,
  C_DECIMAL_MASK  = 0x40,   // decimal digits
  C_HEX_MASK      = 0x80,   // hexidecimal digits
  C_VALUE_MASK    = 0x0F,   // all digits

  // Possible values for the type field in bits 5 and 6.
  C_ALPHA         = 0x10,   // letters
  C_WHITE         = 0x20,   // whitespace
  C_START         = 0x30    // " [ # + - (these all start new tokens)
};

#define C_HEX   (C_ALPHA | C_HEX_MASK)
#define C_DIGIT (C_DECIMAL_MASK | C_HEX_MASK)

static PGM char charCodes[] = {
  0,       0,       0,       0,       0,       0,       0,       0,       
  0,       C_WHITE, C_WHITE, C_WHITE, C_WHITE, C_WHITE, 0,       0,       
  0,       0,       0,       0,       0,       0,       0,       0,       
  0,       0,       0,       0,       0,       0,       0,       0,       

  C_WHITE, 0,       C_START, C_START, 0,       0,       0,       0,       
  0,       0,       0,       C_START, 0,       C_START, 0,       0,       

  C_DIGIT,     C_DIGIT | 1, C_DIGIT | 2, C_DIGIT | 3,
  C_DIGIT | 4, C_DIGIT | 5, C_DIGIT | 6, C_DIGIT | 7, 
  C_DIGIT | 8, C_DIGIT | 9, 0,           0,
  0,           0,           0,           0,       

  0,          C_HEX | 10, C_HEX | 11, C_HEX | 12,
  C_HEX | 13, C_HEX | 14, C_HEX | 15, C_ALPHA,

  C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA,
  C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA,
  C_ALPHA, C_ALPHA, C_ALPHA, C_START, 0,       0,       0,       0,       

  0,          C_HEX | 10, C_HEX | 11, C_HEX | 12,
  C_HEX | 13, C_HEX | 14, C_HEX | 15, C_ALPHA,

  C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA,
  C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA, C_ALPHA,
  C_ALPHA, C_ALPHA, C_ALPHA, 0,       0,       0,       0,       0};

// These evaluate their argument twice, but they are always invoked on
// a variable reference.
#define isAlphabetic(x) ((x) < 128 && (charCodes[x] & C_TYPE_MASK) == C_ALPHA)
#define isWhitespace(x) ((x) < 128 && (charCodes[x] & C_TYPE_MASK) == C_WHITE)
#define isStart(x)      ((x) < 128 && (charCodes[x] & C_TYPE_MASK) == C_START)

// Which bit we use for digits depends on which base we are using.
#define DIGIT_MASK(state) \
(((state)->integerBase == 16) ? C_HEX_MASK : C_DECIMAL_MASK)
#define isDigit(x, state) ((x) < 128 && (charCodes[x] & DIGIT_MASK(state)))

#define isEol(x)        ((x) == '\n' || (x) == '\r')
#define digitValue(x)   (charCodes[x] & C_VALUE_MASK)

//----------------------------------------------------------------
// Initialize the state maachine.

void emberCommandReaderInit(EmberCommandState *state)
{
  state->stringBuffer = EMBER_NULL_MESSAGE_BUFFER;
  state->state = CMD_AWAITING_COMMAND;
  state->commandIndex = 0;
  state->stringLength = 0;
  state->argumentCount = 0;
}

//----------------------------------------------------------------
// This is a state machine for parsing commands.  If 'input' is NULL
// 'sizeOrPort' is treated as a port and characters are read from there.
// 
// There are a couple of goto's in it where one parse state naturally falls
// into another.

void emberProcessCommandString(EmberCommandState *state,
                               int8u *input,
                               int8u sizeOrPort)
{
  while (TRUE) {
    int8u next;
    
    if (input == NULL) {
      switch (emberSerialReadByte(sizeOrPort, &next)) {
      case EMBER_SUCCESS:
        break;
      case EMBER_SERIAL_RX_EMPTY:
        return;
      case EMBER_SERIAL_RX_OVERFLOW:
      case EMBER_SERIAL_RX_FRAME_ERROR:
      case EMBER_SERIAL_RX_PARITY_ERROR:
        readError(state, EMBER_CMD_ERR_PORT_PROBLEM);
        next = 'x';               // anything but EOL works (EOL is the command
        break; }                  // delimiter and resets the error flag).
    } else if (sizeOrPort == 0)
      return;
    else {
      next = *input;
      input += 1;
      sizeOrPort -= 1;
    }

//    fprintf(stderr, "[processing '%c' (%s)]\n", next, stateNames[state]);

    switch (state->state) {

    case CMD_AWAITING_COMMAND:
      if (isAlphabetic(next)) {
        state->state = CMD_READING_COMMAND;
        if ('A' <= next && next <= 'Z')
          state->command[0] = next - 'A' + 'a';
        else
          state->command[0] = next;
        state->stringLength = 1;
      } else if (next == EMBER_COMMENT_CHARACTER)
        state->state = CMD_READING_TO_EOL;
      else if (! isWhitespace(next))
        readError(state, EMBER_CMD_ERR_NO_SUCH_COMMAND);
      break;
      
    CMD_READING_TO_EOL:
    case CMD_READING_TO_EOL:
      if (isEol(next)) {
        if (state->error != EMBER_CMD_SUCCESS) {
          (state->errorHandler)(state->error);
	  state->error = EMBER_CMD_SUCCESS;
        }
        emberCommandReaderInit(state);
      }
      break;
      
    case CMD_READING_COMMAND:
      if (isAlphabetic(next) || next == '_') {
        if (state->stringLength < EMBER_MAX_COMMAND_LENGTH) {
          if ('A' <= next && next <= 'Z')
            state->command[state->stringLength] = next - 'A' + 'a';
          else
            state->command[state->stringLength] = next;
          state->stringLength += 1; }
        else
          readError(state, EMBER_CMD_ERR_NO_SUCH_COMMAND); }
      else if (isWhitespace(next) || isStart(next)) {
        state->command[state->stringLength] = 0;
        if (commandLookup(state)) {
          state->stringLength = 0;
          state->state = CMD_READING_ARGUMENTS;
          goto CMD_READING_ARGUMENTS;
        } else {
          readError(state, EMBER_CMD_ERR_NO_SUCH_COMMAND);
          goto CMD_READING_TO_EOL;
        }
      } else {
        readError(state, EMBER_CMD_ERR_NO_SUCH_COMMAND);
        goto CMD_READING_TO_EOL;
      }
      break;
      
    CMD_READING_WHITESPACE_TO_EOL:
    case CMD_READING_WHITESPACE_TO_EOL:
      if (isEol(next)) {
        callCommandAction(state);
        emberCommandReaderInit(state);
      } else if (next == EMBER_COMMENT_CHARACTER) {
        callCommandAction(state);
        state->state = CMD_READING_TO_EOL;
      } else if (! isWhitespace(next))
        readError(state, EMBER_CMD_ERR_WRONG_COMMAND_ARGUMENTS);
      break;

    CMD_READING_ARGUMENTS:
    case CMD_READING_ARGUMENTS:
      if (isDigit(next, state)) {
        state->integerSign = '+';
        state->integerValue = digitValue(next);
        state->state = CMD_READING_INTEGER;
      } else if (next == '-' || next == '+') {
        state->integerSign = next;
        state->integerValue = 0;
        state->state = CMD_READING_SIGNED_INTEGER;
      } else if (next == '"') {
        state->stringBuffer = emberAllocateStackBuffer();
        if (state->stringBuffer == EMBER_NULL_MESSAGE_BUFFER) {
          readError(state, EMBER_CMD_ERR_NO_BUFFER_AVAILABLE);
          state->state = CMD_READING_BAD_STRING; }
        else {
          emberSetMessageBufferLength(state->stringBuffer, 0);
          state->state = CMD_READING_STRING;
        }
      } else if (next == '[')
        state->state = CMD_READING_BINARY_STRING_LENGTH;
      else
        goto CMD_READING_WHITESPACE_TO_EOL;
      break;
      
    case CMD_READING_INTEGER:
    case CMD_READING_SIGNED_INTEGER:
      if (isDigit(next, state)) {
        int32u temp = (state->integerValue * state->integerBase)
                       + digitValue(next);
        state->state = CMD_READING_INTEGER;
        if (0xFFFF < temp)
          readError(state, EMBER_CMD_ERR_ARGUMENT_OUT_OF_RANGE);
        else
          state->integerValue = temp;
      } else if (state->state == CMD_READING_SIGNED_INTEGER
                 || (! isWhitespace(next) && ! isStart(next)))
        readError(state, EMBER_CMD_ERR_ARGUMENT_SYNTAX_ERROR);
      else if (addIntegerArgument(state)) {
        state->state = CMD_READING_ARGUMENTS;
        goto CMD_READING_ARGUMENTS;
      } else
        goto CMD_READING_TO_EOL;
      break;
      
    case CMD_READING_STRING:
      if (next == '"') {
        if (addStringArgument(state, FALSE))
          state->state = CMD_READING_ARGUMENTS;
      } else if (emberMessageBufferLength(state->stringBuffer) == 255)
        readError(state, EMBER_CMD_ERR_STRING_TOO_LONG);
      else if (emberAppendToLinkedBuffers(state->stringBuffer, &next, 1)
               != EMBER_SUCCESS) {
        readError(state, EMBER_CMD_ERR_NO_BUFFER_AVAILABLE);
        state->state = CMD_READING_BAD_STRING;
      }
      break;
      
    case CMD_READING_BAD_STRING:
      if (next == '"')
        state->state = CMD_READING_TO_EOL;
      break;
      
    case CMD_READING_BINARY_STRING_LENGTH:
      state->binaryLength = next;
      state->stringLength = 0;
      state->stringBuffer = emberFillLinkedBuffers(NULL, next);
      if (state->stringBuffer == EMBER_NULL_MESSAGE_BUFFER) {
        readError(state, EMBER_CMD_ERR_NO_BUFFER_AVAILABLE);
        state->state = CMD_READING_BAD_BINARY_STRING;
      } else
        state->state = CMD_READING_BINARY_STRING;
      break;
    
    case CMD_READING_BINARY_STRING:
      if (state->stringLength < state->binaryLength) {
        emberSetLinkedBuffersByte(state->stringBuffer,
                                  state->stringLength,
                                  next);
        state->stringLength += 1;
      } else if (next == ']') {
        if (addStringArgument(state, TRUE))
          state->state = CMD_READING_ARGUMENTS;
      } else
        readError(state, EMBER_CMD_ERR_NO_BINARY_STRING_TERMINATOR);
      break;
      
    case CMD_READING_BAD_BINARY_STRING:
      if (state->stringLength < state->binaryLength)
          state->stringLength += 1;
      else
        state->state = CMD_READING_TO_EOL;
      break; }}
}

// We preserve the identity of the first error to occur and then move
// forward to the next end-of-line before reporting it.
//
// This is not a macro because:
//  - It occurs in many places, so using a function keeps the image smaller.
//  - If we are discarding the current line anyway we don't care so much about
//      the possibility of dropping a character.

static void readError(EmberCommandState *state, int8u error)
{
  if (state->error == EMBER_CMD_SUCCESS) {
    state->state = CMD_READING_TO_EOL;
    state->error = error;
    if (state->stringBuffer != EMBER_NULL_MESSAGE_BUFFER) {
      emberReleaseMessageBuffer(state->stringBuffer);
      state->stringBuffer = EMBER_NULL_MESSAGE_BUFFER;
    }
    releaseArgumentBuffers(state);
  }
}

//----------------------------------------------------------------
// Command lookup and processing

// Single letter commands get looked up as abbreviations, longer ones are
// matched against the full names.  We use linear search in either case.

static boolean commandLookup(EmberCommandState *state)
{
  int8u *command = state->command;
  
  if (command[0] != 0           // zeros mark commands that have no short form
      && command[1] == 0) {
    int8u i;
    for (i = 0; state->commands[i].action != NULL; i++) {
      if (command[0] == state->commands[i].shortName) {
	state->commandIndex = i;
	goto found; }}
    return FALSE; }
  else {
    int8u i;
    for (i = 0; state->commands[i].action != NULL; i++) {
      PGM_P name = state->commands[i].longName;
      int8u j;
      for (j = 0;; j++) {
	if (command[j] == 0) {
	  if (name[j] == 0) {
	    state->commandIndex = i;
	    goto found; }
	  else
	    break; }
	else if (name[j] != command[j])
	  break; }}
    return FALSE; }

 found:
  state->argumentSpecs = state->commands[state->commandIndex].argumentTypes;
  state->argumentCount = 0;
  return TRUE;
}

static void callCommandAction(EmberCommandState *state)
{
  if (! argumentsDone(state->argumentSpecs))
    (state->errorHandler)(EMBER_CMD_ERR_WRONG_COMMAND_ARGUMENTS);
  else
    (state->commands[state->commandIndex].action)(state);

  releaseArgumentBuffers(state);
  if (state->stringBuffer != EMBER_NULL_MESSAGE_BUFFER)
    emberReleaseMessageBuffer(state->stringBuffer);
}

//----------------------------------------------------------------
// Checking argument types.

static boolean addArgumentValue(EmberCommandState *state,
                                int16u value,
                                int8u specSkip)
{
  if (state->argumentCount == state->maximumArguments) {
    readError(state, EMBER_CMD_ERR_TOO_MANY_COMMAND_ARGUMENTS);
    return FALSE;
  } else {
    state->arguments[state->argumentCount] = value;
    state->argumentCount += 1;
    
    if (state->argumentSpecs[specSkip] != '*')
      state->argumentSpecs += specSkip;
    return TRUE;
  }
}

static boolean addIntegerArgument(EmberCommandState *state)
{
  int8u type = state->argumentSpecs[0];
  int8u size = state->argumentSpecs[1];
  int32s integerValue = state->integerValue;
  int32s high;
  int32s low;

  if (state->integerSign == '-')
    integerValue = - integerValue;

  if (type == 'u') {
    low = 0;
    high = (size == '1') ? 0xFF : 0xFFFF;
  } else if (type != 's') {
    readError(state, EMBER_CMD_ERR_WRONG_COMMAND_ARGUMENTS);
    return FALSE;
  } else if (size == '1') {
    low = -0x80;
    high = 0x7F;
  } else {
    low = -0x8000;
    high = 0x7FFF;
  }

  if (integerValue < low || high < integerValue) {
    readError(state, EMBER_CMD_ERR_ARGUMENT_OUT_OF_RANGE);
    return FALSE;
  }

  return addArgumentValue(state, integerValue, 2);
}

static boolean addStringArgument(EmberCommandState *state, boolean isBinary)
{
  if (state->argumentSpecs[0] != 'b') {
    readError(state, EMBER_CMD_ERR_WRONG_COMMAND_ARGUMENTS);
    return FALSE;
  } else {
    EmberMessageBuffer temp = state->stringBuffer;
    state->stringBuffer = EMBER_NULL_MESSAGE_BUFFER;
    return addArgumentValue(state, HIGH_LOW_TO_INT(isBinary, temp), 1);
  }
}

static void releaseArgumentBuffers(EmberCommandState *state)
{
  PGM_P spec = state->commands[state->commandIndex].argumentTypes;
  int8u i;
  int8u skip;

  for (i = 0; i < state->argumentCount; i++) {
    if (spec[0] == 'b') {
      emberReleaseMessageBuffer(LOW_BYTE(state->arguments[i]));
      skip = 1;
    } else 
      skip = 2;
    if (spec[skip] != '*')
      spec += skip;
  }
  state->argumentCount = 0;
} 

static boolean argumentsDone(PGM_P argumentSpecs)
{
  switch (argumentSpecs[0]) {
  case 0:
    return TRUE;
  case 'b':
    return argumentSpecs[1] == '*';
  default:
    return argumentSpecs[2] == '*';
  }
}

//----------------------------------------------------------------
// Strings for reporting the various possible errors.

PGM_NO_CONST PGM_P emberCommandErrorNames[] =
  {
    "no error",
    "serial port error",
    "command processor busy",
    "no such command",
    "incorrect command arguments",
    "too many command arguments",
    "integer argument out of range",
    "argument syntax error",
    "message too long",
    "missing binary message terminator"
  };

