/** @file hal-serial.h
 * @brief Serial hardware abstraction layer interfaces
 *
 * This file contains the HAL interfaces that must be implemented for the 
 * high-level serial code.
 *
 * Some functions in this file return an EmberStatus value. See 
 * error-def.h for definitions of all EmberStatus return values.
 *
 * <!-- Author(s): Lee Taylor, lee@ember.com -->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->                
 */

#ifndef __HAL_SERIAL_H__
#define __HAL_SERIAL_H__

/** @name Serial Port Definitions
 * Code currently only supports two ports, 0 and 1.
 */
//@{
#ifdef TINY_SERIAL_USART
  #define EM_NUM_SERIAL_PORTS 2
#else
  #define EM_NUM_SERIAL_PORTS 2
#endif
//@}  // End of Serial Port Definitions


/** @name Serial Mode Definitions
 * These are numerical definitions for the possible serial modes so that code
 * can test for the one being used.
 */
//@{
#define EMBER_SERIAL_UNUSED 0
#define EMBER_SERIAL_FIFO   1
#define EMBER_SERIAL_BUFFER 2
#define EMBER_SERIAL_LOWLEVEL 3
//@}  // End of Serial Mode Definitions


#ifndef DOXYGEN_SHOULD_SKIP_THIS

// Configures a specified serial port to operate in debug mode
#ifdef DEBUG_ON
#ifdef EMBER_SERIAL0_DEBUG
  #define EMBER_SERIAL0_MODE EMBER_SERIAL_BUFFER
  #define EMBER_SERIAL0_TX_QUEUE_SIZE 16
  #define EMBER_SERIAL0_RX_QUEUE_SIZE 2
#endif
#ifdef EMBER_SERIAL1_DEBUG
  #define EMBER_SERIAL1_MODE EMBER_SERIAL_BUFFER
  #define EMBER_SERIAL1_TX_QUEUE_SIZE 16
  #define EMBER_SERIAL1_RX_QUEUE_SIZE 2
#endif
#endif

// Determine if FIFO and/or Buffer modes are being used, so those sections of
//  code may be disabled if not
#if (EMBER_SERIAL0_MODE == EMBER_SERIAL_FIFO) || (EMBER_SERIAL1_MODE == EMBER_SERIAL_FIFO)
  #define EM_ENABLE_SERIAL_FIFO
#endif
#if (EMBER_SERIAL0_MODE == EMBER_SERIAL_BUFFER) || (EMBER_SERIAL1_MODE == EMBER_SERIAL_BUFFER)
  #define EM_ENABLE_SERIAL_BUFFER
#endif

// The following tests for setting of an invalid mode
#ifdef EMBER_SERIAL0_MODE
#if (EMBER_SERIAL0_MODE != EMBER_SERIAL_FIFO) && (EMBER_SERIAL0_MODE != EMBER_SERIAL_BUFFER) && (EMBER_SERIAL0_MODE != EMBER_SERIAL_LOWLEVEL)
  #error Invalid Serial 0 Mode
#endif
#else
  #define EMBER_SERIAL0_MODE EMBER_SERIAL_UNUSED
#endif
#ifdef EMBER_SERIAL1_MODE
#if (EMBER_SERIAL1_MODE != EMBER_SERIAL_FIFO) && (EMBER_SERIAL1_MODE != EMBER_SERIAL_BUFFER) && (EMBER_SERIAL1_MODE != EMBER_SERIAL_LOWLEVEL)
  #error Invalid Serial 1 Mode
#endif
#else
  #define EMBER_SERIAL1_MODE EMBER_SERIAL_UNUSED
#endif

#endif //DOXYGEN_SHOULD_SKIP_THIS

/** @brief Structure that maintains the state of the
* FIFO used in EMBER_SERIAL_FIFO mode.
*/
typedef struct {
  /** Index of next byte to send.*/
  int8u head;
  /** Index of where to enqueue next message.*/
  int8u tail;
  /** Number of bytes queued.*/
  volatile int8u used;
  /** FIFO of queue data.*/
  int8u fifo[0];
} EmSerialFifoQueue;

/** @brief Element of the FIFO in an EmSerialBufferQueue that maintains
* the state of the current packet buffer being output.
*/
typedef struct {
  /** Number of bytes in this buffer to write.*/
  int8u length;			
  /** The first linked buffer.*/
  EmberMessageBuffer buffer;	
  /** Starting position within the buffer.*/
  int8u startIndex;		
} EmSerialBufferQueueEntry;

/** @brief Structure that maintains a FIFO of packet buffers used in 
* EMBER_SERIAL_BUFFER mode.
*/
typedef struct {
  /** Index of next message to send */  
  int8u head;        
  /** Index of where to enqueue new messages.*/
  int8u tail;        
  /** Number of active messages.*/
  volatile int8u used;        
  /** Number of messages sent that need cleanup.*/
  volatile int8u dead;        
  /** Packet buffer currently writing from, if any (used to
  step down a chain of buffers).*/
  EmberMessageBuffer currentBuffer;
  /** Pointer to the next byte to send from current packet buffer.*/
  int8u *nextByte;
  /** Pointer to the last byte to send from current packet buffer.*/
  int8u *lastByte;
  /** FIFO of queued messages.*/
  EmSerialBufferQueueEntry fifo[0];
} EmSerialBufferQueue;

/** @name Serial Data Structures
 * These data structures maintain the state of serial communication on each
 * port. They are defined in serial.c and may be accessed by the HAL.
 */
//@{
/** Pointer to the transmit queue data structure for each port.*/
extern void *emSerialTxQueues[EM_NUM_SERIAL_PORTS];
/** Pointer to the receive FIFO queue data structure for each port.*/
extern EmSerialFifoQueue *emSerialRxQueues[EM_NUM_SERIAL_PORTS];
/** Masks used for managing the FIFOs of the transmit queues.*/
extern int8u PGM emSerialTxQueueMasks[EM_NUM_SERIAL_PORTS];
/** Masks used for managing the FIFOs of the receive queues.*/
extern int8u PGM emSerialRxQueueMasks[EM_NUM_SERIAL_PORTS];
/** Last UART receive error to occur.*/
extern int8u emSerialRxError[EM_NUM_SERIAL_PORTS];
/** Index in the FIFO queue where the last receive error occurred.*/
extern int8u emSerialRxErrorIndex[EM_NUM_SERIAL_PORTS];
/** Mode in which the ports are operating.*/
extern int8u PGM emSerialPortModes[EM_NUM_SERIAL_PORTS];

//@}  // end of Serial Data Structures

/** 
 * @description Enqueues a byte of data in a FIFO queue.
 *  
 * @param queue: Pointer to the FIFO queue.
 *  
 * @param data: Data byte to be enqueued.
 *  
 * @param mask: Mask used to control the wrap-around of the FIFO pointers.
 */
#define FIFO_ENQUEUE(queue,data,mask) \
  do {                                \
    queue->fifo[queue->head] = data;      \
    queue->head = ((queue->head + 1) & mask);  \
    queue->used++;                    \
  } while(0)
  
/** 
 * @description De-queues a byte of data from a FIFO queue.
 *  
 * @param queue: Pointer to the FIFO queue.
 *  
 * @param mask: Mask used to control the wrap-around of the FIFO pointers.
 */  
#define FIFO_DEQUEUE(queue,mask) \
  queue->fifo[queue->tail];          \
  queue->tail = ((queue->tail + 1) & mask);  \
  queue->used--

/** @name Serial HAL APIs
 * These functions must be implemented by the HAL in order for the
 * serial code to operate. Only the higher-level serial code uses these 
 * functions, so they should not be called directly. The HAL should also 
 * implement the appropriate interrupt handlers to drain the TX queues and fill 
 * the RX FIFO queue.
 */
//@{

/** @description Initializes the UART to the given settings (same parameters 
 * as  emberSerialInit() ).
 * 
 * @param port: Serial port number (0 or 1).
 *  
 * @param rate: Baud rate (see  SerialBaudRate).
 * 
 * @param parity: Parity value (see  SerialParity).
 * 
 * @param stopBits: Number of stop bits.
 * 
 * @return An error code if initialization failed (such as invalid baud rate),  
 * otherise EMBER_SUCCESS.
 */
EmberStatus emHalSerialInit(int8u port, 
                            SerialBaudRate rate,
                            SerialParity parity,
                            int8u stopBits);


/** @description Called by serial code whenever anything is queued for 
 * transmission, to start any interrupt-driven transmission. May
 * be called when transmission is already in progess.
 *  
 *  @param port: Serial port number (0 or 1).
 */
void emHalStartSerialTx(int8u port);


/** @description Called by serial code to stop any interrupt-driven serial
 * transmission currently in progress.
 * 
 * @param port: Serial port number (0 or 1).
 */
void emHalStopSerialTx(int8u port);


/** @description Directly writes a byte to the UART for transmission, 
 * regardless of anything currently queued for transmission. Should wait 
 * for anything currently in the UART hardware registers to finish transmission 
 * first, and block until \c dataByte is finished being sent.
 *  
 * @param port: Serial port number (0 or 1).
 *  
 * @param dataByte: The byte of data to be transmitted.
 */
void emHalForceWriteSerialByte(int8u port, int8u dataByte);


/** @description Blocks until the UART has finished transmitting any data in 
 * its hardware registers.
 * 
 * @param port: Serial port number (0 or 1).
 */
void emHalWaitSerialTxComplete(int8u port);
//@}  // end of Serial HAL APIs


/** @name Buffered Serial Utility APIs
 * The higher-level serial code implements these APIs, which the HAL uses 
 * to deal with buffered serial output.
 */
//@{

/** @description When new serial transmission is started and  
 * \c bufferQueue->nextByte is equal to NULL, this can be called to set up  
 * \c nextByte and \c lastByte for the next message.
 *  
 * @param q: Pointer to the buffer queue structure for the port.
 */
void emSerialBufferNextMessageIsr(EmSerialBufferQueue *q);


/** @description When a serial transmission is in progress and  
 * \c bufferQueue->nextByte has been sent and incremented leaving it equal to 
 * lastByte, this should be called to set up \c nextByte and \c lastByte 
 * for the next block.
 *  
 * @param q: Pointer to the buffer queue structure for the port.
 *  
 * @param port: Serial port number (0 or 1). 
*/
void emSerialBufferNextBlockIsr(EmSerialBufferQueue *q, int8u port);
//@}  // end of Buffered serial utility APIs


#endif //__HAL_SERIAL_H__
