/**
 * File: serial.c
 * Description: High Level Serial Communications
 *
 * Culprit(s):  Lee Taylor (lee@ember.com)
 *              Richard Kelsey (kelsey@ember.com)
 *              Jeff Mathews (jeff@ember.com)
 *
 * Copyright 2004 by Ember Corporation.  All rights reserved.               *80*
 */

#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "stack/include/error.h"
#include "stack/include/packet-buffer.h"
#include "serial.h"
#include "hal-serial.h"

#include <stdarg.h>


//Documentary comments:
// To conserve precious flash, there is very little validity checking
//  on the given parameters.  Be sure not to use an invalid port number
//  or a port that is unused.
// Blocking routines will always wait for room (but not buffers - if buffers
//  cannot be allocated, no part of the message will be sent)
// Non-blocking routines will never wait for room, and may cause partial
//  messages to be sent.  If a contiguous message needs to be sent, available
//  space should be checked _before_ calling the appropriate write API.


//------------------------------------------------------
// Determine if blocking code needs to be enabled
#if defined(EMBER_SERIAL0_BLOCKING) || defined(EMBER_SERIAL1_BLOCKING)
  #define EM_ENABLE_SERIAL_BLOCKING
#endif
#ifdef EMBER_SERIAL0_BLOCKING
  #define EM_SERIAL0_BLOCKSTATE TRUE
#else
  #define EM_SERIAL0_BLOCKSTATE FALSE
#endif
#ifdef EMBER_SERIAL1_BLOCKING
  #define EM_SERIAL1_BLOCKSTATE TRUE
#else
  #define EM_SERIAL1_BLOCKSTATE FALSE
#endif

//------------------------------------------------------
// Memory allocations for Queue data structures

// Allocate Appropriate TX Queue for port 0
#if EMBER_SERIAL0_MODE == EMBER_SERIAL_FIFO
  // Can't use structure typedef because we need to declare actual size of fifo
  static struct {
    /*! Indexes of next byte to send*/
    int8u head;
    /*! Index of where to enqueue next message*/
    int8u tail;
    /*! Number of bytes queued*/
    volatile int8u used;
    /*! FIFO of queue data*/
    int8u fifo[EMBER_SERIAL0_TX_QUEUE_SIZE];
  } emSerial0TxQueue;
  #define EM_SERIAL0_TX_QUEUE_ADDR (&emSerial0TxQueue)
#elif EMBER_SERIAL0_MODE == EMBER_SERIAL_BUFFER
  static struct {
    /*! Indexes of next byte to send*/
    int8u head;
    /*! Index of where to enqueue next message*/
    int8u tail;
    /*! Number of bytes queued*/
    volatile int8u used;
    /*! FIFO of queue data*/
    int8u dead;
    EmberMessageBuffer currentBuffer;
    int8u *nextByte, *lastByte;
    EmSerialBufferQueueEntry fifo[EMBER_SERIAL0_TX_QUEUE_SIZE];
  } emSerial0TxQueue;
  #define EM_SERIAL0_TX_QUEUE_ADDR (&emSerial0TxQueue)
#elif EMBER_SERIAL0_MODE == EMBER_SERIAL_UNUSED || \
      EMBER_SERIAL0_MODE == EMBER_SERIAL_LOWLEVEL
  #define EM_SERIAL0_TX_QUEUE_ADDR (NULL)
  #define EMBER_SERIAL0_TX_QUEUE_SIZE 0
  #define EMBER_SERIAL0_RX_QUEUE_SIZE 0
#endif

// Allocate Appropriate TX Queue for port 1
#if EMBER_SERIAL1_MODE == EMBER_SERIAL_FIFO
  static struct {
    /*! Indexes of next byte to send*/
    int8u head;
    /*! Index of where to enqueue next message*/
    int8u tail;
    /*! Number of bytes queued*/
    volatile int8u used;
    int8u fifo[EMBER_SERIAL1_TX_QUEUE_SIZE];
  } emSerial1TxQueue;
  #define EM_SERIAL1_TX_QUEUE_ADDR (&emSerial1TxQueue)
#elif EMBER_SERIAL1_MODE == EMBER_SERIAL_BUFFER
  static struct {
    /*! Indexes of next byte to send*/
    int8u head;
    /*! Index of where to enqueue next message*/
    int8u tail;
    /*! Number of bytes queued*/
    volatile int8u used;
    int8u dead;
    EmberMessageBuffer currentBuffer;  
    int8u *nextByte, *lastByte;
    EmSerialBufferQueueEntry fifo[EMBER_SERIAL1_TX_QUEUE_SIZE];
  } emSerial1TxQueue;
  #define EM_SERIAL1_TX_QUEUE_ADDR (&emSerial1TxQueue)
#elif EMBER_SERIAL1_MODE == EMBER_SERIAL_UNUSED || \
      EMBER_SERIAL1_MODE == EMBER_SERIAL_LOWLEVEL 
  #define EM_SERIAL1_TX_QUEUE_ADDR (NULL)
  #define EMBER_SERIAL1_TX_QUEUE_SIZE 0
  #define EMBER_SERIAL1_RX_QUEUE_SIZE 0
#endif

// Allocate RX Queues (Always FIFOs)
#if EMBER_SERIAL0_MODE != EMBER_SERIAL_UNUSED
  static struct {
    /*! Indexes of next byte to send*/
    int8u head;
    /*! Index of where to enqueue next message*/
    int8u tail;
    /*! Number of bytes queued*/
    volatile int8u used;
    int8u fifo[EMBER_SERIAL0_RX_QUEUE_SIZE];
  } emSerial0RxQueue;
  #define EM_SERIAL0_RX_QUEUE_ADDR (&emSerial0RxQueue)
#else
  #define EM_SERIAL0_RX_QUEUE_ADDR (NULL)
#endif
#if EMBER_SERIAL1_MODE != EMBER_SERIAL_UNUSED
  static struct {
    /*! Indexes of next byte to send*/
    int8u head;
    /*! Index of where to enqueue next message*/
    int8u tail;
    /*! Number of bytes queued*/
    volatile int8u used;
    int8u fifo[EMBER_SERIAL1_RX_QUEUE_SIZE];
  } emSerial1RxQueue;
  #define EM_SERIAL1_RX_QUEUE_ADDR (&emSerial1RxQueue)
#else
  #define EM_SERIAL1_RX_QUEUE_ADDR (NULL)
  #define emSerial1RxQueue (NULL)
#endif


//------------------------------------------------------
// Easy access to data structures for a particular port

// Data structure for referencing TX Queues
//  (allows for different modes and queue sizes)
void *emSerialTxQueues[EM_NUM_SERIAL_PORTS] = { (void *)EM_SERIAL0_TX_QUEUE_ADDR,
                                                (void *)EM_SERIAL1_TX_QUEUE_ADDR };

int8u PGM emSerialTxQueueSizes[EM_NUM_SERIAL_PORTS] =
  { EMBER_SERIAL0_TX_QUEUE_SIZE,
    EMBER_SERIAL1_TX_QUEUE_SIZE };

#pragma diag_suppress=pe068  // avoid warning about sign change with 0-1 case
int8u PGM emSerialTxQueueMasks[EM_NUM_SERIAL_PORTS] =
  { EMBER_SERIAL0_TX_QUEUE_SIZE-1,
    EMBER_SERIAL1_TX_QUEUE_SIZE-1 };
#pragma diag_default=pe068

// Data structure for referencing RX Queues
//  (allows for different queue sizes)
EmSerialFifoQueue *emSerialRxQueues[EM_NUM_SERIAL_PORTS] = 
  { (EmSerialFifoQueue *)EM_SERIAL0_RX_QUEUE_ADDR,
    (EmSerialFifoQueue *)EM_SERIAL1_RX_QUEUE_ADDR };

// Only needed by the simulator, so we ifdef it to save two (2) bytes of flash.
#ifdef SIMULATOR
int8u PGM emberSerialRxQueueSizes[EM_NUM_SERIAL_PORTS] =
  { EMBER_SERIAL0_RX_QUEUE_SIZE,
    EMBER_SERIAL1_RX_QUEUE_SIZE };
#endif

#pragma diag_suppress=pe068  // avoid warning about sign change with 0-1 case
int8u PGM emSerialRxQueueMasks[EM_NUM_SERIAL_PORTS] =
  { EMBER_SERIAL0_RX_QUEUE_SIZE-1,
    EMBER_SERIAL1_RX_QUEUE_SIZE-1 };
#pragma diag_default=pe068

// In-flash data structure for determined port mode
int8u PGM emSerialPortModes[EM_NUM_SERIAL_PORTS] = 
  { EMBER_SERIAL0_MODE,
    EMBER_SERIAL1_MODE };

#ifdef EM_ENABLE_SERIAL_BLOCKING
// In-flash data structure for blocking mode
boolean PGM emSerialBlocking[EM_NUM_SERIAL_PORTS] =
  { EM_SERIAL0_BLOCKSTATE,
    EM_SERIAL1_BLOCKSTATE };
#endif

int8u emSerialRxError[EM_NUM_SERIAL_PORTS] = {EMBER_SUCCESS,};
int8u emSerialRxErrorIndex[EM_NUM_SERIAL_PORTS] = {0,};

//------------------------------------------------------
// Private functions

// --------------------------------
// A simple printf() implementation
// Supported format specifiers are:
//  %% - percent sign
//  %c - single byte character
//  %s - ram string
//  %p - flash string  (non-standard)
//  %u - 2-byte unsigned decimal
//  %d - 2-byte signed decimal
//  %x %2x %4x - 1, 2, 4 byte hex value (always 0 padded) (non-standard)
// 
// A few macros and a function help make this readable:
//   - flush the local buffer to the output
//   - ensure that there is some room in the local buffer
//   - add a single byte to the local buffer
//   - convert a nibble to its ascii hex character
//   - convert an int16u to a decimal string
// Most of these only work within the emPrintfInternal() function.

#define flushBuffer() \
do { if (flushHandler(flushVar, localBuffer, i) != EMBER_SUCCESS) \
       goto fail;                                  \
     total += i;                                   \
     i = 0;                                        \
} while (FALSE)                                           

#define makeBufferRoom(need)	\
do { if (16 < i + need) flushBuffer(); } while (FALSE)

#define addByte(byte)	\
do { makeBufferRoom(1); localBuffer[i] = (byte); i += 1; } while (FALSE)

extern int8u PGM emHexChars[16] = "0123456789ABCDEF";
#define nibbleToHexChar(val) (emHexChars[((val) & 0x0F)])

// returns the number of digits added to charBuffer
int8u emPrintfWriteDecimal(int8u* charBuffer, int16u value)
{
  int8u numBuff[5]; // max int16u is 5 characters (65535)
  int8u i,qty;

  if(value == 0) {
    *charBuffer = '0';
    return 1;
  }

  i=0;
  while(value) {
    numBuff[i++] = value % 10;
    value /= 10;
  }
  qty=i;
  while(i--) {
    *charBuffer++ = (numBuff[i] + '0');
  }
  return qty;
}

// Returns number of characters written
int8u emPrintfInternal(emPrintfFlushHandler flushHandler, 
                       int8u flushVar, 
                       PGM_P string, 
                       va_list args)
{
  int8u localBuffer[16];
  int8u i = 0;
  int8u total = 0;

  for (; *string; string++) {
    if (*string != '%')
      addByte(*string);
    else {
      string += 1;
      switch (*string) {
      case '%':
        // escape for printing "%"
        addByte('%');
        break;
      case 'c':
        // character
        addByte(va_arg(args, unsigned int) & 0xFF);
        break;
      case 's': {
        // string
        int8u len;
        int8u *arg = va_arg(args, int8u *);
        flushBuffer();
        for(len=0;arg[len]!='\0';len++)
          ;
        if (flushHandler(flushVar, arg, len) != EMBER_SUCCESS)
          goto fail;
        total += len;
        break; }
      case 'p': {
        // flash string
        PGM_P arg = va_arg(args, PGM_P);
        int8u ch;
        flushBuffer();
        while((ch=*arg++)!='\0') {
          if(i>=16) {
            flushBuffer();
          }
          localBuffer[i++]=ch;
        }
        break; }
      case 'u':
        // unsigned 2-byte
        makeBufferRoom(5);
        i += emPrintfWriteDecimal(localBuffer + i, va_arg(args, int));
        break;
      case 'd': {
        // signed 2-byte
        int16s val = va_arg(args, int);
        makeBufferRoom(6);
        if(val < 0) {
          addByte('-');
          if(val != -32768)
            val = -val;
        }
        i += emPrintfWriteDecimal(localBuffer + i, val);
        break; }
      case 'x':
      case 'X': {
        // single hex byte (always prints 2 chars, ex: 0A)
        int8u data = va_arg(args, int);
        makeBufferRoom(2);
        localBuffer[i] = nibbleToHexChar(data >> 4);
        localBuffer[i+1] = nibbleToHexChar(data);
        i += 2;
        break; }
      case '2':
        // %2x only, 2 hex bytes (always prints 4 chars)
        string += 1;
        if(*string != 'x' && *string != 'X') {
          string -= 1;
          break;
        } else {
          int16u data = va_arg(args, int);
          int8u *localPtr;
          makeBufferRoom(4);
          localPtr = &localBuffer[i];
          *localPtr++ = nibbleToHexChar(data >> 12);
          *localPtr++ = nibbleToHexChar(data >> 8);
          *localPtr++ = nibbleToHexChar(data >> 4);
          *localPtr = nibbleToHexChar(data);
          i += 4;
        break; }
      case '4':
        // %4x only, 4 hex bytes (always prints 8 chars)
        string += 1;
        if(*string != 'x' && *string != 'X') {
          string -= 1;
          break;
        } else {
          int32u data = va_arg(args, int32u);
          int8u *localPtr;
          makeBufferRoom(8);
          // since we're on an 8 bit processor most of the time,
          //   its actually much more optimal to unroll this since
          //   the compiler actually accesses only the byte
          //   being referenced, rather than doing a really ineffiecient
          //   shift through 4 registers, which it would do if it were a loop
          localPtr = &localBuffer[i];
          *localPtr++ = nibbleToHexChar(data >> 28);
          *localPtr++ = nibbleToHexChar(data >> 24);
          *localPtr++ = nibbleToHexChar(data >> 20);
          *localPtr++ = nibbleToHexChar(data >> 16);
          *localPtr++ = nibbleToHexChar(data >> 12);
          *localPtr++ = nibbleToHexChar(data >> 8);
          *localPtr++ = nibbleToHexChar(data >> 4);
          *localPtr = nibbleToHexChar(data);
          i += 8;
        break; }
      case '\0': {
        string -= 1;
        // catch null in for loop
        break; }
      }
    }
  }

  flushBuffer();
  return total;

 fail:
  return 0;
}

//------------------------------------------------------
// Buffered Serial utility APIs

#if defined(EM_ENABLE_SERIAL_BUFFER) || defined(SIMULATOR)
// always executed in interrupt context
void emSerialBufferNextMessageIsr(EmSerialBufferQueue *q) 
{
  EmSerialBufferQueueEntry *e = &q->fifo[q->tail];

  q->currentBuffer = e->buffer;
  q->nextByte = emberLinkedBufferContents(q->currentBuffer) + e->startIndex;
  if((e->length + e->startIndex) > PACKET_BUFFER_SIZE) {
    q->lastByte = q->nextByte + ((PACKET_BUFFER_SIZE-1) - e->startIndex);
    e->length -= PACKET_BUFFER_SIZE - e->startIndex;
  } else {
    q->lastByte = q->nextByte + e->length - 1;
    e->length = 0;
  }
}
#endif

#if defined(EM_ENABLE_SERIAL_BUFFER) || defined(SIMULATOR)
#ifdef DEBUG_ON
  // Debug needs support for time sync messages to be 
  // inserted between normal messages
  extern boolean emDebugSyncNeeded;
  extern int8u emDebugPort;
  int8u *emDebugGetSyncMessage(int8u *len);
#endif
// always executed in interrupt context
void emSerialBufferNextBlockIsr(EmSerialBufferQueue *q, int8u port)
{
  EmSerialBufferQueueEntry *e = &q->fifo[q->tail];
  
  if(e->length != 0) {
    q->currentBuffer = emberStackBufferLink(q->currentBuffer);
    q->nextByte = emberLinkedBufferContents(q->currentBuffer);
    if(e->length > PACKET_BUFFER_SIZE) {
      q->lastByte = q->nextByte + 31;
      e->length -= PACKET_BUFFER_SIZE;
    } else {
      q->lastByte = q->nextByte + e->length - 1;
      e->length = 0;
    }
  } else {
// Low level Debug TimeSync interaction, currently disabled
//    #ifdef DEBUG_ON
//      int8u len;
//      if((emDebugSyncNeeded) && (emDebugPort == port)) {
//        q->nextByte = emDebugGetSyncMessage(&len);
//        q->lastByte = q->nextByte + len - 1;
//        return;
//      }
//    #endif
    q->tail = ((q->tail+1) & emSerialTxQueueMasks[port]);
    q->dead++;
    q->used--;
    if(q->used)
      emSerialBufferNextMessageIsr(q);
    else
      q->nextByte = NULL;
  }
}
#endif

//------------------------------------------------------
// Serial initialization

EmberStatus emberSerialInit(int8u port, 
                            SerialBaudRate rate,
                            SerialParity parity,
                            int8u stopBits)
{
  EmSerialFifoQueue *rq;

#if EMBER_SERIAL0_MODE == EMBER_SERIAL_UNUSED
  assert(port != 0);
#endif
#if EMBER_SERIAL1_MODE == EMBER_SERIAL_UNUSED
  assert(port != 1);
#endif
  assert(port < EM_NUM_SERIAL_PORTS);

  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO: {
    EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
    ATOMIC(
      q->used = 0;
      q->head = 0;
      q->tail = 0;
    )
    break; }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER: {
    EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[port];
    ATOMIC(
      q->used = 0;
      q->head = 0;
      q->tail = 0;
      q->dead = 0;
      q->currentBuffer = EMBER_NULL_MESSAGE_BUFFER;
      q->nextByte = NULL;
      q->lastByte = NULL;
    )
    break; }
#endif
  default:
    assert(FALSE);
    //break;  //statement is unreachable
  }

  rq = emSerialRxQueues[port];
  ATOMIC(
    rq->used = 0;
    rq->head = 0;
    rq->tail = 0;
    emSerialRxError[port] = EMBER_SUCCESS;
  )

  return emHalSerialInit(port, rate, parity, stopBits);
}

//------------------------------------------------------
// Serial Input

// returns # bytes available for reading
int8u emberSerialReadAvailable(int8u port)  
{
  return emSerialRxQueues[port]->used;
}

EmberStatus emberSerialReadByte(int8u port, int8u *dataByte)
{
  int8u retval;
  EmSerialFifoQueue *q = emSerialRxQueues[port];

  if(emSerialPortModes[port] == EMBER_SERIAL_UNUSED) {
    return EMBER_ERR_FATAL;
  }

  if(emSerialRxError[port] != EMBER_SUCCESS) {
    if(emSerialRxErrorIndex[port] == q->tail) {
      ATOMIC(
        retval = emSerialRxError[port];
        emSerialRxError[port] = EMBER_SUCCESS;
      )
      return retval;
    }
  }

  if(q->used > 0) {
    ATOMIC(
      *dataByte = FIFO_DEQUEUE(q,emSerialRxQueueMasks[port]);
    )
    return EMBER_SUCCESS;
  } else {
    return EMBER_SERIAL_RX_EMPTY;
  }
}

EmberStatus emberSerialReadLine(int8u port, char *data, int8u max)
{
  int8u err;
  int8u ch;
  int8u len;

  len = 0;
  data[0] = '\0';

  for (;;) {   
    err = emberSerialReadByte(port, &ch);

    // no new serial port char?, keep looping
    if (err) continue;

    // handle bogus characters
    if ( ch > 0x7F ) continue;

    // handle leading newline - fogBUGZ # 584
    if ((len == 0) && (ch == '\n')) continue;

    // handle backspace
    if ( ch == 0x08 ) {
      if ( len>0 ) {
        // delete the last character from our string
        len--;
        data[len] = '\0';
        // echo backspace
        emberSerialWriteString(port,"\010 \010");
      }
      // don't add or process this character
      continue;
    }

    //if the string is about to overflow, fake in a CR
    if ( len+2>max ) {
      ch = '\r';
    }

    emberSerialWriteByte(port, ch); // term char echo

    //upcase that char
    if ( ch>='a' && ch<='z') ch = ch - ('a'-'A');

    // build a string until we press enter
    if ( ch == '\r' ) {
      emberSerialWriteByte(port, '\n'); // "append" LF
      data[len] = '\0';
      return EMBER_SUCCESS;
    } 
      
    if ( ch != '\r' ) 
      data[len++] = ch;

  }
}

//------------------------------------------------------
// Serial Output

// returns # bytes (if fifo mode)/messages (if buffer mode) that can be written
int8u emberSerialWriteAvailable(int8u port)  
{
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO:
    return emSerialTxQueueSizes[port] - 
      ((EmSerialFifoQueue*)emSerialTxQueues[port])->used;
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER: {
    EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[port];
    return emSerialTxQueueSizes[port] - (q->used + q->dead);
    }
#endif
  }
  return 0;
}

EmberStatus emberSerialWriteByte(int8u port, int8u dataByte)
{
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO:
    {
      EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
      #ifdef EM_ENABLE_SERIAL_BLOCKING
        if(emSerialBlocking[port]) {
          while(q->used >= emSerialTxQueueSizes[port])
            ;
        } else
      #endif
        if(q->used >= emSerialTxQueueSizes[port])
          return EMBER_SERIAL_TX_OVERFLOW;

      ATOMIC(
        FIFO_ENQUEUE(q,dataByte,emSerialTxQueueMasks[port]);
      )
      // make sure the interrupt is enabled so it will be sent
      emHalStartSerialTx(port);
      return EMBER_SUCCESS;
    }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER:
    {
      // There is much room for optimization here
      EmberMessageBuffer buff = emberFillLinkedBuffers(&dataByte,1);
      if(buff != EMBER_NULL_MESSAGE_BUFFER) {
        EmberStatus stat = emberSerialWriteBuffer(port, buff, 0, 1);
        emberReleaseMessageBuffer(buff);
        return stat;
      } else 
        return EMBER_NO_BUFFERS;
    }
#endif
  default:
    return EMBER_ERR_FATAL;
  }
}

EmberStatus emberSerialWriteHex(int8u port, int8u dataByte)
{
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO:
    {
      EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];

      #ifdef EM_ENABLE_SERIAL_BLOCKING
        if(emSerialBlocking[port]) {
          while(q->used >= emSerialTxQueueSizes[port]-1)
            ;
        } else
      #endif
        if(q->used >= emSerialTxQueueSizes[port]-1)
          return EMBER_SERIAL_TX_OVERFLOW;

      ATOMIC(
        FIFO_ENQUEUE(q,nibbleToHexChar(dataByte>>4),emSerialTxQueueMasks[port]);
        FIFO_ENQUEUE(q,nibbleToHexChar(dataByte),emSerialTxQueueMasks[port]);
      )
      // make sure the interrupt is enabled so it will be sent
      emHalStartSerialTx(port);
      return EMBER_SUCCESS;
    }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER:
    {
      // There is much room for optimization here
      EmberMessageBuffer buff;
      int8u hex[2];
      hex[0] = nibbleToHexChar(dataByte>>4);
      hex[1] = nibbleToHexChar(dataByte);
      buff = emberFillLinkedBuffers(hex,2);
      if(buff != EMBER_NULL_MESSAGE_BUFFER) {
        EmberStatus stat = emberSerialWriteBuffer(port, buff, 0, 2);
        emberReleaseMessageBuffer(buff);
        return stat;
      } else 
        return EMBER_NO_BUFFERS;
    }
#endif
  default:
    return EMBER_ERR_FATAL;
  }
}

EmberStatus emberSerialWriteString(int8u port, PGM_P string)
{
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO:
    {
      EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
      while(*string != '\0') {

        #ifdef EM_ENABLE_SERIAL_BLOCKING
          if(emSerialBlocking[port]) {
            while(q->used >= emSerialTxQueueSizes[port])
              ;
          } else
        #endif
          if(q->used >= emSerialTxQueueSizes[port])
            return EMBER_SERIAL_TX_OVERFLOW;
        ATOMIC(
          FIFO_ENQUEUE(q,*string,emSerialTxQueueMasks[port]);
        )
        // make sure the interrupt is enabled so it will be sent
        emHalStartSerialTx(port);
        string++;
      }
      return EMBER_SUCCESS;
    }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER:
    {
      EmberMessageBuffer buff = emberAllocateStackBuffer();
      if(buff != EMBER_NULL_MESSAGE_BUFFER) {
        EmberStatus stat;
        if(emberAppendPgmStringToLinkedBuffers(buff, string) == EMBER_SUCCESS) {
          stat = emberSerialWriteBuffer(port, buff, 0, emberMessageBufferLength(buff));
        }
        emberReleaseMessageBuffer(buff);
        return stat;
      }
      return EMBER_NO_BUFFERS;
    }
#endif
  default:
    return EMBER_ERR_FATAL;
  }
}

EmberStatus emberSerialPrintf(int8u port, PGM_P formatString, ...)
{
   va_list ap;
   EmberStatus stat = EMBER_SUCCESS;

   va_start (ap, formatString);
   switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
   case EMBER_SERIAL_FIFO:
    {
      if(!emPrintfInternal(emberSerialWriteData, port, formatString, ap))
        stat = EMBER_ERR_FATAL;
      break;
    }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER:
    {
      EmberMessageBuffer buff = emberAllocateStackBuffer();
      if(buff == EMBER_NULL_MESSAGE_BUFFER) {
        stat = EMBER_NO_BUFFERS;
        break;
      }
      if(emPrintfInternal(emberAppendToLinkedBuffers, 
                          buff, 
                          formatString, 
                          ap)) {
        stat = emberSerialWriteBuffer(port,buff,0,emberMessageBufferLength(buff));
      } else {
        stat = EMBER_NO_BUFFERS;
      }
      emberReleaseMessageBuffer(buff);

      break;
    }
#endif
  }     
  va_end (ap);

  return stat;
}


EmberStatus emberSerialWriteData(int8u port, int8u *data, int8u length)
{
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO:
    {
      EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
 
      while(length--) {
        #ifdef EM_ENABLE_SERIAL_BLOCKING
          if(emSerialBlocking[port]) {
            while(q->used >= emSerialTxQueueSizes[port])
              ;
          } else
        #endif
          if(q->used >= emSerialTxQueueSizes[port])
             return EMBER_SERIAL_TX_OVERFLOW;
        ATOMIC(
          FIFO_ENQUEUE(q,*data,emSerialTxQueueMasks[port]);
        )
        // make sure the interrupt is enabled so it will be sent
        emHalStartSerialTx(port);
        data++;
      }
      return EMBER_SUCCESS;
    }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER:
    {
      // Note: We must always copy this, even in buffer mode
      //  since it is ram based data and there are no reference counts
      //  or indication of when it is actually written out the serial
      //  we cannot trust that the data won't be changed after this call
      //  but before it was actually written out.
      EmberMessageBuffer buff = emberFillLinkedBuffers(data,length);
      if(buff != EMBER_NULL_MESSAGE_BUFFER) {
        EmberStatus stat = emberSerialWriteBuffer(port, buff, 0, emberMessageBufferLength(buff));
        emberReleaseMessageBuffer(buff);
        return stat;
      } else 
        return EMBER_NO_BUFFERS;
    }
#endif
  default:
    return EMBER_ERR_FATAL;
  }
}

#if !defined(LEAF_NODE)
EmberStatus emberSerialWriteBuffer(int8u port, 
                                   EmberMessageBuffer buffer, 
                                   int8u start, 
                                   int8u length)
{
  if(buffer == EMBER_NULL_MESSAGE_BUFFER)
    return EMBER_ERR_FATAL;

  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO:
    {
      int8u *contents;
      EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
 
      for(;PACKET_BUFFER_SIZE <= start; start-=PACKET_BUFFER_SIZE)
        buffer = emberStackBufferLink(buffer);

      contents = emberMessageBufferContents(buffer) + start;

      while(length--) {
      #ifdef EM_ENABLE_SERIAL_BLOCKING 
        if(emSerialBlocking[port]) {
          while(q->used >= emSerialTxQueueSizes[port])
            ;
        } else
      #endif
        if(q->used >= emSerialTxQueueSizes[port])
          return EMBER_SERIAL_TX_OVERFLOW;
      
        ATOMIC(           
          FIFO_ENQUEUE(q,*contents++,emSerialTxQueueMasks[port]);
        )
        // make sure the interrupt is enabled so it will be sent
        emHalStartSerialTx(port);
        if(++start >= PACKET_BUFFER_SIZE) {
          start=0;
          buffer = emberStackBufferLink(buffer);
          contents = emberMessageBufferContents(buffer);
        }
      }
      break;
    }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER:
    {
      EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[port];
      EmSerialBufferQueueEntry *e;
      
      #ifdef EM_ENABLE_SERIAL_BLOCKING 
      if(emSerialBlocking[port]) {
        while(q->used + q->dead >= emSerialTxQueueSizes[port])
          emberSerialBufferTick();
      } else
      #endif
      if(q->used + q->dead >= emSerialTxQueueSizes[port]) {
        if(q->dead)
          emberSerialBufferTick();
        else
          return EMBER_SERIAL_TX_OVERFLOW;
      }

      for(;PACKET_BUFFER_SIZE <= start; start-=PACKET_BUFFER_SIZE)
        buffer = emberStackBufferLink(buffer);
      emberHoldMessageBuffer(buffer);
      
      e = &q->fifo[q->head];
      e->length = length;
      e->buffer = buffer;
      e->startIndex = start;
      q->head = ((q->head+1) & emSerialTxQueueMasks[port]);
      ATOMIC(
        q->used++;
      )
      emHalStartSerialTx(port);  
      break; 
    }
#endif
  default:
    return EMBER_ERR_FATAL;
  }
  return EMBER_SUCCESS;
}
#endif // if !defined(LEAF_NODE)

EmberStatus emberSerialWaitSend(int8u port)  // waits for all byte to be written out of a port
{
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO: {
    EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
    while(q->used)
      simulatedSerialTimePasses();
    break; }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER: {
    EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[port];
    while(q->used)
      simulatedSerialTimePasses();
    break; }
#endif
  }
  emHalWaitSerialTxComplete(port);
  return EMBER_SUCCESS;
}

//------------------------------------------------------
// Guaranteed output

// Flush function for emberSerialGuaranteedPrintf
static EmberStatus emSerialGuaranteedFlush(int8u port, int8u *data, int8u length)
{
  while(length--) {
    emHalForceWriteSerialByte(port, *data);
    data++;
  }
  return EMBER_SUCCESS;
}

// The _only_ Guaranteed API:  The usage model for this api 
//   Does not require efficiency
EmberStatus emberSerialGuaranteedPrintf(int8u port, PGM_P formatString, ...)
{
  va_list ap;
  
  // prevent interrupt driven transmission from intermixing
  emHalStopSerialTx(port);

  va_start(ap, formatString);
  emPrintfInternal(emSerialGuaranteedFlush, port, formatString, ap);
  va_end(ap);

  // re-enable interrupt driven transmission if needed
  switch(emSerialPortModes[port]) {
#ifdef EM_ENABLE_SERIAL_FIFO
  case EMBER_SERIAL_FIFO: {
    EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[port];
    if(q->used)
      emHalStartSerialTx(port);
    break; }
#endif
#ifdef EM_ENABLE_SERIAL_BUFFER
  case EMBER_SERIAL_BUFFER: {
    EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[port];
    if(q->used)
      emHalStartSerialTx(port);
    break; }
#endif
  }

  return EMBER_SUCCESS;
}

//------------------------------------------------------
// Serial buffer maintenance
void emberSerialFlushRx(int8u port) 
{
  EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialRxQueues[port];
  ATOMIC(
	 q->used = 0;
	 q->head = 0;
	 q->tail = 0;
	 )
    
}

//------------------------------------------------------
// Serial Buffer Cleanup Tick

void emberSerialBufferTick(void)
{
#ifdef EM_ENABLE_SERIAL_BUFFER
  int8u port;
  EmSerialBufferQueue *q;
  int8u numDead, deadIndex;

  for(port=0; port<EM_NUM_SERIAL_PORTS; port++) {
    if(emSerialPortModes[port] == EMBER_SERIAL_BUFFER) {
      q = (EmSerialBufferQueue *)emSerialTxQueues[port];

      if(q->dead) {
        ATOMIC(
          numDead = q->dead;
          q->dead = 0;
          deadIndex = (q->tail - numDead) & emSerialTxQueueMasks[port];
        )
        for(;numDead;numDead--) {
          emberReleaseMessageBuffer(q->fifo[deadIndex].buffer);
          deadIndex = (deadIndex + 1) & emSerialTxQueueMasks[port];
        }
      }
    }
  }
#endif
}
