/** @file serial.h
 * @brief High-level serial communication functions
 *
 * Unless otherwise noted, the EmberNet stack does not use these functions, 
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 *
 * Many functions in this file return an EmberStatus value. See 
 * stack/include/error-defs.h for definitions of all EmberStatus return values.
 *
 * <!--Culprit(s):  Lee Taylor (lee@ember.com)-->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */

#ifndef __SERIAL_H__
#define __SERIAL_H__

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#include <stdarg.h>
#endif // DOXYGEN_SHOULD_SKIP_THIS

typedef enum {
  BAUD_2400 = 0,
  BAUD_4800 = 1,
  BAUD_9600 = 2,
  BAUD_14400 = 3,
  BAUD_19200 = 4,
  BAUD_28800 = 5,
  BAUD_38400 = 6,
  BAUD_50000 = 7,
  BAUD_57600 = 8,
  BAUD_76800 = 9,
  BAUD_100000 = 10,
  BAUD_115200 = 11
} SerialBaudRate;

typedef enum {
  PARITY_NONE = 0,
  PARITY_ODD = 1,
  PARITY_EVEN = 2
} SerialParity;

 /** @name Serial Functions
 *  To use these serial routines, they must be properly configured. 
 *
 *  The functions can work in two ways, depending on how messages waiting for 
 *  transmission are stored:
 *  - Buffered mode: Stack linked buffers are used. This can be more efficient
 *  if many messages received over the air also need to be transmitted
 *  over the serial interface.
 *  - FIFO mode: A statically allocated queue of bytes is used,
 *  and data to be transmitted are copied into it.
 *
 * (These modes deal only with data transmission. Note that data \b reception 
 *  always occurs in a FIFO-like mode.)
 * 
 *  The current version of these sources provides support for up to two serial
 *  ports, but it could be easily extended. The ports are numbered 0
 *  and 1 and should be accessed by those numbers. The ports can be set up 
 *  independently of each other.
 *
 *  To enable a port, a use mode (buffered or FIFO) and a queue size must be 
 *  declared on it. In FIFO mode, the queue size is the size of the FIFO and 
 *  represents the number of bytes that can be waiting for transmission at any 
 *  one time. In buffered mode, the queue size represents the number of whole
 *  "messages" that can be waiting for transmission at any one time. A single 
 *  message is created for each call to any of the serial APIs.
 *  
 *  You can specify a use mode and queue size placing declarations in the compiler 
 *  preprocessor options when building your application:
 *  - <b>Use mode:</b>
 *    -  EMBER_SERIAL0_MODE=EMBER_SERIAL_BUFFER or EMBER_SERIAL_FIFO
 *    -  EMBER_SERIAL1_MODE=EMBER_SERIAL_BUFFER or EMBER_SERIAL_FIFO
 *  - <b>Queue size</b> (note that queue sizes \b MUST always be a power of 2):
 *    -  EMBER_SERIAL0_TX_QUEUE_SIZE=2
 *    -  EMBER_SERIAL0_RX_QUEUE_SIZE=4
 *    -  EMBER_SERIAL1_TX_QUEUE_SIZE=8
 *    -  EMBER_SERIAL1_TX_QUEUE_SIZE=16
 *
 *  Note the following: 
 *  - If buffered mode is declared, emberSerialBufferTick() should be 
 *  called in the application's main event loop.
    - By default, both ports are unused.
 *
 *  You can also use declarations to specify what should be done if an attempt
 *  is made to send more data than the queue can accommodate:
 *  - EMBER_SERIAL0_BLOCKING
 *  - EMBER_SERIAL1_BLOCKING
 *
 *  Be aware that since blocking spins in a loop, doing nothing until space 
 *  is available, it can adversely affect any code that has tight timing 
 *  requirements.
 *
 *  If EMBER_SERIAL0_BLOCKING or EMBER_SERIAL1_BLOCKING is defined,
 *  then the call to the port will block until space
 *  is available, guaranteeing that the entire message is sent. Note that 
 *  in buffered mode, even if blocking mode is in effect entire messages
 *  may be dropped if insufficient stack buffers are available to hold them.
 *  When this happens, EMBER_NO_BUFFERS is returned.
 *  
 *  If no blocking mode is defined, the serial code defaults to "non-blocking"
 *  mode. In this event, when the queue is too short, the data that don't fit 
 *  are dropped. In FIFO mode, this may result bytes being dropped,
 *  starting in the middle of message. In buffered mode, the entire
 *  message is dropped. When data is dropped, EMBER_SERIALTX_OVERFLOW 
 *  is returned. 
 *
 *  To minimize code size, very little error checking is done on
 *  the given parameters. Specifying an invalid or unused serial port may
 *  result in unexplained behavior. In some cases EMBER_ERR_FATAL may be 
 *  returned.
 */

//@{
/** @description Initializes a serial port to a specific baud rate, parity,
 * and number of stop bits. Eight data bits are always used.
 *   
 * @param port: A serial port number (0 or 1).
 *  
 * @param rate: The baud rate (see SerialBaudRate).
 *   
 * @param parity: The parity value (see SerialParity).
 *   
 * @param stopBits: The number of stop bits.
 *   
 * @return An error code if initialization failed (such as invalid baudrate), 
 * or EMBER_SUCCESS.
 */
EmberStatus emberSerialInit(int8u port, 
                            SerialBaudRate rate,
                            SerialParity parity,
                            int8u stopBits);

/** @description Returns the number of bytes currently available for reading
 * in the specified RX queue.
 *  
 * @param port: A serial port number (0 or 1).
 *   
 * @return The number of bytes available.
 */
int8u emberSerialReadAvailable(int8u port);

/** @description Reads a byte from the specified RX queue.
 * 
 * @param port: A serial port number (0 or 1).
 *   
 * @param dataByte: A pointer to storage location for the byte.
 *  
 * @return EMBER_SERIAL_RX_EMPTY if no data is available, or EMBER_SUCCESS.
 */
EmberStatus emberSerialReadByte(int8u port, int8u *dataByte);

/** @description Simulates a terminal interface, reading a line of characters
 * at a time. Supports backspace. Always converts to uppercase.
 * Blocks until a line has been read or max has been exceeded.
 *  
 * @param port: A serial port number (0 or 1).
 *   
 * @param data: A pointer to storage location for the read line. There must be
 * \c max contiguous bytes available at this location.
 *  
 * @param max: The maximum number of bytes to read.
 *  
 * @return  EMBER_SUCCESS
 */
EmberStatus emberSerialReadLine(int8u port, char *data, int8u max);

/** @description Returns the number of bytes (in FIFO mode) or messages 
 * (in buffered mode) that can currently be queued to send without blocking
 * or dropping.
 * 
 * @param port: A serial port number (0 or 1).
 *   
 * @return The umber of bytes or messages available for queueing. 
 */
int8u emberSerialWriteAvailable(int8u port);

/** @description Queues a single byte of data for transmission on the 
 * specified port.
 *  
 * @param port: A serial port number (0 or 1).
 *   
 * @param dataByte: The byte to be queued.
 *  
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data were dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialWriteByte(int8u port, int8u dataByte);

/** @description Converts a given byte of data to its two-character ASCII
 * hex representation and queues it for transmission on the specified port. 
 * Values less than 0xF are always zero padded and queued as "0F".
 *  
 * @param port: A serial port number (0 or 1).
 *
 * @param dataByte: The byte to be converted.
 *  
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data were dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialWriteHex(int8u port, int8u dataByte);

/** @description Queues a string for transmission on the specified port.
 *  
 * @param port: A serial port number (0 or 1).
 *   
 * @param string: The string to be queued.
 *   
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data were dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialWriteString(int8u port, PGM_P string);

/** @description Printf for printing on a specified port. Supports the 
 * following format specifiers:
 * - %% percent sign
 * - %c single-byte character
 * - %s RAM string
 * - %p flash string (nonstandard specifier)
 * - %u 2-byte unsigned decimal
 * - %d 2-byte signed decimal
 * - %x %2x %4x 1-, 2-, 4-byte hex value (always 0 padded) (nonstandard
 * specifier)
 *   
 * @param port: A serial port number (0 or 1).
 *   
 * @param formatString: The string to print.
 *  
 * @param ...: Format specifiers. 
 *   
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data was dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialPrintf(int8u port, PGM_P formatString, ...);

/** @description Queues an arbitrary chunk of data for transmission on a 
 * specified port.
 *   
 * @param port: A serial port number (0 or 1).
 *   
 * @param data: A pointer to data.
 *   
 * @param length: The number of bytes to queue.
 *    
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data was dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialWriteData(int8u port, int8u *data, int8u length);

/** @description Queues data contained in linked stack buffers for 
 * transmission on a specified port. Can specify an arbitrary initial 
 * offset within the linked buffer chain.
 *   
 * @param port: A serial port number (0 or 1).
 *   
 * @param buffer: The starting buffer in linked buffer chain.
 *   
 * @param start: The offset from first buffer in chain.
 *  
 * @param length: The number of bytes to queue.
 *    
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data was dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialWriteBuffer(int8u port, EmberMessageBuffer buffer, int8u start, int8u length);

/** @description Waits for all data currently queued on the specified port to 
 * be transmitted before returning. \b Note: Call this function before serial 
 * reinitialization to ensure that transmission is complete.
 *   
 * @param port: A serial port number (0 or 1).
 * 
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data was dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialWaitSend(int8u port);

/** @description A printf routine that takes over the specified serial 
 * port and immediately transmits the given data regardless of what is 
 * currently queued. Does not return until the transmission is complete. 
 *  
 * @appusage Useful for fatal situations (such as asserts) where the node will 
 * be reset, but information on the cause for the reset needs to be transmitted 
 * first.
 *   
 * @param port: A serial port number (0 or 1).
 *  
 * @param formatString: The string to print.
 *  
 * @param ...: Formatting specifiers. See emberSerialPrintf() for arguments.
 *   
 * @return One of the following (see the Main Page):
 * - EMBER_SERIAL_TX_OVERFLOW when data was dropped. 
 * - EMBER_NO_BUFFERS when insufficient stack buffers were available.
 * - EMBER_SUCCESS.
 */
EmberStatus emberSerialGuaranteedPrintf(int8u port, PGM_P formatString, ...);

/** @description When a serial port is used in buffered mode, this must be
 * called in an application's main event loop, similar to emberTick(). 
 * It handles freeing the buffers that are used in message queueing. \b Note:
 * Has no effect if only FIFO mode is being used.
 */
void emberSerialBufferTick(void);

/** @description Flushes the receive buffer in case none of the
 * incoming serial data is wanted.
 *
 * @param port: A serial port number (0 or 1).
*/
void emberSerialFlushRx(int8u port);

//@}  // end of serial Functions

/** @name Printf Prototypes
 * These prototypes are for the internal printf implementation, in case
 * it is desired to use it elsewhere. See the code for emberSerialPrintf()
 * for an example of printf usage.
 */
//@{

/** @description Typedefine to cast a function into the appropriate format
 * to be used inside the \c emPrintfInternal function below, for performing the
 * actual flushing of a formatted string to a destination such as a serial port.
 *
 * @param flushVar: The destination of the flush, most commonly a serial port
 * number (0 or 1).
 * 
 * @param contents: A pointer to the string to flush.
 *
 * @param length: The number of bytes to flush.
 *
 * @return The EmberStatus value of the typedefined function.
 */
typedef EmberStatus (emPrintfFlushHandler)(int8u flushVar, 
                                           int8u *contents, 
                                           int8u length);


/** @description The internal printf function, which scans the string for the
 * format specifiers and appropriately implants the passed data into the string.
 *
 * @param emPrintfFlushHandler: The name of an internal function, which has
 * parameters matching the function \c emPrintfFlushHandler above, responsible
 * for flushing a string formatted by this function, \c emPrintfInternal, to
 * the appropriate buffer or function that performs the actual transmission.
 * 
 * @param int8u: The destination of the flush performed above, most commonly
 * serial port number (0 or 1).
 *
 * @param PGM_P: The string to print.
 *
 * @param va_list: The list of arguments for the format specifiers.
 *
 * @return The number of characters written.
 */
int8u emPrintfInternal(emPrintfFlushHandler, int8u, PGM_P, va_list);

//@}  // end of Printf Prototypes


#endif // __SERIAL_H__

