/** @file hal/board/board.h
 * @brief General carrier board functions and defines.
 *  
 * Unless otherwise noted, the EmberNet stack does not use these functions, 
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */

#ifndef __BOARD_H__
#define __BOARD_H__

#ifndef DOXYGEN_SHOULD_SKIP_THIS

// A specific board should always be defined
#ifndef BOARD_INCLUDE
  #error No board defined
#endif

#include BOARD_INCLUDE

#endif  // DOXYGEN_SHOULD_SKIP_THIS

#if !defined(__IAR_SYSTEMS_ASM__)


/** @description Initializes carrier board-specific hardware such as LEDs, RS-232
 * transceivers, temperature sensors, buttons, and the accelerometer.
 */
void boardBoot(void);

/** @description Turns off all carrier board-specific hardware. The reverse 
 * of boardReboot().
 * 
 * @appusage Use to shut down the carrier board-specific hardware for deep 
 * sleeping.
 */
void boardPowerDown(void);

/** @description Brings carrier board-specific hardware back up after having been
 * shut down with boardPowerDown(). @note This function is \b DEPRECATED. Use 
 * boardReboot() instead.
 */
void boardPowerUp(void);


/** @description Called implicitly through the standard C language assert() macro. An 
 * implementation where notification is, for instance, sent over the serial port
 * can provide meaningful and useful debugging information. \b Note: Liberal 
 * usage of assert() will consume flash space.
 * 
 * @stackusage The interface must exist, but implementation is optional.
 *     
 * @param filename: Name of the file throwing the assert.
 * 
 * @param linenumber: Line number that threw the assert.
 */
void boardAssertFailed(PGM_P filename, int linenumber);

/**  @description Configures GPIOs pertaining to the control of LEDs.*/
void boardLEDInit(void);

/** @description Atomically wraps an XOR or similar operation for a single GPIO
 * pin attached to an LED.
 * 
 * @param led: Identifier for the LED to be toggled.
 */
void boardLEDToggle(int8u led);

/** @description Turns on (sets) a GPIO pin connected to an LED so that the LED 
 * turns on.
 * 
 * @param led: Identifier for the LED to turn on.
 */
void boardLEDSet(int8u led);

/** @description Turns off (clears) a GPIO pin connected to an LED, which turns off
 * the LED. 
 *  
 * @param led: Identifier for the LED to turn off.
 */
void boardLEDClear(int8u led);

#ifndef DOXYGEN_SHOULD_SKIP_THIS 

//  Notes which can be used in tune buzzer definitions.
//  This is specific to evaluation kit like functionality.

#if defined(AVR7p3278MHZ) || defined(AVR8MHZ)

// Note: Flats are used instead of sharps because # is a special character
#define	NOTE_C3	119
#define	NOTE_Db3	112
#define	NOTE_D3	106
#define	NOTE_Eb3	100
#define	NOTE_E3	94
#define	NOTE_F3	89
#define	NOTE_Gb3	84
#define	NOTE_G3	79
#define	NOTE_Ab3	74
#define	NOTE_A3	70
#define	NOTE_Bb3	66
#define	NOTE_B3	63
#define	NOTE_C4	59
#define	NOTE_Db4	55
#define	NOTE_D4	52
#define	NOTE_Eb4	49
#define	NOTE_E4	46
#define	NOTE_F4	44
#define	NOTE_Gb4	41
#define	NOTE_G4	39
#define	NOTE_Ab4	37
#define	NOTE_A4	35
#define	NOTE_Bb4	33
#define	NOTE_B4	31
#define	NOTE_C5	29
#define	NOTE_Db5	27
#define	NOTE_D5	26
#define	NOTE_Eb5	24
#define	NOTE_E5	23
#define	NOTE_F5	21
#define	NOTE_Gb5	20
#define	NOTE_G5	19
#define	NOTE_Ab5	18
#define	NOTE_A5	17
#define	NOTE_Bb5	16
#define	NOTE_B5	15

#else  // 3p68MHZ or 4MHZ

#define NOTE_C3 220
#define NOTE_Db3 207
#define NOTE_D3 196
#define NOTE_Eb3 184
#define NOTE_E3 174
#define NOTE_F3 164
#define NOTE_Gb3 155
#define NOTE_G3 146
#define NOTE_Ab3 138
#define NOTE_A3 130
#define NOTE_Bb3 122
#define NOTE_B3 116
#define NOTE_C4 109   // Middle C
#define NOTE_Db4 103
#define NOTE_D4 97
#define NOTE_Eb4 91
#define NOTE_E4 86
#define NOTE_F4 81
#define NOTE_Gb4 77
#define NOTE_G4 73
#define NOTE_Ab4 68
#define NOTE_A4 64
#define NOTE_Bb4 61
#define NOTE_B4 57
#define NOTE_C5 54
#define NOTE_Db5 51
#define NOTE_D5 48
#define NOTE_Eb5 45
#define NOTE_E5 43
#define NOTE_F5 40
#define NOTE_Gb5 38
#define NOTE_G5 36
#define NOTE_Ab5 34
#define NOTE_A5 32
#define NOTE_Bb5 30
#define NOTE_B5 28

#endif

#endif // DOXYGEN_SHOULD_SKIP_THIS

//////////////////////////////////////////////////////////////////////
#endif  // __IAR_SYSTEMS_ASM__
//////////////////////////////////////////////////////////////////////

#endif //__BOARD_H__
