/*
 * File: board.c
 * Description: Devkit board specific HAL sources
 *
 * Author(s): Lee Taylor, lee@ember.com,
 *            Jeff Mathews, jm@ember.com
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#define ENABLE_BIT_DEFINITIONS
#include <ioavr.h>

#include PLATFORM_HEADER
#include "stack/include/ember.h"

#include "app/util/serial/serial.h"
#include "micro/adc.h"

#include "stack/include/hal.h"
#include "board/board.h"
#include "micro/micro.h"


void boardBoot(void)
{
  boardLEDInit();
  #ifdef BOARD0220
    connLEDInit();
  #endif
  adcInit();
  boardEnableTempSensor();

  // Enable the RS-232 tranceiver
  // always enabled on dev kit
  
  // Accelerometer
  // none on dev kit
  
  // Buzzer
  DDRB |= 0x80;

  boardButtonInit();
}

void boardReboot(void)
{
  microReboot();
}

void boardPowerDown(void) {
  // Temp sensor
  boardDisableTempSensor();
  
  // RS232
  // n/a
  
  // Accelerometer
  // n /a
  
  // Buzzer
  TCCR2 = BIT(WGM21) | BIT(COM20); // no clock (timer stopped)
  TIMSK &= ~(BIT(OCIE2)); // disable interrupt
  
  // Adc
  adcSleep();
  
  // Uarts
  UCSR0B &= ~(BIT(RXCIE0) | BIT(TXCIE0) | BIT(UDRIE0) | BIT(RXEN0) | BIT(TXEN0));
  UCSR1B &= ~(BIT(RXCIE1) | BIT(TXCIE1) | BIT(UDRIE1) | BIT(RXEN1) | BIT(TXEN1));

  // Leds
  boardLEDInit(); // Turns leds off
  #ifdef BOARD0220
    connLEDInit();
  #endif
}

void boardPowerUp(void) {
  // Temp sensor
  boardEnableTempSensor();
  // RS232
  // n/a
  // Accelerometer
  // n/a
  // Buzzer
  // Don't need to do anything
  // Adc
  adcInit();
  // Uarts
  // emberSerialInit must be called after boardPowerUp to turn the uarts back on
  // Leds
  // Don't need to do anything
}


void boardEnableTempSensor(void) {
  DDRC |= BIT(1);
  PORTC |= BIT(1);
}

void boardDisableTempSensor(void) {
  PORTC &= ~BIT(1);
}


void boardAssertFailed(PGM_P filename, int linenumber)
{
  __watchdog_reset();              // In case we're close to running out.
#if DEBUG_ON
  emberDebugAssert(filename, linenumber);
  __watchdog_reset();              // In case we're close to running out.
  emberDebugMemoryDump((int8u *)0, (int8u *)4095);  // entire ram
#else
  __watchdog_reset();              // In case we're close to running out.
  emberSerialGuaranteedPrintf(1, "\n\r[ASSERT:%p:%d:", filename, linenumber);
  //stack trace:
  {
    int8u i;
    for (i=0; i<6; i++) {
      int8u l,h;
      int16u pc;
      h = *((int8u*)(SP+2*i+1));
      l = *((int8u*)(SP+2*i+2));
      pc = HIGH_LOW_TO_INT(h,l);
      pc = pc << 1;
      emberSerialGuaranteedPrintf(1, " %2x", pc);
    }
  }
  emberSerialGuaranteedPrintf(1, "]\n\r");
#endif
  microReboot();                        
}

// This state is kept track of so the ISRs know whether they are looking
//  for rising or falling edges
int8u button1State, button2State;

void boardButtonInit(void)
{  
  // Enable Pull-ups 
  PORTD |= BIT(0) | BIT(1);
  // Configure for falling edge
  EICRA |= BIT(ISC01) | BIT(ISC11);
  EICRA &= ~(BIT(ISC00) | BIT(ISC10));

  button1State = BUTTON_RELEASED;
  button2State = BUTTON_RELEASED;

  // Enable Interrupts
  EIMSK |= BIT(INT0) | BIT(INT1);

}

int8u boardButtonState(int8u button)
{
  // Note: this returns the "soft" state rather than reading the port
  //  so it jives with the interrupts and their callbacks
  switch(button) {
    case BUTTON_1:
      return button1State;
    case BUTTON_2:
      return button2State;
    default:
      return BUTTON_RELEASED;
  }
}


#pragma vector = INT0_vect                        
__interrupt void but1Interrupt(void)
{
  if(button1State == BUTTON_RELEASED) {
    button1State = BUTTON_PRESSED;
    boardButtonCallback(BUTTON_1, BUTTON_PRESSED);

    // switch Interrupt to trigger on rising edge (release)
    EICRA |= BIT(ISC00);

    // check see if it was released in the meantime, meaning we won't
    //   get the release interrupt
    if(PIND & BIT(0)) {
      // reconfigure back to falling edge (press)
      EICRA &= ~(BIT(ISC00));

      // clear interrupt flag just in case it did get triggered in the short
      //  period between when we set up the interrupt and tested for release
      EIFR = BIT(INTF0);

      // finally, call the released callback since we won't get the interrupt
      button1State = BUTTON_RELEASED;
      boardButtonCallback(BUTTON_1, BUTTON_RELEASED);
    }
  } else {
    button1State = BUTTON_RELEASED;
    boardButtonCallback(BUTTON_1, BUTTON_RELEASED);

    // reconfigure back to falling edge (press)
    EICRA &= ~(BIT(ISC00));
  }
}

#pragma vector = INT1_vect                        
__interrupt void but2Interrupt(void)
{
  if(button2State == BUTTON_RELEASED) {
    button2State = BUTTON_PRESSED;
    boardButtonCallback(BUTTON_2, BUTTON_PRESSED);

    // switch Interrupt to trigger on rising edge (release)
    EICRA |= BIT(ISC10);

    // check see if it was released in the meantime, meaning we won't
    //   get the release interrupt
    if(PIND & BIT(1)) {
      // reconfigure back to falling edge (press)
      EICRA &= ~(BIT(ISC10));

      // clear interrupt flag just in case it did get triggered in the short
      //  period between when we set up the interrupt and tested for release
      EIFR = BIT(INTF1);

      // finally, call the released callback since we won't get the interrupt
      button2State = BUTTON_RELEASED;
      boardButtonCallback(BUTTON_2, BUTTON_RELEASED);
    }
  } else {
    button2State = BUTTON_RELEASED;
    boardButtonCallback(BUTTON_2, BUTTON_RELEASED);

    // reconfigure back to falling edge (press)
    EICRA &= ~(BIT(ISC10));
  }
}


void boardLEDInit(void)
{
  int8u i;

  // Initialize the LEDs ports for output 
  #ifdef BOARD0221
    DDRC |= BIT(2) | BIT(3) | BIT(4) | BIT(5) | BIT(6) | BIT(7) ;
  #else
    DDRC |= BIT(4) | BIT(5) | BIT(6) | BIT(7) ;
  #endif 

  // turn the LEDs off
  for(i=0; i<=BOARDLED_MAX; i++)
    boardLEDClear(i);
}


void boardLEDToggle (int8u led)
{
  // c lanuguage xor on an i/o port doesn't make avr machine code 
  // which is atomic. so to avoid contention with other code using
  // the other pins for other purposes, we disable interrupts to
  // make the read io register, xor and write io atomic.
  
  if(led <= BOARDLED3) {
    ATOMIC(
      PORTC ^= BIT(led + 4);
    )
  }
  #ifdef BOARD0221
  else if(led <= BOARDLED_MAX) {
      ATOMIC(
        PORTC ^= BIT(led - 2);
      )
  }
  #endif
}
  
void boardLEDSet (int8u led)
{
  // this should all compile to an atomic assembly instruction
  if(led <= BOARDLED3) {
      PORTC &= ~BIT(led + 4);
  }
  #ifdef BOARD0221
  else if(led <= BOARDLED_MAX) {
        PORTC &= ~BIT(led - 2);
  }
  #endif
}

void boardLEDClear (int8u led)
{
  // this should all compile to an atomic assembly instruction  
  if(led <= BOARDLED3) {
      PORTC |= BIT(led + 4);
  }
  #ifdef BOARD0221
  else if(led <= BOARDLED_MAX) {
      PORTC |= BIT(led - 2);
  }
  #endif
}

void halIndicateActivity(boolean turnOn)
{
  if(turnOn) {
    boardLEDSet(BOARDLED0);
  } else {
    boardLEDClear(BOARDLED0);
  }  
}

#ifdef BOARD0220
void connLEDInit(void)
{
  int8u i;
  
  // Initialize the LEDs ports for output 
  DDRA = 0xFF;
  DDRC |= BIT(2) | BIT(3);
  
  // turn them all off
  for (i=0; i<=CONNLED_MAX; i++)
    connLEDClear(i);
}

void connLEDClear(int8u led)
{
  if (led<=7)
    PORTA |= BIT(led);
  if (led>=8 && led<=9)
    PORTC |= BIT(led-8+2);
}

void connLEDSet(int8u led)
{
  if (led<=7)
    PORTA &= ~BIT(led);
  if (led>=8 && led<=9)
    PORTC &= ~BIT(led-8+2);
}
#else // other board
void connLEDClear(int8u led) {}
void connLEDSet(int8u led) {}
#endif // BOARD0220

// Tune state
int8u PGM *currentTune = NULL;
int8u tunePos = 0;
int16u currentDuration = 0;
boolean tuneDone=TRUE;

void boardPlayTune_P(int8u PGM *tune, boolean bkg)
{
  TCCR2 = BIT(WGM21) | BIT(COM20); // no clock (timer stopped)
  TIMSK &= ~(BIT(OCIE2));  // No interrupts
  currentTune = tune; 
  OCR2 = tune[0];
  // Magic duration calculation based on frequency
  #ifdef AVR8MHZ
    currentDuration = ((int16u)200*(int16u)tune[1])/(tune[0]/15);
  #else //AVR4MHZ
    currentDuration = ((int16u)200*(int16u)tune[1])/(tune[0]/28);
  #endif
  tunePos = 2;  // First note is already set up
  tuneDone = FALSE;
  ATOMIC(
    TIMSK |= BIT(OCIE2);  // start the interrupts comin'!
    #ifdef AVR8MHZ
      TCCR2 = BIT(WGM21) | BIT(COM20) | BIT(CS22); // Enable Timer clk/256
    #else //AVR4MHZ
      TCCR2 = BIT(WGM21) | BIT(COM20) | BIT(CS21) | BIT(CS20); // Enable Timer clk/64
    #endif
  )
  while( (!bkg) && (!tuneDone)) {
    __watchdog_reset();
  }
}

#pragma vector = TIMER2_COMP_vect
__interrupt void tuneDelayInterrupt(void)
{
  if(currentDuration-- == 0) {
    if(currentTune[tunePos+1]) {
      if(currentTune[tunePos]) {
        // note
        OCR2 = currentTune[tunePos];
        #ifdef AVR8MHZ
          TCCR2 = BIT(WGM21) | BIT(COM20) | BIT(CS22); // Enable Timer clk/256
          // work some magic to determine the duration based upon the frequency
          //  of the note we are currently playing.
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(currentTune[tunePos]/15);
        #else //AVR4MHZ
          TCCR2 = BIT(WGM21) | BIT(COM20) | BIT(CS21) | BIT(CS20); // Enable Timer clk/64
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(currentTune[tunePos]/28);
        #endif
      } else {
        // pause
        OCR2 = 0xFF;
        #ifdef AVR8MHZ
          TCCR2 = BIT(WGM21) | BIT(CS22); // Timer clk/256, no pin toggle
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(0xFF/15);
        #else // AVR4MHZ
          TCCR2 = BIT(WGM21) | BIT(CS21) | BIT(CS20); // Timer clk/64, no pin toggle
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(0xFF/28);
        #endif
      }
      tunePos += 2;
    } else {
      // End tune
      TCCR2 = BIT(WGM21) | BIT(COM20); // no clock (timer stopped)
      TIMSK &= ~(BIT(OCIE2));
      tuneDone = TRUE;
    }
  }
}

int8u PGM hereIamTune[] = {
  NOTE_B4,  1,
  0,        1,
  NOTE_B4,  1,
  0,        1,
  NOTE_B4,  1,
  0,        1,
  NOTE_B5,  5,
  0,        0
};

void halIndicatePresence(void)
{
  boardPlayTune_P(hereIamTune,TRUE);
}

