/** @file hal/board/dev/board.h 
 *  @brief Functions specific to particular carrier boards
 *
 * Currently this header can be used for all combinations of
 * Developer Kit carrier boards and RF communication modules.
 * 
 * There are two types of carrier board:
 * - \b BOARD0221: For use with the Ember RF communication module.
 * - \b BOARD0220: Has an integrated microcontroller, RF-only module.
 * 
 * For the most part, the two boards are electrically compatible.
 *
 * NOTE: Ember uses the term SPD (Start Packet Delimiter) instead of the term 
 *       SFD (Start Frame Delimiter).
 * 
 * <!-- Author(s): Lee Taylor, lee@ember.com, Jeff Mathews, jm@ember.com -->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */

#ifndef __DEV_BOARD_H__
#define __DEV_BOARD_H__

//#ifndef __BOARD_H__
//  #error do not include this file directly - include board/board.h
//#endif

// By default, the BOARD0221 board is assumed.
#if !defined(BOARD0221) && !defined(BOARD0220)
  #define BOARD0221
#endif

// By default, the microcontroller runs at 8MHz.
#if !defined(AVR8MHZ) && !defined(AVR4MHZ)
  #if defined(RADIO_CC1020)
    #define AVR4MHZ
  #elif defined(RADIO_EM2420)
    #define AVR8MHZ
  #endif
#endif


/** @name Radio Connection Items 
 *
 * The following are used to aid in the abstraction with 
 * the radio HAL implementations.  The microcontroller-specific radio 
 * HAL sources use these definitions to work across a variety of boards,
 * each of which have different radio connections.  The names
 * and ports/pins used below are intended to match with a schematic of
 * the system to provide the abstraction.
 * 
 * <b>CC1020 Radio Connection Items</b>
 * - \b radPORT: PORTD
 * - \b radPIN:  PIND 
 * - \b radDDR:  DDRD
 * - \b radPSEL: 7 
 * - For carrier board type 0221:
 *   - \b radPORT_PCLK:  PORTE
 *   - \b radDDR_PCLK:   DDRE
 *   - \b radPCLK:       5
 * - For carrier board type 0220:
 *   - \b radPORT_PCLK:  PORTD
 *   - \b radDDR_PCLK:   DDRD
 *   - \b radPCLK:       6
 * - If no board is defined:
 *   - \b radPDI:        5
 *   - \b radPDO:        4
 *   - \b radDDR_Lock:   DDRE
 *   - \b radPIN_Lock:   PINE
 *   - \b radLockBIT:    4
 *
 * <b>EM2420 Radio Connection Items</b>
 * 
 * - \b radPORT:      PORTB
 * - \b radPIN:       PINB
 * - \b radDDR:       DDRB
 * - \b radPSEL:      0
 * - \b radPCLK:      1
 * - \b radPDI:       2 // MOSI
 * - \b radPDO:       3 // MISO
 * - \b radVregPORT:  PORTB
 * - \b radVregDDR:   DDRB
 * - \b radVreg:      5
 * - \b radRESETPORT: PORTD
 * - \b radRESETPIN:  PIND
 * - \b radDDRD:      DDRD
 * - \b radnReset:    7
 * - \b radFifoPORT:  PORTE
 * - \b radFifoPIN:   PINE
 * - \b radFifoDDR:   DDRE
 * - \b radFIFOP:     4
 * - \b radFIFO:      5
 * - \b radCCAPIN:    PIND
 * - \b radCCA:       5
 * - \b radSPDPIN:    PIND
 * - \b radSPD:       4
 * - \b radPORTFIFOP: PORTE
 * - \b radPINFIFOP:  PINE
 * - \b radDDRFIFOP:  DDRE
 * - \b radFIFOP:     4
 * - \b radPORTFIFO:  PORTE
 * - \b radPINFIFO:   PINE
 * - \b radDDRFIFO:   DDRE
 * - \b radFIFO:      5
*/
//@{

#if defined(RADIO_CC1020) || defined(DOXYGEN_SHOULD_SKIP_THIS)

  #define radPORT PORTD
  #define radPIN  PIND
  #define radDDR  DDRD
  #define radPSEL 7
  #if defined(BOARD0221)
    #define radPORT_PCLK  PORTE
    #define radDDR_PCLK   DDRE
    #define radPCLK       5
  #elif defined(BOARD0220)
    #define radPORT_PCLK  PORTD
    #define radDDR_PCLK   DDRD
    #define radPCLK       6
  #else
    #error NO BOARD DEFINED
  #endif
  #define radPDI  5
  #define radPDO  4
  #define radDDR_Lock  DDRE
  #define radPIN_Lock  PINE
  #define radLockBIT 4

#endif
#if defined(RADIO_EM2420) || defined(DOXYGEN_SHOULD_SKIP_THIS)

  #define radPORT PORTB
  #define radPIN  PINB
  #define radDDR  DDRB
  #define radPSEL 0
  #define radPCLK 1
  #define radPDI  2 // MOSI
  #define radPDO  3 // MISO

  #define radVregPORT PORTB
  #define radVregDDR  DDRB
  #define radVreg     5

  #define radRESETPORT PORTD
  #define radRESETPIN  PIND
  #define radDDRD      DDRD
  #define radnReset     7

  #define radFifoPORT PORTE
  #define radFifoPIN  PINE
  #define radFifoDDR  DDRE
  #define radFIFOP    4
  #define radFIFO     5

  #define radCCAPIN       PIND
  #define radCCA          5
  #define radSPDPIN       PIND
  #define radSPD          4

  #define radPORTFIFOP PORTE
  #define radPINFIFOP  PINE
  #define radDDRFIFOP  DDRE
  #define radFIFOP     4

  #define radPORTFIFO PORTE
  #define radPINFIFO  PINE
  #define radDDRFIFO  DDRE
  #define radFIFO     5

#endif
//@} //END OF RADIO CONNECTION ITEMS


//////////////////////////////////////////////////////////////////////
#if !defined(__IAR_SYSTEMS_ASM__)
//////////////////////////////////////////////////////////////////////


/** @name LED Definitions
 *
 * These macros should be used for changing the state of LEDs. \b 
 * Note: BOARD0221 has six total LEDs, and BOARD0220 has four.
 */
//@{
#define BOARDLED0 0
#define BOARDLED1 1
#define BOARDLED2 2
#define BOARDLED3 3
#define BOARDLED4 4
#define BOARDLED5 5
#ifdef BOARD0221
  #define BOARDLED_MAX 5
#else
  #define BOARDLED_MAX 3 
#endif
//@} //END OF LED DEFINITIONS

/** @description Enables the temperature sensor. This function is automatically
 * called by boardBoot().
 */
void boardEnableTempSensor(void);

/** @description Disables the temperature sensor.*/
void boardDisableTempSensor(void);

// Use the following definitions with the button APIs.
#define BUTTON_1 1
#define BUTTON_2 2
#define BUTTON_PRESSED  1
#define BUTTON_RELEASED 0


/** @description Initializes the buttons. This function is automatically called
 * by boardBoot().
 */
void boardButtonInit(void);

/** @description Returns the current state (pressed or released) of a button.
 *   
 * @param button: The button being queried.
 */
int8u boardButtonState(int8u button);

/** @description A callback called in interrupt context whenever a button
 * changes its state.
 * 
 * @appusage Must be implemented by the application. 
 *  
 * @param button: A button.
 *  
 * @param state: The state the button is in.
 */
void boardButtonCallback(int8u button, int8u state);

/** @description Plays a tune on the piezo buzzer. The tune is played in the 
 * background if \c bkg is TRUE. Otherwise, the API blocks until
 * the playback of the tune is complete.
 *   
 * A tune is implemented as follows:
 * <pre> 
 * int8u PGM hereIamTune[] = {    //All tunes are stored in flash.
 *    NOTE_B4,  1,                //Plays the note B4 for 100 milliseconds.
 *    0,        1,                //Pause for 100 milliseconds.
 *    NOTE_B5,  5,                //Plays the note B5 for 500 milliseconds.
 *    0,        0                 //NULL terminates the tune.
 *  }; 
 * </pre>
 * 
 * @param tune: A pointer to tune to play. 
 *  
 * @param bkg: Determines whether the tune plays in the background.
 * If TRUE, tune plays in background; if FALSE, tune plays in foreground.
 */
void boardPlayTune_P(int8u PGM *tune, boolean bkg);

/** @name Connectivity LED Functions
 *
 * BOARD0220 has a bank of 10 LEDs that are controlled by these functions.
 */
//@{
#define CONNLED0 0
#define CONNLED1 1
#define CONNLED2 2
#define CONNLED3 3
#define CONNLED4 4
#define CONNLED5 5
#define CONNLED6 6
#define CONNLED7 7
#define CONNLED8 8
#define CONNLED9 9
#define CONNLED_MAX 9

/** @description Initializes and clears the LEDs. This function is automatically called
 * by boardBoot().
 */
void connLEDInit(void);

/** @description Turns on a specific LED.
 *   
 * @param led: The LED to turn on.
 */
void connLEDSet(int8u led);

/** @description Turns off a specific LED.
 *   
 * @param led: The LED to turn off.
 */
void connLEDClear(int8u led);

//@}  end of Example/Prototype LED Functions


/** @name Radio Connection Functions
 *     
 * These macros aid in board abstraction from the various radios available for
 * use with the board. The microcontroller-specific radio HAL sources use these
 * definitions to work across a variety of boards with differeing radio connections.
 */
//@{


// A few serial port definitions for ember internal targets
#if defined(HALTEST)
  extern int8u serialPort;
  #define SER232 serialPort
  #define SIGPORT 0   // serial port for signal generator lqi testing
#else
  #define EVAL_KIT_SERIAL_PORT 1
  #define EVAL_KIT_DEBUG_PORT 0
#endif

//////////////////////////////////////////////////////////////////////
#endif  // __IAR_SYSTEMS_ASM__

//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//@}  end of Radio Connection Functions
////////////////////////////////////////////////////////////////////////////////

#endif //__DEV_BOARD_H__
