/*
 * File: board.c
 * Description: HAL sources for an ATmega16 on a 0221 Developer Kit carrier board
 *
 * Author(s): Brooks Barrett, brooks.barrett@ember.com
 *            Lee Taylor, lee@ember.com,
 *            Jeff Mathews, jm@ember.com
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#define ENABLE_BIT_DEFINITIONS
#include <ioavr.h>

#include PLATFORM_HEADER
#include "stack/include/ember.h"

#include "app/util/serial/serial.h"
#include "micro/adc.h"

#include "stack/include/hal.h"
#include "board/board.h"
#include "micro/micro.h"


void boardBoot(void)
{
  boardLEDInit();
  adcInit();
  boardEnableTempSensor();

  // Enable the RS-232 tranceiver
  // always enabled on dev kit
  
  // Accelerometer
  // none on dev kit
  
  // Buzzer
  DDRB |= BIT(3);

  boardButtonInit();
}

void boardReboot(void)
{
  microReboot();
}

void boardPowerDown(void) {
  // Temp sensor
  boardDisableTempSensor();
  
  // RS232
  // n/a
  
  // Accelerometer
  // n /a
  
  // Buzzer
  TCCR0 = BIT(WGM01) | BIT(COM00); // no clock (timer stopped)
  TIMSK &= ~(BIT(OCIE0)); // disable interrupt
  
  // Adc
  adcSleep();
  
  // Uarts
  UCSRB &= ~(BIT(RXCIE) | BIT(TXCIE) | BIT(UDRIE) | BIT(RXEN) | BIT(TXEN));

  // Leds
  boardLEDInit(); // Turns leds off
}

void boardPowerUp(void) {
  // Temp sensor
  boardEnableTempSensor();
  // RS232
  // n/a
  // Accelerometer
  // n/a
  // Buzzer
  // Don't need to do anything
  // Adc
  adcInit();
  // Uarts
  // emberSerialInit must be called after boardPowerUp to turn the uarts back on
  // Leds
  // Don't need to do anything
}

void boardEnableTempSensor(void) {
  DDRA |= BIT(3);
  PORTA |= BIT(3);
}

void boardDisableTempSensor(void) {
  PORTA &= ~BIT(3);
}

//This function does not get used with Leaf Nodes.  The dev16 board files are
//designed for the ATmega16, which is intended to run Leafs.  Leave this function
//here for the off-chance it ends up getting used.  Compiler will strip it out
//if it's not used so we dont need to worry about code size.
void boardAssertFailed(PGM_P filename, int linenumber)
{
  __watchdog_reset();              // In case we're close to running out.
#if DEBUG_ON
  emberDebugAssert(filename, linenumber);
  __watchdog_reset();              // In case we're close to running out.
  emberDebugMemoryDump((int8u *)0, (int8u *)4095);  // entire ram
#else
  __watchdog_reset();              // In case we're close to running out.
  emberSerialGuaranteedPrintf(1, "\n\r[ASSERT:%p:%d:", filename, linenumber);
  //stack trace:
  {
    int8u i;
    for (i=0; i<6; i++) {
      int8u l,h;
      int16u pc;
      h = *((int8u*)(SP+2*i+1));
      l = *((int8u*)(SP+2*i+2));
      pc = HIGH_LOW_TO_INT(h,l);
      pc = pc << 1;
      emberSerialGuaranteedPrintf(1, " %2x", pc);
    }
  }
  emberSerialGuaranteedPrintf(1, "]\n\r");
#endif
  microReboot();                        
}

void boardButtonInit(void)
{
  // Enable internal Pull-ups
  PORTB |= BIT(0);  //BUTTON1
  PORTC |= BIT(1);  //BUTTON2
  
  //MEGA128 and MEGA64 on dev boards have the buttons tied to interrupts.
  //16L doesn't have any free interrupts for buttons.  This is where we
  //used to enable button interrupts, nothing to do for the 16L though.  The
  //button states are now polling only.
}

int8u boardButtonState(int8u button)
{
  // Note: Unlike a MEGA128 or MEGA64 on the dev board, this returns the port
  // value since we dont have buttons on interrupts.
  switch(button) {
    case BUTTON_1:
      if( (PINB & BIT(0)) == BIT(0) ) {
        return BUTTON_RELEASED;
      } else {
        return BUTTON_PRESSED;
      }
    //always returning, don't need a break
    case BUTTON_2:
      if( (PINC & BIT(1)) == BIT(1)) {
        return BUTTON_RELEASED;      
      } else {
        return BUTTON_PRESSED;      
      }
    //always returning, don't need a break
    default:
      return BUTTON_RELEASED;
  }
}

void boardLEDInit(void)
{
  int8u i;

  // Initialize the LEDs ports for output 
  DDRA |= BIT(4) | BIT(5) | BIT(6) | BIT(7) ;

  // turn the LEDs off
  for(i=0; i<=BOARDLED_MAX; i++)
    boardLEDClear(i);
}


void boardLEDToggle (int8u led)
{
  // c lanuguage xor on an i/o port doesn't make avr machine code 
  // which is atomic. so to avoid contention with other code using
  // the other pins for other purposes, we disable interrupts to
  // make the read io register, xor and write io atomic.
  ATOMIC(
    PORTA ^= BIT(led + 4);
  )
}
  
void boardLEDSet (int8u led)
{
  // this should compile to an atomic assembly instruction  
  PORTA &= ~BIT(led + 4);
}

void boardLEDClear (int8u led)
{
  // this should compile to an atomic assembly instruction  
  PORTA |= BIT(led + 4);
}

void halIndicateActivity(boolean turnOn)
{
  if(turnOn) {
    boardLEDSet(BOARDLED0);
  } else {
    boardLEDClear(BOARDLED0);
  }  
}

// Tune state
int8u PGM *currentTune = NULL;
int8u tunePos = 0;
int16u currentDuration = 0;
boolean tuneDone=TRUE;

void boardPlayTune_P(int8u PGM *tune, boolean bkg)
{
  TCCR0 = BIT(WGM01) | BIT(COM00); // no clock (timer stopped)
  TIMSK &= ~(BIT(OCIE0));  // No interrupts
  currentTune = tune; 
  OCR0 = tune[0];
  // Magic duration calculation based on frequency
  #ifdef AVR8MHZ
    currentDuration = ((int16u)200*(int16u)tune[1])/(tune[0]/15);
  #else //AVR4MHZ
    currentDuration = ((int16u)200*(int16u)tune[1])/(tune[0]/28);
  #endif
  tunePos = 2;  // First note is already set up
  tuneDone = FALSE;
  ATOMIC(
    TIMSK |= BIT(OCIE0);  // start the interrupts comin'!
    #ifdef AVR8MHZ
      TCCR0 = BIT(WGM01) | BIT(COM00) | BIT(CS02); // Enable Timer clk/256
    #else //AVR4MHZ
      TCCR0 = BIT(WGM01) | BIT(COM00) | BIT(CS01) | BIT(CS00); // Enable Timer clk/64
    #endif
  )
  while( (!bkg) && (!tuneDone)) {
    __watchdog_reset();
  }
}

#pragma vector = TIMER0_COMP_vect
__interrupt void tuneDelayInterrupt(void)
{
  if(currentDuration-- == 0) {
    if(currentTune[tunePos+1]) {
      if(currentTune[tunePos]) {
        // note
        OCR0 = currentTune[tunePos];
        #ifdef AVR8MHZ
          TCCR0 = BIT(WGM01) | BIT(COM00) | BIT(CS02); // Enable Timer clk/256
          // work some magic to determine the duration based upon the frequency
          //  of the note we are currently playing.
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(currentTune[tunePos]/15);
        #else //AVR4MHZ
          TCCR0 = BIT(WGM01) | BIT(COM00) | BIT(CS01) | BIT(CS00); // Enable Timer clk/64
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(currentTune[tunePos]/28);
        #endif
      } else {
        // pause
        OCR0 = 0xFF;
        #ifdef AVR8MHZ
          TCCR0 = BIT(WGM01) | BIT(CS02); // Timer clk/256, no pin toggle
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(0xFF/15);
        #else // AVR4MHZ
          TCCR0 = BIT(WGM01) | BIT(CS01) | BIT(CS00); // Timer clk/64, no pin toggle
          currentDuration = ((int16u)200*(int16u)currentTune[tunePos+1])/(0xFF/28);
        #endif
      }
      tunePos += 2;
    } else {
      // End tune
      TCCR0 = BIT(WGM01) | BIT(COM00); // no clock (timer stopped)
      TIMSK &= ~(BIT(OCIE0));
      tuneDone = TRUE;
    }
  }
}

int8u PGM hereIamTune[] = {
  NOTE_B4,  1,
  0,        1,
  NOTE_B4,  1,
  0,        1,
  NOTE_B4,  1,
  0,        1,
  NOTE_B5,  5,
  0,        0
};

void halIndicatePresence(void)
{
  boardPlayTune_P(hereIamTune,TRUE);
}

