/** @file hal/board/dev16/board.h 
 *  @brief Functions specific to an ATmega16 on a 0221 Developer Kit carrier board
 *
 * NOTE: Ember uses the term SPD (Start Packet Delimiter) instead of the term 
 *       SFD (Start Frame Delimiter).
 * 
 * <!-- Author(s): Brooks Barrett, brooks.barrett@ember.com
 *                 Lee Taylor, lee@ember.com, 
 *                 Jeff Mathews, jm@ember.com -->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */

#ifndef __DEV_BOARD_H__
#define __DEV_BOARD_H__

#ifndef __BOARD_H__
  #error do not include this file directly - include board/board.h
#endif

/**@description Only the BOARD0221 board can be used with the ATmega16. */
#define BOARD0221

// By default, the microcontroller runs at 8MHz.
#if !defined(AVR8MHZ) && !defined(AVR4MHZ)
  #if defined(RADIO_CC1020)
    #define AVR4MHZ
  #elif defined(RADIO_EM2420)
    #define AVR8MHZ
  #endif
#endif


/** @name Radio Connection Items 
 *
 * The following are used to aid in the abstraction with 
 * the radio HAL implementations. The microcontroller-specific radio 
 * HAL sources use these definitions to work across a variety of boards,
 * each of which have different radio connections. The names
 * and ports/pins used below are intended to match with a schematic of
 * the system to provide the abstraction.
 * 
 * <b>EM2420 Radio Connection Items</b>
 * 
 * - \b radPORT: PORTB
 * - \b radPIN:  PINB
 * - \b radDDR:  DDRB
 * - \b radPSEL: 4
 * - \b radPCLK: 7
 * - \b radPDI:  5 // MOSI
 * - \b radPDO:  6 // MISO
 * - \b radVregPORT: PORTC
 * - \b radVregDDR:  DDRC
 * - \b radVreg:  0
 * - \b radRESETPORT:  PORTD
 * - \b radRESETPIN:  PIND
 * - \b radDDRD:  DDRD
 * - \b radnReset:  7
 * - \b radCCAPIN:  PIND
 * - \b radCCA:  5
 * - \b radSPDPIN:  PIND
 * - \b radSPD:  6
 * - \b radPORTFIFOP:  PORTD
 * - \b radPINFIFOP:  PIND
 * - \b radDDRFIFOP:  DDRD
 * - \b radFIFOP:  2
 * - \b radPORTFIFO:  PORTD
 * - \b radPINFIFO:  PIND
 * - \b radDDRFIFO:  DDRD
 * - \b radFIFO:  4
*/
//@{

#if defined(RADIO_EM2420) || defined(DOXYGEN_SHOULD_SKIP_THIS)

  #define radPORT PORTB
  #define radPIN  PINB
  #define radDDR  DDRB
  #define radPSEL 4
  #define radPCLK 7
  #define radPDI  5 // MOSI
  #define radPDO  6 // MISO

  #define radVregPORT PORTC
  #define radVregDDR  DDRC
  #define radVreg     0

  #define radRESETPORT PORTD
  #define radRESETPIN  PIND
  #define radDDRD      DDRD
  #define radnReset     7

  #define radCCAPIN       PIND
  #define radCCA          5

  #define radSPDPIN       PIND
  #define radSPD          6

  #define radPORTFIFOP PORTD
  #define radPINFIFOP  PIND
  #define radDDRFIFOP  DDRD
  #define radFIFOP     2

  #define radPORTFIFO PORTD
  #define radPINFIFO  PIND
  #define radDDRFIFO  DDRD
  #define radFIFO     4

#endif
//@} //END OF RADIO CONNECTION ITEMS


//////////////////////////////////////////////////////////////////////
#if !defined(__IAR_SYSTEMS_ASM__)
//////////////////////////////////////////////////////////////////////


/** @name LED Definitions
 *
 * Use these macros to change the state of LEDs. \b 
 * Note: BOARD0221 has six LEDs, but the ATmega16 only connects to four LEDs.
 */
//@{
#define BOARDLED0 0
#define BOARDLED1 1
#define BOARDLED2 2
#define BOARDLED3 3
#define BOARDLED_MAX 3
//@} //END OF LED DEFINITIONS

/** @description Enables the temperature sensor. This function is automatically
 * called by boardBoot().
 */
void boardEnableTempSensor(void);

/** @description Disables the temperature sensor.*/
void boardDisableTempSensor(void);

// Use the following definitions with the button APIs.
#define BUTTON_1 1
#define BUTTON_2 2
#define BUTTON_PRESSED  1
#define BUTTON_RELEASED 0

/** @description Initializes the buttons. This function is automatically called
 * by boardBoot().
 */
void boardButtonInit(void);

/** @description Returns the current state (pressed or released) of a button.
 *   
 * @param button: The button being queried.
 */
int8u boardButtonState(int8u button);

/** @description Plays a tune on the piezo buzzer. The tune is played in the 
 * background if \c bkg is TRUE. Otherwise, the API blocks until
 * the playback of the tune is complete.
 *   
 * A tune is implemented as follows:
 * <pre> 
 * int8u PGM hereIamTune[] = {    //All tunes are stored in flash.
 *    NOTE_B4,  1,                //Plays the note B4 for 100 milliseconds.
 *    0,        1,                //Pause for 100 milliseconds.
 *    NOTE_B5,  5,                //Plays the note B5 for 500 milliseconds.
 *    0,        0                 //NULL terminates the tune.
 *  }; 
 * </pre>
 * 
 * @param tune: A pointer to tune to play. 
 *  
 * @param bkg: Determines whether the tune plays in the background.
 * If TRUE, tune plays in background; if FALSE, tune plays in foreground.
 */
void boardPlayTune_P(int8u PGM *tune, boolean bkg);


/** @name Radio Connection Functions
 *     
 * These macros aid in board abstraction from the various radios available for
 * use with the board. The microcontroller-specific radio HAL sources use these
 * definitions to work across a variety of boards with differeing radio connections.
 */
//@{


// A few serial port definitions for ember internal targets
#if defined(HALTEST)
  extern int8u serialPort;
  #define SER232 serialPort
  #define SIGPORT 0   // serial port for signal generator lqi testing
#else
  #define EVAL_KIT_SERIAL_PORT 1
  #define EVAL_KIT_DEBUG_PORT 0
#endif

//////////////////////////////////////////////////////////////////////
#endif  // __IAR_SYSTEMS_ASM__

//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//@}  end of Radio Connection Functions
////////////////////////////////////////////////////////////////////////////////

#endif //__DEV_BOARD_H__
