/** @file /hal/micro/adc.h
 * @brief A/D converter driver, for multiple users
 * <!-- Author(s): Andy Wheeler, andy@ember.com -->
 *
 * Unless otherwise noted, the EmberNet stack does not use these functions, 
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 *
 * To use the ADC system, include this file and ensure that
 * adcInit() is called whenever the microcontroller is started. Call
 * adcSleep() to sleep the module and adcInit() to wake up the module.
 *
 * A "user" is a separate thread of execution and usage. That is, 
 * internal Ember code is one user and clients are a different
 * user. But a client that is calling the ADC in two different functions
 * constitutes only one user, as long as the ADC access is not interleaved.
 *
 * \b Note: Global interrupts must be enabled for the ADC system to work.
 *
 * This code does not allow:
 * - Access to the continuous reading mode of the ADC, which some 
 * clients may require. 
 * - Users to switch the reference (but obviously this is easy to change).
 *
 * Many functions in this file return an \c EmberStatus value. See 
 * error-def.h for definitions of all \c EmberStatus return values.
 *
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
*/

#ifndef __ADC_H__
#define __ADC_H__
/** @description ADC functions employ a user ID to keep different users separate.
 * Avoid many users because each user requires some amount of state storage.
 */
typedef enum
{
  ADC_USER_LQI = 0,
  ADC_USER_APP = 1
} ADCUser;

/** @description Be sure to update if additional users are added to the 
 * ADCUser list.
 */
#define NUM_ADC_USERS 2 // make sure to update if the above is adjusted

/** @description ADC references. */
typedef enum
{
  ADC_REF_AREF = 0x00,
  ADC_REF_AVCC = 0x40,
  ADC_REF_INT  = 0xC0
} ADCReferenceTypes;

/** @description Initializes and powers up the ADC. Should also be called
 * to wake from sleep.
 */
void adcInit(void);

/** @description Shuts down the voltage reference and ADC system to minimize
 * power consumption in sleep.
 */
void adcSleep(void);

/** @description Starts an ADC conversion for the user specified by \c id.
 * 
 * @appusage The application must set \c reference to the voltage reference 
 * desired (see the ADC references enum) and set \c channel to the channel 
 * number required. 
 *
 * @param id: An ADC user.
 * 
 * @param reference: Voltage reference to use, chosen from 
 * ADCReferenceTypes.
 * 
 * @param channel: Microprocessor channel number. For differential and amped 
 * channel numbers,see the ATMega128 datasheet. For basic single-ended channels, 
 * use 0 to 7.
 * 
 * @return EMBER_ADC_CONVERSION_DEFERRED or EMBER_ADC_CONVERSION_BUSY
 */
EmberStatus adcStartConversion(ADCUser id, int8u reference, int8u channel);

/** @description Returns the status of a pending conversion previously started 
 * by adcStartConversion(). If the conversion is complete, writes the value
 * of the conversion into \c value. 
 *
 * @param id: An ADC user.
 *
 * @param value: Pointer to an int8u to be loaded with the new value.
 * 
 * @return One of the following: 
 * - EMBER_ADC_CONVERSION_DONE if the conversion is complete.
 * - EMBER_ADC_CONVERSION_DEFERRED if the conversion is still waiting 
 * to start.
 * - EMBER_ADC_CONVERSION_BUSY if the conversion is currently taking 
 * place.
 * - EMBER_ADC_NO_CONVERSION_PENDING if \c id does not have a pending 
 * conversion.
 */
EmberStatus adcRequestData(ADCUser id, int8u *value);

/** @description Waits for the user's request to complete and then, if a
 * conversion was done, writes its value into \c value and returns
 * EMBER_ADC_CONVERSION_DONE, or immediately returns 
 * EMBER_ADC_NO_CONVERSION_PENDING.
 *
 * @param id: An ADC user.
 *
 * @param value: Pointer to an int8u to be loaded with the new value.
 * 
 * @return One of the following: 
 * - EMBER_ADC_CONVERSION_DONE if the conversion is complete.
 * - EMBER_ADC_NO_CONVERSION_PENDING  if \c id does not have a pending 
 * conversion.
 */
EmberStatus adcRequestDataBlocking(ADCUser id, int8u *value);

/** @description Operates identically to adcRequestDataBlocking() 
 * except that it works when global interrupts are disabled.
 *
 * @appusage Use this function <b>ONLY IF INTERRUPTS ARE DISABLED</b>. Its
 * operation is not defined otherwise.
 *
 * @param id: An ADC user.
 *
 * @param value: Pointer to an int8u to be loaded with the new value.
 * 
 * @return One of the following: 
 * - EMBER_ADC_CONVERSION_DONE if the conversion is complete.
 * - EMBER_ADC_NO_CONVERSION_PENDING  if \c id does not have a pending
 * conversion.
 */
EmberStatus adcRequestDataBlockingNoInt(ADCUser id, int8u *value);

#endif // __ADC_H__


