/** @file bootloader.h
 * @brief Bootloader functions
 *
 * Unless otherwise noted, the EmberNet stack does not use these functions, 
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 *
 * <b> About the Bootloader Trap Entry Points</b>
 * These functions should never be called directly from an application. 
 * Instead, microBootloaderTrap() should be called with the associated trap
 * number and the corresponding parameters from these prototypes:
 * - flashWriteBlock()
 * - flashAdjustBlock()
 *
 * <!-- Author(s): Jeff Mathews, jm@ember.com -->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */

#ifndef __BOOTLOADER_H__
#define __BOOTLOADER_H__

/** @description Erases and then writes a block of flash with new data. 
 *   
 * @note This function can only be used to write to one flash block at a time. 
 * If \c dst points to an address inside the block, make sure that \c dst plus
 * \c size does not exceed the boundaries of the block.
 *  
 * @param dst: Byte address pointer. Must be an even address and within the first 
 * 64k bytes of flash address space. Usually, use to point to the beginning of a
 * flash block. 
 *     
 * @param src: Byte address pointer. Must be within the first 64k bytes of 
 * flash address space.  
 * 
 * @param size: Size of block, in words. Normally set to 128 or one full block.
 * Remaining data in that block is erased.
 */
void flashWriteBlock(int16u* dst, int16u* src, int8u size);

/** @description Modifies a block of flash without losing existing data.
 *   
 * @note \c dst and \c size may be set so that they span multiple pages. Any
 * remaining portion of those pages will remain unchanged other than having 
 * been erased and rewritten.
 * 
 * @param dst: Byte address pointer. Must be an even address and within the first 
 * 64k bytes of flash address space. Usually use to point to the beginning of a
 * flash block.
 *     
 * @param src: Byte address pointer. Must be within the first 64k bytes of 
 * flash address space. 
 *     
 * @param size: Size of block, in words. Must be 0 to 128.
 */
void flashAdjustBlock(int16u* dst, int16u* src, int8u size);

#ifndef DOXYGEN_SHOULD_SKIP_THIS

/** @description Launches the bootloader.
 * @note This function is \b DEPRECATED. 
 */
void bootloadAppStart(void); 
#endif // DOXYGEN_SHOULD_SKIP_THIS


#ifndef DOXYGEN_SHOULD_SKIP_THIS

// Automatically used by the trap functions for triggering the bootloader
#define BOOTLOADER_APP_REQUEST_KEY 0xFA27

#endif // DOXYGEN_SHOULD_SKIP_THIS

#endif // __BOOTLOADER_H__

