/*
 * File: micro.c
 * Description: AVR Atmega128 micro specific HAL functions
 *
 * Author(s): Lee Taylor, lee@ember.com,
 *            Jeff Mathews, jm@ember.com
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */

#define ENABLE_BIT_DEFINITIONS
#include <ioavr.h>

#include <stdarg.h>

#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "include/error.h"

#include "stack/include/hal.h"
#include "board/board.h"
#include "micro/micro.h"
#include "micro/avr-atmega/128/bootloader/bootloader.h"
#include "app/util/serial/serial.h"

// battery lo level set to 1.0V on ADC, which is 2.0V on the battery line.
#define BATTERY_LEVEL_LO  0x60

#ifndef EMBER_APPLICATION_HAS_MICRO_CUSTOM_SYSTEM_TIMER
static int32u systemTick;
#endif
#ifdef SIXTEEN_BIT_BEACONS
static int16u syncTimeOffset;
#else //SIXTEEN_BIT_BEACONS
static int16u syncTick;
static int32u syncTimeOffset;
static boolean syncCaptured;
static int32u syncCapturedTime;

static boolean symbolDelayBPending = FALSE;
static int16u symbolDelayBHigh;
static int16u symbolDelayBLow;
#endif //SIXTEEN_BIT_BEACONS

//I decided to invent my own units of measure:
//Hick = hardware tick
//Mick = MAC tick = symbol
//Sick = System tick
//Depending on the MHz of system, we have different Hicks.
//Depending on the type of radio, we have different Micks.
//We need consistant Micks/symbols for the MAC to run off of.
//Define all the possible conversions for radio/speed types up here.
//Use the pair of defines below to make the code easier to read.
//  convertSymbolsToHicks(x)
//  convertHicksToSymbols(x)
// -WBB
#if defined(RADIO_EM2420)
  #if defined(AVR3p68MHZ) || defined(AVR4MHZ)
    #define convertSymbolsToHicks(x) x = x
    #define convertHicksToSymbols(x) x = x
  #elif defined(AVR7p3278MHZ) || defined(AVR8MHZ)
    #define convertSymbolsToHicks(x) x = x << 1
    #define convertHicksToSymbols(x) x = x >> 1
  #else
    #error MCU system clock must be defined!
  #endif
#elif defined(RADIO_CC1020)
  #if defined(AVR3p68MHZ) || defined(AVR4MHZ)
    #define convertSymbolsToHicks(x) x = x
    #define convertHicksToSymbols(x) x = x
  #elif defined(AVR7p3278MHZ) || defined(AVR8MHZ)
    #define convertSymbolsToHicks(x) x = x << 1
    #define convertHicksToSymbols(x) x = x >> 1
  #else
    #error MCU system clock must be defined!
  #endif
#else
  #error Radio type must be defined!
#endif

// reserved ram is special, see micro.h
__no_init __root int16u bootloaderApplicationRequestKey @ RESERVED_RAM_BOOTLOADER_KEY;
__no_init __root int16u micro_last_known_pc   @ RESERVED_RAM_PC_DIAG;
__no_init __root int8u  gatewayID[8]          @ RESERVED_RAM_GATEWAY_ID;
__no_init __root int8u  eui64[8]              @ RESERVED_RAM_EUI64_ID;
__no_init __root int16u panID                 @ RESERVED_RAM_PAN_ID;
__no_init __root int16u nodeID                @ RESERVED_RAM_NODE_ID;
__no_init __root int8u  radioChannel          @ RESERVED_RAM_CHANNEL;
__no_init __root int8u  radioPower            @ RESERVED_RAM_POWER;
__no_init __root int8u  radioBands            @ RESERVED_RAM_BANDS;
__no_init __root int8u  bootloaderMode        @ RESERVED_RAM_MODE;
__no_init __root volatile int8u  emResetCause          @ RESERVED_RAM_RESET_CAUSE;
__no_init __root volatile int8u  emResetCauseMarker    @ RESERVED_RAM_RESET_CAUSE_B;
__no_init __root volatile int16u emResetRStack         @ RESERVED_RAM_RSTACK_L;
__no_init __root volatile int16u emResetCStack         @ RESERVED_RAM_CSTACK_L;

#define INTERNAL_RC_1MHZ_MODE 0x01
#define INTERNAL_RC_2MHZ_MODE 0x02
#define INTERNAL_RC_4MHZ_MODE 0x03
#define INTERNAL_RC_8MHZ_MODE 0x04
#define CKSEL_BITS_MASK       0x07

static boolean usingInternalOscillator(void)
{
  int8u clockSelect = (_SPM_GET_FUSEBITS() & CKSEL_BITS_MASK);

  if ((clockSelect >= INTERNAL_RC_1MHZ_MODE) && 
      (clockSelect <= INTERNAL_RC_8MHZ_MODE))
    return TRUE;

  return FALSE;
  
}
void microBoot(void)
{
  if (usingInternalOscillator() == TRUE)
    rcCalibrate();

  // enable watchdog
  // Note:  the watchdog timer is a *separate* 1 MHz clock that is independent
  // of the processor clock other timers.  Setting WDP0-2 will set off the
  // watchdog interrupt after 2 million cycles, or every 2 seconds.
  WDTCR = BIT(WDCE) + BIT(WDE);
  WDTCR = BIT(WDE) + BIT(WDP2) + BIT(WDP1) + BIT(WDP0);   // 2secs

  // start timers
#ifndef EMBER_APPLICATION_HAS_MICRO_CUSTOM_SYSTEM_TIMER
  microStartSystemTick();
#endif

#ifndef EMBER_APPLICATION_HAS_MICRO_CUSTOM_MAC_TIMER
  microStartMacTick();
#endif
}

void microReboot(void) 
{
  // we're self wdog'in, record that fact
  emResetCauseMarker = BIT(RESET_CAUSE_SELF_WDOG)|BIT(RESET_CAUSE_ASSERT_BIT);

   // instead of jumping to the reset vector and having to 
   // re init all the atmel control registers ourselves, we
   // force a watchdog reset with a minimum timeout
   WDTCR = BIT(WDCE) + BIT(WDE); 
   WDTCR = BIT(WDE);    // ~18ms
   for (;;) { }   // wait for the WDT to get us
}

void microDisableWatchDog(int8u magicKey)
{
  if (magicKey == MICRO_DISABLE_WATCH_DOG_KEY) {
    WDTCR = BIT(WDCE) + BIT(WDE); 
    WDTCR = 0;
  }
}

EmberResetType microGetResetInfo(void)
{
  static int8u cause_mark = 0xFF;
  if ( cause_mark == 0xFF ) {
    cause_mark = emResetCauseMarker;
    emResetCauseMarker = 0;
  }
  switch ( emResetCause ) {
    case BIT(PORF):  return RESET_POWERON;
    case BIT(EXTRF): return RESET_EXTERNAL;
    case BIT(WDRF):  
      if ( cause_mark & BIT(RESET_CAUSE_SELF_WDOG) ) {
        if ( cause_mark & BIT(RESET_CAUSE_ASSERT_BIT) ) {
          return RESET_ASSERT;
        } else if ( cause_mark & BIT(RESET_CAUSE_CSTACK_BIT) ) {
          return RESET_CSTACK;
        } else if ( cause_mark & BIT(RESET_CAUSE_RSTACK_BIT) ) {
          return RESET_RSTACK;
        } else if ( cause_mark & BIT(RESET_CAUSE_BOOTLOADER_BIT) ) {
          return RESET_BOOTLOADER;
        } else {
          return RESET_UNKNOWN;
        }
      }
      return RESET_WATCHDOG;
    case BIT(BORF):  return RESET_BROWNOUT;
#ifdef JTDR
    case BIT(JTDR):  return RESET_JTAG;
#else 
    case BIT(JTRF):  return RESET_JTAG;
#endif
  }
  return RESET_UNKNOWN;
}


void* microGetResetString(void)
{   
  static PGM_P resetString[] = { "PCZERO", "PWR", "EXT", "WDOG", "BOD", "JTAG", "RSTK", "CSTK", "ASSERT", "BL" };
  switch ( microGetResetInfo() ) {
    case RESET_POWERON:
      return((void*)resetString[1]);
    case RESET_EXTERNAL:
      return((void*)resetString[2]);
    case RESET_WATCHDOG:
      return((void*)resetString[3]);
    case RESET_BROWNOUT:
      return((void*)resetString[4]);
    case RESET_JTAG:
      return((void*)resetString[5]);
    case RESET_RSTACK:
      return((void*)resetString[6]);
    case RESET_CSTACK:
      return((void*)resetString[7]);
    case RESET_ASSERT:
      return((void*)resetString[8]);
    case RESET_BOOTLOADER:
      return((void*)resetString[9]);
  }         
  return((void*)resetString[0]);
}


#ifndef EMBER_APPLICATION_HAS_MICRO_CUSTOM_SYSTEM_TIMER
/**
 * Set up the 8-bit timer 0 to use external 32.768kHz (31.25kHz) crystal. By
 * using a prescalar value of 1, one tick represents 30.52 us (32 us), and an
 * overflow occurs every 7.8125 ms (8.192 ms).
 */
int16u microStartSystemTick(void)
{
  int16u asyncTime = 0;
  DECLARE_INTERRUPT_STATE;
  DISABLE_INTERRUPTS();
  __watchdog_reset(); // if the wdog is enabled, give us 2 seconds starting now.
  TIMSK &= ~(BIT(OCIE0)|BIT(TOIE0)); // disable tim0 ints
  TCNT0 = 0;                         // tim0 counter register
  TCCR0 = 1;                         // 1:1 prescale
  ASSR |= BIT(AS0);                  // clock source (async ext xtal)
  TIFR = (BIT(OCF0)|BIT(TOV0));      // clear (write 1) int flags
  // We use output compare with match set to FF rather than oveflow
  //  because the asynchronous operation of the timer requires 1 timer
  //  cycle to trigger the interrupt.
  OCR0 = 0xFF;
  while ( ASSR & 0x7 ) { // wait for asych update of TCNT0/OCR0/TCCR0    
    microDelayMicroSeconds(100); 
#if defined(HALTEST)
    //Watchdog may not always be enabled, so we are capable of ending up
    //in an infinite loop waiting - the assert kicks us out
    assert( asyncTime++ < 65535 );  // 6.5 seconds (max before 16bit overflow)
#else
    //Watchdog may not always be enabled, so we are capable of ending up
    //in an infinite loop waiting - the assert kicks us out
    assert( asyncTime++ < 15000 );  // 1.5 seconds (must be < wdog timeout)
#endif
  }
  TIMSK |= BIT(OCIE0);  // enable tim0 output compare int  
  RESTORE_INTERRUPTS();
    
  systemTick = 0;
  return asyncTime;
}

/**
 * Return a 16 bit real time clock value. The clock's increment is 0.977 ms
 * (1.024 ms) and it rolls over every 64 sec (67.1 sec).
 */
int16u microGetSystemTick(void)
{
  return (int16u)microGetSystemTime();
}

int16u microGetQuarterSeconds(void)
{
  return (int16u)(microGetSystemTime() >> 8);
}

#pragma vector = TIMER0_COMP_vect
__interrupt void TIMER0_COMP_interrupt (void)
{
  ++systemTick;
}

int32u microGetSystemTime(void)
{
  int32u t;
  int8u c;
  ATOMIC(
    // read before testing if the interrupt is pending, in case it goes pending
    // right after we read it
    c = TCNT0;
    if (TIFR & BIT(OCF0)) {
      // account for interrupt that hasn't yet been able to occur
      //  have to re-read TCNT0 because it has wrapped.
      c = TCNT0;
      t = ((systemTick + 1) << 3) | (c >> 5);
    } else {
      t = (systemTick << 3) | (c >> 5);
    }
  )
  return t;
}
#endif //EMBER_APPLICATION_HAS_MICRO_CUSTOM_SYSTEM_TIMER

#ifndef EMBER_APPLICATION_HAS_MICRO_CUSTOM_MAC_TIMER
/**
 * Orders a callback after a certain delay.
 * This function is designed expressly for the MAC backoff algorithm.
 * Delay is specified in symbol times
 */
// Uses timer 1 compare interrupt A
void microOrderSymbolDelayA(int16u symbols)
{
  // make sure there is no currently pending delay
  //The MAC can only handle a single CSMA packet being transmitted at a time.
  //A second delay means the MAC has fatally errored.
  assert( (TIMSK & BIT(OCIE1A)) == 0 );
  TIFR = BIT(OCF1A);  // clear any pending interrupts
  convertSymbolsToHicks(symbols);
  // minimum delay required, or else compare will not happen until wrap-around
  if(symbols == 0) symbols = 1;
  OCR1A = (TCNT1 + symbols);
  TIMSK |= BIT(OCIE1A); // enable timer 1 compare interrupt A
}

void microCancelSymbolDelayA(void)
{
  ATOMIC(
    TIMSK &= ~BIT(OCIE1A);
    TIFR = BIT(OCF1A);
  )
}

#pragma vector = TIMER1_COMPA_vect
__interrupt void TIMER1_COMPA_interrupt (void)
{
  // disable timer 1 compare interrupt
  TIMSK &= ~BIT(OCIE1A);
  emMicroSymbolDelayAIsr();
}
#ifndef SIXTEEN_BIT_BEACONS
static void microReallyOrderSymbolDelayB(int16u symbols)
{
  // make sure there is no currently pending delay
  //The MAC can only handle a single non-CSMA packet being transmitted at a time.
  //A second delay means the MAC has fatally errored.
  assert( (TIMSK & BIT(OCIE1B)) == 0 );
  TIFR = BIT(OCF1B);  // clear any pending interrupts
  if(symbols == 0) symbols = 1;
  OCR1B = (TCNT1 + symbols);
  TIMSK |= BIT(OCIE1B); // enable timer 1 compare interrupt B
}
#endif //SIXTEEN_BIT_BEACONS

// uses timer 1 compare interrupt B
void microOrderSymbolDelayB(EmberBeaconTime symbols)
{
  // make sure there is no currently pending delay
#ifdef SIXTEEN_BIT_BEACONS
  //The MAC can only handle a single non-CSMA packet being transmitted at a time.
  //A second delay means the MAC has fatally errored.
  assert( (TIMSK & BIT(OCIE1B)) == 0 );
  TIFR = BIT(OCF1B);  // clear any pending interrupts
  convertSymbolsToHicks(symbols);
  // minimum delay required, or else compare will not happen until wrap-around
  if(symbols == 0) symbols = 1;
  OCR1B = (TCNT1 + symbols);
  TIMSK |= BIT(OCIE1B); // enable timer 1 compare interrupt B

#else //SIXTEEN_BIT_BEACONS

  // make sure there is no currently pending delay
  //The MAC can only handle a single non-CSMA packet being transmitted at a time.
  //A second delay means the MAC has fatally errored.
  assert( !symbolDelayBPending );

  convertSymbolsToHicks(symbols);
  ATOMIC (
    symbolDelayBPending = TRUE;
    if (symbols < 0x10000) {
      microReallyOrderSymbolDelayB((int16u)(symbols));
    } else {
      symbols -= (0xFFFF - TCNT1);
      // add one to account for the first overflow interrupt
      symbolDelayBHigh = (int16u)(symbols >> 16) + 1;
      symbolDelayBLow = (int16u)(symbols);
    }
  ) // ATOMIC
#endif //SIXTEEN_BIT_BEACONS
}

#pragma vector = TIMER1_COMPB_vect
__interrupt void TIMER1_COMPB_interrupt (void)
{
#ifndef SIXTEEN_BIT_BEACONS
  symbolDelayBPending = FALSE;
#endif //SIXTEEN_BIT_BEACONS
  // disable timer 1 compare interrupt
  TIMSK &= ~BIT(OCIE1B);
  emMicroSymbolDelayBIsr();
}

// Timer 1 is used for the mac timer
void microStartMacTick(void)
{
  // disable all interrupts
  TIMSK &= ~(BIT(TICIE1)|BIT(OCIE1A)|BIT(OCIE1B)|BIT(TOIE1));
  ETIMSK &= ~(BIT(OCIE1C));
  // disable any special modes
  TCCR1A = 0;
  // set prescale to clk/64
  // This creates a resolution of 16us on 4Mhz boards and 8us on 8Mhz boards
  TCCR1B = (BIT(CS11) | BIT(CS10));
  // reset counter register (16 bit)
  TCNT1 = 0;
  // clear (write 1) any pending interrupt flags
  TIFR = (BIT(ICF1)|BIT(OCF1A)|BIT(OCF1B)|BIT(TOV1));
  ETIFR = (BIT(OCF1C));
  // enable overflow interrupt
  TIMSK |= BIT(TOIE1);
#ifdef SIXTEEN_BIT_BEACONS
  syncTimeOffset = 0;
#else //SIXTEEN_BIT_BEACONS
  syncTick = 0;
  syncTimeOffset = 0;
  syncCaptured = FALSE;
  syncCapturedTime = 0;
#endif //SIXTEEN_BIT_BEACONS
}
#ifndef SIXTEEN_BIT_BEACONS
#pragma vector = TIMER1_OVF_vect
__interrupt void TIMER1_OVF_interrupt (void)
{
  ++syncTick;
  if (symbolDelayBPending) {
    --symbolDelayBHigh;
    if (symbolDelayBHigh == 0) {
      microReallyOrderSymbolDelayB(symbolDelayBLow);
    }
  }
}

void microSpdIsr(void)
{
  syncCapturedTime = microGetSyncTime();
  syncCaptured = TRUE;
}
#endif //SIXTEEN_BIT_BEACONS

EmberBeaconTime microGetSyncTime(void)
{
#ifdef SIXTEEN_BIT_BEACONS
  int16u t;

  t = TCNT1;
  convertHicksToSymbols(t);
  t += syncTimeOffset;
  return t;
  
#else //SIXTEEN_BIT_BEACONS
  int32u t;
  int16u c;
  ATOMIC(
    // read before testing if the interrupt is pending, in case it goes pending
    // right after we read it
    c = TCNT1;
    if (TIFR & BIT(TOV1)) {
      // account for interrupt that hasn't yet been able to occur
      //  have to re-read TCNT1 because it has wrapped.
      c = TCNT1;
      t = syncTick + 1;
    } else {
      t = syncTick;
    }
    t = t << 16;
    t = (t | c);
  )
  convertHicksToSymbols(t);
  t += syncTimeOffset;
  return t;
#endif //SIXTEEN_BIT_BEACONS
}

void microIncreaseSyncOffset(EmberBeaconTime increase)
{
  syncTimeOffset += increase;
}

#ifndef SIXTEEN_BIT_BEACONS
int32u microGetSyncCapture(void)
{
  int32u t;
  ATOMIC(
    // make an atomic copy in case microSpdIsr interrupts us
    t = syncCapturedTime;
  )
  return t;
}

void microClearSyncCapture(void)
{
  syncCaptured = FALSE;
}

boolean microSyncCaptured(void)
{
  return syncCaptured;
}
#endif //SIXTEEN_BIT_BEACONS
#endif //EMBER_APPLICATION_HAS_CUSTOM_MAC_TIMER

void microSleep(SleepModes sleepMode)
{
  
  // enable sleeping
  MCUCR |= BIT(SE);


  switch ( sleepMode )
  {
    case SLEEPMODE_IDLE:       // idle - sleep mode        (uart ints, timer ints, and external ints can wake us)
      MCUCR &= ~(BIT(SM1)|BIT(SM0));
      break;
    case SLEEPMODE_RESERVED:   // reserved - sleep mode
      MCUCR &= ~BIT(SM1); MCUCR |= BIT(SM0);
      break;
    case SLEEPMODE_POWERDOWN:  // power down - sleep mode  (only external ints can wake us)
      MCUCR |= BIT(SM1); MCUCR &= ~BIT(SM0);
      break;
    case SLEEPMODE_POWERSAVE:  // power save - sleep mode  (wake on ext ints, and optionally timer0)
      MCUCR |= BIT(SM1)|BIT(SM0);
      break;
  }


  __sleep();   // command micro into sleep


  // disable sleeping
  MCUCR &= ~BIT(SE);

}


void microMemCopy(void *dest, void *src, int8u bytes)
{
  int8u *d=dest;
  int8u *s=src;

  while(bytes>=4) {
    bytes -= 4;
    *d++ = *s++;
    *d++ = *s++;
    *d++ = *s++;
    *d++ = *s++;
  }
  for(;bytes;bytes--) {
    *d++ = *s++;
  }
}

void microMemSet(void *dest, int8u val, int8u bytes)
{
  int8u *d=dest;

  for(;bytes;bytes--) {
    *d++ = val;
  }
}

int8s emMicroMemCompare(void *source0, void *source1, int8u bytes)
{
  int8u *s0 = source0;
  int8u *s1 = source1;

  for(; 0 < bytes; bytes--, s0++, s1++) {
    int8u b0 = *s0;
    int8u b1 = *s1;
    if (b0 != b1)
      return b0 - b1;
  }
  return 0;  
}

// Test code for emMicroMemCompare().  There is no good place for unit tests
// for this file.  If you change the function you should probably rerun the
// test.
//  {
//    int8u s0[3] = { 0, 0, 0};
//    int8u s1[3] = { 0, 0, 0};
//    int8u i;
//    assert(emMicroMemCompare(s0, s1, 0) == 0);
//    assert(emMicroMemCompare(s0, s1, 3) == 0);
//    for (i = 0; i < 3; i++) {
//      s0[i] = 1;
//      assert(emMicroMemCompare(s0, s1, 3) > 0);
//      s1[i] = 2;
//      assert(emMicroMemCompare(s0, s1, 3) < 0);
//      s0[i] = 2;
//    }
//  }

// Same again, except that the second source is in program space.

int8s emMicroMemPGMCompare(void *source0, void PGM *source1, int8u bytes)
{
  int8u *s0 = source0;
  int8u PGM *s1 = source1;

  for(; 0 < bytes; bytes--, s0++, s1++) {
    int8u b0 = *s0;
    int8u b1 = *s1;
    if (b0 != b1)
      return b0 - b1;
  }
  return 0;  
}

EmberStatus microBootloaderTrap(BootloaderTraps trapNumber, ...)
{
  int16u trapTabAddr;
  int16u signature;
  int16u funcAddr;
  int8u high, low;

  LPM_HIGH((PGM_PU)FLASH_TAB_POINTERS + 
    FLASH_TAB_POINTERS_BOOT_TRAP_TAB_PTR + 1, 
    high);
  LPM_HIGH((PGM_PU)FLASH_TAB_POINTERS + 
    FLASH_TAB_POINTERS_BOOT_TRAP_TAB_PTR + 0, 
    low);

  trapTabAddr = HIGH_LOW_TO_INT(high, low);

  LPM_HIGH((PGM_PU)(trapTabAddr + 1), high);
  LPM_HIGH((PGM_PU)(trapTabAddr + 0), low);

  signature = HIGH_LOW_TO_INT(high, low);

  LPM_HIGH((PGM_PU)(trapTabAddr + (trapNumber<<1) + 1), high);
  LPM_HIGH((PGM_PU)(trapTabAddr + (trapNumber<<1)), low);

  funcAddr = HIGH_LOW_TO_INT(high, low);
  
//  INTERRUPTS_ON(); //req for serial prints
//  emberSerialPrintf(SER232,"trapTabAddr:%x\r\n",trapTabAddr);
//  emberSerialPrintf(SER232,"signature:%x\r\n",signature);
//  emberSerialPrintf(SER232,"funcAddr:%x\r\n",funcAddr);
  
  if (signature != BOOTLOADER_TRAPTAB_SIGNATURE )
    return EMBER_ERR_BOOTLOADER_TRAP_TABLE_BAD;
  
  switch (trapNumber) {
    case BOOT_TRAP_APP_START: {
      void (*bootloadAppStart)(void) = (void(*)(void))funcAddr;
      bootloadAppStart();
      break; }
    case BOOT_TRAP_FLASH_WRITE: {
      break;}
    case BOOT_TRAP_FLASH_ADJUST: {
      va_list args;
      int16u* dst;
      int16u* src;
      int8u size;
      void (*flashAdjustBlock)(int16u* dst, int16u* src, int8u size) = 
        (void(*)(int16u* dst, int16u* src, int8u size))funcAddr;
      va_start(args, trapNumber);
      dst = va_arg(args, int16u*);
      src = va_arg(args, int16u*);
      size = va_arg(args, int8u);
      va_end(args);
//emberSerialPrintf(SER232,"dst:%x,src:%x,size:%d",dst,src,size);
//emberSerialWaitSend(SER232);
      flashAdjustBlock(dst, src, size);
      break; }
    default:
      return EMBER_ERR_BOOTLOADER_TRAP_UNKNOWN;
  }
  
  return EMBER_SUCCESS;
}


int16u microGetUpperFlashStart(void)
{
  int16u tableAddress;
  int8u high, low;
  
  LPM_HIGH((PGM_PU)FLASH_TAB_POINTERS + 
    FLASH_TAB_POINTERS_UPPER_TOKEN_TAB_PTR + 1, 
    high);
  LPM_HIGH((PGM_PU)FLASH_TAB_POINTERS + 
    FLASH_TAB_POINTERS_UPPER_TOKEN_TAB_PTR + 0, 
    low);
  
  tableAddress = HIGH_LOW_TO_INT(high, low);


  return tableAddress;
}

static void loadTokensToReservedRAM(void)
{
  getToken( eui64, (void*)TOKEN_EUI_64, sizeof(eui64));
  radioPower = emberGetRadioPower();
  radioBands = GET_TOKEN( TOKEN_RADIO_BANDS_SUPPORTED );

  // Can't rely on the tokens because as of flare branch, the radioChannel
  // token isn't updated in real time.
  radioChannel = emberGetRadioChannel();
}

// this is called for over the air bootloading.
void runBootload(int8u *eui)
{
  int8u i;

  INTERRUPTS_OFF();
  microDisableWatchDog(MICRO_DISABLE_WATCH_DOG_KEY);
  // using this technique to get into the bootloader, means that
  // it's segment initialization has occurred correctly.
  bootloaderApplicationRequestKey = BOOTLOADER_APP_REQUEST_KEY;

  loadTokensToReservedRAM();
  
  // gateway ID is passed as the argument eui
  for(i=0; i<8; i++) {
    gatewayID[i] = eui[i];
  }

  bootloaderMode = eui[8];

  #if __HAS_ELPM__
    asm("jmp 0x1E000"); // jump to bootloader-reset-vec
  #else
    asm("jmp 0xe000");
  #endif
}
    

void microLaunchBootloader(int8u magicKey)
{
  int8u i;

  if (magicKey == MICRO_LAUNCH_BOOTLOADER_KEY) {
    INTERRUPTS_OFF();
    microDisableWatchDog(MICRO_DISABLE_WATCH_DOG_KEY);
    // using this technique to get into the bootloader, means that
    // it's segment initialization has occurred correctly.
    bootloaderApplicationRequestKey = BOOTLOADER_APP_REQUEST_KEY;

    loadTokensToReservedRAM();
    
    // no gateway, so we need to initialize to defaults.
    for(i=0; i<8; i++)
      gatewayID[i] = i;

    bootloaderMode = 0;

  #if __HAS_ELPM__
    asm("jmp 0x1E000"); // jump to bootloader-reset-vec
  #else
    asm("jmp 0xe000");
  #endif
  }
}


#pragma vector = TIMER3_OVF_vect
__interrupt void microStoreCurrentPC_interrupt (void)
{
  // The stack:
  //   SP  : empty
  //   SP+1: PC1 (high byte)
  //   SP+2: PC1 (low byte)
  //   SP+3: PC2 (high byte)
  //   SP+4: PC2 (low byte)
  // PC1 is the instruction that will be executed when this ISR returns
  // PC2 is the instruction after the call to the function this ISR interrupted
  int8u l,h;
  int16u pc;

#if MEGA64
  TCNT3L = 0;
  TCNT3H = 0;
#else
  TCNT3 = 0;
#endif
  // Store PC1
  h = *((int8u*)(SP+1));
  l = *((int8u*)(SP+2));
  pc = HIGH_LOW_TO_INT(h,l);
  // convert from word to byte address
  pc = pc << 1;
  micro_last_known_pc = pc;
}

int16u microGetPCDiagnostics(void)
{
  return micro_last_known_pc; 
}

void microStartPCDiagnostics(void)
{
   // Use 16 bit timer3 with a 1:8 prescalar
   TCCR3B = BIT(CS31);  // prescaler clk/8: ea tick is 2us. (ovf is 131ms)

   ATOMIC(
     // ensure interrupt is cleared
     ETIFR = BIT(TOV3);  // write logic one to clear interrupts on AVRs
#if MEGA64
     TCNT3L = 0;
     TCNT3H = 0;
#else
     TCNT3 = 0;
#endif
     ETIMSK |= BIT(TOIE3);  // enable timer 1 overflow interrupt
   )
}

void microStopPCDiagnostics(void)
{
  TCCR3B = 0;              // disable timer 3 to save power
  ETIMSK &= ~BIT(TOIE3);   // disable timer 3 overflow interrupt
}

void microDelayMicroSeconds(int16u us)
{
#if defined(AVR8MHZ)
  while ( (us-=1) > 0 ) {  // 6 cycles
#elif defined(AVR4MHZ) || defined(AVR3p68MHZ) || defined(AVR7p3728MHZ)
  while ( (us-=2) > 1 ) {  // 6 cycles
#else 
  #error no mcu system clock defined.
#endif
    
  #if defined(AVR7p3728MHZ)  
    // 6 for the loop +
    asm("mul r18,r18"); // 2
    asm("mul r18,r18"); // 2
    asm("mul r18,r18"); // 2
    asm("mul r18,r18"); // 2
    asm("nop\n"); // 1
    // want ~15 (14.74) total (for 2 us)
  #else
    // 6 for the loop +
    asm("mul r18,r18"); // 2
    // want 8 total (for 2 us)
  #endif
  }
      
  // return is 4 cycles (but not accounted for)
  
}

/*****************************************************************************
*
*   Function name : rcCalibrate
*
*   Returns :       None
*
*   Parameters :    None
*
*   Purpose :       Calibrate the internal OSCCAL byte, using the external 
*                   32,768 kHz crystal as reference
*
*   Atmel application which incluedes an internal runtime oscillator calibration
*   routine. Ported to 128 from Atmel code, timer 2 to timer 0 for application.
*   1 MHz to 8 MHz internal oscillator.
*   http://www.atmel.com/dyn/products/tools_card.asp?family_id=607&family_name=AVR+8%2DBit+RISC+&tool_id=3146
*   Referred to us by the Atmel FAE Ken Yasi.
*   Redesigned by DC Seward to take out the magic numbers.
*   
*   Needs to be called in microBoot(); otherwise internal oscillator will not
*   be accurate enough for serial port baud rate. Haltest calls it after
*   boardBoot();. Fuses must be set CLK0, CLK1, CLK3 and CLK2 is
*   unchecked. Could read out fuses in runtime and flag a variable to see if
*   internal oscillator is used and if the routine needs to be run.  High and
*   Low values are a range around the number of processor clock oscillations
*   for the given number of 32 kHz cycles as follows:
*       Number = OCR0 * Fproc / Fosc
*   where Number is the desird number of processor clock oscillations, OCR0 is
*   the number of 32 kHz cycles, Fproc is the desired processor frequency,
*   usually 8 MHz, and
*   Fosc is the oscillator speed, usually 32kHz.
*
*   Starts timer 1 (16 bit) and timer 0 (8 bit asynch). Timer 0 measures 32K
*   external crystal. Timer 1 measures internal RC. If external 32K changes
*   frequency, the 'magic' numbers must change. Cycles through 200 ticks of the
*   32K crystal, compare then stops timer 1. Timer 1 is compared to 'magic'
*   numbers and if too high, it decrements OSCCAL, if too high or overflow it
*   lowers OSCCAL. It continues until the timer 1 value fits in between the
*   high and low and is calibrated.
*****************************************************************************/
void rcCalibrate(void)
{
    boolean calibrate = FALSE, previousSubtract;
    int16u temp;
    int16u delta, previousDelta;

    previousDelta = 0xffff;
    
    TIMSK &= ~((BIT(OCIE0) | BIT(TOIE0)));    //disable OCIE0 and TOIE0

    ASSR = BIT(AS0);  //select asynchronous operation of timer0 (32,768kHz)
    
    // Actual number of cycles, minus one for the single cycle delay between
    // the compare being true and the compare flag being set. 
    #define NUMBER_32K_CYCLES 3
    OCR0 = (NUMBER_32K_CYCLES - 1);      // set timer0 compare value
        
    TCCR1B = BIT(CS10);    // start timer1 with no prescaling
    TCCR0  = BIT(CS00);    // start timer0 with no prescaling

    // need to wait for the OCR0 and TCCR0 registers to be latched, although
    // technically this step is unnecessary since this happens at the start of
    // the next clock cycle, and we wait again for the next clock cycle below. 
    //while((ASSR & BIT(OCR0UB)) | (ASSR & BIT(TCR0UB))); 

    microDelayMicroSeconds(1000);    // wait for external crystal to stabilise
    
    while(!calibrate)
    {
        __disable_interrupt();  // disable global interrupt
        
        TIFR |= BIT(TOV0) | BIT(OCF0);
        TIFR |= BIT(ICF1) | BIT(OCF1A) | BIT(OCF1B) | BIT(TOV1);

        TCNT0 = 0;      // clear timer0 counter
        // Wait for TCNT0 register to get set.  This happens at the start of
        // the next 32kHz crystal cycle.
        while((ASSR & BIT(TCN0UB)));

        TCNT1H = 0;     // clear timer1 counter
        TCNT1L = 0;
           
        while ( !((TIFR & BIT(OCF0)) == BIT(OCF0))); // wait for timer0 compareflag
    
        TCCR1B &= ~((BIT(CS12)) | (BIT(CS11)) | (BIT(CS10))); // stop timer1

        __enable_interrupt();  // enable global interrupt
    
        if ( ((TIFR & BIT(TOV1)) == BIT(TOV1))) {
            temp = 0xFFFF;      // if timer1 overflows, set the temp to 0xFFFF
        }
        else {   // read out the timer1 counter value
            temp = TCNT1L;
            temp |= (TCNT1H << 8);
        }
    
        // number of cycles 
        #define NUMBER_CODE_AMBIGUITY   4
#ifdef AVR8MHZ
        #define NUMBER_PROCESSOR_CYCLES (((NUMBER_32K_CYCLES * 8000000) / 32768)+NUMBER_CODE_AMBIGUITY) // 8 MHz
#elif defined(AVR4MHZ)
        #define NUMBER_PROCESSOR_CYCLES (((NUMBER_32K_CYCLES * 4000000) / 32768)+NUMBER_CODE_AMBIGUITY) // 4 MHz
#else
        // doesn't matter because we aren't doing RC calibration anyway.  
        #define NUMBER_PROCESSOR_CYCLES (((NUMBER_32K_CYCLES * 8000000) / 32768)+NUMBER_CODE_AMBIGUITY) // 8 MHz
#endif
        #define DELTA_PER_STEP          ((NUMBER_PROCESSOR_CYCLES+189) / 379)
        #define DELTA_THRESH            (3 * DELTA_PER_STEP)

        if (temp > NUMBER_PROCESSOR_CYCLES) {
          // We are fast.  See if our delta is better than before.
          delta = temp - NUMBER_PROCESSOR_CYCLES;
          if(delta < previousDelta || delta > DELTA_THRESH) {
            previousDelta = delta;
            previousSubtract = TRUE;
            //OSCCAL = OSCCAL - (int8u) (delta / DELTA_PER_STEP);
            OSCCAL--;

#if defined(HALTEST)
            emberSerialGuaranteedPrintf(SER232, "+Delta: %x %x\r\n", 
                                        (int8u) (delta >> 8),
                                        (int8u) (delta & 0x00ff));
#endif
          } else if(delta < DELTA_THRESH) {
            calibrate = TRUE;
          }
        } else {
          // We are slow.  See if our delta is better than before.
          delta = NUMBER_PROCESSOR_CYCLES - temp;
          if(delta < previousDelta || delta > DELTA_THRESH) {
            previousDelta = delta;
            previousSubtract = FALSE;
            //OSCCAL = OSCCAL + (int8u) (delta / DELTA_PER_STEP);
            OSCCAL++;

#if defined(HALTEST)
            emberSerialGuaranteedPrintf(SER232, "-Delta: %x %x\r\n", 
                                        (int8u) (delta >> 8),
                                        (int8u) (delta & 0x00ff));
#endif
          } else if(delta < DELTA_THRESH) {
            calibrate = TRUE;
          }
        }
    
        TCCR1B = BIT(CS10); // start timer1*/
    }

    // exit loop.  Note that we are one OSCCAL value away from ideal.  Need to
    // subtract or add depending on the last step.
    if(previousSubtract == TRUE) {
      OSCCAL++;
    } else {
      OSCCAL--;
    }
}
