/** @file hal/micro/avr-atmega/128/micro.h
 * @brief Utility and convenience functions for ATmega128 microcontroller
 *
 * Unless otherwise noted, the EmberNet stack does not use these functions,
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 *
 * \b Note: These HAL files/functions are also used for the ATmega64 
 * microcontroller since the only difference between the ATmega128 and ATmega64 
 * is the size of the flash.
 *
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */


#ifndef __128_MICRO_H__
#define __128_MICRO_H__

#ifndef DOXYGEN_SHOULD_SKIP_THIS

#ifndef __MICRO_H__
#error do not include this file directly - include micro/micro.h
#endif

#define ENABLE_BIT_DEFINITIONS
#include <ioavr.h>

#include "micro/avr-atmega/128/reserved-ram.h"

#endif // DOXYGEN_SHOULD_SKIP_THIS


/** \b DEPRECATED. Use CLEARBIT() in ember.h. */
#define INPUTPIN(reg, bit)	  reg &= ~(1<<bit)

/** \b DEPRECATED. Use CLEARBITS() in ember.h. */
#define INPUTPINS(reg, bits)	reg &= ~(bits)

/** \b This function is \b DEPRECATED. Use SETBIT() in ember.h. */
#define OUTPUTPIN(reg, bit)	  reg |= 1<<bit

/** \b DEPRECATED. Use SETBITS() in ember.h. */
#define OUTPUTPINS(reg, bits)	reg |= bits


/** @name Flash Accessor Macros and Functions
 * \b Note: All int16u values are little-endian.
 */
//@{

#if __HAS_ELPM__
  #define microReadFlashByte(addr)   __extended_load_program_memory(addr)
  #define microReadFlashWord(addr)   HIGH_LOW_TO_INT( \
                                      __extended_load_program_memory((addr)+1), \
                                      __extended_load_program_memory(addr) )
#else // __HAS_ELPM__

/** @description Reads a byte from flash.
 * 
 * @param addr: A 16-bit address in program space.
 *  
 * @return An int8u value.
 */
  #define microReadFlashByte(addr)   __load_program_memory(addr)

/** @description Reads a word from flash.
 * 
 * @param addr: A 16-bit address in program space.
 * 
 * @return An int16u value.
 */
#define microReadFlashWord(addr)   HIGH_LOW_TO_INT( \
                                      __load_program_memory((addr)+1), \
                                      __load_program_memory(addr) )
#endif  // __HAS_ELPM__

/** @description Returns the address of the end of flash memory. */ 
#define microGetFlashEnd()         FLASHEND

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** @description DEPRECATED: returns address of the start of the token 
 * system's Upper Flash Override Token Table.  Implementation still
 * required for token-upgrade capability.
 */ 
int16u microGetUpperFlashStart(void); 

// Constants used by the bootloader
#define FLASH_TAB_POINTERS                        0x00F0
#define FLASH_TAB_POINTERS_DEFAULT_TOKEN_TAB_PTR  0       //DEPRECATED
#define FLASH_TAB_POINTERS_UPPER_TOKEN_TAB_PTR    2       //DEPRECATED
#define FLASH_TAB_POINTERS_BOOT_TRAP_TAB_PTR      4
#define FLASH_TAB_POINTERS_HELPERFUNC_TAB         6
#define BOOTLOADER_TRAPTAB_SIGNATURE              0xBEEF
#endif // DOXYGEN_SHOULD_SKIP_THIS

//@}  // end of Flash Accessor Macros, Functions, and Constants


/** @name EEPROM Accessor Functions
 * \b Note: All int16u values are little-endian.
 */
//@{

/** @description Returns a byte from EEPROM memory space.
 * 
 * @param addr: A 16-bit address in EEPROM memory space.
 * 
 * @param var: The int8u read. 
 */
#define microReadEEPROMByte(addr,var)  __EEGET(var,addr)

/** @description Returns a word from EEPROM memory space.
 * 
 * @param addr: A 16-bit address in EEPROM memory space.
 * 
 * @param var: The int16u read.
 */
#define microReadEEPROMWord(addr,var)   \
        do {                            \
        int8u h,l;                      \
        __EEGET(h,(addr)+1);            \
        __EEGET(l,addr);                \
        var = HIGH_LOW_TO_INT( h,l );   \
        } while(0)

/** @description Writes a byte to EEPROM memory space.
 * 
 * @param addr: A 16-bit address in program space.
 *
 * @param var: The int8u value of the byte to write to EEPROM.
 */
#define microWriteEEPROMByte(addr,var)  \
        do {                            \
        __EEPUT(addr,var);              \
        } while(0)

/** @description Writes a word to EEPROM memory space.
 * 
 * @param addr: A 16-bit address in program space.
 * 
 * @param var: The int16u value of the word to write to EEPROM.
 */
#define microWriteEEPROMWord(addr,var)  \
        do {                            \
        __EEPUT(addr+1,HIGH_BYTE(var)); \
        __EEPUT(addr,LOW_BYTE(var));    \
        } while(0)

/** @description Returns the address of the end of EEPROM memory space */ 
#define microGetEEPROMEnd()        (E2END - 1)

/** @description Returns the address of the start of EEPROM memory space. */ 
#define microGetEEPROMStart()      0

//@}  // end of EEPROM Accessor Functions




/** @description Gives the watchdog reset a friendlier name. */
#define watchdogReset() __watchdog_reset()

#ifndef DOXYGEN_SHOULD_SKIP_THIS

// this should at the least, be renamed.
void rcCalibrate(void);

// Reserved RAM addresses for communication with the bootloader
#define RESERVED_RAM_BOOTLOADER_KEY  0x10FE
#define RESERVED_RAM_PC_DIAG         0x10FC
#define RESERVED_RAM_GATEWAY_ID      0x10F4
#define RESERVED_RAM_EUI64_ID        0x10EC
#define RESERVED_RAM_PAN_ID          0x10EA
#define RESERVED_RAM_NODE_ID         0x10E8
#define RESERVED_RAM_CHANNEL         0x10E7
#define RESERVED_RAM_POWER           0x10E6
#define RESERVED_RAM_BANDS           0x10E5
#define RESERVED_RAM_MODE            0x10E4
#define RESERVED_RAM_UNUSED3         0x10E3
#define RESERVED_RAM_UNUSED2         0x10E2
#define RESERVED_RAM_UNUSED1         0x10E1
// note: you must actually reserve the ram in low-level-init.c and 
// all the linker scripts. this just defines the addresses
//#define NORMAL_RAM_END             0x10E0 

#if __HAS_ELPM__
  #define LPM_HIGH(x,y) \
    do {                \
      int8u reg;        \
      reg = RAMPZ;      \
      y = __extended_load_program_memory(x); \
      RAMPZ = reg;      \
    } while(0)

//  #define LPM_HIGH __extended_load_program_memory
  #define FLASH_HIGH __farflash
#else
  #define LPM_HIGH(x,y) \
    do { \
      y = __load_program_memory(x); \
    } while(0)

  #define FLASH_HIGH __flash
#endif

// ATMega128 / 64 differences

#ifdef MEGA64
#define ADCSR ADCSRA  // For ATMEGA64, it is called ADCSRA
#endif

#endif // DOXYGEN_SHOULD_SKIP_THIS



#endif //__128_MICRO_H__
