/** @file hal/micro/avr-atmega/16/micro.h
 * @brief Utility and convenience functions for ATmega16 microcontroller
 *
 * Unless otherwise noted, the EmberNet stack does not use these functions,
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 *
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */


#ifndef __16_MICRO_H__
#define __16_MICRO_H__

#ifndef DOXYGEN_SHOULD_SKIP_THIS

#ifndef __MICRO_H__
#error do not include this file directly - include micro/micro.h
#endif

#define ENABLE_BIT_DEFINITIONS
#include <ioavr.h>

#include "micro/avr-atmega/16/reserved-ram.h"

#endif // DOXYGEN_SHOULD_SKIP_THIS


/** \b Note: \b DEPRECATED. Use CLEARBIT() in ember.h. */
#define INPUTPIN(reg, bit)	  reg &= ~(1<<bit)

/** \b Note: \b DEPRECATED. Use CLEARBITS() in ember.h. */
#define INPUTPINS(reg, bits)	reg &= ~(bits)

/** \b Note: \b DEPRECATED. Use SETBIT() in ember.h. */
#define OUTPUTPIN(reg, bit)	  reg |= 1<<bit

/** \b Note: \b DEPRECATED. Use SETBITS() in ember.h. */
#define OUTPUTPINS(reg, bits)	reg |= bits

/** @name Flash Accessor Macros and Functions
 * \b Note: All int16u values are little-endian.
 */
//@{

#if __HAS_ELPM__
  #define microReadFlashByte(addr)   __extended_load_program_memory(addr)
  #define microReadFlashWord(addr)   HIGH_LOW_TO_INT( \
                                      __extended_load_program_memory((addr)+1), \
                                      __extended_load_program_memory(addr) )
#else // __HAS_ELPM__

/* @description Reads a byte from flash.
 * 
 * @param addr: A 16-bit address in program space.
 *  
 * @return An int8u value of the byte in flash.
 */
  #define microReadFlashByte(addr)   __load_program_memory(addr)

/* @description Reads a word from flash.
 * 
 * @param addr: A 16-bit address in program space.
 * 
 * @return An int16u value of the word in flash.
 */
#define microReadFlashWord(addr)   HIGH_LOW_TO_INT( \
                                      __load_program_memory((addr)+1), \
                                      __load_program_memory(addr) )
#endif  // __HAS_ELPM__

/* @description Returns the address of the end of flash memory. */ 
#define microGetFlashEnd()         FLASHEND

//@}  // end of Flash Accessor Macros, Functions, and Constants


/* @name EEPROM Accessor Macros, Functions, and Constants
 * \b Note: All int16u values are little-endian.
 */
//@{

/* @description Returns a byte from EEPROM memory space.
 * 
 * @param addr: A 16-bit address in EEPROM memory space.
 * 
 * @param var: An int8u value of the byte in EEPROM. 
 */
#define microReadEEPROMByte(addr,var)  __EEGET(var,addr)

/* @description Returns a word from EEPROM memory space.
 * 
 * @param add: A 16-bit address in EEPROM memory space.
 * 
 * @param var: An int16u value of the word in EEPROM.
 */
#define microReadEEPROMWord(addr,var)   \
        do {                            \
        int8u h,l;                      \
        __EEGET(h,(addr)+1);            \
        __EEGET(l,addr);                \
        var = HIGH_LOW_TO_INT( h,l );   \
        } while(0)

/* @description Writes a byte to EEPROM memory space.
 * 
 * @param addr: A 16-bit address in program space.
 *
 * @param var: The int8u value of the byte to write to EEPROM.
 */
#define microWriteEEPROMByte(addr,var)  \
        do {                            \
        __EEPUT(addr,var);              \
        } while(0)

/* @description Writes a word to EEPROM memory space.
 * 
 * @param addr: A 16-bit address in program space.
 * 
 * @param var: The int16u value of the word to write to EEPROM.
 */
#define microWriteEEPROMWord(addr,var)  \
        do {                            \
        __EEPUT(addr+1,HIGH_BYTE(var)); \
        __EEPUT(addr,LOW_BYTE(var));    \
        } while(0)

/* @description Returns the address of the end of EEPROM memory space */ 
#define microGetEEPROMEnd()        (E2END - 1)

/* @description Returns the address of the start of EEPROM memory space. */ 
#define microGetEEPROMStart()      0

//@}  // end of EEPROM Accessor Functions


/* @description Gives the watchdog reset a friendlier name. */
#define watchdogReset() __watchdog_reset()

#ifndef DOXYGEN_SHOULD_SKIP_THIS

// executes full calibration
void rcCalibrate(void);

#if __HAS_ELPM__
  #define LPM_HIGH(x,y) \
    do {                \
      int8u reg;        \
      reg = RAMPZ;      \
      y = __extended_load_program_memory(x); \
      RAMPZ = reg;      \
    } while(0)

//  #define LPM_HIGH __extended_load_program_memory
  #define FLASH_HIGH __farflash
#else
  #define LPM_HIGH(x,y) \
    do { \
      y = __load_program_memory(x); \
    } while(0)

  #define FLASH_HIGH __flash
#endif

#endif // DOXYGEN_SHOULD_SKIP_THIS

#endif //__16_MICRO_H__
