/*
 * File: iar.h
 * Description: Compiler/Platform specific definitions and typedefs for the
 *  IAR AVR C compiler
 *
 * This file should be included first in all source files by setting the
 *  preprocessor macro PLATFORM_HEADER to point to it
 *
 * Author(s): Lee Taylor, lee@ember.com
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */
#ifndef __IAR_H__
#define __IAR_H__

#ifndef __ICCAVR__
  #error Improper PLATFORM_HEADER
#endif

  #include <inavr.h>

  // Typedefs to make the size of variable declarations well known
  typedef unsigned char  boolean;
  typedef unsigned char  int8u;
  typedef signed   char  int8s;
  typedef unsigned int   int16u;
  typedef signed   int   int16s;
  typedef unsigned long  int32u;
  typedef signed   long  int32s;
  typedef unsigned int  PointerType;
  #if __HAS_ELPM__
      #define PGM     __farflash
      #define PGM_P   const char __farflash *
      #define PGM_PU  const unsigned char __farflash *
    #else
      #define PGM     __flash
      #define PGM_P   const char __flash *
      #define PGM_PU  const unsigned char __flash *
    #endif
  // Sometimes a second PGM is needed in a declaration.  Having two 'const'
  // declarations generates a warning so we have a second PGM that turns
  // into nothing under gcc.
  #define PGM_NO_CONST PGM

  #define NO_STRIPPING  __root
  #define EEPROM __eeprom

  // The __SOURCEFILE__ macro is used by asserts to list the filename
  //  If it isn't otherwise defined, set it to the compiler intrinsic
  //  which specifies the whole filename and path of the sourcefile
  #ifndef __SOURCEFILE__
    #define __SOURCEFILE__ __FILE__
  #endif

  /*! see micro.h */
  #undef assert
  /*! redefined to call boardAssertFailed() see micro.h */
  #ifdef SIMPLER_ASSERT_REBOOT
      #define assert(condition) \
            do { if( !(condition) ) while(1){} } while(0)
  #else
    #ifndef LEAF_NODE
      void boardAssertFailed(PGM_P filename, int linenumber);
      #define assert(condition)                                   \
            do {                                          \
              if (! (condition))                              \
                boardAssertFailed(__SOURCEFILE__, __LINE__); } while(0)
    #else //LEAF
      void microReboot(void);
      #define assert(condition) \
            do { if( !(condition) ) microReboot(); } while(0)
    #endif
  #endif
            
#ifndef BOOTLOADER
  #undef __delay_cycles
  /*!
    __delay_cycles() is an intrinsic IAR WB call; however, we have explicity
    disallowed it since it is too specific to the system clock.
    \note Please use microDelayMicroSeconds() instead, because it correctly
      accounts for various system clock speeds.
  */
  #define __delay_cycles(x)  please_use_microDelayMicroSeconds_instead_of_delay_cycles
#endif

  /*!
    Macro to reset the watchdog timer.  Note:  be very very careful when
    using this as you can easily get into an infinite loop if you are not
    careful.
  */
  #define RESET_WATCHDOG_TIMER()  __watchdog_reset()
  

////////////////////////////////////////////////////////////////////////////////
/*! \name Global Interrupt Manipulation Macros
*/
////////////////////////////////////////////////////////////////////////////////
//@{


#ifdef BOOTLOADER

#ifndef DOXYGEN_SHOULD_SKIP_THIS
  // The bootloader does not use interrupts

  #define DECLARE_INTERRUPT_STATE int8u _emIsrState
  #define DISABLE_INTERRUPTS() do { } while(0)
  #define RESTORE_INTERRUPTS() do { } while(0)
  #define INTERRUPTS_ON() do { } while(0)
  #define INTERRUPTS_OFF() do { } while(0)
  #define INTERRUPTS_ARE_OFF() do { } while(0)
  #define ATOMIC(blah) { blah }
  #define HANDLE_PENDING_INTERRUPTS() do { } while(0)

#endif  // DOXYGEN_SHOULD_SKIP_THIS

#else


#ifndef DOXYGEN_SHOULD_SKIP_THIS
  // A series of macros for the diagnostics of the global interrupt state

  //#define INTERRUPT_DEBUGGING

  #ifdef INTERRUPT_DEBUGGING
    #define INT_DEBUG(x)  x
    #define ENABLE_BIT_DEFINITIONS
    #include <ioavr.h>
  #else
    #define INT_DEBUG(x)
  #endif


  #define DDR_I DDRG
  #define PORT_I PORTG
  #define IPIN0 BIT(0)    // we're inside a critical section
  #define IPIN1 BIT(1)    // state restored

#endif  // DOXYGEN_SHOULD_SKIP_THIS

  /*!
    This macro should be called in the local variable declarations section 
    of any function which calls DISABLE_INTERRUPTS() or RESTORE_INTERRUPTS().
  */
  #define DECLARE_INTERRUPT_STATE int8u _emIsrState

  /*!
    Disable interrupts, saving the previous state so it can be later restored
    with RESTORE_INTERRUPTS().
    \note Do not fail to call RESTORE_INTERRUPTS().
    \note It is safe to nest this call.
  */
  #define DISABLE_INTERRUPTS()  \
    _emIsrState = __save_interrupt(); \
    do {                      \
    INT_DEBUG(                \
      DDR_I |= (IPIN0|IPIN1);\
      PORT_I |= IPIN0;        \
      PORT_I &= ~IPIN1;       \
    )                         \
    __disable_interrupt();    \
    } while(0)
  
  /*! 
    Restore the global interrupt state previously saved by DISABLE_INTERRUPTS()
    \note Do not call without having first called DISABLE_INTERRUPTS() to have
      saved the state.
    \note It is safe to nest this call.
  */
  #define RESTORE_INTERRUPTS()  \
    do {                        \
    INT_DEBUG(                  \
      PORT_I &= ~IPIN0;         \
      if (_emIsrState & BIT(7)) \
        PORT_I |= IPIN1;        \
    )                           \
    __restore_interrupt(_emIsrState); \
    } while(0)

  /*! Enable global interrupts without regard to the current or previous state */
  #define INTERRUPTS_ON() \
    do {                            \
      INT_DEBUG(                    \
        PORT_I |= IPIN1;            \
      )                             \
    __enable_interrupt();           \
    } while(0)

  /*! Disable global interrupts without regard to the current or previous state */
  #define INTERRUPTS_OFF() \
    do {                            \
      INT_DEBUG(                    \
        PORT_I &= ~IPIN1;           \
      )                             \
    __disable_interrupt();          \
    } while(0)
 
  /*! \returns TRUE if the current global interrupt flag is enabled. */
  #define INTERRUPTS_ARE_OFF() (!(__save_interrupt() & BIT(7))) 

  /*!
    A block of code may be made atomic by wrapping it with this macro. 
    Something which is atomic cannot be interrupted by interrupts.
  */
  #define ATOMIC(blah)         \
  {                            \
    DECLARE_INTERRUPT_STATE;   \
    DISABLE_INTERRUPTS();      \
    { blah }                   \
    RESTORE_INTERRUPTS();      \
  }

  /*!
    Allows any pending interrupts to be executed. Usually this would be
    called at a safe point while interrupts are disabled (such as within 
    an ISR).
    
    Takes no action if interrupts are already enabled.
    
    The implementation accounts for an AVR related oddity, wherein one 
    instruction must be executed (a NOP in this instance) before global
    interrupts are allowed to occur.
  */
  #define HANDLE_PENDING_INTERRUPTS() \
    do {                          \
      if (INTERRUPTS_ARE_OFF()) { \
        INTERRUPTS_ON();          \
        __no_operation();         \
        INTERRUPTS_OFF();         \
      }                           \
   } while (0)

#endif // BOOTLOADER


////////////////////////////////////////////////////////////////////////////////
//@}  // end of Global Interrupt Manipulation Macros
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/*! \name Miscellaneous Macros
*/
////////////////////////////////////////////////////////////////////////////////
//@{

/*!
  define __attribute__ to nothing since it isn't handled by IAR
*/
#define __attribute__(foo)


/*!
  Declare a variable as unused to avoid a warning.
  Has no effect in IAR builds
*/
#define UNUSED

/*!
  Stubs for code not running in simulation.
*/
#define simulatedTimePasses()
#define simulatedTimePassesMs(x)
#define simulatedSerialTimePasses()

////////////////////////////////////////////////////////////////////////////////
//@}  // end of Miscellaneous Macros
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/*! \name C Standard Library Memory Utilities
    These should be used in place of the standard library functions.
    
    These functions have the same parameters and expected results as their C
    Standard Library equivalents but may take advantage of certain implementation
    optimizations.
    
    Unless otherwise noted, these functions are utilized by the EmberStack and are 
    therefore required to be implemented in the HAL. Additionally, unless otherwise
    noted, applications that find these functions useful may utilze them.
*/
////////////////////////////////////////////////////////////////////////////////
//@{

  /*! see c stdlib memcpy() */
  void microMemCopy(void *dest, void *src, int8u bytes);
  /*! see c stdlib memset() */
  void microMemSet(void *dest, int8u val, int8u bytes);
  /*! see c stdlib memcmp() */
  int8s emMicroMemCompare(void *source0, void *source1, int8u bytes);
  /*! works like c stdlib memcmp() but takes a flash space source parameter */
  int8s emMicroMemPGMCompare(void *source0, void PGM *source1, int8u bytes);
  #define MEMSET(d,v,l)  microMemSet(d,v,l)
  #define MEMCOPY(d,s,l) microMemCopy(d,s,l)
  #define MEMCOMPARE(s0,s1,l) emMicroMemCompare(s0, s1, l)
  #define MEMPGMCOMPARE(s0,s1,l) emMicroMemPGMCompare(s0, s1, l)

////////////////////////////////////////////////////////////////////////////////
//@}  // end of C Standard Library Memory Utilities
////////////////////////////////////////////////////////////////////////////////

#endif // __EMBER_CONFIG_H__
