/*
 * File: token.c
 * Description:  See token.h for details.
 *
 * Author(s): Jeff Mathews, jm@ember.com
 *
 * Copyright 2004 by Ember Corporation. All rights reserved.                *80*
 */


#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "stack/include/error.h"
#include "stack/include/hal.h"
#include "stack/include/packet-buffer.h"
#include "board/board.h"
#include "micro/micro.h"

#include "hal/micro/token.h"

/*
 normally, we would store the defaults into flash and write them into eeprom
 when we detected at run time that there was an issue with the values in eeprom.
 however, we are very code space constrained and we opt instead to store this
 data *only* in eeprom. we rely on the developer's ability to load the eeprom
 image and carefully avoid overwriting the vital manufacturing data
 */

#if defined(SIMULATOR)
#include <stdio.h>

// When running in the simulator, the tokens are defined here.
// When running on hardware, the tokens are defined in token-def.c.
// See token-def.c for the full explanation.
#define TOKEN_DEF(name, type, size, loc, ...)    \
  type TOKEN_ ## name size = __VA_ARGS__;
#include "stack/config/token-stack.h"
#undef TOKEN_DEF

#endif

#if defined(DEBUG_ON) || defined(SIMULATOR)
static void resetSystemTokens(void) {
/*
  this routine will prove useful when the developer changes the format of the
  stack related non-volatile data. for example while the developer is working on
  the code base.
  please note however, that this functionality isn't necessarily required in the
  embedded space, since an external script could easily handle the process of
  carefully reading, modifying, and re-uploading the critical manufacturing
  and/or stack related non-volatile data. thus - why it's only in the debug
  builds.
*/

// Must be careful here not to reset the manufacturing tokens to their default
// values.
#define TOKEN_MFG(name, type, size, loc, ...)
#define TOKEN_DEF(name, type, size, loc, ...)    \
{                                                \
  type data size = __VA_ARGS__;                  \
  setToken( (void*)&TOKEN_ ## name,              \
            (void*)&data,                        \
            SIZEOF_TOKEN(TOKEN_ ## name) );      \
}
#include "stack/config/token-stack.h"
#undef TOKEN_DEF
}
#endif

EmberStatus initTokens(void) {
#if !defined(HALTEST) && !defined(BOOTLOADER)

//Applications are never supposed to touch the mfg tokens,
//and I don't think we want to require customers to run 
//another "token-upgrade" sort of thing, so we wind up needing
//to put a small backward compatibility blip in initTokens()
//where the version numbers are checked.
  if(CURRENT_MFG_TOKEN_VERSION != GET_TOKEN(TOKEN_MFG_NVDATA_VERSION)) {
    if( GET_TOKEN(TOKEN_MFG_NVDATA_VERSION) == 0x01FE ) {
      SET_TOKEN(TOKEN_RADIO_CRYSTAL_OFFSET, 0);  //resonable default value
      SET_TOKEN(TOKEN_MFG_NVDATA_VERSION, 0x02FD);  //up the version
    }
  }
  
  if (CURRENT_MFG_TOKEN_VERSION != GET_TOKEN(TOKEN_MFG_NVDATA_VERSION)
      || CURRENT_STACK_TOKEN_VERSION != GET_TOKEN(TOKEN_STACK_NVDATA_VERSION)) {
#if defined(DEBUG_ON)
    // It would be nice if there were an emDebugError call.
    emberDebugPrintf("Err %x", EMBER_EEPROM_DATA_VERSION_MISMATCH);
    resetSystemTokens();
#elif defined(SIMULATOR)
    resetSystemTokens();
#else //  node release image
    // we have no other options, since the data is *only* in eeprom (never flash)
    return EMBER_EEPROM_DATA_VERSION_MISMATCH;
#endif
  }
#endif    
  return EMBER_SUCCESS;
}

void getToken(void *dst, void *src, int8u len)
{
  int8u i;
  int8u *ram = (int8u*)dst;
  int8u EEPROM *ee = (int8u EEPROM *)src;
  for ( i = 0; i < len; i++ )
    ram[i] = ee[i];
}

void setToken(void *dst, void *src, int8u len)
{
  int8u i;
  int8u *ram = (int8u*)src;
  int8u EEPROM *ee = (int8u EEPROM *)dst;
  for ( i = 0; i < len; i++ )
    ee[i] = ram[i];
}


// the following interfaces are admittedly code space hogs but serve
// only as temporary legacy glue interfaces for monitor service, studio and haltest

void * getTokenAddress( int16u creator ) {
  switch ( creator ) {

 #define CREATOR_DEF(tokenName, creator) \
    case CREATOR_TOKEN_ ## tokenName: return (void*)&TOKEN_ ## tokenName;
 #define TOKEN_DEF(name, type, size, loc, ...) 
#include "config/token-stack.h" 
  
  };    
#undef CREATOR_DEF  
#undef TOKEN_DEF
  return (void*)INVALID_EE_ADDRESS;
}

int8u getTokenSize( int16u creator ) {
  switch ( creator ) {
#define CREATOR_DEF(tokenName, creator)        \
    case CREATOR_TOKEN_ ## tokenName: return sizeof(TOKEN_ ## tokenName);
#define TOKEN_DEF(name, type, size, loc, ...)  
#include "config/token-stack.h"
  };
#undef CREATOR_DEF
#undef TOKEN_DEF
  return 0;  
}

/////////////////////////////////////////////
#if !defined(HALTEST) && !defined(LEAF_NODE)
/////////////////////////////////////////////

void copyLinkedBufferIntoToken( void *dst, EmberMessageBuffer buff, 
                                int8u messageOffset, int8u size )
{
  int8u j;
  int8u EEPROM *ee = (int8u EEPROM *)dst;  
  for (j = 0; j < size; j++) {
    int8u temp = emberGetLinkedBuffersByte(buff, messageOffset + j);
    ee[j] = temp;
  }  
}

void copyTokenIntoLinkedBuffer( EmberMessageBuffer buff, int16u creator,
                                   int8u messageOffset, int8u size )
{
  int8u j;
  int8u EEPROM *ee = (int8u EEPROM *)getTokenAddress( creator );
  if ( ee == (int8u EEPROM *)INVALID_EE_ADDRESS ) return;
  for (j = 0; j < size; j++) {
    int8u temp = ee[j];
    emberSetLinkedBuffersByte(buff, messageOffset + j, temp);
  }
}

/////////////////////////////////////////////
#endif
/////////////////////////////////////////////
