/**
 * File: uart.c
 * Description: HAL Uart abstraction
 *
 * Culprit(s):  Lee Taylor (lee@ember.com)
 *              Jeff Mathews (jeff@ember.com)
 *
 * Copyright 2004 by Ember Corporation.  All rights reserved.               *80*
 */

#include PLATFORM_HEADER
#include "stack/include/ember.h"
#include "stack/include/error.h"
#include "app/util/serial/serial.h"
#include "app/util/serial/hal-serial.h"
#include "hal/board/board.h"
#include "hal/micro/avr-atmega/uart.h"


#define INVALID 255 // frequency error exceeds +/-0.5%, given sysclock

// Set up for LOWLEVEL driver mode -- these defines are
// for convenience.
#if (EMBER_SERIAL0_MODE != EMBER_SERIAL_FIFO) && \
    (EMBER_SERIAL0_MODE != EMBER_SERIAL_BUFFER)
#define EM_SERIAL0_LOWLEVEL
#endif
     
#if (EMBER_SERIAL1_MODE != EMBER_SERIAL_FIFO) && \
    (EMBER_SERIAL1_MODE != EMBER_SERIAL_BUFFER)
#define EM_SERIAL1_LOWLEVEL
#endif
     
// Save flash if both are undefined
#if !defined(EM_SERIAL0_LOWLEVEL) || !defined(EM_SERIAL1_LOWLEVEL)
     
// avr datasheet table 83. U2X=0
#if defined(AVR4MHZ)
  const int8u PGM baudSettings[12] = { 
    103,      // 2400 0.2%
    51,       // 4800 0.2% 
    25,       // 9600 0.2%
    INVALID,  // 14.4k 2.1% 
    12,       // 19.2k 0.2% 
    INVALID,  // 28.8k -3.5%
    INVALID,  // 38.4k -7.0%
    4,        // 50.0k
    INVALID,  // 57.6k 8.5%
    INVALID,  // 76.8k 8.5%
    INVALID,  // 100.0k
    INVALID   // 115.2k 8.5%
  };
#elif defined(AVR8MHZ)
const int8u PGM baudSettings[12] = { 
  207,      // 2400 0.2%
  103,      // 4800 0.2% 
  51,       // 9600 0.2%
  INVALID,  // 14.4k -0.8% 
  25,       // 19.2k 0.2% 
  INVALID,  // 28.8k 2.1%
  12,       // 38.4k 0.2%
  9,        // 50.0k
  INVALID,  // 57.6k -3.5%
  INVALID,  // 76.8k -7.0%
  4,        // 100.0k 2.4%
  INVALID   // 115.2k 8.5%
};
#elif defined(AVR7p3728MHZ)
const int8u PGM baudSettings[12] = { 
  191,  // 2400 0.0%
  95,   // 4800 0.0% 
  47,   // 9600 0.0%
  31,   // 14.4k 0.0% 
  23,   // 19.2k 0.0% 
  15,   // 28.8k 0.0%
  11,   // 38.4k 0.0%
  INVALID, // 50.0k
  7,    // 57.6k 0.0%
  5,    // 76.8k 0.0%
  INVALID, // 100.0k
  3     // 115.2k 0.0%
};
#elif defined(AVR3p68MHZ)
const int8u PGM baudSettings[12] = { 
  95,  // 2400 0.0%
  47,   // 4800 0.0% 
  23,   // 9600 0.0%
  15,   // 14.4k 0.0% 
  11,   // 19.2k 0.0% 
  7,   // 28.8k 0.0%
  5,   // 38.4k 0.0%
  INVALID, // 50.0k
  3,    // 57.6k 0.0%
  2,    // 76.8k 0.0%
  INVALID, // 100.0k
  1     // 115.2k 0.0%
};
#elif defined(AVR1MHZ)
const int8u PGM baudSettings[12] = { 
  25,      // 2400 0.2%
  12,      // 4800 0.2% 
  INVALID, //6,       // 9600 -7.0%
  INVALID, //3,       // 14.4k 8.5%% 
  INVALID, //2,       // 19.2k 8.5%% 
  INVALID, //1,       // 28.8k 8.5%%
  INVALID, //1,       // 38.4k -18.6%
  INVALID, //0,       // 50.0k 25%
  INVALID, //0,       // 57.6k 8.5%%
  INVALID, // 76.8k
  INVALID, // 100.0k
  INVALID  // 115.2k
};
#else 
  #error no processor speed defined in board.h
#endif

#endif // LOWLEVEL

// Bit-vector to indicate which ports have transmitted data.
// The transmit complete flag (TXC) is only useful after the first transmission.
#if !defined(EM_SERIAL0_LOWLEVEL) || !defined(EM_SERIAL1_LOWLEVEL)
static int8u txStarted = 0;
#endif

#if !defined(EM_SERIAL0_LOWLEVEL) || !defined(EM_SERIAL1_LOWLEVEL)
EmberStatus emHalSerialInit(int8u port,
                            SerialBaudRate rate,
                            SerialParity parity,
                            int8u stopbits)
{
#if !defined(EM_SERIAL0_LOWLEVEL)
  int8u ucsrc;  // support for the mega8 (see below)
#endif

  if(rate > BAUD_115200 || baudSettings[rate] == INVALID) 
    return EMBER_SERIAL_INVALID_BAUD_RATE;

  // on the MEGA8, the UBRRH and the UCSRC registers share the same memory
  // location.  To write to the UCSRC, bit 7 must be a one.  The read cycle is very
  // cryptic by comparison.  The only way to read the UCSRC register on the MEGA8
  // is to read from this memory location in two successive clock cycles.  The
  // first read is always from UBRRH, the second read is from UCSRC.  
#if !defined(EM_SERIAL0_LOWLEVEL)
  if(port == 0) {
#if defined(MEGA8) || defined(MEGA16)
    ucsrc = UBRRH; // required on the MEGA 8.  Do not delete!
#endif
    ucsrc = UCSR0C;
#if defined(MEGA8) || defined(MEGA16)
    ucsrc |= BIT(URSEL);
#endif
    
    if(parity == PARITY_ODD) {
      ucsrc |= (BIT(UPM00) | BIT(UPM01));
    } else if(parity == PARITY_EVEN) {
      ucsrc |= BIT(UPM01);
      ucsrc &= ~(BIT(UPM00));
    }
    if(stopbits == 2) {
      ucsrc |= BIT(USBS0);
    }

    #if (MEGA8 || MEGA16)
      ucsrc |= BIT(UCSZ1) | BIT(UCSZ0);
    #else //everything else
      ucsrc |= BIT(UCSZ01) | BIT(UCSZ00);
    #endif
    
    UCSR0C = ucsrc;

    UBRR0L  = baudSettings[rate]; 
    // Enable UART transmitter and receiver
    UCSR0B = BIT(RXCIE0) | BIT(TXEN0) | BIT(RXEN0);
    // Clear(one) UART transmit complete bit (TXC)
    UCSR0A = BIT(TXC0);  // Flag cleared by writing a one
    // Clear txStarted flag for port 0
    txStarted &= ~BIT(0);
    return EMBER_SUCCESS;
  } 
#endif // SERIAL0

#if !defined(MEGA8) && !defined(MEGA168) && !defined(MEGA16) && !defined(EM_SERIAL1_LOWLEVEL)  
  if(port == 1) {
    if(parity == PARITY_ODD) {
      UCSR1C |= (BIT(UPM10) | BIT(UPM11));
    } else if(parity == PARITY_EVEN) {
      UCSR1C |= BIT(UPM11);
      UCSR1C &= ~(BIT(UPM10));
    }
    if(stopbits == 2) {
      UCSR1C |= BIT(USBS1);
    }
    UCSR1C |= BIT(2);
    UCSR1C |= BIT(1);
    UBRR1L  = baudSettings[rate];
//    UBRR1L  = (baudSettings[rate]/2);
    // Enable UART transmitter and receiver
    UCSR1B = BIT(RXCIE1) | BIT(TXEN1) | BIT(RXEN1);
    // Clear(one) UART transmit complete bit (TXC)
    UCSR1A = BIT(TXC1);  // Flag cleared by writing a one
    // Clear txStarted flag for port 1
    txStarted &= ~BIT(1);
    return EMBER_SUCCESS;
  } 
#endif

  return EMBER_SERIAL_INVALID_PORT;
}
#endif // If both ports are low-level

void emHalStartSerialTx(int8u port)
{
#if !defined(EM_SERIAL0_LOWLEVEL)
  if (port == 0) {
    // Enable Interrupt
    UCSR0B |= BIT(UDRIE0);
    txStarted |= BIT(0);
    return;
  }
#endif

#if !defined(EM_SERIAL1_LOWLEVEL)
  if (port == 1) {
    UCSR1B |= BIT(UDRIE1);
    txStarted |= BIT(1);
    return;
  }
#endif
}

void emHalStopSerialTx(int8u port)
{
  switch(port) {
#if EMBER_SERIAL0_MODE != EMBER_SERIAL_UNUSED
  case 0:
    // Disable Interrupt
    UCSR0B &= ~BIT(UDRIE0);
    break;
#endif
#if EMBER_SERIAL1_MODE != EMBER_SERIAL_UNUSED
  case 1:
    UCSR1B &= ~BIT(UDRIE1);
    break;
#endif
  }
}

#if !defined(EM_SERIAL0_LOWLEVEL)
// If there is a serial transmission error, the location in the q where
//  it occurred is remembered so that the error can be reported to the
//  application at the appropriate time.  If a second error occurs before the
//  first is reported we discard everything that arrived between the two errors
#pragma vector = USART0_RXC_vect
__interrupt void emHalUart0RxIsr(void)
{
  EmSerialFifoQueue *q = emSerialRxQueues[0];
  int8u errors = UCSR0A & (BIT(FE0) | BIT(DOR0) | BIT(UPE0));
  int8u incoming = UDR0;

  if((errors == 0) && (q->used < (EMBER_SERIAL0_RX_QUEUE_SIZE-1))) {
    FIFO_ENQUEUE(q,incoming,emSerialRxQueueMasks[0]);
  } else {
    //translate error code
    if( (errors == 0) || (errors & BIT(DOR0)) )
      errors = EMBER_SERIAL_RX_OVERFLOW;
    else if( (errors & BIT(FE0)) )
      errors = EMBER_SERIAL_RX_FRAME_ERROR;
    else if( (errors & BIT(UPE0)) )
      errors = EMBER_SERIAL_RX_PARITY_ERROR;
    else // unknown
      errors = EMBER_ERR_FATAL;
    
    // save error code & location in queue
    if(emSerialRxError[0] == EMBER_SUCCESS) {
      emSerialRxErrorIndex[0] = q->head;
      emSerialRxError[0] = errors;
    } else {
      // Flush back to previous error location & update value
      q->head = emSerialRxErrorIndex[0];
      emSerialRxError[0] = errors;
      q->used = ((q->head - q->tail) & emSerialRxQueueMasks[0]);
    }
  }
}
#endif

#if EMBER_SERIAL0_MODE == EMBER_SERIAL_FIFO
#pragma vector = USART0_UDRE_vect
__interrupt void emHalUart0TxIsr(void)
{
  EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[0];

  if(q->used > 0) {
    // Clear TX complete flag (TXCn) before transmission (before UDR is written)
    // so that it can be checked by emHalWaitSerialTxComplete.
    UCSR0A = BIT(TXC0);  // Flag cleared by writing a one
    UDR0 = FIFO_DEQUEUE(q,emSerialTxQueueMasks[0]);
  } else {
    // nothing to send, disable interrupt
    UCSR0B &= ~(BIT(UDRIE0));
  }    
}
#elif EMBER_SERIAL0_MODE == EMBER_SERIAL_BUFFER
#pragma vector = USART0_UDRE_vect
__interrupt void emHalUart0TxIsr(void)
{
  EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[0];

  if(q->nextByte == NULL) {
    if(q->used > 0) {
      // new message pending, but nextByte not set up yet
      emSerialBufferNextMessageIsr(q);
    } else {
      // Nothing to send, disable interrupt & return
      UCSR0B &= ~(BIT(UDRIE0));
      return;
    }
  }

  UCSR0A = BIT(TXC0);  // Flag cleared by writing a one
  UDR0 = *q->nextByte;
  if(q->lastByte != q->nextByte)
    q->nextByte++;
  else
    emSerialBufferNextBlockIsr(q,0);
}
#endif

#if !defined(EM_SERIAL1_LOWLEVEL)
#pragma vector = USART1_RXC_vect
__interrupt void emHalUart1RxIsr(void)
{
  EmSerialFifoQueue *q = emSerialRxQueues[1];
  int8u errors = UCSR1A & (BIT(FE1) | BIT(DOR1) | BIT(UPE1));
  int8u incoming = UDR1;

  if((errors == 0) && (q->used < (EMBER_SERIAL1_RX_QUEUE_SIZE-1))) {
    FIFO_ENQUEUE(q,incoming,emSerialRxQueueMasks[1]);
  } else {
    //translate error code
    if( (errors == 0) || (errors & BIT(DOR1)) )
      errors = EMBER_SERIAL_RX_OVERFLOW;
    else if( (errors & BIT(FE1)) )
      errors = EMBER_SERIAL_RX_FRAME_ERROR;
    else if( (errors & BIT(UPE1)) )
      errors = EMBER_SERIAL_RX_PARITY_ERROR;
    else // unknown
      errors = EMBER_ERR_FATAL;
    
    // save error code & location in queue
    if(emSerialRxError[1] == EMBER_SUCCESS) {
      emSerialRxErrorIndex[1] = q->head;
      emSerialRxError[1] = errors;
    } else {
      // Flush back to previous error location & update value
      q->head = emSerialRxErrorIndex[1];
      emSerialRxError[1] = errors;
      q->used = ((q->head - q->tail) & emSerialRxQueueMasks[1]);
    }
  }
}
#endif

#if EMBER_SERIAL1_MODE == EMBER_SERIAL_FIFO
#pragma vector = USART1_UDRE_vect
__interrupt void emHalUart1TxIsr(void)
{
  EmSerialFifoQueue *q = (EmSerialFifoQueue *)emSerialTxQueues[1];
  if(q->used > 0) {
    // Clear TX complete flag (TXCn) before transmission (before UDR is written)
    // so that it can be checked by emHalWaitSerialTxComplete.
    UCSR1A = BIT(TXC1);  // Flag cleared by writing a one
    UDR1 = FIFO_DEQUEUE(q,emSerialTxQueueMasks[1]);
  } else {
    // nothing to send, disable interrupt
    UCSR1B &= ~(BIT(UDRIE1));
  }    
}
#elif EMBER_SERIAL1_MODE == EMBER_SERIAL_BUFFER
#pragma vector = USART1_UDRE_vect
__interrupt void emHalUart1TxIsr(void)
{
  EmSerialBufferQueue *q = (EmSerialBufferQueue *)emSerialTxQueues[1];

  if(q->nextByte == NULL) {
    if(q->used > 0) {
      // new message pending, but nextByte not set up yet
      emSerialBufferNextMessageIsr(q);
    } else {
      // Nothing to send, disable interrupt & return
      UCSR1B &= ~(BIT(UDRIE1));
      return;
    }
  }

  UCSR1A = BIT(TXC1);  // Flag cleared by writing a one
  UDR1 = *q->nextByte;
  if(q->lastByte != q->nextByte)
    q->nextByte++;
  else
    emSerialBufferNextBlockIsr(q,1);
}
#endif



// full blocking, no queue overflow issues, can be used in or out of int context
// does not return until character is transmitted.
// this is useful for things like assert messages
void emHalForceWriteSerialByte(int8u port, int8u dataByte)
{
#if !defined(EM_SERIAL0_LOWLEVEL)
  if (port == 0) {
    // Block until data register empty
    while ( !( UCSR0A & BIT(UDRE0)) ) {}
    // Set txStarted flag for port 0
    txStarted |= BIT(0);
    // Clear TX complete flag (TXCn) before UDR is written
    UCSR0A = BIT(TXC0);  // Flag cleared by writing a one
    UDR0 = dataByte;
    // Block until TX complete
    while( !(UCSR0A & BIT(TXC0)) ) { }
    return;
  }
#endif

#if !defined(EM_SERIAL1_LOWLEVEL)
  if (port == 1) {
    // Block until data register empty
    while ( !( UCSR1A & BIT(UDRE1)) ) {}
    // Set txStarted flag for port 1
    txStarted |= BIT(1);
    // Clear TX complete flag (TXCn) before UDR is written
    UCSR1A = BIT(TXC1);  // Flag cleared by writing a one
    UDR1 = dataByte;
    // Block until TX complete
    while( !(UCSR1A & BIT(TXC1)) ) { }
    return;
  }
#endif
}


// blocks until the text actually goes out
// Checks the flag TXCn (USART Transmit Complete)
// This flag bit is set when the entire frame in the Transmit Shift Register has
// been shifted out and there are no new data currently present in the transmit 
// buffer (UDRn). The TXCn flag bit is automatically cleared when a transmit 
// complete interrupt is executed, or it can be cleared by writing a one to its 
// bit location. The TXCn flag can generate a Transmit Complete interrupt.
void emHalWaitSerialTxComplete(int8u port)
{
   __watchdog_reset();
#if !defined(EM_SERIAL0_LOWLEVEL)
   if (port == 0) {
     // if no data has been transmitted on this port, then TXC will never be set
     // so blocking on it would be bad.
     if (txStarted & BIT(0)) {
       // last byte was shifted out of shift register
       while( !(UCSR0A & BIT(TXC0)) ) { }
     }
     return;
   }
#endif

#if !defined(EM_SERIAL1_LOWLEVEL) 
   if (port == 1) {
     if (txStarted & BIT(1)) {
       while( !(UCSR1A & BIT(TXC1)) ) { }
     }
     return;
   }
#endif
}

void emberUart0WriteByteTiny(int8u byte)
{
  while ( !( UCSR0A & BIT(UDRE0)) ) {}
  UDR0 = byte;
}

EmberStatus emberUart0ReadByteTimeoutTiny(int8u *dataByte)
{
//a timer costs 20 bytes
//error detection costs 6 bytes
/*
    while( !(UCSR0A & BIT(RXC0)) ) {}
    if( !(UCSR0A & (BIT(FE0) | BIT(DOR0) | BIT(UPE0))) ) {
      *dataByte = UDR0;
      return EMBER_SUCCESS;
    } else {
      *dataByte = UDR0;
      return EMBER_SERIAL_RX_EMPTY;
    }
*/
  int16u count = 0;
  int8u bah = 0;

  // We do this because the do/while loop is not as optimized at
  // the top end, which is where we really need it to be
 try_again:
  if (UCSR0A & BIT(RXC0)) {
    int8u error = UCSR0A & (BIT(FE0) | BIT(DOR0) | BIT(UPE0));
    *dataByte = UDR0;
    return error;
  }
  if (++count == 0) {
    __watchdog_reset();
    if (++bah == 4) { // 80ms
      return EMBER_SERIAL_RX_EMPTY;
    } 
  }
  goto try_again;
}

