/** @file hal/micro/micro.h
 * @brief Functions common across all microcontroller-specific
 * files.
 *
 * Unless otherwise noted, the EmberNet stack does not use these functions, 
 * and therefore the HAL is not required to implement them. However, many of
 * the supplied example applications do use them.
 *
 * <b>About the Program Counter (PC) Diagnostics Utilities</b>
 *  
 * Use these utilities with microcontrollers whose watchdog circuits do not make 
 * it possible to find out where they were executing when a watchdog was triggered.
 * Enabling the PC diagnostics enables a periodic timer 
 * interrupt that saves the current PC to a reserved, unitialized memory address.  
 * This address can be read after a  reset.
 *
 * Some functions in this file return an EmberStatus value. See 
 * error-def.h for definitions of all EmberStatus return values.
 *
 * NOTE: Ember uses the term SPD (Start Packet Delimiter) instead of the term 
 *       SFD (Start Frame Delimiter).
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */

#ifndef __MICRO_H__
#define __MICRO_H__

#ifndef DOXYGEN_SHOULD_SKIP_THIS


#if defined( SIMULATOR )
  #include "avr-atmega/128/micro-sim.h"
#else
  #if defined(MEGA64) || defined(MEGA128)
    #include "avr-atmega/128/micro.h"
  #elif defined(MEGA16)
    #include "avr-atmega/16/micro.h"
  #elif defined(MEGA169)
    #include "avr-atmega/169/micro.h"
  #elif defined(MEGA8)
    #include "avr-atmega/8/micro.h"
  #elif defined(MEGA168)
    #include "avr-atmega/168/micro.h"
  #else
    // default, assume 128
    #include "avr-atmega/128/micro.h"
  #endif
#endif

#endif // DOXYGEN_SHOULD_SKIP_THIS

/** @description Initializes microcontroller-specific peripherals.
*/
void microBoot(void);

/** @description Gets information about what caused the microcontroller to reset.  
 * This information is provided to the stack when it is initialized.
 *  
 * @appusage Call this function as soon after reset as possible, to ensure 
 * that the proper information is returned.
 * 
 * @return An enumeration defined in ember.h.
 */
EmberResetType microGetResetInfo(void);

/** @description Calls microGetResetInfo() and supplies a string describing it.
 *
 * @appusage Useful for diagnostic printing of text just after program 
 * initialization.
 *
 * @return A pointer to a program space string.
 */
void* microGetResetString(void);

/** @description Restarts the microcontroller and therefore everything else. */
void microReboot(void);

/** @description Initializes the system tick.
 *
 * @return Time to update the async registers after RTC is started (units of 100 
 * microseconds).
 */
int16u microStartSystemTick(void);

/** @description Initializes the MAC timer. */
void microStartMacTick(void);

/** \showinitializer */ 
#define MICRO_DISABLE_WATCH_DOG_KEY 0xA5

/** @description Disables the watchdog timer.
 *
 * \b Note: To prevent the watchdog from being disabled accidentally, 
 * a magic key must be provided.
 * 
 * @param magicKey: A value (MICRO_DISABLE_WATCH_DOG_KEY) that enables the function.
 */
void microDisableWatchDog(int8u magicKey);

/** @description  Should be called by the radio HAL whenever the
 * SPD interrupt or radio that supports it is triggered.
 */
void microSpdIsr(void);

/** @description Enumerations for the possible microcontroller sleep modes. */
typedef enum
{
   SLEEPMODE_IDLE,
   SLEEPMODE_RESERVED,
   SLEEPMODE_POWERDOWN,
   SLEEPMODE_POWERSAVE
} SleepModes;

/** @description Puts the microcontroller to sleep in a specified mode.
 * 
 * @param sleepMode: A microcontroller sleep mode (see SleepModes).
 */
void microSleep(SleepModes sleepMode);

/** @description Starts a timer that initiates sampling the 
 * microcontroller's program counter.
 */
void microStartPCDiagnostics(void);

/** @description Stops the timer that was started by 
 * microStartPCDiagnostics().
 */
void microStopPCDiagnostics(void);

/** @description Provides a captured PC location from the last run. 
 *   
 * @appusage Call this function when an application starts, before
 * calling microStartPCDiagnostics().
 *  
 * @return The most recently sampled program counter location. 
 */
int16u microGetPCDiagnostics(void);

/** @description Enumerations for possible bootloader trap 
 * identifiers.
 */
typedef enum
{
  /** \b DEPRECATED */
  BOOT_TRAP_APP_START = 1,    
  /** See flashWriteBlock(). */
  BOOT_TRAP_FLASH_WRITE = 2,
  /** \b DEPRECATED */
  BOOT_TRAP_FLASH_ADJUST = 3
} BootloaderTraps;

/** @description Calls the bootloader trap functions.
 *  
 * @param trap: Trap identifier.
 * 
 * @param ...: Parameter list corresponding to the trap.
 *  
 * @return  An error if the trap identifier is invalid, or EMBER_SUCCESS.
 */
EmberStatus microBootloaderTrap(BootloaderTraps trap, ...);

/** \showinitializer */ 
#define MICRO_LAUNCH_BOOTLOADER_KEY 0x5A

/** @description Quits the current application and launches the bootloader
 * in a non-over-the-air mode (usually the serial port).
 *   
 * @param magicKey: A value (MICRO_LAUNCH_BOOTLOADER_KEY) that enables
 * the bootloader, preventing it from being accidentally launched.
 */
void microLaunchBootloader(int8u magicKey);

#endif //__MICRO_H__

