/*
 File: token.h

 Description:
 Tokens are indexed by a two byte "creator code".
  Token query takes some time, so it is envisioned that code will load the 
  tokens they are interested in only once through their init/boot functions.
  For memory conservation reasons, tokens are not cached to RAM (although
  the location of the last token is).
  Any necessity a user might have for ram cache will be implemented on their
  part. If there are multiple and active users of a particular token, a cache
  accessor function should be made available by a third party.

  There are three token storage locations:
  1. Default Token Table. This is a relocatable segment constant table in the
     program space of flash.  Tokens which are stored here are only adjusted 
     at build time.  *All tokens DO NOT have to be listed here*
  2. EEPROM Override Token Table. These are adjustable programmatically. Tokens
     which are "often" adjusted should be stored here. (boot count, channel). 
     This is also where Application Tokens are stored.
  3. Upper Flash Override Token Table. These are not adjustable programmatically.
     These are programmed via the bootloader functionality usually just at 
     manufacture time. This is a repository for things like the node id and IEEE
     64bit address. Things we want to force and/or set and forget.

  If a Token exists in multiple tables, the Read Token Priority Rules apply.
  Users can override emberstack tokens in EEPROM and use the EEPROM for general
  use. However, the Upper Flash Override Token Table always takes precedence.

  None of these interfaces should be called from interrupt context.

  Note, the last byte in the eeprom is reserved for a low level boot_count

  Author(s): Jeff Mathews, jm@ember.com
  Copyright 2004 by Ember Corporation. All rights reserved.                 *80*
*/


#ifndef TOKEN_H
#define TOKEN_H

#define TOKEN_CREATOR_INVALID_MASK  0xFFF0
#define TOKEN_CREATOR_DESTROYED     0xFFF8
#define TOKEN_CREATOR_CACHE_INVALID 0xFFF9
#define TOKEN_CREATOR_SIGNATURE     0xFFFA
#define TOKEN_CREATOR_SEARCH_FIRST  0xFFFB
#define TOKEN_CREATOR_SEARCH_RESET  0xFFFC
#define TOKEN_CREATOR_SEARCH_NEXT   0xFFFD
#define TOKEN_CREATOR_DELETED       0xFFFE
#define TOKEN_CREATOR_END_1         0xFFFF
#define TOKEN_CREATOR_END_0         0x0000


// token permissions as used in token-names.h
#define TOKEN_READABLE    0x01
#define TOKEN_WRITEABLE   0x02
#define TOKEN_NO_AIR      0x04
#define TOKEN_INTERNAL_ONLY     0
#define TOKEN_READ_ONLY         TOKEN_READABLE
#define TOKEN_WRITE_ONLY        TOKEN_WRITEABLE
#define TOKEN_READ_WRITE        (TOKEN_READABLE|TOKEN_WRITEABLE)

typedef struct {
  int16u creator;   // secret
  int8u* addr;      // secret
  int8u  table;     // secret
  int8u  size;      // secret
} TokenIterator;

// Use of a token cache allows the user to avoid looking up a token every time
// it is used.

typedef struct {
  int8u* addr;      // secret
  int8u  table;     // secret
  int8u  size;      // secret
} TokenCache;

/*
  Read Token priority rules: 
    1.Upper Flash Override Token Table
    2.EEPROM Override Token Table
    3.Default Token Table
*/
#define TOKEN_TABLE_UPPER_FLASH     0
#define TOKEN_TABLE_EEPROM          1
#define TOKEN_TABLE_DEFAULT_FLASH   2
#define TOKEN_TABLE_LAST            2


// returns true if the EEPROM format seems to be the creator system style.
boolean eeFormatIsOld(void);

// Use this api to determine where the token's values are coming from. 
// For instance, if you're setting the channel in EEPROM, but there is an 
// override in the upper flash table that's taking precedence - it's 
// informative to get this knowledge out to a human via ember-view.
EmberStatus retrieveTokenTableType(int16u creator, int8u* table);

// Search all tables for the token, and return the size.
EmberStatus retrieveTokenSize(int16u creator, int8u* size);

// Search all tables for the token, and copy qty token.size
// bytes into the destination RAM.
// Returns err if token can't be found, or if token.size
// doesn't match the expected checkSize.
EmberStatus retrieveTokenData(int16u creator, int8u* dst, int8u checkSize);

// Search all tables for the token, and copy length number bytes of data from 
// the token starting at offset length bytes.
// Returns err if token can't be found, or if length+offset > token.size
EmberStatus retrieveTokenDataChunk(int16u creator, int8u* dst, int8u offset, int8u length);

// Use this api to retrieve the list of token creators currently in use.
// Call initially with creator set to TOKEN_CREATOR_SEARCH_RESET
// Call repeated until an error is returned.
// Note: will return multiple creators if they exist in multiple tables.
EmberStatus retrieveNextTokenCreator(int16u* creator, TokenIterator* iterator);

// Writes token.size bytes from src RAM addr into the token.
// Returns err,  if token.size doesn't match the expected checkSize.
// If the token doesn't exist, it creates it.
// Note: operates only on EEPROM table (the existence of a token of the
// same creator in the Upper Flash Override Token Table will take precedence)
EmberStatus modifyToken(int16u creator, int8u *src, int8u checkSize);

// Writes length bytes from src RAM addr into the token at offset.
// Returns err,  if the token doesn't exist or if length+offset > token.size
// Note: operates only on EEPROM table (the existence of a token of the
// same creator in the Upper Flash Override Token Table will take precedence)
EmberStatus modifyTokenChunk(int16u creator, int8u *src, int8u offset, int8u length);

// Marks a token deleted  (watch out for fragmentation)
// Returns err if the token isn't found
// Note: operates only on EEPROM table.
EmberStatus deleteToken(int16u creator);

// Create new token entry.
// Returns err if the token already exists, or there is no space.
// Note: operates only on EEPROM table.
EmberStatus createToken(int16u creator, int8u size);

// Flushes the EEPROM of all tokens which are not Emberstack related.
// Returns the number of tokens removed.
int8u deleteApplicationTokens(void);

// Scrubs and reinit the EEPROM. (note, takes about 45sec for the scrub)
void scrubTokenEEPROM(void);


#endif

