/** 
 * @file hal/micro/token.h
 * @brief Token system for storing non-volatile information.
 *
 * The token system stores such non-volatile information as the manufacturing ID, channel
 * number, transmit power, and various pieces of information that the 
 * application developer needs to be persistent between power cycles of the 
 * device. Tokens are stored in the EEPROM of the ATmega.
 *
 * There are three main types of tokens: 
 * - <b>Manufacturing tokens:</b> Tokens that are set at the factory and 
 *   cannot be changed through normal software operations.
 * - <b>Stack-level tokens:</b> Tokens that can be changed via the 
 *   appropriate stack API calls.  
 * - <b>Application level tokens:</b> Tokens that can be set via the  
 *   token system API calls in this file.
 *
 * The token system API controls writing non-volatile data to
 * and reading non-volatile data from the EEPROM. If an application wishes to
 * use non-volatile memory, it must do so by creating its own token header file
 * similar to this one. The macro APPLICATION_TOKEN_HEADER (see token-stack.h) 
 * should be defined to equal the name of the header file in which application 
 * tokens are defined.
 * 
 * Because the token system is based on memory location within the ATmega
 * EEPROM, it is necessary to keep track of the token version number.  Otherwise,
 * the token information can get out of sync. If the application or stack changes
 * tokens, the token version should change as well, and all affected tokens
 * should be reinitialized. The application must specify the memory location for
 * any tokens that it defines. The location of the first application token must
 * be 384 (0x180). All locations below 384 are reserved for manufacturing and
 * stack-level tokens.
 *
 * A token definition is constructed as follows:
 * TOKEN_DEF( TOKEN_NAME, TYPE, SIZE, LOCATION, DEFAULT_VALUES )
 * where TYPE is the type of the data in the token, SIZE is the array size and
 * should be left blank for scalar tokens, LOCATION is the address at which
 * the token will be stored, and DEFAULT_VALUES is the value to
 * which the token is set upon initialization.
 * 
 * Here are two example token definitions:
 * - TOKEN_DEF( CHANNEL, int8u, , 134, 0x04 ): Defines a token called
 *   TOKEN_CHANNEL, which is an int8u.  Because it is
 *   a scalar, we do not include the array size.  The token will be stored at
 *   address 134 (0x86) in the EEPROM. If this token is initialized, it will be
 *   set to the value 0x04.
 *
 * - TOKEN_DEF( ENCRYPTION_KEY, int8u, [16], 135, {0,} ):  Defines a token 
 *   called TOKEN_ENCRYPTION_KEY, which is a 16-element array
 *   of type int8u. The token will be stored at address 135 (0x87) in the
 *   EEPROM.  If this token is initialized, it will be set to all zeros.  
 *
 * See token-stack.h to see the default set of tokens and their values.
 *
 * A note on token initialization: The compiler will create an ebin file
 * that can be uploaded serially using the bootloader. It is also possible to
 * create an application that will write the initialized values to the EEPROM
 * (see the sample application token-init for more details), which can be used
 * to initialize tokens over the air.  
 *
 * Some functions in this file return an EmberStatus value. See 
 * stack/include/error-def.h for definitions of all EmberStatus values.
 *
 * <!--DC Seward, dc@ember.com-->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved. -->
 */

#ifndef __TOKEN_H__
#define __TOKEN_H__

#if defined(BOOTLOADER)

#define TOKEN_DEF(name, type, size, loc, ...)       \
  TOKEN_ ## name,
  enum {
    #include "config/token-stack.h"
  };
#undef TOKEN_DEF

int16u GET_TOKEN( int index );
                  
#else

/**
 * @description Macro for translating token definitions into local variables
 * that point to the correct location in the EEPROM.
 *
 * @param name: The name of the token.
 *
 * @param type: The type definition of the token.
 *
 * @param size: An optional parameter for arrays that will declare the size.
 * \b Note: You must put square braces around this parameter.  See 
 * token-stack.h for an example.
 *
 * @param loc: The address in EEPROM at which the token will be stored.
 *
 * @param ...: The default token value used at compile time to create an
 * .ebin file to be uploaded via the bootloader, to initialize the tokens.
 */
#define TOKEN_DEF(name, type, size, loc, ...)    \
  extern type EEPROM TOKEN_ ## name size;
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#include "stack/config/token-stack.h"
#endif //DOXYGEN_SHOULD_SKIP_THIS
#undef TOKEN_DEF

/** @description Asserts if the EEPROM tokens have been corrupted. */
EmberStatus initTokens(void);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// These defines provide a small bit of abstraction, 
// but still allow fast, easy, and tight access to simple datatypes 
// stored in the EEPROM.
#endif //DOXYGEN_SHOULD_SKIP_THIS
 
/**
 * @description Macro that gets the token's value, to be used for scalar tokens
 * only. To get data from more complex tokens (such as array tokens), use the
 * getToken() function.
 *
 * @param token: The name of the token.  
 *
 * @return Either the value of the token (scalar case) or a pointer to the
 * token (array case).
 */
#define GET_TOKEN( token ) \
  token
  
/**
 * @description Gets the size of a token.
 *
 * @param token: The name of the token.
 *
 * @return The size of the token, in bytes.
 */
#define SIZEOF_TOKEN( token ) \
  sizeof( token )
  
/**
 * @description Sets the value of a token. To be used for scalar tokens only.
 * To set more complex tokens (such as array tokens), use the setToken() function.
 *
 * @param token: The name of the token.
 *
 * @param val: The value to set the token to.
 */
#define SET_TOKEN( token, val ) \
  do { token = val; } while (FALSE)

/**
 * @description Copies the token value from EEPROM into a RAM location.
 * Note that with the inclusion of an appropriate length parameter, 
 * these interfaces support more complex datatypes, such as arrays of bytes. 
 *
 * @param dst: A pointer to RAM storage where the token value should be stored.
 *
 * @param src: A pointer to the EEPROM location of token value.
 *
 * @param len: Token length, in bytes.
 */
void getToken( void *dst, void *src, int8u len);

/**
 * @description Writes the token value from a RAM location 
 * and puts it into the EEPROM location \c dst.  Note that with the 
 * inclusion of an appropriate length parameter, these interfaces 
 * support more complex datatypes, such as arrays of bytes. 
 *
 * @param dst: A pointer to an EEPROM location.
 *
 * @param src: A pointer to RAM storage holding a token value.
 *
 * @param len: Token length, in bytes.
 */
void setToken( void *dst, void *src, int8u len);


#ifndef DOXYGEN_SHOULD_SKIP_THIS
// all interfaces that follow are DEPRECATED and serve only as a glue layer
// between the old token system and the new simple strategy. mainly this 
// supports the monitor-service/studio interface and haltest 

#define CREATOR_DEF(tokenName, creator)        \
  CREATOR_TOKEN_ ## tokenName = creator,
#define TOKEN_DEF(name, type, size, loc, ...)
enum {
  #include "config/token-stack.h"
};
#undef CREATOR_DEF
#undef TOKEN_DEF

void copyLinkedBufferIntoToken( void *dst, EmberMessageBuffer buff, 
                             int8u messageOffset, int8u size );

void copyTokenIntoLinkedBuffer( EmberMessageBuffer buff, int16u creator,
                            int8u messageOffset, int8u size );

#define INVALID_EE_ADDRESS 0xFFFF
void * getTokenAddress( int16u creator );

int8u getTokenSize( int16u creator );

#endif // DOXYGEN_SHOULD_SKIP_THIS

#endif  // BOOTLOADER

#endif // __TOKEN_H__

