//  Token-Upgrade application
//  programatically reformats the eeprom to switch between the old token system
//  and the new simplified format.
//  note: unlike all other applications, you probably shouldn't use the eeprom
//  image from this build, since the translation work is all programatic.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pgmspace.h> // sscanf_P 

#include "core/ember-stack.h"
#include "hal/micro/token.h"
#include "hal/micro/token-deprecated.h"
#include "framework/eui64.h"
#include "framework/nodeid.h"
#include "mac/802.15.4/mac.h"
#include "phy/security.h"
#include "include/error.h"
#include "stack/include/hal.h"
#define extern
#include "board/board.h"
#undef extern
#include "stack/phy/phy.h"
#include "stack/phy/phytest.h"
#include "micro/micro.h"
#include "app/util/serial/serial.h"
#include "haltest.h"

#define CMDHEADER "token-upgrade-cmd.h"
#include "test-def.h"
#undef CMDHEADER

extern int8u emInitAndSelectSerialPort(void);

// buffer stubs (for serial)
int8u *emberMessageBufferContents(EmberMessageBuffer buffer) {return NULL;}
const int8u emApplicationBufferCount = 0;
int8u emPacketBufferLinks[1];

// stubs (other)
void boardButtonCallback(int8u button, int8u state){}
void emMicroDelayedCallback(void) {}
void emMicroSymbolDelayAIsr(void) {}
void emMicroSymbolDelayBIsr(void) {}
//int8u emberGetLinkedBuffersByte(EmberMessageBuffer buffer, int8u index) {return 0;}
//void emberSetLinkedBuffersByte(EmberMessageBuffer buffer, int8u index, int8u byte) {}

// protos
static boolean choiceIsNo(PGM char string[]);


void printEOL(void)
{
  emberSerialPrintf(SER232,"\r\n");
}

void fillEEPROM(int8u fill)
{
  int16u i;
  emberSerialPrintf(SER232,"EEPROM scrub... wait.");
  for ( i = microGetEEPROMStart(); i <= microGetEEPROMEnd(); i++) {
    int8u r;
    microReadEEPROMByte( i, r );
    if ( r != fill )
      microWriteEEPROMByte( i, fill );
  }
  emberSerialPrintf(SER232,"done.");
  printEOL();
}

// Simple function to switch the EUI64 between big endian and little endian.
// Note:  works only on the new token system.
void reverseEui64( void )
{
  int8u oldEui64[8], newEui64[8], i;

  getToken( oldEui64, (void *) TOKEN_EUI_64, 8 );

  for(i=0; i<8; i++) {
    newEui64[i] = oldEui64[7-i];
  }

  setToken( (void *) TOKEN_EUI_64, newEui64, 8 );
}

static boolean eeFormatIsNew(void)
{
  return ( CURRENT_MFG_TOKEN_VERSION == GET_TOKEN(TOKEN_MFG_NVDATA_VERSION) );
}

PGM_P tokenTableString(int8u table) 
{
  switch (table) {
    case TOKEN_TABLE_UPPER_FLASH: {
      static PGM char s[] = "(UF)";
      return s; }
    case TOKEN_TABLE_EEPROM: {
      static PGM char s[] = "(EE)";
      return s; }
    case TOKEN_TABLE_DEFAULT_FLASH: {
      static PGM char s[] = "(DF)";
      return s; }
    default: {
      static PGM char s[] = "(??)";
      return s; }
  }
}

PGM_P tokenNameString(int16u creator)
{
  switch (creator) {
#define CREATOR_DEF(name, creator)        \
    case CREATOR_TOKEN_ ## name : {       \
      static PGM char s[] = #name ;       \
        return s; }
#define TOKEN_DEF(name, type, size, loc, ...)  
#include "stack/config/token-stack.h"
#undef CREATOR_DEF
#undef TOKEN_DEF
    default: {
      static PGM char s[] = "?"; 
      return s; } 
  }
}


void tokenRead(int16u creator)
{
  EmberStatus err;
  int8u size, data, offset;
  err = retrieveTokenSize(creator, &size);
  if (err) { 
    emberSerialPrintf(SER232,"not found");
  } else {
    emberSerialPrintf(SER232,"[%2x]",creator);
    emberSerialPrintf(SER232, tokenNameString(creator) );
    emberSerialPrintf(SER232," =");
    for (offset=0; offset<size; offset++) {
      retrieveTokenDataChunk(creator, &data, offset, 1/*len*/);
      emberSerialPrintf(SER232," %x", data);
    }
  }    
  printEOL();  
}

void tokenDump(void)
{
  EmberStatus err;
  int16u creator;
  
  creator = TOKEN_CREATOR_SEARCH_RESET;
  do {
    int8u table;
    TokenIterator iterator;
    err = retrieveNextTokenCreator(&creator, &iterator);
    if (!err) {
      emberSerialPrintf(SER232, tokenTableString(iterator.table));
      err = retrieveTokenTableType(creator, &table);
      if (table != iterator.table) {
        emberSerialPrintf(SER232, "[OVERRIDE IN %d", table);
        emberSerialPrintf(SER232, tokenTableString(table));
        emberSerialPrintf(SER232, "]");
      }
      tokenRead( creator );
    }
  } while (!err);
}

void newTokenDump(void)
{
  emberSerialPrintf(SER232,"(data shown little endian)\r\n");
    
#define TOKEN_DEF(name, type, size, loc, ...)                               \
  {                                                                         \
    int8u i, dst[ SIZEOF_TOKEN(TOKEN_ ## name) ];                           \
    emberSerialPrintf(SER232, "[%2X]", CREATOR_TOKEN_ ## name );            \
    emberSerialWriteString(SER232, #name );                                 \
    getToken( dst, (void*)&TOKEN_ ## name, SIZEOF_TOKEN(TOKEN_ ## name) );  \
    for( i = 0; i < SIZEOF_TOKEN( TOKEN_ ## name ); i++)                    \
      emberSerialPrintf(SER232, " %X", dst[i]);                             \
    printEOL();                                                             \
  }
#include "stack/config/token-stack.h"
#undef TOKEN_DEF
}

void reformatLoadData( void* name, int16u creator, int8u size, int8u* data)
{                                     
  EmberStatus err;                    
  int8u i, sz;                      
  emberSerialPrintf(SER232, (PGM_P)name);
  emberSerialPrintf(SER232, ": ");
  err = retrieveTokenSize( creator, &sz);           
  if ( err ) 
    emberSerialPrintf(SER232, "[not found]");                
  else if ( sz != size )                      
    emberSerialPrintf(SER232, "[size mismatch]");                     
  if ( !err && sz == size ) { 
    emberSerialPrintf(SER232, "[found]");                             
    retrieveTokenData( creator, data, size );
  } else {                                                            
    emberSerialPrintf(SER232, "[new default:");                       
    for( i = 0; i < size; i++)              
      emberSerialPrintf(SER232, " %X", data[i]);
    emberSerialPrintf(SER232, "]");                                   
  }                                                                   
  printEOL();
}


void reformatNewCommand(int16u param1, int16u param2)
{
#define TOKEN_DEF(name, type, size, loc, ...)                \
  static type VALUE_ ## name size = __VA_ARGS__ ;
#include "stack/config/token-stack.h"
#undef TOKEN_DEF
  
  if ( ! eeFormatIsOld() ) {
    emberSerialPrintf(SER232,"eeprom is not old format\r\n");
    return;
  }
    
#define TOKEN_DEF(name, type, size, loc, ...)             \
  reformatLoadData( (void*)#name, CREATOR_TOKEN_ ## name, \
                   SIZEOF_TOKEN(TOKEN_ ## name),          \
                   (int8u*)&VALUE_ ## name );
#include "stack/config/token-stack.h"
#undef TOKEN_DEF

  if ( choiceIsNo( "convert to new format?(y/n)" ) )
    return;

  fillEEPROM(0xFF);

#define TOKEN_DEF(name, type, size, loc, ...)               \
  emberSerialPrintf(SER232, #name ": converted\r\n");       \
  setToken( (void*)&TOKEN_ ## name,                         \
              &VALUE_ ## name,                              \
              SIZEOF_TOKEN(TOKEN_ ## name));
#include "stack/config/token-stack.h"
#undef TOKEN_DEF
  // Must handle the endian switch from the old to the new system.
  reverseEui64();

}

void reformatOldCommand(int16u param1, int16u param2)
{
#define TOKEN_DEF(name, type, size, loc, ...)                \
  static type VALUE_ ## name size = __VA_ARGS__ ;
#include "stack/config/token-stack.h"
#undef TOKEN_DEF

  if( ! eeFormatIsNew() ) {
    emberSerialPrintf(SER232, "eeprom is not new format\r\n");
    return;
  }

  // Must handle the endian switch from the new to the old system.
  reverseEui64();

#define TOKEN_DEF(name, type, size, loc, ...)               \
  emberSerialPrintf(SER232, #name ": read\r\n");            \
  getToken( &VALUE_ ## name,                                \
            (void*)&TOKEN_ ## name   ,                      \
            SIZEOF_TOKEN(TOKEN_ ## name));                  \
  emberSerialWaitSend(SER232);
#include "stack/config/token-stack.h"
#undef TOKEN_DEF

  if ( choiceIsNo( "convert to old format?(y/n)" ) ) {
    reverseEui64();
    return;
  }

  emberSerialPrintf(SER232, "Initializing EEPROM for old token format.\r\n"); 
  scrubTokenEEPROM();

#define TOKEN_DEF(name, type, size, loc, ...)               \
  emberSerialPrintf(SER232, #name ": created\r\n");         \
  modifyToken(CREATOR_TOKEN_ ## name,                       \
              (int8u*)&VALUE_ ## name,                      \
               SIZEOF_TOKEN(TOKEN_ ## name));
#include "stack/config/token-stack.h"
#undef TOKEN_DEF
}

void reverseEui64Command(int16u param1, int16u param2)
{
  int8u i, eui64[8];

  if( ! eeFormatIsNew() ) {
    emberSerialPrintf(SER232, "eeprom is not new format\r\n");
    return;
  }

  getToken((void *) eui64, (void *) TOKEN_EUI_64, 8);

  emberSerialPrintf(SER232, "EUI64:");
  
  for(i=0; i<8; i++) {
    emberSerialPrintf(SER232, " %X", eui64[i]);
  }

  printEOL();
  
  if ( choiceIsNo( "Reverse the EUI64?(y/n)" ) )
    return;

  reverseEui64();

  emberSerialPrintf(SER232, "EUI64 has been reversed.\r\n");
}

void newTokenSet(int16u creator)
{
  char buff[3];
  int8u i, data, size;
  int8u EEPROM *ee;

  size = getTokenSize( creator );
  ee = (int8u EEPROM *)getTokenAddress( creator );
  if ( size == 0  || ee == (int8u EEPROM *)INVALID_EE_ADDRESS ) {
    emberSerialPrintf(SER232, "not found\r\n");
    return;
  }
    
  for(i = 0; i < size; i++) {
    emberSerialPrintf(SER232, "data byte[%d of %d]:", i, size);
    emberSerialReadLine(SER232, buff, sizeof(buff));
    sscanf_P(buff, "%x", &data);
    ee[ i ] = data;
  } 
}

void newTokenGet(int16u creator)
{
  int8u i, size;
  int8u EEPROM *ee;
  
  size = getTokenSize( creator );
  ee = (int8u EEPROM *)getTokenAddress( creator );
  if ( size == 0  || ee == (int8u EEPROM *)INVALID_EE_ADDRESS ) {
    emberSerialPrintf(SER232, "not found");
    printEOL();
    return;
  }
  
  emberSerialPrintf(SER232,"[%2x](%d) =",creator,size);
  for(i = 0; i < size; i++) {
    int8u data;
    data = ee[ i ];
    emberSerialPrintf(SER232, " %X", data);
  }   
  printEOL();
}


void tokenDumpCommand(int16u p1, int16u p2)
{
  if ( eeFormatIsOld() ) {
    emberSerialPrintf(SER232,"[eeprom old format]\r\n");
    tokenDump();
  } else {
    if ( eeFormatIsNew() ) {
      emberSerialPrintf(SER232,"[eeprom new format]\r\n");
      newTokenDump();
    } else {
      emberSerialPrintf(SER232,"[eeprom unknown format]\r\n");
    }
  }
}

void tokenGetCommand(int16u creator, int16u param2)
{
  if ( eeFormatIsOld() ) {
    emberSerialPrintf(SER232,"[eeprom old format]\r\n");
    tokenRead( creator );
  } else {
    if ( eeFormatIsNew() ) {
      emberSerialPrintf(SER232,"[eeprom new format]\r\n");
      newTokenGet( creator );
    } else {
      emberSerialPrintf(SER232,"[eeprom unknown format]\r\n");
    }
  }
}

void tokenSetCommand(int16u creator, int16u param2)
{
  if ( eeFormatIsOld() ) {
    emberSerialPrintf(SER232,"[eeprom old format]\r\n");
    emberSerialPrintf(SER232,"not supported\r\n");    
  } else {
    if ( eeFormatIsNew() ) {
      emberSerialPrintf(SER232,"[eeprom new format]\r\n");
      newTokenSet( creator );
    } else {
      emberSerialPrintf(SER232,"[eeprom unknown format]\r\n");
    }
  }
}

void scrubCommand(int16u p1, int16u p2)
{
  if ( choiceIsNo("erase eeprom?(y/n)") )
    return;
  fillEEPROM(0xFF);
}

void formatCommand(int16u p1, int16u p2)
{
  if ( choiceIsNo("format eeprom?(y/n)") )
    return;

  fillEEPROM(0x00);
  SET_TOKEN( TOKEN_STACK_NVDATA_VERSION, CURRENT_STACK_TOKEN_VERSION );
}

//////////////////////////////////////////////////////////////////////

// Support for uploading a file to EEPROM using XMODEM

#define TIMEOUT_3SEC 3000
#define TIMEOUT_1SEC 1000
#define EEPROM_BLOCK_SIZE 128
// Upload the file to eeprom, ignoring the first 128 bytes (0x080), or
// manufacturing data.
#define  MEMORY_MODE_EEPROM1 1
// Upload the file to eeprom, ignoring the first 384 bytes (0x180), or
// everything but the app data.
#define  MEMORY_MODE_EEPROM2 2  

int8u expectedBlock;
int8u eepromBlock[EEPROM_BLOCK_SIZE];

static EmberStatus serWaitChar(int8u* ch, int16u timeout)
{
  int16u startTime= microGetSystemTick();
  do {
    if(emberSerialReadByte(SER232, ch) == EMBER_SUCCESS) {
      return EMBER_SUCCESS;
    }
  } while(microGetSystemTick() - startTime < timeout);
  return EMBER_SERIAL_RX_EMPTY;
}

static void serGetFlush(void)
{
  int8u err=0;
  int8u scrap;
  do {
    err = emberSerialReadByte(SER232, &scrap);
  } while(!err);
}

// note:  these flags do not start at 0 so that they can be outputted via the
// serial port during debug and easily screened out of normal xmodem traffic
// where I am only interested in the ack (0x06) and the NAK (0x15).
#define TIMEOUT           0x16
#define FILEDONE          0x17
#define FILEABORT         0x18
#define BLOCKOK           0x19
#define BLOCKERR_MASK     0x20
#define BLOCKERR_SOH      0x21
#define BLOCKERR_CHK      0x22
#define BLOCKERR_CRCH     0x23
#define BLOCKERR_CRCL     0x24
#define BLOCKERR_SEQUENCE 0x25
#define BLOCKERR_PARTIAL  0x26


#define SOH    0x01 // Start of Header 
#define EOT    0x04 // End of Transmission
#define ACK    0x06 // Acknowledge
#define NAK    0x15 // Not Acknowledge 
#define CAN    0x18 // Cancel (from sender)
#define C      0x43 // ASCII 'C'
#define CTRL_C 0x03 // Cancel (from sender, user)

static int8u XModemBlockReceive(int8u* data) {
  int8u err,i;
  int8u header, seq_1,seq_2, crc_h, crc_l;
  int16u crc=0x0000;

  err = serWaitChar(&header, TIMEOUT_3SEC);
  if (err) return TIMEOUT;
  if (header==CTRL_C) return FILEABORT;
  if (header==CAN) return TIMEOUT;
  if (header==EOT) return FILEDONE;
  if (header!=SOH) return BLOCKERR_SOH;

  err = serWaitChar(&seq_1, TIMEOUT_1SEC);
  if (err) return BLOCKERR_PARTIAL;

  err = serWaitChar(&seq_2, TIMEOUT_1SEC);
  if (err) return BLOCKERR_PARTIAL;

  for(i=0; i<EEPROM_BLOCK_SIZE; i++) {
    err = serWaitChar(&data[i], TIMEOUT_1SEC);
    if (err) return BLOCKERR_PARTIAL;
  }

  err = serWaitChar(&crc_h, TIMEOUT_1SEC);
  if (err) return BLOCKERR_PARTIAL;

  err = serWaitChar(&crc_l, TIMEOUT_1SEC);
  if (err) return BLOCKERR_PARTIAL;
  
  for(i=0; i<EEPROM_BLOCK_SIZE; i++)
    crc = emHalCrc16(data[i], crc);

  if (seq_1+seq_2!=0xFF) return BLOCKERR_CHK;
  if (crc_h!=HIGH_BYTE(crc)) return BLOCKERR_CRCH;
  if (crc_l!=LOW_BYTE(crc)) return BLOCKERR_CRCL;
  if (seq_1!=expectedBlock) return BLOCKERR_SEQUENCE;
  
  return BLOCKOK;
}

// TBD:  following segment, from the define through eepromWriteBlock, should be
// in a shared file with the equivalent function from bootloader.c_.
#define EEPROM_END ((int8u EEPROM *) 0x1000)

static void eepromWriteBlock(int8u* eepromAddr, int8u* eepromBlock, int8u memoryMode)
{
  int8u i;
  int8u EEPROM *ee = (int8u EEPROM *) eepromAddr;
  int8u EEPROM *eeStart;

  if(memoryMode == MEMORY_MODE_EEPROM1) {
    eeStart = (int8u EEPROM *) 0x80;
  } else {
    eeStart = (int8u EEPROM *) 0x180;
  }

  for(i=0; i<128; i++) {
    if(ee >= eeStart &&
       ee < EEPROM_END) {
      *ee = *eepromBlock;
    }

    ee++;
    eepromBlock++;
  }
}

static int8u XModemReceive( int8u memoryMode ) {
  int8u status;
  int8u* eepromAddr;
  
  for (;;)  {
  
    expectedBlock = 1;
    eepromAddr = 0x00;
     
    emberSerialWriteByte(SER232, C);
      
    do {
      status = XModemBlockReceive(eepromBlock);
      if (status == BLOCKOK) {
        eepromWriteBlock(eepromAddr, eepromBlock, memoryMode);
        eepromAddr += 0x80;
        expectedBlock++;
        serGetFlush();
        emberSerialWriteByte(SER232, ACK);
      }
      if (status & BLOCKERR_MASK)  {
        serGetFlush();
        emberSerialWriteByte(SER232, NAK);
      }
      if (status == FILEDONE) {
        serGetFlush();
        emberSerialWriteByte(SER232, ACK);
      }
      if (status == FILEABORT || status == FILEDONE)
        return status;     
    } while(status != TIMEOUT);
  } 
}

void uploadEepromCommand(int16u memoryMode, int16u param2)
{
  int8u status = FILEABORT;
  if (memoryMode == 1) {
    emberSerialPrintf(SER232, "uploading stack and app sections\r\n");
    emberSerialPrintf(SER232, "\r\nbegin upload\r\n");
    status = XModemReceive( MEMORY_MODE_EEPROM1 );
  } else if (memoryMode == 2) {
    emberSerialPrintf(SER232, "uploading app section\r\n");
    emberSerialPrintf(SER232, "\r\nbegin upload\r\n");
    status = XModemReceive( MEMORY_MODE_EEPROM2 );
  } else {
    emberSerialPrintf(SER232, "eeprom 1: upload stack and app sections\r\n");
    emberSerialPrintf(SER232, "eeprom 2: upload app section\r\n");
    return;
  }
  emberSerialPrintf(SER232, "\r\nupload ");
  if (status == FILEDONE) {
    emberSerialPrintf(SER232, "complete\r\n");
  } else {
    emberSerialPrintf(SER232, "aborted\r\n");
  }
}

//////////////////////////////////////////////////////////////////////

static boolean choiceIsNo(PGM char string[]) {
  char buff[2];
  do {
    emberSerialPrintf(SER232, string);
    emberSerialReadLine(SER232, buff, sizeof(buff));  // does upcase
  } while ( buff[0] != 'Y' && buff[0] != 'N' );
  return( buff[0] == 'N' );
}

static void printHelp(int16u param1, int16u param2)
{
  printCommands(baseCommandList);
  emberSerialPrintf(SER232,"    ** NOTE: all input values are in hex!");
  printEOL();
}

void printCommands(commandType PGM *list)
{
  int8u i;

  for(i=0; list[i].func != NULL; i++) {
    emberSerialPrintf(SER232,"  ");
    emberSerialPrintf(SER232,list[i].command);
    emberSerialPrintf(SER232," ");
    emberSerialPrintf(SER232,list[i].params);
    emberSerialPrintf(SER232," - ");
    emberSerialPrintf(SER232,list[i].helpString);
    printEOL();
  }
}



static boolean runCommand(char *cmd, char *params,commandType PGM *list)
{
  int8u i;
  int32u param1, param2;

  param1=0; param2=0;
  for(i=0; list[i].func != NULL; i++) {
    if(strcmp_P(cmd, list[i].command) == 0) {
      sscanf_P(params, "%x %x", &param1, &param2);
      list[i].func(param1,param2);
      return TRUE;
    }
  }
  return FALSE;
}


//  -------------main-----------------


int main() {
  char buff[30];
  int8u paramStart;

  boardBoot();

  SER232 = emInitAndSelectSerialPort();

  initTokens(); 

  INTERRUPTS_ON();
  printEOL();
  emberSerialPrintf(SER232,"Ember EEPROM Transition Application v1.0");
  printEOL();
  emberSerialPrintf(SER232,__DATE__);
  emberSerialPrintf(SER232,":");
  emberSerialPrintf(SER232,__TIME__);
  
  while(1) {
    printEOL();
    emberSerialPrintf(SER232,"> ");
    emberSerialReadLine(SER232,buff, 30);

    // scan for end of first word
    for(paramStart=0; (buff[paramStart] != ' ') && (buff[paramStart] != '\0'); paramStart++) {
      ; // nada
    }
    buff[paramStart] = '\0';

    runCommand(buff, &buff[paramStart+1], baseCommandList);
  }

}

int8s emberGetRadioPower( void )
{
  return -1;
}

int8u emberGetRadioChannel( void )
{
  return 2;
}
