/**
 * @file config.h
 * @brief Public stack configuration settings.
 *
 * This file is a repository for static stack configuration parameters that are
 * visible to the HAL and the application as well as to the stack. These values 
 * are set at build time and cannot be adjusted.
 *
 * Please note that because the build number changes frequently, the Doxygen file for 
 * config.h may not show the correct number. Check the file to view the 
 * correct build number.
 *
 * <!--Tech Pubs: Must manually update this table for each release!-->
 * 
 * <h2>General Configurations</H2>
 * <table border=1>
 * <tr><td>SOFTWARE_VERSION</td><td>0x3347</td><td>High byte = release number, low byte
 *          = build number</td></tr>
 * <tr><td>BEST_PAYLOAD_SIZE</td><td>64</td><td>Recommended maximum application payload 
 *         size. The actual best payload size is unclear, as it depends on many network 
 *         dynamics. Given that Ember's packets are so short and the header overhead is
 *         so large, the optimum likely  occurs on the boundary, namely, the max payload 
 *         size.  For now, the value is being left at 64.</td></tr>
 * </table>
 *
 * <h2>Payload Size Factors</H2>
 *
 * <b>About the settings:</b>
 * - MAC frame: 9 bytes for 15.4, 4 bytes for the Ember MAC
 * - Packet information frame: 1 byte
 * - Worst case routing frame: Unicast, 9 bytes
 * - Worst case dispatch frame : Datagram, 13 bytes
 * 
 * \b Caveats:
 * - For aggregation endpoints, subtract 16 bytes.
 * - If security is enabled, subtract 16 bytes.
 *
 * <table border=1>
 * <tr><td><b>Parameter</b></td><td><b>em2420 Value</b></td><td><b>cc1020 Value</b></td>
 *     <td><b>Comment</b></td></tr>
 * <tr><td>MAX_PHY_PAYLOAD_SIZE</td><td>125</td><td>127</td><td>127 bytes is the  
 *         maximum payload size for our current PHY. The EM2420 counts the 2-byte CRC 
 *         as part of the PHY payload, so from the standpoint of somebody 
 *         calling radioTransmit, you have to subtract 2.</td></tr>
 * <tr><td>MAX_APP_PAYLOAD_SIZE</td><td>93</td><td>100</td><td>Maximum size of the actual 
 *         data payload carried in a radio packet in bytes. This is MAX_PHY_PAYLOAD_SIZE, 
 *         minus the worst-case header size.</td></tr>
 * <tr><td>AGGREGATION_PAYLOAD_REDUCTION</td><td>16</td><td>16</td><td>None</td></tr>
 * <tr><td>SECURITY_PAYLOAD_REDUCTION</td><td>16</td><td>0</td><td>None</td></tr>
 * </table>
 *
 * <h2>MAC Configuration Values</H2>
 * <table border=1>
 * <tr><td><b>Parameter</b></td><td><b>em2420 Value</b></td><td><b>cc1020 Value</b></td></tr>
 * <tr><td>MAC_BACKOFF_SIZE</TD><td>20</td><td>22</td></tr>
 * <tr><td>PHY_SYMBOL_RATE</TD><td>16</td><td>24</td></tr>
 * <tr><td>MAC_ACK_WAIT_DURATION</TD><td>75</td><td>300</td></tr>
 * </table>
 *
 * <!--Author(s): Many-->
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.                *80*-->
 */


/**
 * Software version. High byte = release #, low byte = build #
 */
BUILD_DEF( SOFTWARE_VERSION, 0x3347 )

/**
 * MAX_PHY_PAYLOAD_SIZE
 * 127 bytes is the maximum payload size for our current PHY. The EM2420 
 * counts the 2-byte CRC as part of the  phy payload, so from the standpoint of 
 * somebody calling \c radioTransmit, you have to subtract 2.
 * 
 * MAX_APP_PAYLOAD_SIZE
 * Maximum size of the actual data payload carried in a radio packet in bytes.
 * This is MAX_PHY_PAYLOAD_SIZE, minus the worst-case header size.
 * 
 * CURRENTLY:
 * - MAC frame: 9 bytes for 15.4, 4 bytes for the Ember MAC.
 * - Packet info frame: 1 byte.  
 * - Worst case routing frame: Unicast, 9 bytes.
 * - Worst case dispatch frame: Datagram, 13 bytes.
 * 
 * \b Caveats:
 * - For aggregation endpoints, subtract 16 bytes.
 * - If security is enabled, subtract 16 bytes.
 */

#ifdef RADIO_EM2420
  BUILD_DEF( MAX_PHY_PAYLOAD_SIZE, 125 )
  BUILD_DEF( MAX_APP_PAYLOAD_SIZE, 93 )
  BUILD_DEF( AGGREGATION_PAYLOAD_REDUCTION, 16 )
  BUILD_DEF( SECURITY_PAYLOAD_REDUCTION, 16 )
#else // RADIO_CC1020
  BUILD_DEF( MAX_PHY_PAYLOAD_SIZE, 127 )
  BUILD_DEF( MAX_APP_PAYLOAD_SIZE, 100 )
  BUILD_DEF( AGGREGATION_PAYLOAD_REDUCTION, 16 )
  // security not supported on cc1020
  BUILD_DEF( SECURITY_PAYLOAD_REDUCTION, 0 )
#endif

/** 
 * @description Recommended maximum application payload size.
 *
 *<!--[from MNP]-->The actual best payload size is unclear, as it depends 
 * on many network dynamics. Given that Ember's packets are so short and the 
 * header overhead is so large, the optimum likely  occurs on the boundary,
 * namely, the max payload size.  For now, the value is being left at 64.
 */
BUILD_DEF( BEST_PAYLOAD_SIZE, 64 )


//----------------------------------------------------------------
// MAC 

/**
 * MAC_BACKOFF_SIZE indicates how large the slot times are.
 * Times are defined as MAC_BACKOFF_SIZE * 16us (actually, 17.39us on 3.68MHz 
 * boards). Note that these values are further adjusted for system clock in 
 * micro.c.
 */
#if defined( RADIO_MICROLIN )
  // 128us
  BUILD_DEF( MAC_BACKOFF_SIZE, 8)
  BUILD_DEF( MAC_ACK_WAIT_DURATION, 54 )
#elif defined( RADIO_CC1020 )
  // 512us 
  BUILD_DEF( MAC_BACKOFF_SIZE, 32)  // 32*16us = 512
  BUILD_DEF( PHY_SYMBOL_RATE, 24) // us
  BUILD_DEF( MAC_ACK_WAIT_DURATION, 300 ) //symbols
#elif defined( RADIO_EM2420 )
  // 320us
  BUILD_DEF( MAC_BACKOFF_SIZE, 20)
  BUILD_DEF( PHY_SYMBOL_RATE, 16) // us
     // Note:  extra 21 symbols is needed for manual acking.
     // Note:  needed to extend this to 40 symbols as 21 was insufficient.
  BUILD_DEF( MAC_ACK_WAIT_DURATION, (54 + 40) ) //symbols
#elif defined( RADIO_UNIX )
  // 128us
  // should probably make this so that it varies with the radio you are simulating
  // AJW - 8/27/2003
  BUILD_DEF( MAC_BACKOFF_SIZE, 0) 
#endif
