/** 
* @file token-stack.h
* 
* \brief Definitions for tokens.
*
* This file should not be included directly. It is included with the file
* token.h.
*
* BECKY: When integrating, I noticed this descriptive text was removed, but
*  I think some of it should stay -lee
*
* The user application should include its own tokens in a header file similar
* to this one. The macro APPLICATION_TOKEN_HEADER should be defined to equal
* the name of the header file in which application tokens are defined. 
* 
* DC: PLEASE REVISE THE FOLLOWING PARAGRAPH SO THAT IT JUST DESCRIBES TOKEN VERSION CONTROL.
* USERS DON'T NEED TO KNOW ANYTHING ABOUT THE CHANGE.
* Because the new token system no longer labels each token individually with 
* a creator, we needed to introduce the idea of token version control. If the 
* application or stack change tokens, the token version should change as well 
* and all tokens should be reinitialized.  
*
* A note on initialization of tokens: The compiler will create an ebin file
* that can be uploaded serially using the bootloader. It is also possible to
* create an application that will write the initialized values to the EEPROM
* (see the sample application token-init for more details), that can be used
* to initialize tokens over the air.  
*
* DC: SO THE USER APP MUST SUPPLY THE TOKEN DEFINITIONS? LET'S SAY THAT PLAINLY, IF YES.
* A token definition is constructed as follows:
* TOKEN_DEF( TOKEN_NAME, TYPE, SIZE, DEFAULT_VALUES )
* where TYPE is the type of the data in the token, SIZE is the array size and
* should be left blank for scalar tokens, and DEFAULT_VALUES is the value to
* wHich the token is set upon initialization.  
* 
* Here are two examples:
* TOKEN_DEF( CHANNEL, int8u, , 0x04 )
* This defines a token called TOKEN_CHANNEL, which is an int8u.  Because it is
* a scalar, we do not include the array size.  If/when this token is
* initialized, it will be set to the value 0x04.
*
* TOKEN_DEF(ENCRYPTION_KEY, int8u, [16], {0,})
* This defines a token called TOKEN_ENCRYPTION_KEY, which is a 16-element array
* of type int8u.  If/when this token is initialized, it will be set to all
* zeros.  
* <!--Author(s): Jeff Matthews, jm@ember.com, -->
*        <!--DC Seward, dc@ember.com-->
* Copyright 2004 by Ember Corporation. All rights reserved.
*/

// NOTE: The location in EEPROM must be specified for every token. If the
// location is changed then the appropriate version number should be
// incremented. If a token is inserted or changes size then the location of all
// subsequent tokens must be adjusted. The linker will produce an error for
// tokens whose locations overlap but will not warn about gaps between tokens.

// By default, manufacturing tokens are treated the same as all other tokens.
#ifndef TOKEN_MFG
  #define TOKEN_MFG TOKEN_DEF
#endif

#define CURRENT_MFG_TOKEN_VERSION 0x02FD  // msb is version, lsb is complement
// this identifier should be incremented each time the format of the
// non-volatile data below is changed.
TOKEN_MFG(MFG_NVDATA_VERSION, int16u, , 1, CURRENT_MFG_TOKEN_VERSION)

///////////////// 
// MANUFACTURING DATA
// *these must come first, and their order may not change*

// IEEE 64 bit ID (big endian)
TOKEN_MFG(EUI_64, int8u, [EUI64_SIZE], 3, {3,0,0,0,0,0,0,3})
// IEEE 64 bit ID (copy protection hash)
TOKEN_MFG(EUI_64_HASH, int8u, [EUI64_SIZE], 11, {3,0,0,0,0,0,0,0})
// device's bar code label
TOKEN_MFG(MANUFACTURING_STRING, int8u, [16], 19, {0,})
// indicator for radio type and allowed channels
TOKEN_MFG(RADIO_BANDS_SUPPORTED, int8u, , 35, 0 )
// offset to optimize register settings on the CC1020 radio
TOKEN_MFG(RADIO_CRYSTAL_OFFSET, int8s, , 36, 0)

#if defined(MEGA8) || defined(MEGA168)
// Threshold for battery readings.  Manufacturing can dial the micro to a
// specific voltage, read the adc and set this token.  Periodically, the app
// can read the adc and if the returned value is below the token, trigger
// some sort of battery low/fail condition.
TOKEN_MFG(BATTERY_THRESHOLD, int8u, , 37, 0)
//Put all the CC1020 configuration bytes into EEPROM
//TOKEN_MFG(RADIO_CHANNEL_SETTINGS, int8u, [30], 37, {0,})
//TOKEN_MFG(RADIO_REGISTER_CONFIG, int8u, [46], 67, {0,})
#endif //MEGA8||MEGA168


#define CURRENT_STACK_TOKEN_VERSION 0x02FD  // msb is version, lsb is complement
// this identifier should be incremented each time the format of the
// non-volatile data below is changed.
// Stack section starts at 0x80 = 128
TOKEN_DEF(STACK_NVDATA_VERSION, int16u, , 128, CURRENT_STACK_TOKEN_VERSION)

//////////////////
// APPLICATION DEFINED
// at some point emberInit() might instead take a struct parameter with these values.

// 15.4 PAN ID   (little endian)
TOKEN_DEF(PAN_ID, int16u, , 130, 0xFFFF)
// allow/inhibit relaying of packets
TOKEN_DEF(GRAD_ENABLE_RELAY, int8u, , 132, TRUE)
// radio transmit power and channel settings
#if defined(RADIO_CC1020)
  TOKEN_DEF(RADIO_TX_POWER, int8s, , 133, 5) // 5dBm (max, such that emRadioSetTransmitPower() does not fail)
  TOKEN_DEF(RADIO_FREQ_CHANNEL, int8u, , 134, 4)
#elif defined(RADIO_EM2420)
  TOKEN_DEF(RADIO_TX_POWER, int8s, , 133, -1) // -1dBm (max, such that emRadioSetTransmitPower() does not fail)
  TOKEN_DEF(RADIO_FREQ_CHANNEL, int8u, , 134, 0) // 15.4 default
#elif SIMULATOR
  TOKEN_DEF(RADIO_TX_POWER, int8s, , 133, 2) 
  TOKEN_DEF(RADIO_FREQ_CHANNEL, int8u, , 134, 6) 
#else
  #error unknown radio type
#endif
// 128bit encryption key (set to all 0's is disabled)
TOKEN_DEF(ENCRYPTION_KEY, int8u, [16], 135, {0,})
// enable dynamic short id assignment
TOKEN_DEF(ID_ASSIGNMENT_ENABLED, int8u, , 151, TRUE)

//////////////////
// NON-VOLATILE STORAGE FOR THE STACK

// reset count
TOKEN_DEF(BOOT_COUNT, int16u, , 152, 0x0000)
// current value of dynamic grad node id  (little endian)
TOKEN_DEF(GRAD_NODE_ID, int16u, , 154, 0xFFFF)
// non-volatile counter for use by encryption
TOKEN_DEF(ENCRYPTION_NONCE_COUNTER, int32u, , 156, 0x00000000)
// reserved
TOKEN_DEF(RESERVED, int8u, , 160, 0x00)
// analysis
TOKEN_DEF(ANALYSIS_REBOOT, int16u, , 161, 0x0000)
// id assignment token
TOKEN_DEF(ID_AUTHORITY_TAG, int8u, [6], 163, {0,})
// binding table storage is the responsibility of the application since the
// number of non-volatile entries in the table can be set by the application.

//////////////////
// END MARKER
TOKEN_DEF(STACK_NVDATA_END, int8u, , 169, 0x00)


#ifndef DOXYGEN_SHOULD_SKIP_THIS
// the use of creator ids for tokens is DEPRECATED and serves only as a glue
// layer between the old token system and the new simple strategy. mainly this 
// supports the monitor-service/studio interface and haltest

#ifndef CREATOR_DEF
  #define CREATOR_DEF(name, creator)
#endif

#define SYS_CREATOR(a,b) (int16u)HIGH_LOW_TO_INT((a)|0x80,b)
#define APP_CREATOR(a,b) (int16u)HIGH_LOW_TO_INT(a,b)

CREATOR_DEF(EUI_64,                       SYS_CREATOR('6','4'))
CREATOR_DEF(EUI_64_HASH,                  SYS_CREATOR('6','5'))
CREATOR_DEF(MANUFACTURING_STRING,         SYS_CREATOR('m','s'))
CREATOR_DEF(RADIO_BANDS_SUPPORTED,        SYS_CREATOR('r','b'))
CREATOR_DEF(RADIO_CRYSTAL_OFFSET,         SYS_CREATOR('c','o'))
CREATOR_DEF(PAN_ID,                       SYS_CREATOR('p','i'))
CREATOR_DEF(GRAD_ENABLE_RELAY,            SYS_CREATOR('g','r'))
CREATOR_DEF(RADIO_TX_POWER,               SYS_CREATOR('r','p'))
CREATOR_DEF(RADIO_FREQ_CHANNEL,           SYS_CREATOR('r','c'))
CREATOR_DEF(ENCRYPTION_KEY,               SYS_CREATOR('e','k'))
CREATOR_DEF(ID_ASSIGNMENT_ENABLED,        SYS_CREATOR('i','e'))
CREATOR_DEF(BOOT_COUNT,                   SYS_CREATOR('b','c'))
CREATOR_DEF(GRAD_NODE_ID,                 SYS_CREATOR('g','i'))
CREATOR_DEF(ENCRYPTION_NONCE_COUNTER,     SYS_CREATOR('e','c'))
CREATOR_DEF(ANALYSIS_REBOOT,              SYS_CREATOR('a','b'))
CREATOR_DEF(ID_AUTHORITY_TAG,             SYS_CREATOR('i','t'))

#if defined(MEGA8) || defined(MEGA168)
CREATOR_DEF(BATTERY_THRESHOLD,            SYS_CREATOR('t','h'))
//Put all the CC1020 configuration bytes into EEPROM
//CREATOR_DEF(RADIO_CHANNEL_SETTINGS,       SYS_CREATOR('r','s'))
//CREATOR_DEF(RADIO_REGISTER_CONFIG,        SYS_CREATOR('r','r'))
#endif //MEGA8||MEGA168

CREATOR_DEF(STACK_NVDATA_VERSION,         0xFF01)
CREATOR_DEF(STACK_NVDATA_END,             0xFF03)
CREATOR_DEF(RESERVED,                     0xFF07)
CREATOR_DEF(MFG_NVDATA_VERSION,           0xFF09)

#endif // DOXYGEN_SHOULD_SKIP_THIS 


//////////////////
// NON-VOLATILE STORAGE FOR THE APPLICATION
// note: the application itself must detect if these are somehow corrupt, or
// leftover from some previous application. it must then flush them.

// The application must specify the location of its tokens. The location of the
// first application token must be 0x180 = 384.

// EMBER_BINDING_ENTRY_SIZE is defined here for convenience since the
// BINDING_TABLE token must be defined by the application. It is also defined
// in ember-static-memory.h
#define EMBER_BINDING_ENTRY_SIZE 11

#ifdef APPLICATION_TOKEN_HEADER
  #include APPLICATION_TOKEN_HEADER
#else
  // If the application does not use any non-volatile binding entries and does
  // not have any other application tokens, the BINDING_TABLE token is defined
  // here for convenience.
  // If the application does use non-volatile binding entries then it must
  // define the token itself in the APPLICATION_TOKEN_HEADER and set the size
  // correctly.
  TOKEN_DEF(BINDING_TABLE, int8u, [8][EMBER_BINDING_ENTRY_SIZE], 384, {0,})
  CREATOR_DEF(BINDING_TABLE,                SYS_CREATOR('b','t'))
#endif

#undef CREATOR_DEF
#undef TOKEN_MFG
