/** @file ember-common.h
 * @brief Ember's common application API
 *
 * Some functions in this file return an EmberStatus value. See 
 * error-def.h for definitions of all EmberStatus return values.
 *
 * <!--Author(s): Lee Taylor, lee@ember.com, Richard Kelsey, richard@ember.com-->
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.                *80*-->
 */

#ifndef __EMBER_COMMON_H__
#define __EMBER_COMMON_H__

#ifndef __EMBER_H__
#error Do not include ember-common.h directly
#endif 

#define TRUE  1
#define FALSE 0

#ifndef NULL
#define NULL ((void *)0)
#endif

#define EUI64_SIZE 8
#define EMBER_ENCRYPTION_KEY_SIZE 16

/** @description  Return type for Ember functions. */
typedef int8u EmberStatus;

/** @description Value representing the cost of sending a message. The 
 * word "cost" is specific to the EmberNet routing algorithm and gives a 
 * measure of distance between nodes. The cost takes both hop count and 
 * link reliability into account. Less reliable links are given higher 
 * costs.
 */
typedef int8u EmberCost;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
typedef int8u EmberCostValue;      // deprecated
#endif //DOXYGEN_SHOULD_SKIP_THIS

/** @description IEEE EUI 64-bit ID. */
typedef int8u EmberEUI64[EUI64_SIZE];

/** @description 802.15.4 PAN ID. 
 */
typedef int16u EmberPanId;

/** @description Defines reset conditions.
 */
typedef enum
{
   /** Cause for the reset is not known.*/
   RESET_UNKNOWN,
   /** A low level was present on the reset pin.*/
   RESET_EXTERNAL,
   /** The supply voltage was below the power-on threshold.*/   
   RESET_POWERON,
   /** The Watchdog is enabled, and the Watchdog Timer period expired.*/
   RESET_WATCHDOG,
   /** The Brown-out Detector is enabled, and the supply voltage was 
    * below the brown-out threshold.*/
   RESET_BROWNOUT,
   /** A logic one was present in the JTAG Reset Register. */ 
   RESET_JTAG,
   /** A self-check within the code failed unexpectedly. */
   RESET_ASSERT,
   /** The return address stack (RSTACK) went out of bounds. */
   RESET_RSTACK,
   /** The data stack (CSTACK) went out of bounds. */
   RESET_CSTACK,
   /** The bootloader deliberately caused a reset. */
   RESET_BOOTLOADER
} EmberResetType;

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** @description Defines beacons as using either 16- or 32-bit times.
 */
#ifdef SIXTEEN_BIT_BEACONS
typedef int16u EmberBeaconTime;
#else
typedef int32u EmberBeaconTime;
#endif
#endif

/** @brief Holds network parameters. For information about power 
 * settings and radio channels, see the technical specification for the
 * RF communication module in your Developer Kit. 
 */

typedef struct {
  /** The networks's PAN identifier.*/
  int16u  panId;
  /** True if this node will relay messages for other nodes. */
  boolean enableRelay;
  /** A power setting, in dBm.*/
  int8s   radioTxPower;
  /** A radio channel. Be sure to specify a channel supported by the radio. */
  int8u   radioChannel;
} EmberNetworkParameters;

/** @description Defines binding types. See EmberBindingTableEntry
 * for descriptions of these values.
 */
enum {
  EMBER_UNUSED_BINDING    = 0,
  EMBER_UNICAST_BINDING   = 1,
  EMBER_MULTICAST_BINDING = 2,
  EMBER_LINK_BINDING      = 3
};

typedef int8u EmberBindingType;

#define EMBER_NULL_BINDING 0xFF

/** @description A pairing of one endpoint with one interface. Terminals are 
 * represented as a single byte that contains the index of the endpoint 
 * in the higher-order 5 bits and the index of the interface 
 * (among the interfaces implemented by that endpoint) in the 
 * lower-order 3 bits.
 * <table border=0><tr><td><img src="terminal.gif"></td></tr></table>
 *
 * Use these functions to define terminals:
 * - emberTerminalEndpoint()
 * - emberTerminalInterface()
 * - emberMakeTerminal()
 */
typedef int8u EmberTerminal;

/** @description Returns the index of a terminal's endpoint in the node's
 * array of endpoints.
 *
 * @param terminal: A terminal.
 *
 * @return An index.
 */
#define emberTerminalEndpoint(terminal)  (((terminal) >> 3) & 0x1F)

/** @description Returns the index of a terminal's interface, among the
 * interfaces implemented by the terminal's endpoint.
 *
 * @param terminal: A terminal.
 *
 * @return An index.
 */
#define emberTerminalInterface(terminal) ((terminal) & 0x7)

/** @description Creates a terminal.
 *
 * @param endpoint: The index of the endpoint to include in the terminal.
 *
 * @param interface: The index of the interface to include in the terminal.
 *
 * @return A terminal.
 *
 * @see EmberTerminal
 */
#define emberMakeTerminal(endpoint, interface)  \
 ((((endpoint) & 0x1F) << 3) | ((interface) & 0x7))

/**@brief Specifies a binding table entry. 
 * @description Defines an entry in the binding table. A binding specifies
 * a destination to which a message can be sent and a source from which 
 * messages can arrive. The structure binds two endpoints and indicates which 
 * interface will be used for communication. Each paired endpoint and interface 
 * index is specified as a terminal. \b Note: For multicast bindings 
 * the remote terminal has no meaning.
 */
typedef struct {
  /** ID for an entry in the binding table. One of the following:
   * - EMBER_UNUSED_BINDING: A binding indicating that the entry is 
   *   currently not in use.
   * - EMBER_UNICAST_BINDING: A binding whose destination address for 
   *   the entry is the 64-bit IEEE address of a single node.
   * - EMBER_MULTICAST_BINDING: A binding whose destination address for the
   *   entry is a 2-byte multicast group address (high-order byte first).
   * - EMBER_LINK_BINDING: A binding whose destination address for the entry is 
   *   the 64-bit IEEE address of a single node.
   */
  EmberBindingType type;
  /** A node ID. Must match \c type:
   * - A 64-bit IEEE address, for EMBER_UNICAST_BINDING.
   * - A 2-byte multicast group address, for EMBER_MULTICAST_BINDING.
   */
  int8u identifier[8];
  /** The terminal that is the local side of this binding. */
  EmberTerminal local;
  /** The terminal that is the remote side of this binding 
   * (on the node specified by \c identifier)  
   */
  EmberTerminal remote;
} EmberBindingTableEntry;

/** @description Initializes the radio and the EmberNet stack. The 
 * value of \c reset is returned by the HAL function microGetResetInfo(). 
 * (See the <i>Hardware Abstraction Layer Interface Guide</i>.) 
 *
 * Device configuration functions must be called before emberInit() 
 * is called.
 *
 * \b Note: The application must check the return value of this function. If the
 * initialization fails, normal messaging functions will not be available.
 * Some failure modes (for example, EMBER_UNINITIALIZED_EUI64) are not
 * fatal, but the application must follow certain procedures to permit recovery.
 * Ignoring the return code will result in unpredictable radio and API behavior.
 * (In particular, problems with association will occur.)
 *
 * @param reset: Cause of the microprocessor reset.
 *
 * @return An EmberStatus value indicating successful initialization or the
 * reason for failure.
 */
EmberStatus emberInit(EmberResetType reset);

/** @description Causes the stack to associate with the network using the
 * specified network parameters. It may take several seconds for the stack to
 * associate with the local network. Do not send messages until a call to the \c
 * emberStackStatusHandler() callback informs you that the stack is up.
 *
 * \b Note: If emberInit() returns an error, and your application fails to handle
 * it and calls emberAssociate(), emberAssociate() may block forever.
 *
 * @param parameters: Specification of the network with which the node
 *   should associate.
 *
 * @return An EmberStatus value indicating successful beginning of the
 * association process or reason for failure. 
 */
#ifdef DOXYGEN_SHOULD_SKIP_THIS
EmberStatus emberAssociate(EmberNetworkParameters *parameters);
#else
// To save code space we define emberAssociate() and emberReassociate() as
// macros that call emAssociate().  We hide this fact from doxygen.

#define emberAssociate(parameters) emAssociate(parameters, TRUE)
EmberStatus emAssociate(EmberNetworkParameters *parameters,
                        boolean isReassociation);
#endif // DOXYGEN_SHOULD_SKIP_THIS

/** @description A version of emberAssociate() that brings the stack up
 * more quickly by assuming that this node's previous association used the
 * same network parameters.  emberAssociate() should be used instead if
 * the node was not previously associated or was associated using different
 * network parameters.
 *
 * @param parameters: Specification of the network with which the node
 *   should associate.
 *
 * @return An EmberStatus value indicating successful beginning of the
 * association process or reason for failure. 
 */
#ifdef DOXYGEN_SHOULD_SKIP_THIS
EmberStatus emberReassociate(EmberNetworkParameters *parameters);
#else
#define emberReassociate(parameters) emAssociate(parameters, FALSE)
#endif // DOXYGEN_SHOULD_SKIP_THIS

/** @description Causes the stack to disassociate from the current network. This
 * generates a call to the emberStackStatusHandler() callback to indicate
 * that the network is down. The radio will not be used until after a later call
 * to emberAssociate(). 
 *
 * @return An EmberStatus value indicating success or reason for failure. 
 */
EmberStatus emberDisassociate(void);

/** @description A periodic tick routine that should be called in 
 * the application's main event loop and also as soon as possible after 
 * any radio interrupts and after emberInit().
 *
 * @return The number of clock ticks (approximately, milliseconds) until 
 * the next node event. The value of the return specifies when to call 
 * emberTick() again. Note that an interrupt may change the value of 
 * the return.
 */
int16u emberTick(void);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
extern EmberEUI64 emLocalEui64;
#endif

/** @description Returns the EUI64 ID of the local node.
 *
 * @return The 64-bit ID.
 */
#define emberGetEui64() (emLocalEui64)

/** @description Determines whether \c eui64 is the local node's EUI64 ID.
 *
 * @param eui64: An EUI64 ID.
 *
 * @return TRUE if \c eui64 is the local node's ID, otherwise FALSE.
 */
#define emberIsLocalEui64(eui64)                           \
(MEMCOMPARE(eui64, emLocalEui64, EUI64_SIZE) == 0)

/** @description Copies the current network parameters into the structure
 * provided by the caller.
 *
 * @param parameters: A pointer to an EmberNetworkParameters value
 *  into which the current network parameters will be copied.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberGetNetworkParameters(EmberNetworkParameters *parameters);

/** @description Gets the range of channels available to use. The range
 * depends on the radio in use and the value of \c TOKEN_RADIO_BANDS_SUPPORTED.
 *
 * @param channelMin: A pointer to storage where the minimum channel value
 * can be placed.
 *
 * @param channelMax: A pointer to storage where the maximum channel value
 * can be placed.
 */
void emberGetRadioChannelRange(int8u* channelMin, int8u* channelMax);

/** @description Gets the radio channel to which a node is set. The 
 * possible return values depend on the radio in use. For a list of 
 * available radio channels, see the technical specification for 
 * the RF communication module in your Developer Kit.
 *
 * @return Current radio channel.
 */
int8u emberGetRadioChannel(void);

/** @description DEPRECATED: Changes the radio channel for a node after it is 
 * associated with the network. Be sure to specify a channel supported by the radio.
 * The node may not associate properly with the network on the new channel,
 * and the transition to the new network will take several seconds in any
 * case. It is recommended that new applications call emberDisassociate(),
 * change the network parameters, and then call emberAssociate() instead of
 * making this call. If this call is used, the application is responsible for 
 * updating its copy of the network parameters.
 *
 * @param channel: A radio channel. For a list of available channels,
 *  see the technical specification for the RF communication module 
 *  in your Developer Kit.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
#ifdef DOXYGEN_SHOULD_SKIP_THIS
EmberStatus emberSetRadioChannel (int8u channel);
#else
// emberSetRadioChannel() no longer wraps emberDisassociate()
EmberStatus emRadioSetChannel(int8u channel);
#define emberSetRadioChannel emRadioSetChannel
#endif

/** @description Gets the radio output power at which a node is operating. Ember 
 * radios have discrete power settings. For a list of available power settings,
 * see the technical specification for the RF communication module in 
 * your Developer Kit. 
 *
 * @return Current radio output power, in dBm.
 */
int8s emberGetRadioPower(void);

/** @description DEPRECATED: Changes the radio output power for a node after it is
 * associated with the network. Be sure to specify a power setting supported by
 * the radio. It is recommended that new applications call emberDisassociate(),
 * change the network parameters, and then call emberAssociate() instead of
 * making this call. If this call is used, the application is responsible for updating
 * its copy of the network parameters.
 *
 * @param power: A power setting, in dBm. For the available settings,
 *  see the technical specification for the RF communication module 
 *  in your Developer Kit.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. See error-def.h for EmberStatus value
 *  definitions.
 */
#ifdef DOXYGEN_SHOULD_SKIP_THIS
EmberStatus emberSetRadioPower(int8s power);
#else
EmberStatus emRadioSetTransmitPower(int8s power);
#define emberSetRadioPower emRadioSetTransmitPower
#endif

/** @description Returns a node's 802.15.4 PAN identifier.
 *
 * @return A PAN ID.
 */
EmberPanId emberGetPanId(void);

/** @description DEPRECATED: Sets the node's 802.15.4 PAN identifier after
 * it is associated with the network. It is recommended that new applications
 * call emberDisassociate(), change the network parameters, and then call
 * emberAssociate() instead of making this call. If this call is used, the 
 * application is responsible for updating its copy of the network parameters.
 *
 * @param id: A 2-byte number that is a PAN ID.
 */
extern EmberPanId emLocalPanId;
#define emberSetPanId(id) do { emLocalPanId = (id); } while(0)



/** @description A callback invoked when the status of the stack changes. If the
 * status parameter equals EMBER_NETWORK_UP then the emberGetNetworkParameters() 
 * function can be called to obtain the new network parameters. If any of the 
 * parameters are being stored in nonvolatile memory by the application, the 
 * stored values should be updated.
 *
 * @param status: Stack status. One of the following:
 * - EMBER_NETWORK_UP
 * - EMBER_NETWORK_DOWN
 */
void emberStackStatusHandler(EmberStatus status);

#define RADIO_BANDS_SUPPORTED_868   BIT(0)
#define RADIO_BANDS_SUPPORTED_915   BIT(1)
#define RADIO_BANDS_SUPPORTED_433   BIT(2)
#define RADIO_BANDS_SUPPORTED_2400  BIT(3)
#define RADIO_BANDS_SUPPORTED_408   BIT(4)

/** @description Puts the EmberNet stack to sleep and turns the radio's
 * power off to conserve energy. If the stack has pending tasks, the call 
 * will return FALSE and should be made again. Once the stack is asleep, 
 * the application must call emberWakeUp() before calling any other 
 * stack function.
 *
 * @return TRUE if the stack went to sleep, and FALSE if there were pending 
 *  tasks and the stack could not sleep.
 */
boolean emberSleep(void);

/** @description Wakes the EmberNet stack from sleep and turns radio
 * power back on. If a node is sleeping, emberWakeUp() must be called on 
 * it before any other stack function can be called.
 *
 */
void emberWakeUp(void);
//@} //END POWER MGT FUNCTIONS

/** @description Useful to reference a single bit of a byte. 
 * <!--EVERYBODY defines this.--> 
 */
#define BIT(x) (1 << (x))

/** @description Sets \c bit in the \c reg register or byte. \b Note: 
 * Assuming \c reg is an IO register, some platforms (such as the
 * AVR) can implement this in a single atomic operation.
*/
#define SETBIT(reg, bit)      reg |= BIT(bit)

/** @description Sets the bits in the \c reg register or the byte 
 * as specified in the bitmask \c bits. \b Note: This is never a single 
 * atomic operation.
 */
#define SETBITS(reg, bits)    reg |= (bits)

/** @description Clears a bit in the \c reg register or byte. \b Note:
 * Assuming \c reg is an IO register, some platforms (such as the AVR) 
 * can implement this in a single atomic operation.
 */
#define CLEARBIT(reg, bit)    reg &= ~(BIT(bit))

/** @description Clears the bits in the \c reg register or byte 
 * as specified in the bitmask \c bits. \b Note: This is never a 
 * single atomic operation.
 */
#define CLEARBITS(reg, bits)  reg &= ~(bits)

/** @description Returns the value of \c bit within the register or 
 * byte \c reg.
*/
#define READBIT(reg, bit)     (reg & (BIT(bit)))

/** @description Returns the value of the bitmask \c bits within 
 * the register or byte \c reg.
 */
#define READBITS(reg, bits)   (reg & (bits))

// <!--Byte manipulation.  Similarly, everyone defines this.-->

#define LOW_BYTE(n)                     (int8u)((n) & 0xFF)
#define HIGH_BYTE(n)                    (int8u)(LOW_BYTE((n) >> 8))
#define HIGH_LOW_TO_INT(high, low)      (((high) << 8) + (low))

#ifndef DOXYGEN_SHOULD_SKIP_THIS

#define BUILD_DEF(name, value) \
    name = value,
  enum {
    #include "config/config.h"
  };
#undef BUILD_DEF

#endif //DOXYGEN_SHOULD_SKIP_THIS

/** @description Tells the stack to track beacons from the sender of the current
 * incoming beacon. \b Note: This function must be called from within
 * emberIncomingBeaconHandler() to begin tracking. \b Note: This function is
 * paired with the EMBER_MAC_BEACON_LOSS status value.  If tracked beacons are
 * not received within four beacon periods, the EMBER_MAC_BEACON_LOSS status value
 * is passed to the application through the emberStackStatusHandler() callback.
 *
 * @param track: Passing TRUE from inside emberIncomingBeaconHandler() will
 * cause the stack to track beacons matching the beacon that is currently being
 * handled. Passing TRUE from outside of emberIncomingBeaconHandler() will
 * return an error code without initiating tracking. Passing FALSE from anywhere
 * will disable tracking.
 *
 * @return An EmberStatus value indicating the success or
 * failure of the command. 
*/
EmberStatus emberBeaconTrack(boolean track);

/** @description Gets the time in symbols of the last beacon reception.
 *
 * @return The time in symbols of the last beacon reception.
*/
EmberBeaconTime emberGetLastBeaconReceiveTime(void);

#endif //__EMBER_COMMON_H__

