/** @file ember-leaf.h
 * @brief Leaf Node stack API
 *
 * Some functions in this file return an EmberStatus value. See 
 * error-defs.h for definitions of all EmberStatus return values.
 *
 * <!-- Author(s): Lee Taylor, lee@ember.com-->
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.                *80*-->
*/
 
/** @mainpage Leaf Node Stack Interface Guide
 * 
 * The Leaf Node stack is a simpler subset of the full EmberNet stack.
 * The primary API differences are that leaf nodes do not use binding tables and
 * use only datagram messages.
 *
 * \b Note: Except for the bootloader, leaf stacks can use the same HAL 
 * code as full stacks. 
 
 * The board files for the ATmega16 are different from those used for the ATmega128 
 * and the ATmega64. Use hal/dev16/board.h for the ATmega16, and use hal/dev/board.h 
 * for the ATmega128 and ATmega64. \b Note: The bootloader is not supported on the 
 * ATmega16 because there is not enough flash space to store both a bootloader and
 * a leaf node application. You must use an ISP programmer to update the application
 * on an ATmega16.
 *
 * The following #defines can be used to write smaller, simpler leaf node code.
 * They must be defined by the user application.
 *
 * - LEAF_NODE: Selects the leaf node stack code from common source files.
 * This define is the primary trigger differentiating full-stack node code from 
 * leaf node code. Some core functionality is drawn from different files, depending 
 * on node type (full-stack node, leaf node), while other core functionality is 
 * chosen using this define.
 * 
 * - SIMPLER_ASSERT_REBOOT: Simplifies asserting, to save code space. Normal 
 * asserting calls microReboot() for a little handling when asserting. When this 
 * define is used, microReboot() is not called, and the system goes into an
 * infinite loop instead.
 *
*/ 

#ifndef __EMBER_LEAF_H__
#define __EMBER_LEAF_H__

#ifdef __EMBER_H__
#error Do not include both ember.h and ember-leaf.h
#endif 

#define __EMBER_H__  // for ember-common.h
#include "ember-common.h"

typedef int8u *EmberMessageBuffer;

/** @description Sends a datagram to the node specified by the EUI64.
 *
 * @param eui64: The EUI64 of the intended recipient.
 *
 * @param localTerminal: The terminal the message is sent from.
 *
 * @param remoteTerminal: The terminal the message is sent to.
 *
 * @param message: A pointer to the contents of the datagram. The first
 * byte of the datagram gives the message length.                                                         
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSendDatagram(EmberEUI64 eui64,
                              EmberTerminal localTerminal,
                              EmberTerminal remoteTerminal,
                              int8u *message);

/** @description Sends a datagram to all full-stack nodes that are
 * one hop away.
 *
 * @param localTerminal: The terminal the message is sent from.
 *
 * @param remoteTerminal: The terminal the message is sent to.
 *
 * @param message: A pointer to the contents of the datagram. The first
 * byte of the datagram gives the message length.                                                         
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSendLinkBroadcast(EmberTerminal localTerminal,
                                   EmberTerminal remoteTerminal,
                                   int8u *message);
                              
/** @description A callback invoked by the Leaf Node stack either when a datagram 
 * has been received or the stack has given up trying to send one. 
 *
 * @param message: A pointer to the message that was sent. The first
 * byte of the datagram gives the message length.                              
 *
 * @param status: An EmberStatus value indicating whether the message was 
 * received (that is, whether the receiver ACKed the message) 
 * or the stack gave up trying to send it.
*/
void emberMessageSent(int8u *message,
                      EmberStatus status);

enum {
  EMBER_INCOMING_LINK
};
typedef int8u EmberIncomingMessageType;


/** @description A callback invoked by the Leaf Node stack when a message is
 * received.
 *
 * @param type: The type of the incoming message. Only one type currently
 * exists for the Leaf Node stack:
 *  - EMBER_INCOMING_LINK
 *
 * @param eui64: The EUI64 of the sender.
 *
 * @param localTerminal: The terminal the message was sent to.
 * 
 * @param remoteTerminal: The terminal the message was sent from.
 * 
 * @param message: A pointer to message that was sent. The first
 * byte of the datagram gives the message length.             
*/ 
void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 EmberEUI64 eui64,
                                 EmberTerminal localTerminal,
                                 EmberTerminal remoteTerminal,
                                 int8u *message);

/** @description A callback invoked by the stack when a beacon is received.
 * \b Note: The shorter the beacon interval, the greater the processor resources
 * beaconing will consume.  When using very short beacons, please be careful to
 * limit processing in the emberIncomingBeaconHandler() since very little
 * processing time may be available between beacons.
 *
 * @param sender: The EUI64 of the beacon sender.
 *
 * @param payload: A pointer to the beacon's payload, with the first byte
 * the length. If payload is NULL, there is no payload.
*/ 
void emberIncomingBeaconHandler(EmberEUI64 sender, int8u* payload);

#endif //__EMBER_LEAF_H__

