/** @file ember-static-configuration.h
 * @brief User-configurable stack memory allocation and convenience stubs 
 * for little-used callbacks. 
 *
 * \b Note: Application developers should \b not modify any portion
 * of this file. Doing so may lead to mysterious bugs. Allocations should be 
 * adjusted only with the macros, as described below. 
 *
 * The main source file of an application should include this file in
 * order to allocate memory required by the stack. <b>It should only be 
 * included a single time for any given application.</b>
 *
 * All allocations have default sizes, so simply including the file is
 * enough for many applications. You can, however, override these defaults
 * by #defining the appropriate macro to a different size.  For example the 
 * following reduces the number of allocated packet buffers to 8:
 *
 * #define EMBER_PACKET_BUFFER_COUNT 8
 *
 * The convenience stubs provided in this file can be overridden by defining
 * the appropriate macro and providing the corresponding callback function.
 * For example, an application employing beacons needs to implement 
 * emberIncomingBeaconHandler() to process incoming beacons. Along with
 * the function definition, the application should provide the following
 * line before the inclusion of this file:
 *
 * #define EMBER_APPLICATION_HAS_USER_INCOMING_BEACON_HANDLER
 *
 * <!-- Author(s): Lee Taylor, lee@ember.com -->
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.            *80*-->
 */

#ifdef __EMBER_STATIC_CONFIG_H__ 
  #error ember-static-configuration.h should not be included more than once
#endif
#define __EMBER_STATIC_CONFIG_H__

#ifndef DOXYGEN_SHOULD_SKIP_THIS 
#include "ember-static-struct.h" // Required typedefs
#endif //DOXYGEN_SHOULD_SKIP_THIS

#ifndef EMBER_PACKET_BUFFER_COUNT
  #define EMBER_PACKET_BUFFER_COUNT 24
#endif
#ifndef EMBER_TRANSPORT_CONNECTION_COUNT
  #define EMBER_TRANSPORT_CONNECTION_COUNT 10
#endif
#ifndef EMBER_TRANSPORT_PACKET_COUNT
  #define EMBER_TRANSPORT_PACKET_COUNT 10
#endif
#ifndef EMBER_BINDING_TABLE_SIZE
  #define EMBER_BINDING_TABLE_SIZE 8
#endif
#ifndef EMBER_TEMPORARY_BINDING_ENTRIES
  #define EMBER_TEMPORARY_BINDING_ENTRIES 8
#endif
#ifndef EMBER_MAXIMUM_AGGREGATION_ROUTES
  #define EMBER_MAXIMUM_AGGREGATION_ROUTES 0
#endif
#ifndef EMBER_COST_TABLE_SIZE
  #define EMBER_COST_TABLE_SIZE 64
#endif

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// *****************************************
// Memory Allocations & declarations
// *****************************************

// ------------- Packet Buffers ------------------
// used to have static sys creator 'bl'
const int8u emPacketBufferCount = EMBER_PACKET_BUFFER_COUNT;

#define EMBER_MESSAGE_BUFFER_COUNT EMBER_PACKET_BUFFER_COUNT

// The actual memory for buffers.
int8u emPacketBufferData[EMBER_PACKET_BUFFER_COUNT * 32];

// All message buffers have lengths and reference counts.  Application buffers
// also have a data pointer and a size, packet buffers also have a link field.

int8u emMessageBufferLengths[EMBER_PACKET_BUFFER_COUNT];
int8u emMessageBufferReferenceCounts[EMBER_PACKET_BUFFER_COUNT];
int8u emPacketBufferLinks[EMBER_PACKET_BUFFER_COUNT];

// ---------------- Binding Table --------------------

// The size is taken from stack/transport/binding-table.c
// This is also defined in token-stack.h
#define EMBER_BINDING_ENTRY_SIZE 11

const int8u emberBindingTableSize = EMBER_BINDING_TABLE_SIZE;
const int8u emberBindingTokenEntries = (sizeof(TOKEN_BINDING_TABLE) / 
                                         EMBER_BINDING_ENTRY_SIZE);
const int8u emberTemporaryBindingEntries = EMBER_TEMPORARY_BINDING_ENTRIES;
const int8u emberMaximumAggregationRoutes = EMBER_MAXIMUM_AGGREGATION_ROUTES;

#if EMBER_TEMPORARY_BINDING_ENTRIES
int8u emTemporaryBindingTable[EMBER_TEMPORARY_BINDING_ENTRIES
                              * EMBER_BINDING_ENTRY_SIZE];
#else
int8u emTemporaryBindingTable[1];
#endif

// Copied from stack/core/route-record.h.
#define EMBER_ROUTE_RECORD_MAX_ENTRY_COUNT 6
#define EMBER_ROUTE_RECORD_ENTRY_SIZE 2

int8u emBindingLinks[EMBER_BINDING_TABLE_SIZE];
int8u emBindingConnection[EMBER_BINDING_TABLE_SIZE];
int16u emBindingRemoteNode[EMBER_BINDING_TABLE_SIZE];
int8u emBindingGradientState[EMBER_BINDING_TABLE_SIZE];
int8u emBindingRemoteIndex[EMBER_BINDING_TABLE_SIZE];

#if EMBER_MAXIMUM_AGGREGATION_ROUTES
int8u emReplyRoutes[EMBER_MAXIMUM_AGGREGATION_ROUTES
                    * EMBER_ROUTE_RECORD_MAX_ENTRY_COUNT
                    * EMBER_ROUTE_RECORD_ENTRY_SIZE];
int8u emReplyRouteLengths[EMBER_MAXIMUM_AGGREGATION_ROUTES];
#else
int8u emReplyRoutes[1];
int8u emReplyRouteLengths[1];
#endif

// ------------- Transport Records ---------------
const int8u emMaxTransportConnections = EMBER_TRANSPORT_CONNECTION_COUNT;
EmConnectionData emConnectionData[EMBER_TRANSPORT_CONNECTION_COUNT];

const int8u emMaxTransportPackets = EMBER_TRANSPORT_PACKET_COUNT;
EmPacketData emPacketData[EMBER_TRANSPORT_PACKET_COUNT];

// ------------- Cost Table ------------------

EmCostTableEntry emCostTable[EMBER_COST_TABLE_SIZE];
const int8u emberCostTableSize = EMBER_COST_TABLE_SIZE;

// ------------- Convenience Stubs -----------
#ifndef EMBER_APPLICATION_HAS_INCOMING_BEACON_HANDLER
// Note that there are two different versions of the beacon handler because
// the leaf stack doesn't provide buffer support.  
void emberIncomingBeaconHandler(EmberEUI64 sender, EmberMessageBuffer payload) {}

#endif // USER_INCOMING_BEACON_HANDLER

#ifndef EMBER_APPLICATION_HAS_INCOMING_MFG_TEST_MESSAGE_HANDLER
void emberIncomingMfgTestMessageHandler(int8u messageType, 
                                        int8u dataLength, 
                                        int8u *data) {}
#endif

#ifndef EMBER_APPLICATION_HAS_INCOMING_MONITOR_SERVICE_MESSAGE_HANDLER
void emberIncomingMonitorServiceMessageHandler(EmberMessageBuffer message,
                                               EmberEUI64 sender) {}
#endif

#endif //DOXYGEN_SHOULD_SKIP_THIS

