/**
 * @file ember.h
 * @brief Basic EmberNet stack APIs.
 *
 * See config.h for the EmberNet stack build configuration settings.
 *
 * Many functions in this file return an EmberStatus value. See 
 * error-def.h for definitions of all EmberStatus return values.
 *
 * <!-- Author(s): Lee Taylor, lee@ember.com,        -->
 * <!--            Richard Kelsey, richard@ember.com -->
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.         *80*-->
 */
 
/** @mainpage EmberNet Stack Interface Guide
 * 
 * The EmberNet stack APIs can be sorted as follows:
 * - General stack functions
 * - Power functions
 * - Device configuration functions
 * - Binding table functions
 * - Stack buffer functions
 * - Message buffer functions
 * - Discovery functions
 * - Connection functions
 * - Message functions
 * - Callback functions
 * - Debug functions
 */ 

#ifndef __EMBER_H__
#define __EMBER_H__

#ifdef __EMBER_LEAF_H__
#error Do not include both ember.h and ember-leaf.h
#endif 

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// Common definitions
#include "ember-common.h"
#endif

/** @name Configuration Data */
//@{
/** @description Defines interface types. See EmberInterfaceId for 
 * descriptions of these values. Application developers may add others
 * as needed.
 */
enum {
  ENDPOINT_ZERO_INTERFACE = 0,
  AD_HOC_INTERFACE        = 1,             
  ZIGBEE_INTERFACE        = 2,
  EUI64_INTERFACE         = 3
};

typedef int8u EmberInterfaceDomain;

/** @description Defines endpoint types. See EmberInterfaceId for
 * descriptions of these values.
 */
enum { 
  SYMMETRIC_ENDPOINT          = 0,
  SERVER_ENDPOINT             = 1,
  CLIENT_ENDPOINT             = 2,
  AGGREGATION_SERVER_ENDPOINT = 3,
  AGGREGATION_CLIENT_ENDPOINT = 4
};

typedef int8u EmberEndpointType;

/** @description The specification field of an interface ID (see 
 * EmberInterfaceId); may contain up to nine bytes. This maximum occurs in 
 * an AD_HOC_INTERFACE that uses eight bytes of application-assigned ID plus 
 * the required one-byte length prefix (which is eight in this case). 
 */
#define EMBER_MAXIMUM_INTERFACE_ID_SPECIFICATION 9

#define EMBER_MAXIMUM_INTERFACE_ID_SIZE                 \
  (2 + EMBER_MAXIMUM_INTERFACE_ID_SPECIFICATION)

/**@brief Identifies an interface in an EmberNet stack application.
 * @description Every interface in an application must be assigned an ID
 * that allows specific interfaces to be addressed in discovery. 
 * This ID consists of an ID type, or domain, and a unique number whose 
 * length and format are specific to the domain. The EmberNet stack API 
 * defines 4 interface domains, but others may be added.
 */
typedef struct {
  /** The type (domain) of the interface ID. Can take one of the following
   * values or an application-specific domain value:
   * - EUI64_INTERFACE (8-bytes): Interface that uses a fixed 64-bit 
   *   IEEE address as its unique number. (Note that these IDs are 
   *   completely different from the node EUI IDs.)
   * - ZIGBEE_INTERFACE (2-bytes): Interface that uses a Zigbee-assigned 
   *   ID as its unique number. 
   * - AD_HOC_INTERFACE (up to 9 bytes, with the first byte specifying 
   *   the number of bytes that follow):  Interface that uses an 
   *   application-assigned ID as its unique number; for use in closed 
   *   networks where interoperability with devices from other manufacturers
   *   is not desired.
   * - ENDPOINT_ZERO_INTERFACE (2 bytes): Administrative interface; 
   *   defined on endpoint 0.*/
  EmberInterfaceDomain domain;
  
  /** The endpoint type. Can take one of the following values: 
   * - SYMMETRIC_ENDPOINT: An endpoint that implements a symmetric interface 
   *   and thereby expects to send and receive the same type of message. This 
   *   message may be of any type.
   * - SERVER_ENDPOINT: An endpoint that implements an asymmetric interface 
   *   and thereby expects to send one type of message and to receive a 
   *   different type. A server endpoint must be paired to a client endpoint
   *   that implements the reciprocal side of the server's interface.
   * - CLIENT_ENDPOINT: A client endpoint must be paired to a server endpoint 
   *   that implements the reciprocal side of the client's interface. 
   * - AGGREGATION_SERVER_ENDPOINT: A server endpoint that communicates 
   *   with a large number of aggregation client endpoints. 
   * - AGGREGATION_CLIENT_ENDPOINT: A client endpoint that communicates with 
   *   an aggregation server endpoint.*/
  EmberEndpointType type;
  /** An interface ID. The value of this parameter is determined by the 
   * ID's domain. */
  int8u specification[EMBER_MAXIMUM_INTERFACE_ID_SPECIFICATION];
} EmberInterfaceId;

#define EMBER_MAXIMUM_ENDPOINT_INTERFACES 8

/** @brief Defines the interfaces that an endpoint implements.*/
typedef struct {
  /** Defines the application-specific interface number (0 to 7). */
  int8u number;
  /** A pointer to an interface. */
  PGM EmberInterfaceId *interface;
} EmberEndpointDescription[EMBER_MAXIMUM_ENDPOINT_INTERFACES];

/** @brief A structure that defines an endpoint. 
 * @description 
 * An application must maintain an array of all its endpoints. 
 * Every endpoint in this array must be indexed in an EmberEndpoint 
 * struct and associated with an EmberEndpointDescription.
 */
typedef struct {
  /** The index of the endpoint in the application's endpoint array.*/
  int8u number;
  /** A pointer to an EmberEndpointDescription, which identifies an 
   * interface that this endpoint will implement */
  PGM EmberEndpointDescription *description;
} EmberEndpoint;
//@} //END CONFIG DATA



/** @name Public EmberNet Stack Functions */
//@{

/** @description Changes the routing participation of the node after it is
 * associated with the network. By default, nodes forward messages on behalf of
 * their neighbors, but such considerations as power consumption and node
 * density may make forwarding undesirable. The node will disassociate and then
 * associate with the network using the new routing setting. A call to the \c
 * emberStackStatusHandler() callback informs you that the node is associated
 * using the new setting.
 *
 * @param enable: TRUE to enable routing, FALSE to disable routing.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberEnableRouting(boolean enable);

/** @description Defines the number of endpoints used by an application.
 * This variable must be between 0 and 31, inclusive. The EmberNet stack 
 * uses this to access the emberEndpoints variable. If it is not set 
 * correctly, stack behavior is undefined. The application must provide a 
 * definition for this variable. 
 */
extern int8u emberEndpointCount;

/** @description Allows the stack to access user-defined endpoints. This 
 * array must be at least "1" long (even if no endpoints are defined).
 * The application must provide a definition for this variable. 
 */
extern EmberEndpoint emberEndpoints[];

/** @description Sets an entry in the binding table by copying the structure 
 * pointed to by \c value into the binding table. \b Note: You do not need to 
 * reserve memory for \c value.
 *
 * @param index: The index of a binding table entry.
 *
 * @param value: A pointer to a structure.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSetBinding(int8u index, EmberBindingTableEntry *value);

/** @description Copies a binding table entry to the structure that 
 * \c result points to.
 *
 * @param index: The index of a binding table entry.
 *
 * @param result: A pointer to the location to which to copy the binding table entry.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberGetBinding(int8u index, EmberBindingTableEntry *result);

/** @description Deletes a binding table entry.
 *
 *
 * @param index: The index of a binding table entry.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberDeleteBinding(int8u index);

/** @description Indicates whether a binding table entry is active
 * --that is, whether a connection to it is open or any messages are 
 * enroute from it. Note that this function does not indicate whether a binding
 * is clear. To determine whether a binding is clear, check the 
 * EmberBindingTableEntry structure that defines the binding. The type 
 * field should have the value EMBER_UNUSED_BINDING.
 *
 * @param index: The index of a binding table entry.
 *
 * @return TRUE if the binding table entry is active, FALSE if it is not.
 */
boolean emberBindingIsActive(int8u index);

/** @description Deletes all binding table entries.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberClearBindingTable(void);

/** @description Returns the cost associated with sending a message
 * to a binding table entry. The term "cost" is specific to the 
 * EmberNet routing algorithm and gives a measure of distance between 
 * nodes. The cost takes both hop count and link reliability into 
 * account. Less reliable links are given higher costs.
 *
 * @param index: The index of a binding table entry.
 *
 * @param returnLoc: A pointer to place to store the cost.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberGetCost(int8u index, EmberCost *returnLoc);

/** @description Gets the size of an interface.  
 * 
 * @param interfaceId: A pointer to an EmberInterfaceID.
 * 
 * @return The size of the interface ID.
 */
int8u emberInterfaceIdSize(EmberInterfaceId *interfaceId);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// These are defined in ember-static-memory.h.
extern const int8u emberBindingTableSize;
extern const int8u emberTemporaryBindingEntries;
extern const int8u emberMaximumAggregationRoutes;
#endif

#define emberNonvolatileBindingEntries               \
 (emberBindingTableSize - emberTemporaryBindingEntries)
  
/** @description Determines whether a binding is permanent.
 * This function has been implemented as a macro for efficiency.
 *
 * @param index: The index of a binding table entry.
 *
 * @return TRUE if the binding table entry is nonvolatile, 
 *  otherwise FALSE.
 */
#define emberBindingIsNonvolatile(index)                \
 ((index) < emberNonvolatileBindingEntries)

/** @description Determines whether a binding is temporary. 
 * This function has been implemented as a macro for efficiency.
 *
 * @param index: The index of a binding table entry.
 *
 * @return TRUE if the binding table entry is temporary, 
 *  otherwise FALSE.
 */
#define emberBindingIsTemporary(index)                  \
 (emberNonvolatileBindingEntries <= (index))

/** @description Starts a local search for endpoints that implement 
 * a specified set of interfaces. 
 *
 * @param desiredInterfaces: A pointer to an array of pointers to the
 *  interfaces being searched for.
 *
 * @param interfaceCount: The number of interfaces listed in the given array.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberStartEndpointSearch(EmberInterfaceId **desiredInterfaces,
                                     int8u interfaceCount);

/** @description Stops the ongoing local endpoint search.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberStopEndpointSearch(void);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/** @description Set the encryption key to be used in encrypting and decrypting
 * radio messages. This function does not work if the stack is already associated with
 * a network.
 *  
 * @param key: A pointer to a 16-byte encryption key (EMBER_ENCRYPTION_KEY_SIZE).
 *
 * @return An EmberStatus value indicating the success or failure of the command.
 */
EmberStatus emberSetEncryptionKey(int8u *key);
#endif

//@} //END PUBLIC EMBERNET STACK FUNCTIONS 

/** @name Common Message Buffers Functions
 * See packet-buffer.h for APIs related to stack and linked buffers.
 */
//@{

typedef int8u EmberMessageBuffer;

/** @description Provides the message buffer equivalent of NULL.*/
#define EMBER_NULL_MESSAGE_BUFFER 0xFF

/** @description Returns buffer contents in the form of a pointer to 
 * the buffer. This pointer can be used to access only the first 
 * PACKET_BUFFER_SIZE bytes in the buffer. To read a message composed
 * of multiple buffers, use emberCopyFromLinkedBuffers().
 *
 * @param buffer: A buffer.
 *
 * @return A pointer to the first byte in the buffer, or NULL if the 
 *  buffer does not have any associated storage.
 */
int8u *emberMessageBufferContents(EmberMessageBuffer buffer);

/** @description Returns the length of a buffer. Implemented as a macro 
 * for improved efficiency.
 *
 * @param buffer: A buffer.
 *
 * @return Buffer length.
 */
#define emberMessageBufferLength(buffer)   (emMessageBufferLengths[buffer])

/** @description Sets the length of a buffer. When asserts are enabled, 
 * attempting to set a length greater than the size of the buffer triggers 
 * an assert.
 *
 * @param buffer: A buffer
 *
 * @param newLength: The length to set the buffer to.
 */
void emberSetMessageBufferLength(EmberMessageBuffer buffer, int8u newLength);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#ifdef EM_DEBUG_BUFFER_USE
#define EM_BUFFER_DEBUG(X) X
#define EM_BUFFER_FILE_LOC , __FILE__, __LINE__
#define EM_BUFFER_FILE_DECL , char *file, int line
#define EM_BUFFER_FILE_VALUE , file, line
#else
#define EM_BUFFER_DEBUG(X)
#define EM_BUFFER_FILE_LOC
#define EM_BUFFER_FILE_DECL
#define EM_BUFFER_FILE_VALUE
#endif
#endif //DOXYGEN_SHOULD_SKIP_THIS

/** @description Holds a message buffer by incrementing its reference count.
 * Implemented as a macro for improved efficiency.
 *
 * @param buffer: A buffer.
 */
#define emberHoldMessageBuffer(buffer)                                        \
do {                                                                          \
  EmberMessageBuffer EM_HOLD_BUFFER_TEMP_XXX = (buffer);                      \
  EM_BUFFER_DEBUG(                                                            \
    fprintf(stderr, "[holding %d (%s %d), refs %d]\n",                        \
            EM_HOLD_BUFFER_TEMP_XXX, __FILE__, __LINE__,                      \
            emMessageBufferReferenceCounts[EM_HOLD_BUFFER_TEMP_XXX]);         \
    assert(emMessageBufferReferenceCounts[EM_HOLD_BUFFER_TEMP_XXX] > 0);      \
  )                                                                           \
  emMessageBufferReferenceCounts[EM_HOLD_BUFFER_TEMP_XXX]++;                  \
} while (FALSE) 

/** @description Releases a message buffer by decrementing its reference 
 * count. Implemented as a macro for improved efficiency.
 *
 * @param buffer: A buffer.
 */
#define emberReleaseMessageBuffer(buffer)                                     \
do {                                                                          \
  EmberMessageBuffer EM_RELEASE_BUFFER_TEMP_XXX = (buffer);                   \
  EM_BUFFER_DEBUG(                                                            \
    fprintf(stderr, "[releasing %d (%s %d), refs %d]\n",                      \
            EM_RELEASE_BUFFER_TEMP_XXX, __FILE__, __LINE__,                   \
            emMessageBufferReferenceCounts[EM_RELEASE_BUFFER_TEMP_XXX]);      \
    assert(emMessageBufferReferenceCounts[EM_RELEASE_BUFFER_TEMP_XXX] > 0);   \
  )                                                                           \
  if ((--(emMessageBufferReferenceCounts[EM_RELEASE_BUFFER_TEMP_XXX])) == 0)  \
    emFreeMessageBuffer(EM_RELEASE_BUFFER_TEMP_XXX EM_BUFFER_FILE_LOC);       \
} while (FALSE)

#ifndef DOXYGEN_SHOULD_SKIP_THIS
void emFreeMessageBuffer(EmberMessageBuffer buffer EM_BUFFER_FILE_DECL);
extern int8u emMessageBufferLengths[];
extern int8u emMessageBufferReferenceCounts[];
#endif //DOXYGEN_SHOULD_SKIP_THIS
//@} //END Common Message Buffers APIs

/** @name Message Functions */
 //@{

 /** @description Sends a datagram to the node and terminal specified in a 
 * binding table entry.
 *
 * @param bindingTableIndex: The index of the binding table entry specifying 
 *  the node and terminal to which a datagram is being sent.
 *
 * @param message: Contents of the datagram.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberSendDatagram(int8u bindingTableIndex,
                              EmberMessageBuffer message);
                              
/** @description Sends a multicast message to all endpoints that share
 * a specific multicast ID, to nodes a specified number of hops away. 
 *
 * @param bindingTableIndex: Index for a binding table entry (<b>must be 
 *  multicast type</b>).
 *
 * @param message: A message.
 *
 * @param hops: The number of hops to send the message.
 * - For the value of 0, the multicast is not relayed and thus is only heard
 *   by the sender's neighbors. 
 * - For the value of N, the multicast is relayed N times and reaches all 
 *   nodes up to N + 1 hops away.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSendLimitedMulticast(int8u bindingTableIndex,
                                      EmberMessageBuffer message,
                                      int8u hops);
                                      
#ifndef EMBER_MULTICAST_HOPS
/** @description By default multicasts only go out a limited distance. */
#define EMBER_MULTICAST_HOPS 12
#endif

/** @description Sends a multicast message to all endpoints that share
 * a specific multicast ID. Implemented as a macro for improved efficiency.
 *
 * @param bindingTableIndex: The index of the binding table entry specifying 
 *  the multicast group.
 *
 * @param message: A message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
#define emberSendMulticast(bindingTableIndex, message) \
(emberSendLimitedMulticast(bindingTableIndex, message, EMBER_MULTICAST_HOPS))

/** @description Sends a sequenced message over the connection associated
 * with a specified binding table entry.
 *
 * @param bindingTableIndex: The index of the binding table entry specifying the 
 *  message destination.
 *
 * @param message: A message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberSendSequenced(int8u bindingTableIndex,
                               EmberMessageBuffer message);

/** @description Sends a datagram to the binding table entry whose local
 * terminal is \c sender. \b Note: Use this function only when the terminal has 
 * one binding table entry.
 *
 * @param sender: A terminal.
 *
 * @param message: A datagram.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
// Same again with destination given by the sending (local) terminal.
EmberStatus emberSendDatagramFromTerminal(EmberTerminal sender,
                                          EmberMessageBuffer message);
                                          
/** @description Sends a multicast message to all endpoints that share
 * the sending terminal's multicast ID. \b Note: Use this function only when 
 * the sending terminal has one binding table entry.
 *
 * @param sender: A terminal.
 *
 * @param message: A message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSendMulticastFromTerminal(EmberTerminal sender,
                                           EmberMessageBuffer message);
                                           
/** @description Sends a sequenced message to the remote partner of a 
 * specified terminal. \b Note: Use this function only when the sending
 * terminal has one binding table entry.
 *
 * @param sender: A terminal.
 *
 * @param message: A message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberSendSequencedFromTerminal(EmberTerminal sender,
                                           EmberMessageBuffer message);

/** @description Cancels an outgoing message.
 *
 * @param message: A message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberCancelMessage(EmberMessageBuffer message);

/** @description Sends a reply for an application that has received a datagram
 * message.  The reply will be included with the ACK
 * that the stack automatically sends back.
 *
 * \b Note: This function may be called only from within 
 * emberIncomingMessageHandler().
 * 
 * @param reply: A reply message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSendReply(EmberMessageBuffer reply);

/** @description Provides the cost to send a message to the sender of 
 * the current message. The term "cost" is specific to the EmberNet 
 * routing algorithm and gives a measure of distance between nodes. 
 * The cost takes both hop count and link reliability into account. 
 * Less reliable links are given higher costs.
 *
 * @param returnLoc: A pointer to a location to store the cost.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberGetCostToSender(EmberCost *returnLoc);

/** @description Gets the instantaneous link quality indication (LQI) based
 * upon the first 8 symbol periods of the current message as received
 * from the last relaying node. 
 * \b Note: This function may only be called from within
 * emberIncomingMessageHandler().
 *
 * @param lastHopLqi: The instantaneous quality of the link.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberGetLastHopLqi(int8u *lastHopLqi);

/** @description Gets the instantaneous received signal strength indicator
 * (RSSI) based upon the first 8 symbol periods of the current message as 
 * received from the last relaying node. 
 * \b Note: This function may only be called from within
 * emberIncomingMessageHandler().
 *
 * @param lastHopRssi: The instantaneous received signal strength 
 * of the current message.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberGetLastHopRssi(int8s *lastHopRssi);


/** @description Creates a binding table entry for the sender of a datagram 
 * or multicast, which can be used to send messages to the sender. 
 * This function is identical to emberSetBinding() except that calling it
 * tells the stack that this binding corresponds to the sender of the 
 * current message. The stack uses this information to associate the 
 * sender's routing info with the binding table entry. \b Note: This 
 * function may only be called from within emberIncomingMessageHandler().
 *
 * @param index: The index of the binding to set.
 *
 * @param entry: A pointer to data for the binding.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberSetReplyBinding(int8u index, EmberBindingTableEntry *entry);

/** @description Updates the routing information associated with a binding
 * table entry for the sender of a datagram or multicast.  This function
 * should be used in place of emberSetReplyBinding() when a multicast or
 * datagram arrives from a remote terminal for which a binding already exists.
 *
 * @param index: The index of the binding to update.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberNoteSendersBinding(int8u index);

/** @description Returns the terminal from which the current message
 * was sent. This function may only be called from within 
 * emberIncomingMessageHandler(), and only when the current message 
 * is a datagram.
 *
 * @param returnLoc: A pointer to a place to store the terminal.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberGetSendingTerminal(EmberTerminal *returnLoc);

/** @description Returns the EUI64 of the sender of the current
 * incoming link-level message. \b Note: This function may be called
 * only from within emberIncomingMessageHandler() with
 * EMBER_INCOMING_LINK and EMBER_INCOMING_LINK_BROADCAST message
 * types.
 *
 * @param eiu64: A pointer to a variable into which to store the EUI64.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberGetSenderEui64(EmberEUI64 eui64); // For link messages

/** @description Opens a connection to a node.
 *
 * @param bindingTableIndex: The index of the binding table entry for 
 *  the node to which a connection is being opened.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberOpenConnection(int8u bindingTableIndex);

/** @description Returns the connection status of a binding table entry. 
 *
 * @param bindingTableIndex: The index of the binding table entry 
 *  for the node whose connection is being closed.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberConnectionStatus(int8u bindingTableIndex);

/** @description Closes a connection. Explicitly closing unused 
 * connections improves the performance of the network and frees 
 * up resources on the node.
 *
 * @param bindingTableIndex: The index of the binding table entry whose 
 *  status is being queried.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command. 
 */
EmberStatus emberCloseConnection(int8u bindingTableIndex);

/** @description A callback invoked by the stack when it has completed
 * sending a datagram or sequenced message. \b Note: Make sure to check
 *\c status first because \c bindingTableIndex and \c message
 * are meaningless if the status is EMBER_MAC_BEACON_SENT.
 *
 * @param bindingTableIndex: The index of the binding table entry to which
 *  the message was sent.
 *
 * @param message: The message that was sent.
 *
 * @param status: An EmberStatus value of EMBER_SUCCESS if an ACK was received
 * from the destination, EMBER_DELIVERY_FAILED if no ACK was received, or
 * EMBER_MAC_BEACON_SENT if a beacon was just sent.
 */
void emberMessageSent(int8u bindingTableIndex,
                      EmberMessageBuffer message,
                      EmberStatus status);

/** @description Type for an incoming message.*/
enum {
  EMBER_INCOMING_DATAGRAM,
  EMBER_INCOMING_DATAGRAM_REPLY,
  EMBER_INCOMING_SEQUENCED,
  EMBER_INCOMING_MULTICAST,
  EMBER_INCOMING_SHARED_MULTICAST,
  EMBER_INCOMING_MULTICAST_LOOPBACK,
  EMBER_INCOMING_LINK,
  EMBER_INCOMING_LINK_BROADCAST
};

typedef int8u EmberIncomingMessageType;

/** @description A callback invoked by the EmberNet stack when a message is
 * received. Incoming datagrams may arrive from senders for which there 
 * is no entry in the local binding table, in which case bindingTableIndex 
 * takes the value EMBER_NULL_BINDING.
 *
 * If the binding table has multiple entries with the same multicast ID, 
 * emberIncomingMessageHandler() is called once for each such entry. 
 * All but the last have type EMBER_INCOMING_SHARED_MULTICAST. The last has 
 * type EMBER_INCOMING_MULTICAST.
 * 
 * If a node sends a multicast to a multicast ID for which it has multiple 
 * bindings, emberIncomingMessageHandler() is called for the entries other
 * than the one being sent to. These calls have the type 
 * EMBER_INCOMING_MULTICAST_LOOPBACK.
 *
 * The following functions may be called from within 
 * emberIncomingMessageHandler():
 * - emberGetCostToSender()
 * - emberSendReply() (for incoming datagrams only)
 * - emberGetSendingTerminal() (for incoming datagrams only)
 * - emberSetReplyBinding() (for incoming datagram and multicasts only)
 * - emberNoteSendersBinding() (for incoming datagram and multicasts only)
 *
 * @param type: The type of the incoming message. One of the following:
 *  - EMBER_INCOMING_DATAGRAM
 *  - EMBER_INCOMING_DATAGRAM_REPLY
 *  - EMBER_INCOMING_SEQUENCED
 *  - EMBER_INCOMING_MULTICAST
 *  - EMBER_INCOMING_SHARED_MULTICAST
 *  - EMBER_INCOMING_MULTICAST_LOOPBACK
 *  - EMBER_INCOMING_LINK
 *  - EMBER_INCOMING_LINK_BROADCAST
 *
 * @param terminal: The terminal the message was sent to.
 *
 * @param index: The index of the binding table entry to which the message was
 *  sent, or EMBER_NULL_BINDING.
 *
 * @param message: The message that was sent.
 */
void emberIncomingMessageHandler(EmberIncomingMessageType type,
                                 EmberTerminal terminal,
                                 int8u index,
                                 EmberMessageBuffer message);

/** @description A handler that the stack calls when the connection
 * status for a binding changes.
 *
 * @param bindingTableIndex: The index of the binding table entry for 
 * the connection.
 *
 * @param status: The new connection status for the binding table 
 * entry. One of the following:
 *  - EMBER_CONNECTION_OPEN
 *  - EMBER_CONNECTION_OPEN_FAILED
 *  - EMBER_CONNECTION_FAILED
 *  - EMBER_CONNECTION_CLOSING
 *  - EMBER_CONNECTION_CLOSED
 */
void emberConnectionStatusHandler(int8u bindingTableIndex, int8u status);

/** @description A callback invoked when a remote node requests a 
 * change to the local binding table (via endpoint 0, the 
 * administrative interface). The application is free to make the 
 * requested change, ignore the request, or take some other action.
 *
 * @param index: The index of the binding to set.
 *
 * @param entry: A pointer to a new binding table entry.
 *
 * @return An EmberStatus value indicating the success or
 *  failure of the command.
 */
EmberStatus emberRemoteSetBindingHandler(int8u index,
                                         EmberBindingTableEntry *entry);

/** @description Value representing the distance between neighboring nodes. */
typedef int8u EmberNetworkDistance;

/** @description A callback invoked during an interface search initiated 
 * by a call to emberStartEndpointSearch(), as appropriate nodes are 
 * found. The node exposes the given interfaces. These interfaces may 
 * comprise all or a subset of those passed to emberStartEndpointSearch(). 
 * Found endpoints are provided approximately in order of distance from 
 * the searching node, from nearest to farthest.
 *
 * @param desiredInterface: A pointer to the EmberInterfaceId structure 
 *  containing the found interface.
 *
 * @param node: The 64-bit ID of the node on which the interface was found.
 *
 * @param distance: An 8-bit value representing the distance of \c node from 
 * the searching node.
 */
void emberEndpointFoundHandler(EmberInterfaceId *desiredInterface,
                               EmberEUI64 node,
                               EmberNetworkDistance distance);

//@} // END MESSAGE FUNCTIONS

/** @description A callback invoked during an interface search initiated 
 * by a call to emberStartEndpointSearch(), as appropriate nodes are 
 * found. The node exposes the given interfaces. These interfaces may 
 * comprise all or a subset of those passed to emberStartEndpointSearch(). 
 * Found endpoints are provided approximately in order of distance from 
 * the searching node, from nearest to farthest.
 *
 * @param desiredInterface: A pointer to the EmberInterfaceId structure 
 *  containing the found interface.
 *
 * @param node: The 64-bit ID of the node on which the interface was found.
 *
 * @param distance: An 8-bit value representing the distance of \c node from 
 * the searching node.
 */
EmberStatus emberRadioReBoot( void );

//@} // END MESSAGE FUNCTIONS

/**
 * @name Backchannel Debug Functions
 */
//@{
#if DEBUG
  #define DEBUG_ON 1 
#endif
#if defined(DEBUG_ON) || defined(DOXYGEN_SHOULD_SKIP_THIS)
/** @description Initializes the debug system. If you want to use this 
 * function, it must be called before emberInit() is called. Also, 
 * you must configure the debug system to use one of the two serial ports 
 * on the RF communication module by setting either 
 * EMBER_SERIAL0_DEBUG or EMBER_SERIAL1_DEBUG. For more information
 * on configuring the RF communication module serial ports, see the 
 * <i>Ember Developer Kit User's Guide</i>.
 *
 * @param port: The RF communication module serial port to use to output data
 *  to the debug backchannel.
 */
  void emberDebugInit(int8u port);
 
 /** @description Prints the filename and line number to the
  * debug serial port.
  *
  * @param filename: The name of the file where the assert occurred.
  * 
  * @param linenumber: The line number in the file where the assert occurred.
  */
  void emberDebugAssert(PGM_P filename, int linenumber);
 

 /** @description Prints the contents of RAM to the debug serial port.
  *
  * @param start: The start address of the block of RAM to dump.
  * 
  * @param end: The end address of the block of RAM to dump (address of the 
  *  last byte).
  */
 void emberDebugMemoryDump(int8u *start, int8u *end);

 /** @description Prints an EmberStatus return code to the serial port.
  *
  * @param code: The EmberStatus code to print.
  */
 void emberDebugError(EmberStatus code);

 /** @description Turns off all debug output.  
  * 
  * @return The current state (TRUE for on, FALSE for off).
  */
 boolean emberDebugReportOff();

 /** @description Restores the state of the debug output.
  *
  * @param state: The state returned from emberDebugReportOff(). This is
  * done so that debug output is not blindly turned on.
  */
 void emberDebugReportRestore(boolean state);
#else  // DEBUG OFF
  // Calls are compiled out
  #define emberDebugInit(port) do { ; } while(FALSE)
  
  #define emberDebugAssert(filename, linenumber) do { ; } while(FALSE)

  #define emberDebugMemoryDump(start, end) do { ; } while(FALSE)
 
  #define emberDebugError(code) do {} while(FALSE)

  #define emberDebugReportOff() (FALSE)

  #define emberDebugReportRestore(state) do {} while(FALSE)
#endif // DEBUG_ON

// Format: Same as emberSerialPrintf
// emberDebugPrintf("format string"[, parameters ...])
/** @description Prints text debug messages.
 *
 * @param formatString: Takes the following:
 * <table border=0>
 * <tr><td align="right">%%</td><td>Percent sign</td></tr>
 * <tr><td align="right">%%c</td><td>Single-byte char</td></tr>
 * <tr><td align="right">%%s</td><td>RAM string</td></tr>
 * <tr><td align="right">%%p</td><td>Flash string (does not follow the printf standard)</td></tr>
 * <tr><td align="right">%%u</td><td>Two-byte unsigned decimal</td></tr>
 * <tr><td align="right">%%d</td><td>Two-byte signed decimal</td></tr>
 * <tr><td align="right">%%x, %%2x, %%4x </td><td>1-, 2-, 4-byte hex value (always 0 padded;
 *         does not follow the printf standard)</td></tr>
 * </table>
 */
void emberDebugPrintf(PGM_P formatString, ...);
//@}


#define EMBER_STOP_BEACON_TRANSMIT 15
/* @description Tells the stack to begin transmitting beacons if it is not already
 * doing so. If the stack is already transmitting beacons, updates the beacon configuration.
 * Beacons require a single message buffer is available. When beacons begin
 * transmitting, a message buffer is used and held until beacons stop transmitting.
 * If a message buffer is not available, beacons will not begin transmitting
 * and the function will return EMBER_NO_BUFFERS.
 *
 * @param payload: The payload. Payload size is capped at 100 bytes (this is the
 * actual data bytes limit and does not include a size byte). If a payload size
 * greater than 100 bytes is attempted, this function will return the
 * error EMBER_MAC_BEACON_BAD_PARAM.
 *
 * @param beaconInterval: A value from 0 through 14 specifying the transmission
 * interval of the beacons. A value of EMBER_STOP_BEACON_TRANSMIT (defined as 
 * decimal 15) disables beacon transmission. The formula for the beacon interval
 * and symbol time follows the 802.15.4 specification. This formula is:
 * ( (960)*(2^beaconInterval) )*(16 microseconds).  \b Note: The shorter the
 * beacon interval, the greater the processor resources beaconing will consume.
 * When using very short beacons, please be careful to limit processing in
 * emberIncomingBeaconHandler() since very little processing time may be
 * available between beacons.
 *
 * @param activeTime: A value from 0 through 14 and less than or equal to
 * \c beaconInterval that specifies the length of time during which the superframe
 * is active.
 *
 * @return An EmberStatus value indicating the success or
 * failure of the command.
*/ 
EmberStatus emberTransmitBeacons(EmberMessageBuffer payload,
                                 int8u beaconInterval,
                                 int8u activeTime);

/* @description A callback invoked by the stack when a beacon is received.
 * \b Note: The shorter the beacon interval, the greater the processor resources
 * beaconing will consume.  When using very short beacons, please be careful to
 * limit processing in emberIncomingBeaconHandler() since very little
 * processing time may be available between beacons.
 *
 * @param sender: The EUI64 of the beacon sender.
 *
 * @param payload: The beacon's payload. If there is no payload,
 * payload will equal EMBER_NULL_MESSAGE_BUFFER.
*/ 
void emberIncomingBeaconHandler(EmberEUI64 sender, EmberMessageBuffer payload);

/* @description Gets the time in symbols of the last beacon transmission.
 *
 * @return The time in symbols of the last beacon transmission.
*/
EmberBeaconTime emberGetLastBeaconTransmitTime(void);

/* @description Used to set the counters that trigger emberMessageSent()
 * with an EmberStatus of EMBER_MAC_BEACON_SENT after the specified
 * number of beacons have been sent. By default, the notification is disabled.
 *
 * @param resetBeaconCount: Passing TRUE resets the current beacon count.
 * Passing FALSE leaves the current beacon count untouched.
 *
 * @param beaconCount: Specifies the number of beacons to transmit before
 * triggering emberMessageSent(). A value of 0 is the default and disables
 * the notification. emberMessageSent() is triggered when the count
 * is equal to or greater than the specified number of beacons, in case the
 * user sets the specified number to less than the previous number without
 * reseting the counter.
*/ 
void emberSetOutgoingBeaconNotifyCounter(boolean resetBeaconCount,
                                         int8u beaconCount);

/** @description Reports the peak energy level detected on the current 
  * channel over the scan time. Energy is detected over sample periods 8
  * symbols in length.  
  *
  * @param duration: Sets the exponent of the number of scan periods,
  * where a scan period is 960 symbols per superframe, at a symbol rate
  * (in microseconds) that is defined by the stack constant PHY_SYMBOL_RATE.
  * The scan will occur for ((2^duration) + 1) scan periods.
  *
  * @param returnValue: A pointer to the variable where the peak energy scan
  * value will be written. Please see your radio's documentation for an
  * explanation and the meaning of the RSSI value.
  *
  * @return Returns EMBER_SUCCESS if successful, EMBER_ERR_FATAL otherwise.
  */
EmberStatus emberEnergyScan(int8u duration, int8u *returnValue);

#endif // __EMBER_H__
