/**
 * @file error-def.h
 * @brief Return code definitions for EmberNet stack API functions.
 *
 * See error.h for information on how the values for the return codes are
 * built up from these definitions.
 * This file is separated from error.h because other utilities (such as Ember
 * Studio) can use this file to parse the return codes.
 *
 * \b Note: Do not include this file directly.  This file is included by error.h
 * inside an enum typedef. Include error.h for return code definitions.
 * 
 * <!--TECH WRITER: PASTE LIST OF OUTPUT VALUES INSIDE <PRE> TAGS IN THE OUTPUT FILE-->
 * <!--FOR THIS PAGE.-->
 *
 * <!--Update Note: For the sake of error reporting to the UI users, these--> 
 *           <!-- macros have been changed to allow for a third argument--> 
 *           <!--which indicates how the specific error should be reported --> 
 *           <!--on the front end.-->
 * <!--Author(s): Richard Kelsey (richard@ember.com), Cliff --> 
 * <!--Bowman (cliff@ember.com), Lee Taylor (lee@ember.com) -->
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.                *80*-->
 */

/*
 * @description The generic "no error" message.
 */
DEFINE_ERROR(SUCCESS, SYSTEM, 0)

/**
 * @description The generic fatal error message.
 */
DEFINE_ERROR(ERR_FATAL, SYSTEM, 1)

/**
 * @descripion This is an uninitialized node.  The
 * stack operates in configuration mode only.
 */
DEFINE_ERROR(UNINITIALIZED_EUI64, SYSTEM, 2)

/**
 * @description The token format in non-volatile memory is different 
 * from what the stack expects (returned at system initialization).
 */
DEFINE_ERROR(EEPROM_DATA_VERSION_MISMATCH, SYSTEM, 3)

/**
 * @description The static memory definitions in ember-static-memory.h
 * are incompatible with this stack version.
 */
DEFINE_ERROR(INCOMPATIBLE_STATIC_MEMORY_DEFINITIONS, SYSTEM, 4)

/**
 * @description There is an illegal value in the GRAD segment subtype field.
 */
DEFINE_ERROR(GRAD_UNKNOWN_SUBTYPE, GRAD, 0)
/**
 * @description A GRAD cost value overflowed while being incremented.
 */
DEFINE_ERROR(GRAD_COST_OVERFLOW, GRAD, 1)
/**
 * @description GRAD doesn't know how to reach a destination node.
 */
DEFINE_ERROR(GRAD_UNKNOWN_ROUTE, GRAD, 2)

/**
 * @description GRAD found no more room in the retry queue.
 */
DEFINE_ERROR(GRAD_RETRY_QUEUE_FULL, GRAD, 3)
  
/**
 * @description A neighbor node has the same node ID.
 */
DEFINE_ERROR(GRAD_NODE_ID_CONFLICT, GRAD, 4)

/**
 * @description Returned by GRAD functions on receiving a packet when
 * the GRAD module is asleep.
 */
DEFINE_ERROR(GRAD_ASLEEP, GRAD, 5)

/**
 * @description There are no more buffers.
 */
DEFINE_ERROR(NO_BUFFERS, PACKET, 0)

/**
 * @description Specified an invalid baud rate.
 */
DEFINE_ERROR(SERIAL_INVALID_BAUD_RATE, SERIAL, 0)
  
/**
 * @description Specified an invalid serial port.
 */
DEFINE_ERROR(SERIAL_INVALID_PORT, SERIAL, 1)

/**
 * @description Tried to send too much data.
 */
DEFINE_ERROR(SERIAL_TX_OVERFLOW, SERIAL, 2)

/**
 * @description The receive interrupt was not handled in time, and a 
 * character was dropped.
 */
DEFINE_ERROR(SERIAL_RX_OVERFLOW, SERIAL, 3)

/**
 * @description Detected a UART framing error.
 */
DEFINE_ERROR(SERIAL_RX_FRAME_ERROR, SERIAL, 4)

/**
 * @description Detected a UART parity error.
 */
DEFINE_ERROR(SERIAL_RX_PARITY_ERROR, SERIAL, 5)
 
/**
 * @description There is no received data to process.
 */
DEFINE_ERROR(SERIAL_RX_EMPTY, SERIAL, 6)

/**
 * @description The input queue is locked until a reset is ACKed.
 */ 
DEFINE_ERROR(HSS_INPUT_QUEUE_LOCKED_AFTER_RESET, HSS, 0)
/**
 * @description Received an unrecognizable command.
 */
DEFINE_ERROR(HSS_INVALID_COMMAND, HSS, 1)

/**
 * @description A buffer for MAC retry was erroneously freed.
 */

DEFINE_ERROR(MAC_BUFFER_LOST, MAC1, 0)
/**
 * @description The MAC transmit queue is full.
 */
DEFINE_ERROR(MAC_TRANSMIT_QUEUE_FULL, MAC1, 1)
             
/**
 * @description MAC header FCR error on receive.
 */
DEFINE_ERROR(MAC_UNKNOWN_HEADER_TYPE, MAC1, 2)
/**
 * @description The MAC is already busy tracking beacons.
 */
DEFINE_ERROR(MAC_TRACKING_BEACON, MAC1, 3)
/**
 * @description The MAC is already busy sending beacons.
 */
DEFINE_ERROR(MAC_SENDING_BEACON, MAC1, 4)
/**
 * @description The MAC is already busy scanning.
 */
DEFINE_ERROR(MAC_ALREADY_SCANNING, MAC1, 5)
/**
 * @description A passed parameter violates acceptable beacon parameters
 */
DEFINE_ERROR(MAC_BEACON_BAD_PARAM, MAC1, 6)
/**
 * @description Beacon tracking has lost the beacons and tracking has stopped.
 */
DEFINE_ERROR(MAC_BEACON_LOSS, MAC1, 7)
/**
 * @description A number of beacons the application has specified have been sent.
 */
DEFINE_ERROR(MAC_BEACON_SENT, MAC2, 8)

/**
 * @description The decryption request failed, and Message Integrity 
 * Code failed.
 */
DEFINE_ERROR(SECURITY_MIC_TEST_FAILED, SECURITY, 0)
/**
 * @description The system failed in backing up the Nonce Frame Counter.
 */
DEFINE_ERROR(SECURITY_FRAME_COUNTER_BACKUP_FAILURE, SECURITY, 1)
/** 
 * @description Received an unsecured message on a secured network.
 */
DEFINE_ERROR(SECURITY_UNSECURED_MESSAGE_RECEIVED, SECURITY, 2)

/** 
 * @description An unknown flash token was specified.
 */
DEFINE_ERROR(ERR_TOKEN_UNKNOWN, TOKEN, 0)

/** 
 * @description Could not create new flash token because it already exists.
 */
DEFINE_ERROR(ERR_TOKEN_EXISTS, TOKEN, 1)

/** 
 * @description An incorrect size was specified when retrieving token data.
 */
DEFINE_ERROR(ERR_TOKEN_INVALID_SIZE, TOKEN, 2)

/** 
 * @description The EEPROM was reformatted, and all tokens were reset.
 */
DEFINE_ERROR(ERR_TOKEN_INVALID_TABLE_SIGNATURE, TOKEN, 3)

/** 
 * @description Couldnt write token because it is marked read-only.
 */
DEFINE_ERROR(ERR_TOKEN_READ_ONLY, TOKEN, 4)

/** 
 * @description Token data verification failed on token write.
 */
DEFINE_ERROR(ERR_TOKEN_WRITE_VERIFY, TOKEN, 5)

/** 
 * @description Flash memory is full.
 */
DEFINE_ERROR(ERR_TOKEN_FLASH_FULL, TOKEN, 6)

/** 
 * @description EEPROM memory is full.
 */
DEFINE_ERROR(ERR_TOKEN_EEPROM_FULL, TOKEN, 7)

 /**
  * @description The Postmaster message queue is full on message receive.
  */
DEFINE_ERROR(ERR_MESSAGE_QUEUE_FULL, POSTMASTER, 0)
 /**
  * @description The Postmaster message queue was empty on trying to get 
  * next message.
  */
DEFINE_ERROR(ERR_MESSAGE_QUEUE_EMPTY, POSTMASTER, 1)

/**
 * @description The bootloader received an invalid message (failed attempt 
 * to go into bootloader).
 */
DEFINE_ERROR(ERR_BOOTLOADER_TRAP_TABLE_BAD, BOOTLOADER, 0)
             
/**
 * @description Bootloader received an invalid message (failed attempt to go 
 * into bootloader).
 */
DEFINE_ERROR(ERR_BOOTLOADER_TRAP_UNKNOWN, BOOTLOADER, 1)

/**
 * @description The transport layer connection table is full.
 */
DEFINE_ERROR(TOO_MANY_CONNECTIONS, TRANSPORT1, 0)

/**
 * @description The application is trying to open a connection to a node to 
 * which it is already connected.
 */
DEFINE_ERROR(CONNECTION_OPEN, TRANSPORT1, 1)

/**
 * @description The transport layer could not open a connection.
 */
DEFINE_ERROR(CONNECTION_OPEN_FAILED, TRANSPORT1, 2)

/**
 * @description The transport layer connection was open but now has failed.
 */
DEFINE_ERROR(CONNECTION_FAILED, TRANSPORT1, 3)

/**
 * @description The transport layer successfully closed a connection.
 */
DEFINE_ERROR(CONNECTION_CLOSED, TRANSPORT1, 4)

/**
 * @description The transport layer is in process of closing a connection
 * (waiting for a response from the remote device).
 */
DEFINE_ERROR(CONNECTION_CLOSING, TRANSPORT1, 5)

/**
 * @description The transport layer attempted to send or deliver a message, but
 * it failed.
 */
DEFINE_ERROR(DELIVERY_FAILED, TRANSPORT1, 6)

/**
 * @description Cant start an endpoint search because one is already in
 * progress.
 */
DEFINE_ERROR(ENDPOINT_SEARCH_IN_PROGRESS, TRANSPORT1, 7)

/**
 * @description Cant stop an endpoint search because one has not 
 * been started.
 */
DEFINE_ERROR(NO_ENDPOINT_SEARCH_IN_PROGRESS, TRANSPORT2, 0)

/**
 * @description This binding index is out of range of the current binding table.
 */
DEFINE_ERROR(BINDING_INDEX_OUT_OF_RANGE, TRANSPORT2, 1)

/**
 * @description Could not set or delete a binding table entry.
 */
DEFINE_ERROR(BINDING_TABLE_WRITE_FAILED, TRANSPORT2, 2)

/**
 * @description Could not set or find a binding index given the specified
 * terminal.
 */
DEFINE_ERROR(INVALID_BINDING_TERMINAL, TRANSPORT2, 3)

/**
 * @description An invalid binding table index was given to a transmit function.
 */
DEFINE_ERROR(INVALID_BINDING_INDEX, TRANSPORT2, 4)

/**
 * @description An invalid binding table index given to an open or close function.
 */
DEFINE_ERROR(INVALID_BINDING_INDEX_ERROR, TRANSPORT2, 5)

/**
 * @description Tried to open a connection using a binding to a local
 * terminal, but the endpoint doesnt match.
 */
DEFINE_ERROR(BINDING_INTERFACE_NOT_FOUND, TRANSPORT2, 6)

/**
 * @description Multiple binding table entries were found for the specified terminal.
 */
DEFINE_ERROR(TERMINAL_HAS_MULTIPLE_BINDINGS, TRANSPORT2, 7)

/**
 * @description An incoming message was received but is invalid owing to the
 * condition of the system (for example, trying to close a connection before 
 * it is open).
 */
DEFINE_ERROR(INVALID_CALL, TRANSPORT3, 0)

/**
 * @description The link cost to a node is not known.
 */
DEFINE_ERROR(COST_NOT_KNOWN, TRANSPORT3, 1)

/**
 * @description The maximum number of messages has been reached.
 */
DEFINE_ERROR(MAX_MESSAGE_LIMIT_REACHED, TRANSPORT3, 2)

/**
 * @description A connection is not open yet.
 */
DEFINE_ERROR(CONNECTION_NOT_YET_OPEN, TRANSPORT3, 3)

/**
 * @description The message to be transmitted is too big to fit into a
 * single over-the-air packet.
 */
DEFINE_ERROR(MESSAGE_TOO_LONG, TRANSPORT3, 4)

/**
 * @description The application is trying to delete or overwrite a binding
 * that is in use.
 */
DEFINE_ERROR(BINDING_IS_ACTIVE, TRANSPORT3, 5)

/**
 * @description The EUI64 is not available in the current packet.
 */
DEFINE_ERROR(EUI64_NOT_AVAILABLE, TRANSPORT3, 6)

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// Deprecated
DEFINE_ERROR(NOT_SYNCHRONIZED, SYNC, 0)
#endif //DOXYGEN_SHOULD_SKIP_THIS

/**
 * @description Conversion is complete.
 */
DEFINE_ERROR(ADC_CONVERSION_DONE, HAL, 0)

/**
 * @description Conversion cannot be done because a request is being 
 * processed.
 */
DEFINE_ERROR(ADC_CONVERSION_BUSY, HAL, 1)

/**
 * @description Conversion is deferred until the current request has been 
 * processed.
 */
DEFINE_ERROR(ADC_CONVERSION_DEFERRED, HAL, 2)

/**
 * @description Invalid channel was supplied.
 */
  // the channel requested does not exist
DEFINE_ERROR(ADC_INVALID_CHANNEL, HAL, 3)

/**
 * @description No results are pending.
 */
DEFINE_ERROR(ADC_NO_CONVERSION_PENDING, HAL, 4)

/**
 * @description The transmit hardware buffer underflowed.
 */
DEFINE_ERROR(PHY_TX_UNDERFLOW, PHY, 0)

/**
 * @description The transmit hardware did not finish transmitting a packet.
 */
DEFINE_ERROR(PHY_TX_INCOMPLETE, PHY, 1)

/**
 * @description An unsupported channel setting was specified.
 */
DEFINE_ERROR(PHY_INVALID_CHANNEL, PHY, 2)

/**
 * @description An unsupported power setting was specified.
 */
DEFINE_ERROR(PHY_INVALID_POWER, PHY, 3)

/**
 * @description The packet cannot be transmitted because the physical MAC 
 * layer is currently transmitting a packet. (This is used for the MAC 
 * backoff algorithm.)
 */
DEFINE_ERROR(PHY_TX_BUSY, PHY, 4)

/**
 * @description The software installed on the hardware doesn't recognize the
 * hardware radio type.
 */
DEFINE_ERROR(PHY_UNKNOWN_RADIO_TYPE, PHY, 5)

/**
 * @description The software installed on the hardware doesn't recognize the
 * hardware radio type.
 */
DEFINE_ERROR(PHY_OSCILLATOR_CHECK_FAILED, PHY, 6)

/**
 * @description The stack software has completed initialization and is ready
 * to send and receive packets over the air.
 */
DEFINE_ERROR(NETWORK_UP, STACK, 0)

/**
 * @description The network is not operating.
 */
DEFINE_ERROR(NETWORK_DOWN, STACK, 1)

/**
 * @description The network has activity pending and shouln't be shut down.
 */
DEFINE_ERROR(NETWORK_PENDING_ACTIVITY, STACK, 2)

/**
 * @description The node is not associated with a network.
 */
DEFINE_ERROR(NOT_ASSOCIATED, STACK, 3)

/**
 * @description Binding information for the node and host is mismatched.
 */
DEFINE_ERROR(RNAP_MISMATCH, RNAP, 0)

/**
 * @description The host received an RNAP message ACK from a node, but
 * the status returned is not \c EMBER_SUCCESS.
 */
DEFINE_ERROR(RNAP_ACK_WITH_ERR, RNAP, 1)

/**
 * @description The application running on the host
 * has lost communication with the node RNAP application.
 */
DEFINE_ERROR(RNAP_COMMUNICATION_FAILED, RNAP, 2)

/**
 * @description The application failed to deliver an RNAP message to a node.
 */
DEFINE_ERROR(RNAP_DELIVERY_FAILED, RNAP, 3)

/**
 * @description The serial problem related to RNAP message process from
 * from node.
 */
DEFINE_ERROR(RNAP_SERIAL_ERROR, RNAP, 4)
