/**
 * @file error.h
 * @brief Return codes for Ember API functions, and module definitions.
 *
 * Return codes consist of a single byte, which is divided into two parts.
 * The most significant 5 bits are used to designate a "module" for the
 * return code. The least significant 3 bits are used for a specific return 
 * code within a specified module.
 *
 * The separation of return codes into modules allows for easier decoding of
 * the source of the return code. 
 * 
 * <!--Author(s): Lee Taylor, lee@ember.com -->
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.-->
 */

#ifndef __ERRORS_H__
#define __ERRORS_H__

/** 
 * @description Module definitions used for return codes. A module is a 
 * logical subset of API functions. Comments are provided 
 * only for modules whose names are not self-explanatory. Note that multiple 
 * modules are required for transport function return values because there are 
 * so many of them.
 */
//@{
typedef enum {
  /** Basic system-wide return codes. */
  EMBER_MODULE_SYSTEM     =  0,
  /** Basic system-wide return codes. */
  EMBER_MODULE_GRAD       =  1,
  /** Packet handling functions. */
  EMBER_MODULE_PACKET     =  3,
  /** Serial port handling functions. */
  EMBER_MODULE_SERIAL     =  4,
  /** High-speed serial port handling functions. */
  EMBER_MODULE_HSS        =  5,
  /** MAC functions (either Ember or 802.15.4) */
  EMBER_MODULE_MAC2       =  6,
  EMBER_MODULE_MAC1       =  7,
  EMBER_MODULE_SECURITY   =  8,
  /** Deprecated. */
  EMBER_MODULE_TOKEN      =  9,
  /** Postmaster functions. */
  EMBER_MODULE_POSTMASTER = 10,
  /** Bootloader functions. */
  EMBER_MODULE_BOOTLOADER = 11,
  /** Transport functions. */
  EMBER_MODULE_TRANSPORT1 = 12,
  /** Transport functions. */
  EMBER_MODULE_TRANSPORT2 = 13,
  /** Transport functions. */
  EMBER_MODULE_TRANSPORT3 = 14,
  #ifndef DOXYGEN_SHOULD_SKIP_THIS
  EMBER_MODULE_SYNC       = 15,  // Deprecated
  #endif //DOXYGEN_SHOULD_SKIP_THIS
  EMBER_MODULE_HAL        = 16,
  EMBER_MODULE_PHY        = 17,
  EMBER_MODULE_STACK      = 18,
  EMBER_MODULE_RNAP       = 19
} EmberModuleType;
//@} 

/**
 * @description Macro used by error-def.h to define all the return codes.
 * As described above, the actual value of the return code is a combination of
 * the module number and the value given as the third parameter.
 *
 * @param symbol: The name of the constant being defined, which will be 
 *  prepended with EMBER_
 * 
 * @param module: The name of the module that the return code 
 *  belongs to.
 *
 * @param value: The value of the return code within the module.
 */
#define DEFINE_ERROR(symbol, module, value) \
 EMBER_ ## symbol = DEFINE_ERROR_CODE(EMBER_MODULE_ ## module, value),

/**
 * @description The number of bits taken by the per-module errors.
 * 
 * \b Note: You \b cannot number return codes greater than MODULE_ERROR_MASK.
 * If you do, you will wrap error codes within the module! For example, 
 * if MODULE_ERROR_BITS = 3, each module can handle up to 8 values,
 * numbered from 0 to 7.
 */
//@{
#define MODULE_ERROR_BITS 3  // Allows 32 modules, 8 errors each.
#define MODULE_ERROR_MASK ((1 << MODULE_ERROR_BITS) - 1) // = 7
//@}

/**
 * @description Combines the module number with the value given
 * as the second argument (\c moduleError) to form the complete return code.
 * This macro is only used by DEFINE_ERROR.
 *
 * @param module: The module number passed in by DEFINE_ERROR.
 *
 * @param moduleError: The value passed in by DEFINE_ERROR.
 */
#define DEFINE_ERROR_CODE(module, moduleError) \
  ((module << MODULE_ERROR_BITS) + (moduleError & MODULE_ERROR_MASK))

typedef enum {
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#include "include/error-def.h"
#endif //DOXYGEN_SHOULD_SKIP_THIS
  /** Gets defines as a count of all the possible return codes in the 
  * EmberNet stack API. 
  */
  EMBER_ERROR_CODE_COUNT

} EmberErrorCode;

#undef DEFINE_ERROR

#endif // __ERRORS_H__
