/** @file hal-em2420.h
 * @brief HAL interfaces for the EM2420 radio
 *
 * The interfaces described in this file should not be called directly
 * by any applications. They are strictly for internal stack use but
 * must be implemented by the HAL to provide an abstraction across
 * differing hardware implementations.
 *  
 * <!--Author(s): Lee Taylor, lee@ember.com-->
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */
#ifndef __HAL_EM2420_H__
#define __HAL_EM2420_H__



/** @name Radio Power Functions */
//@{

/** @description Initializes the SPI and and ports required for 
 * the EM2420.  
 */
void emberHalRadioInit(void);

/** @description Powers down the radio by disabling the voltage
 * regulator and setting ports so that they don't leak any current. \b 
 * Note: Sleeping the radio via radioSleep() is usually preferable 
 * to powering it down because powering the radio back up requires much 
 * more time than waking it from sleep. However, powering down is the 
 * best choice when radio power consumption must be minimized. 
 */ 
void emberHalRadioShutdown(void);

//@}  // end of Radio Power Functions


/** @name Pin Functions */
//@{

/** @description Returns the state of the FIFO pin.
 *   
 * @return TRUE if asserted.
 */
boolean emberHalGetRadioFifoPin(void);

/** @description Returns the state of the FIFOP pin.
 *   
 * @return TRUE if asserted.
 */
boolean emberHalGetRadioFifopPin(void);

/** @description Returns the state of the CCA pin.
 *   
 * @return TRUE if asserted.
 */
boolean emberHalGetRadioCcaPin(void);

/** @description Returns the state of the SPD pin.
 *   
 * @return TRUE if asserted.
 */
#ifdef HALTEST 
   boolean emberHalGetRadioSpdPin(void);
#endif
//@}  // end of Pin Functions


/** @name Interrupt Functions */
//@{
/** @description Enables both radio interrupts (FIFOP and SPD). */
void emberHalRadioEnableInts(void);

/** @description Enables the FIFOP radio interrupt. */
void emberHalRadioEnableRxInt(void);

/** @description Disables both radio interrupts (FIFOP and SPD). */
void emberHalRadioDisableInts(void);

/** @description Disables the FIFOP radio interrupt. */
void emberHalRadioDisableRxInt(void);

/** @description The HAL should call this whenever the FIFOP interrupt
 * is triggered.
 */
void emberRadioReceiveIsr(void);
//@}  // end of Interrupt Functions


/** @name SPI Interface Functions */
//@{

/** @description Writes a byte over the SPI without changing the mode. 
 * 
 * @param data: The byte to be sent out over the SPI.
*/
void radSpiWrite(int8u data);

/** @description Reads a byte over the SPI without changing the mode.
 *   
 * @return The byte read.
 */
int8u radSpiRead(void);

/** @description Reads and writes a byte over the SPI without changing 
 * the mode.
 * 
 * @param data: The byte to be sent out over the SPI.
 * 
 * @return The byte read.
 */
int8u radSpiReadWrite(int8u data);

/** @description Performs all SPI operations required to set a register
 * with a value.
 * 
 * @param addr: The register to set.
 *   
 * @param data: The value to write.
 * 
 * @return The status from radSpiReadWrite() operation.
 */
int8u radSetReg(int16u addr, int16u data);

/** @description Performs all SPI operations required to get the value stored 
 * in a register.
 *   
 * @param addr: The register to read.
 *   
 * @return The register's value.
 */
int16u radGetReg(int8u addr);

/** @description Performs all SPI operations required to send a strobe command.
 * 
 * @param addr: The strobe to send.
 * 
 * @return The status from radSpiReadWrite() operation.
 */
int8u radStrobe(int8u addr);

/** @description Performs all SPI operations required to write data to the 
 * TX FIFO.
 * 
 * @param buff: A pointer to the data to write.
 *   
 * @param addr: The address to write.
 *   
 * @param numbytes: The number of bytes to write.
 *   
 * @return The status from radSpiReadWrite() operation.
 */
int8u radWriteFifo(int8u* buff, int8u addr, int8u numbytes);

/** @description Performs all SPI operations required to read a specified 
 * amount of data from the RX FIFO into a buffer.
 * 
 * @param buff: A pointer to the data to read.
 *  
 * @param addr: The address to read.
 *  
 * @param numbytes: The number of bytes to read.
 *  
 * @return The status from radSpiReadWrite() operation.
 */
int8u radReadFifo(int8u *buff, int8u addr, int8u numbytes);

/** @description Performs all the initial operations required to put the EM2420 
 * into a mode to read data from the RX FIFO, and leaves it in that mode
 * so that future calls to radSpiRead() will return data from the FIFO.
 *  
 * @param addr: The address to read.
 *  
 * @return The status from radSpiReadWrite() operation.
 */ 
int8u radStartReadFifo(int8u addr);

/** @description Returns both the SPI interface and the EM2420 to a normal 
 * mode after being in the mode set by radStartReadFifo().   
*/
void radEndReadFIFO(void);

/** @description Performs all initial operations to put the EM2420 into a mode
 * to write data to an arbitrary EM2420 RAM address, and leaves it in that
 * mode so that future calls to radSpiWrite() will write data to the
 * next RAM address.
 * 
 * @param addr: The address to read.
 *  
 * @return The status from radSpiReadWrite() operation.
 */ 
int8u radStartWriteRAM(int16u addr); 

/** @description Performs all initial operations to put the EM2420 into a mode
 * to read data from an arbitrary EM2420 RAM address, and leaves it in 
 * that mode so that future calls to radSpiRead() will read data from 
 * the next RAM address.
 * 
 * @param addr: The address to read.
 *    
 * @return The status from radSpiReadWrite() operation.
 */
int8u radStartReadRAM(int16u addr);

/** @description Returns both the SPI interface and the EM2420 to a normal
 * mode after being in the mode set by radStartReadRAM() or 
 * radStartWriteRAM(). 
 */ 
void radEndRAM(void);

/** @description Performs all SPI operations required to write data to an 
 * EM2420 RAM address.
 * 
 * @param buff: A pointer to the data to read.
 * 
 * @param addr: The address to read.
 *   
 * @param numbytes: The number of bytes to read.
 *   
 * @return The status from radSpiReadWrite() operation.
 */ 
int8u radWriteRAM(int8u* buff, int16u addr, int8u numbytes);

/** @description Performs all SPI operations required to read a specified amount
 * of data from an EM2420 RAM address into a buffer.
 * 
 * @param buff: A pointer to the data to read.
 *    
 * @param addr: The address to read.
 *     
 * @param numbytes: The number of bytes to read.
 *   
 * @return The status from radSpiReadWrite() operation.
 */  
int8u radReadRAM(int8u* buff, int16u addr, int8u numbytes);
//@}  // end of SPI interface functions


/** @description On slower systems the FIFO cannot be stuffed as 
 * fast as the radio transmits. This function tells the PHY when to start 
 * stuffing the FIFO.
 * 
 * @param prestuff: Set TRUE for faster systems, to prestuff the FIFO.
 * TRUE is the default choice since it is the most common.
 * This really only needs to be set FALSE for slow systems like 1MHz.
 */  
void emberPhyPrestuffFIFO(boolean prestuff);

#endif // __HAL_EM2420_H__
