/** @file hal.h
 * @brief HAL interfaces required by the EmberNet stack
 * 
 * Unless otherwise noted, the EmberNet stack uses these functions, and 
 * therefore the HAL must implement them. Additionally, unless otherwise
 * noted, applications that find these functions useful may use them.
 *
 * <b>About the Time Support Functions</b>
 *
 * The time support functions include:
 * - microGetSystemTick()
 * - microGetSystemTime()
 * - microGetQuarterSeconds()
 * - microGetSyncTime()
 * - microIncreaseSyncOffset()
 * - microGetSyncCapture()
 * - microClearSyncCapture()
 * - microSyncCaptured()
 * - microDelayMicroSeconds()
 *
 * The time support functions use units that are based on the low-frequency
 * crystal, which may operate at either 32.768kHz or 31.25kHz.
 * 
 * A single system tick (as returned by microGetSystemTick() and
 * microGetSystemTime() ) is roughly approximated as 1 millisecond.
 * When used with a 32.768kHz crystal, a system tick is actually 0.976 
 * milliseconds. When used with a 31.25kHz crystal, a system tick is actually 
 * 1.024 milliseconds.
 * 
 * A single quarter-second tick (as returned by microGetQuarterSeconds() )
 * is roughly approximated as 0.25 seconds. When used with a 32.768kHz crystal, 
 * a system tick is actually 0.25 seconds. When used with a 31.25kHz crystal, 
 * a system tick is actually 0.262 seconds.
 * 
 * A single synchronization tick (as returned by microGetSyncTime() ) is
 * roughly approximated as 32 microseconds. When used with a 32.768kHz crystal, 
 * a system tick is actually 30.5 microseconds. When used with a 31.25kHz 
 * crystal, a system tick is actually 32 microseconds.
 *  
 * The values used by the time support functions will wrap after an interval. 
 * The length of the interval depends on the length of the tick and the number 
 * of bits in the value. However, there is no issue when comparing time deltas 
 * of less than half this interval with a subtraction, if all data types are the 
 * same.
 *
 * <b>About the MAC Timer Support Functions</b>
 *
 * \b Note: Applications should not directly call the following MAC timer support functions. 
 * are used internally by the operation of the stack.
 * - microOrderSymbolDelayA()
 * - microCancelSymbolDelayA()
 * - emMicroSymbolDelayAIsr()
 * - microOrderSymbolDelayB()
 * - emMicroSymbolDelayBIsr()
 * <!-- Copyright 2004 by Ember Corporation. All rights reserved.-->   
 */
#ifndef __HAL_H__
#define __HAL_H__


/** @mainpage Hardware Abstraction Layer Interface Guide
 *  
 * This documentation set provides a specification for the interfaces that 
 * applications must supply to the EmberNet stack, plus some additional
 * application convenience functions.
 *   
 * The interfaces consist of the following:
 *   - <b>EmberNet stack-required interfaces:</b> Includes functions that
 *      must be implemented by the hardware abstraction layer (HAL) for proper 
 *      operation of the EmberNet stack. There are two classes of these functions:
 *        - <b>EmberNet stack-only interfaces:</b> Functions that are not
 *          useful to applications or that should not be used by the 
 *          application programmer.
 *        - <b>EmberNet stack and application interfaces:</b> Functions used 
 *          by the EmberNet stack and the application.
 *      .   
 *      Developers should review these functions to learn what they must
 *      support. The specific implementation of the functions has been supplied
 *      to provide only examples, and not to indicate requirements.
 *    
 *  - <b>Application convenience interfaces:</b> Includes functions that the 
 *     EmberNet stack does not use. Application developers can freely revise 
 *     these functions or delete them. Only the supplied sample applications 
 *     use them.
 */


/** @appusage Call if a clean restart is desired. For example, the Ember 
 * Evaluation Kit application calls this function when a user presses the 
 * "wake" button.
 *     
 * @stackusage Called when the stack's configuration services receive a request 
 * to reboot (such as from Ember Studio).
 *     
 * @implementation Shut down any hardware necessary, then call microReboot().
 */
void boardReboot(void);

/** @stackusage Called to indicate activity over the radio (for both transmission
 * and reception). The stack calls this once with \c turnOn TRUE and shortly
 * thereafter with \c turnOn FALSE.
 *   
 * @implementation Should do something interesting, such as change the state of 
 * an LED.
 *   
 * @param turnOn: See Usage.
 */
void halIndicateActivity(boolean turnOn);

/** @stackusage Called to indicate the presence of a node on the network.
 *     
 * @implementation Should do something obvious, such as beep.
 */
void halIndicatePresence(void);



#ifndef DOXYGEN_SHOULD_SKIP_THIS
//A few macros which make the return values slightly more obvious
#define CURRENTTIME_MS_INT16 (microGetSystemTick())
#define CURRENTTIME_QS_INT16 (microGetQuarterSeconds())
#define CURRENTTIME_MS_INT32 (microGetSystemTime())
#endif


/** @description Returns the current system time in system ticks, as a 16-bit
 * value.
 *   
 * @return The least significant 16 bits of the current system time, in system 
 * ticks.
 */
int16u microGetSystemTick(void);  

/** @description Returns the current system time in system ticks, as a 32-bit
 * value.
 *   
 * @nostackusage
 * 
 * @return The least significant 32 bits of current system time, in system ticks.
 */
int32u microGetSystemTime(void);

/** @description Returns the current system time in quarter second ticks, as a
 * 16-bit value.
 *   
 * @nostackusage
 *   
 * @return The least significant 16 bits of the current system time, in system 
 * ticks, multiplied by 256.
 */
int16u microGetQuarterSeconds(void);

/** @description Returns the current synchronization time in synchronization 
 * ticks, as a 32-bit value. This free-running local time is combined with an 
 * offset to produce the network time on synchronized nodes.
 *    
 * @return The least significant 32 bits of current synchronization time, in
 * synchronization ticks. 
 */
EmberBeaconTime microGetSyncTime(void);
void microIncreaseSyncOffset(EmberBeaconTime increase);

/** @description Returns the time (as returned by microGetSyncTime() )
 * of the  most recent SPD signal from the radio.
 *  
 * @return The time of the most recent SPD signal from the radio. \b Note: 
 * Returns 0 if no SPD signal has been received from the radio.
 */
EmberBeaconTime microGetSyncCapture(void);

/** @description Clears the flag that records whether an SPD signal has been
 * received from the radio.
 */
void microClearSyncCapture(void);

/** @description Indicates whether an SPD signal has been received from the 
 * radio since the most recent call to microClearSyncCapture().
 *   
 * @return TRUE if an SPD signal has been received from the radio since the 
 * most recent call to microClearSyncCapture().
 */
boolean microSyncCaptured(void);

/** @description Blocks the current thread of execution for the specified 
 * amount of time, in microseconds.
 *  
 * @implementation The function is implemented with cycle-counted busy loops 
 * and is intended to create the short delays required when interfacing with 
 * hardware peripherals.
 *   
 * @param us: The specified time, in microseconds. Should be between 1 and 
 * 32767 microseconds.
 */
void microDelayMicroSeconds(int16u us);

/**@description Sets up a timer and calls an interrupt-context callback 
 * when it expires.
 *  
 * @stackusage Used by the MAC to request a callback at a specified amount 
 * of time in the future.
 * 
 * @param symbols: The delay, in units of PHY_SYMBOL_RATE (defined in 
 * config.h).
 */
void microOrderSymbolDelayA(int16u symbols);

/**@description Cancels the timer set up by microOrderSymbolDelayA().
 *  
 * @stackusage Used by the MAC to cancel a scheduled callback.
*/
void microCancelSymbolDelayA(void);

/** @description Private to the EmberNet stack.
 *  
 * @stackusage An EmberNet stack-implemented callback function corresponding to 
 * the event generated by microOrderSymbolDelayA(). 
 */
void emMicroSymbolDelayAIsr(void);

/** @description Sets up a timer and calls an interrupt-context callback 
 * when it expires.
 *  
 * @stackusage Used by the MAC to request a callback a specified amount 
 * of time in the future.
 *   
 * @param symbols: The delay, in units of PHY_SYMBOL_RATE (defined in 
 * config.h).
 */
void microOrderSymbolDelayB(EmberBeaconTime symbols);

/** @description Private to the EmberNet stack.
 * 
 * @stackusage An EmberNet stack-implemented callback function corresponding to the 
 * event generated by microOrderSymbolDelayB(). 
 */
void emMicroSymbolDelayBIsr(void);

/** @description Seeds the firmGetRandom() pseudorandom number 
 * generator.
 * 
 * @stackusage Called during initialization with a seed from the radio. 
 * Therefore, the application should never need to call this.
 *  
 * @param seed: A seed for the pseudorandom number generator.
 */
void firmSeedRandom(int32u seed);

/** @description Runs a standard LFSR to generate pseudorandom numbers.
 *  
 * @stackusage Called by the MAC to choose random backoff slots.
 *  
 * @implementation More complicated implementations may improve the MAC's 
 * ability to avoid collisions in large networks, but it is \b critical to 
 * implement this function such that it can return quickly.
 */ 
#ifdef SIMULATOR
#define firmGetRandom() simulatedFirmGetRandom(__FILE__, __LINE__)
int16u simulatedFirmGetRandom(char *file, int line);
#else
int16u firmGetRandom(void);
#endif

/** @description Calculates cyclic redundancy code (CITT CRC 16).
 * 
 * @stackusage On some radios, the cyclic redundancy code (CRC) 
 * for error detection on packet data is calculated in software.
 *   
 * @implementation Should apply the standard CITT CRC 16 polynomial to a 
 * single byte. It should support being called first with an initial 
 * remainder, then repeatedly until a whole packet is processed.
 *   
 * @param newByte: The new byte to be run through CRC.
 *   
 * @param prevRemainder: The previous CRC remainder.
 *   
 * @return The new CRC remainder.
 */
int16u emHalCrc16(int8u newByte, int16u prevRemainder);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
//token system functions
#include "hal/micro/token.h"
#endif //DOXYGEN_SHOULD_SKIP_THIS


/** @description Transfers control to the bootloader. \b Note: This function 
 * call is not valid for leaf nodes since they do not have a bootloader.
 * 
 * @param eui: A pointer to the EUI64 number.
 */
void runBootload(int8u *eui);

#endif //__HAL_H__
