/** @file mfg-test.h
 * @brief API for manufacturing test and configuration applications
 *
 * The Golden Node mentioned in many of the function comments is the
 * manufacturing test and configuration node.
 *
 * Some functions in this file return an EmberStatus value. See 
 * error-def.h for definitions of all EmberStatus return values.
 *
 * <!-- Copyright 2001-2004 by Ember Corporation. All rights reserved. *80* -->
 */

#ifndef MANUFACTURING_TEST_H
#define MANUFACTURING_TEST_H

#ifndef DOXYGEN_SHOULD_SKIP_THIS
/* Manufacturing Configuration Packets
 *
 *   Basic Packet Structure
 *   
 *   Byte 0   - length (N +3) 
 *   Byte 1   - manufacturing config mode indicator 1 (MFG_TEST_MODE_INDICATOR)
 *   Byte 2   - manufacturing config mode indicator 2 (~MFG_TEST_MODE_INDICATOR)
 *   Byte 3   - message type
 *   Byte 4   - data 0
 *    ...       ...
 *   Byte N+3 - data N-1
 *
 *   ===============================================================
 *
 *   Type        N   Data
 *
 *   ACK         1   [ACKed Type]
 *   Reboot      0
 *   Set EUID    16  [ID0][ID1]...[ID7][Hash0][Hash1]...[Hash7]
 *   Set String  16  [MS0][MS1]...[MS15]
 *   Set Radio   2   [SupportedBands][XTAL Offset]
 */


#define HASH_SIZE                            8
#define MANUFACTURING_STRING_SIZE            16


#define MFG_TEST_MODE_INDICATOR         0xEC

#define INDEX_MFG_TEST_LENGTH           0
#define INDEX_MFG_TEST_MODE_INDICATOR_1 1
#define INDEX_MFG_TEST_MODE_INDICATOR_2 2
#define INDEX_MFG_TEST_TYPE             3
#define INDEX_MFG_TEST_DATA             4

#define MFG_TEST_DATA_EUID_OFFSET       0
#define MFG_TEST_DATA_HASH_OFFSET       EUI64_SIZE
#define MFG_TEST_DATA_ACK_TYPE_OFFSET   0
#define MFG_TEST_DATA_BANDS_OFFSET      0
#define MFG_TEST_DATA_XTAL_OFFSET       1

#define MFG_TEST_ACK_DATA_LENGTH        1
#define MFG_TEST_SET_EUID_LEGACY_LENGTH (EUI64_SIZE + HASH_SIZE)
#define MFG_TEST_SET_EUID_DATA_LENGTH   EUI64_SIZE
#define MFG_TEST_SET_STRING_DATA_LENGTH 16
#define MFG_TEST_SET_RADIO_DATA_LENGTH  2
#define MFG_TEST_REBOOT_COMMAND_LENGTH  0

#define MFG_TEST_PACKET_OVERHEAD        4
#define MAX_MFG_TEST_DATA_LENGTH        MFG_TEST_SET_STRING_DATA_LENGTH
#define MAX_MFG_TEST_PACKET_LENGTH     (MAX_MFG_TEST_DATA_LENGTH + \
                                      MFG_TEST_PACKET_OVERHEAD)

#define INVERT_BITS(x) ((~(x)) & 0xFF)

#define isCorrectLength(l, dl) ((l) == ((dl) + MFG_TEST_PACKET_OVERHEAD))


// the following ends the non-doxygen part of the file.
#endif 

/** @description Message types exchanged during manufacturing configuration. */
enum {
  MFG_TEST_TYPE_ACK                       = 0x00,
  MFG_TEST_TYPE_REBOOT_COMMAND            = 0x01,
  MFG_TEST_TYPE_SET_EUID                  = 0x02,
  MFG_TEST_TYPE_SET_MANUFACTURING_STRING  = 0x03,
  MFG_TEST_TYPE_SET_RADIO_PARAMETERS      = 0x04,
  MFG_TEST_TYPE_ANY                       = 0xFF
};

typedef int8u ConfigMessageType;

/** @description An 8-byte array for the EUID initialization validation
 * code.
 */
typedef int8u EmberHashCode[HASH_SIZE];

/** @description A 16-byte array for the manufacturing string. */
typedef int8u EmberManufacturingString[MANUFACTURING_STRING_SIZE];

/** @description A callback to be implemented on the Golden Node to process
 * acknowledgements. If you supply a custom version of this handler, you
 * must define EMBER_APPLICATION_HAS_INCOMING_MFG_TEST_MESSAGE_HANDLER in 
 * order to override the stub in ember-static-configuration.h.
 *
 * @param messageType: The type of the incoming message. Currently, the
 * only possibility is MFG_TEST_TYPE_ACK.
 *
 * @param dataLength: The length of the incoming message.
 *
 * @param data: A pointer to the data received in the current message.
 */
void emberIncomingMfgTestMessageHandler(int8u messageType, 
                                        int8u dataLength, 
                                        int8u *data);

/** @description A function used on the Golden Node to switch between 
 * normal network operation (for testing) and manufacturing 
 * configuration. Like emberSleep(), it may not be possible to execute this
 * command due to pending network activity. For the transition
 * from normal network operation to manufacturing configuration,
 * it is customary to loop, calling this function alternately
 * with emberTick() until the mode change succeeds.
 *
 * @param beginConfiguration: Determines the new mode of operation.
 * TRUE causes the node to enter manufacturing configuration. FALSE 
 * causes the node to return to normal network operation.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
EmberStatus emberMfgTestSetPacketMode(boolean beginConfiguration);

/** @description A function used during manufacturing configuration
 * on the Golden Node to send the DUT a reboot command. The usual 
 * practice is to execute this command at the end of manufacturing 
 * configuration, to place the DUT into normal network operation for
 * testing. This function executes only during manufacturing configuration
 * mode and returns an error otherwise. If successful, the DUT
 * acknowledges the reboot command within 20 milliseconds and then 
 * reboots.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
EmberStatus emberMfgTestSendRebootCommand(void);

/** @description DEPRECATED: use emberMfgTestSendEui64() instead.
 *
 * @param newId: The 8-byte EUID for the DUT.
 *
 * @param newHashCode: The 8-byte validation code for the DUT, which
 *  is no longer used.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define emberMfgTestSendEui64AndHash(newId, newHashCode) \
  emberMfgTestSendEui64((newId))
#else
EmberStatus emberMfgTestSendEui64AndHash(EmberEUI64 newId,
                                        EmberHashCode newHashCode);
#endif

/** @description A function used during manufacturing configuration
 * on the Golden Node to set the DUT's 8-byte EUI ID.
 * This function executes only during manufacturing configuration
 * mode and returns an error otherwise. If successful,
 * the DUT acknowledges the new EUI ID within 150 milliseconds.
 *
 * @param newId: The 8-byte EUID for the DUT.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
EmberStatus emberMfgTestSendEui64(EmberEUI64 newId);



/** @description A function used during manufacturing configuration
 * on the Golden Node to set the DUT's 16-byte configuration string.
 * This function executes only during manufacturing configuration
 * mode and will return an error otherwise. If successful,
 * the DUT will acknowledge the new string within 150 milliseconds.
 *
 * @param newString: The 16-byte manufacturing string.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
EmberStatus emberMfgTestSendManufacturingString(EmberManufacturingString
                                               newString);

/** @description A function used during manufacturing configuration
 * on the Golden Node to set the DUT's radio parameters. This function 
 * executes only during manufacturing configuration
 * mode and returns an error otherwise. If successful,
 * the DUT acknowledges the new parameters within 25 milliseconds.
 *
 * @param supportedBands: Sets the radio band for the DUT. See 
 * ember-common.h for possible values.
 *
 * @param crystalOffset: Sets the CC1020 crystal offset. This
 * parameter has no effect on the EM2420, and it may safely be
 * set to 0 for this RFIC.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
EmberStatus emberMfgTestSendRadioParameters(int8u supportedBands,
                                           int8s crystalOffset);

/** @description A function used in each of the manufacturing
 * configuration API calls. Most implementations will not need
 * to call this function directly. See mfg-test.c for more detail. 
 * This function executes only during manufacturing configuration
 * mode and returns an error otherwise.
 *
 * @param command: A pointer to the outgoing command string.
 *
 * @return An EmberStatus value indicating success or failure of the
 * command.
 */
EmberStatus emberMfgTestSendCommand(int8u *command);


#endif
