/**
 * @file packet-buffer.h
 * @brief Packet buffer allocation and management routines.
 * <!-- Author(s): Richard Kelsey, kelsey@ember.com -->
 *
 * These functions implement a fixed-block-size memory management scheme 
 * to store and manipulate EmberNet packets. Buffers are identified to 
 * clients with a 1-byte ID.
 *
 * Buffer information is distributed in three arrays indexed by buffer IDs:
 * - <code>int8u packetBufferLengths[]</code> (amount of data in the buffer)
 * - <code>EmberMessageBuffer packetBufferLinks[]</code> (following buffer)
 * - <code>int8u packetBufferReferenceCounts[]</code> (reference counts)
 *
 * Separate arrays are used for the various parts of a packet (instead of an
 * array of structs) because the AVR compiler is not good at indexing
 * into arrays whose elements are not a power of 2 in size.
 * 
 * Buffers can be linked together to create longer packets.
 * \c packetBufferLinks[someBuffer] contains the buffer that follows
 * \c someBuffer, or EMBER_NULL_MESSAGE_BUFFER is \c someBuffer that has no 
 * follower. The utility procedures allow you to treat a linked chain
 * of buffers as a single buffer.
 *
 * If \c someBuffer is the head of a chain of buffers,
 * packetBufferLengths[someBuffer] has the total length of all the
 * data, not just those bytes in \c someBuffer.  The trailing buffers
 * in the chain do not have meaningful lengths<!-- (this may change in
 * the future if we find it is necessary to avoid fragmenting
 * header segments)-->.
 *
 * Freeing a buffer automatically decrements the reference count of
 * any following buffer, possibly freeing the following buffer as well.
 *
 * Packet buffers may be allocated in ISRs. Buffers that are known not
 * to be shared may be held or released in ISRs. It is not safe to hold
 * or release shared buffers in an ISR.
 *
 * Some functions in this file return an EmberStatus value. See 
 * error-defs.h for definitions of all EmberStatus return values.
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.          *80*-->
 */

#ifndef __PACKET_BUFFER_H__
#define __PACKET_BUFFER_H__

// Buffers hold 32 bytes.  Defined as the log to ensure it is a power of 2.
// If this is changed it needs to also change in ember-static-memory.h.
#define LOG_PACKET_BUFFER_SIZE 5
#define PACKET_BUFFER_SIZE (1 << LOG_PACKET_BUFFER_SIZE)

#ifndef DOXYGEN_SHOULD_SKIP_THIS
//----------------
// These are defined in /include/ember-static-memory.h
extern const int8u emPacketBufferCount;
extern int8u emPacketBufferData[];
extern EmberMessageBuffer emPacketBufferLinks[];
#endif //DOXYGEN_SHOULD_SKIP_THIS


/**
 * @description Gets a pointer to a buffer's contents.  This pointer can be
 * used to access only the first PACKET_BUFFER_SIZE bytes in the buffer.
 *
 * @param buffer: The buffer whose contents are desired.
 *
 * @return Returns a pointer to the contents of \c buffer.
 */
#define emberLinkedBufferContents(buffer) \
(&(emPacketBufferData[(buffer) << LOG_PACKET_BUFFER_SIZE]))

/**
 * @description Returns the buffer that follows this one in the message.
 *  EMBER_NULL_MESSAGE_BUFFER is returned if there is no following buffer.
 *
 * @param buffer: The buffer whose following buffer is desired.
 *
 * @return Returns the buffer that follows \c buffer in the message.
 *  EMBER_NULL_MESSAGE_BUFFER is returned if there is no following buffer.
 */
#define emberStackBufferLink(buffer) \
(emPacketBufferLinks[(buffer)])

/**
 * @description Sets the buffer following this one in the message. The
 *  final buffer in the message has EMBER_NULL_MESSAGE_BUFFER as its link.
 *
 * @param buffer: The buffer whose link is to be set.
 *
 * @param newLink: The buffer that is to follow \c buffer.
 */
#define emberSetStackBufferLink(buffer, newLink) \
(emPacketBufferLinks[(buffer)] = (newLink))

/**
 * @description Allocates a stack buffer.
 *
 * @return A newly allocated buffer, or EMBER_NULL_MESSAGE_BUFFER if no buffer
 * is available.
 */
#define emberAllocateStackBuffer() (emberAllocateLinkedBuffers(1))

/**
 * @description Allocates one or more linked buffers.
 *
 * @param count: The number of buffers to allocate.
 *
 * @return The first buffer in the newly allocated chain of buffers, or 
 * EMBER_NULL_MESSAGE_BUFFER if insufficient buffers are available.
 */
EmberMessageBuffer emberAllocateLinkedBuffers(int8u count);

/**
 * @description Allocates a stack buffer and fills in its contents.
 *
 * @param count: Buffer length.
 *
 * @param ...: \c count bytes, which will be placed in the buffer.
 *
 * @return A newly allocated buffer, or EMBER_NULL_MESSAGE_BUFFER if no buffer
 * is available.
 */
EmberMessageBuffer emberFillStackBuffer(int8u count, ...);

/**
 * @description Allocates a chain of stack buffers sufficient to hold 
 * \c length bytes of data and fills the buffers with the data in contents. 
 * If the value of contents is NULL, the buffers are allocated but 
 * not filled.
 *
 * @param contents: A pointer to data to place in the allocated buffers.
 *
 * @param length: The buffer length.
 *
 * @return The first buffer in a series of linked stack buffers, or
 * EMBER_NULL_MESSAGE_BUFFER if insufficient buffers are available.
 */
EmberMessageBuffer emberFillLinkedBuffers(int8u *contents, int8u length);

#ifndef DOXYGEN_SHOULD_SKIP_THIS
// Copy 'length' bytes between 'buffer' and 'contents', starting from
// 'startIndex' in the buffer.  Buffer links are followed as required
// to get enough bytes.  No buffers are allocated or released.
//
// Macroized to save code space.
// void
// emberCopyToLinkedBuffers(int8u *contents,
//                             EmberMessageBuffer buffer,
//                             int8u startIndex,
//                             int8u length);
#endif //DOXYGEN_SHOULD_SKIP_THIS

/**
 * @description Copies a specified number of bytes of data into a buffer, 
 * starting at a specified index in the buffer array. Buffer links are 
 * followed as required. No buffers are allocated or released.
 *
 * @param contents: A pointer to data to copy into the buffer.
 *
 * @param buffer: The buffer to copy data into.
 *
 * @param startIndex: The buffer index at which copying should start.
 *
 * @param length: The number of bytes of data to copy.
 */
#define emberCopyToLinkedBuffers(contents, buffer, startIndex, length) \
emReallyCopyToLinkedBuffers(contents, buffer, startIndex, length, FALSE)

#ifndef DOXYGEN_SHOULD_SKIP_THIS
void
emReallyCopyToLinkedBuffers(int8u *contents,
                            EmberMessageBuffer buffer,
                            int8u startIndex,
                            int8u length,
                            boolean isPgm);
#endif //DOXYGEN_SHOULD_SKIP_THIS                         
     
/**
 * @description Copies \c length bytes of data from a buffer
 * to \c contents, starting at a specified index in the buffer array.
 * Buffer links are followed as required.
 *
 * @param buffer: The buffer to copy data from.
 *
 * @param startIndex: The buffer index at which copying should start.
 *
 * @param contents: A pointer to data to copy from the buffer.
 *
 * @param length: The number of bytes of data to copy.
 */
void
emberCopyFromLinkedBuffers(EmberMessageBuffer buffer,
                           int8u startIndex,
                           int8u *contents,
                           int8u length);

/** 
 * @description Appends \c length bytes from \c contents onto a buffer.
 * Combines the functionality of setPacketBuffersLength() and 
 * copyToPacketBuffers().
 *
 * @param buffer: The buffer to append data to.
 *
 * @param contents: A pointer to data to append.
 *
 * @param length: The number of bytes of data to append.
 *
 * @return EMBER_SUCCESS if sufficient buffers are available, and
 *  EMBER_NO_BUFFERS if not.
 */
EmberStatus emberAppendToLinkedBuffers(EmberMessageBuffer buffer,
                                       int8u *contents,
                                       int8u length);

/**
 * @description Appends \c length bytes from \c contents, 
 * a pointer into program space (flash) to buffer. 
 *
 * @param buffer: The buffer to append data to.
 *
 * @param contents: The data to append.
 *
 * @param length: The number of bytes of data to append.
 *
 * @return EMBER_SUCCESS if sufficient buffers are available, and
 *  EMBER_NO_BUFFERS if not.
 */
EmberStatus emberAppendPgmToLinkedBuffers(EmberMessageBuffer buffer,
                                       PGM_P contents,
                                       int8u length);

/**
 * @description Appends a string from program space (flash) to a buffer. 
 *
 * @param buffer: The buffer to append data to.
 *
 * @param suffix: The string in program space to append. 
 *
 * @return EMBER_SUCCESS if sufficient buffers are available, and
 *  EMBER_NO_BUFFERS if not.
 */
EmberStatus emberAppendPgmStringToLinkedBuffers(EmberMessageBuffer buffer, PGM_P suffix);

/**
 * @description Sets the length of a chain of buffers, adding or removing 
 * trailing buffers as needed.
 *
 * @param buffer: The buffer whose length is to be set.
 *
 * @param length: The length to set.
 *
 * @return EMBER_SUCCESS if changing \c buffer's length by \c length bytes
 *  does not require additional buffers or if sufficient buffers are available,
 *  and EMBER_NO_BUFFERS if not.
 */
EmberStatus emberSetLinkedBuffersLength(EmberMessageBuffer buffer, int8u length);

/**
 * @description Gets a pointer to a specified byte in a linked list of buffers.
 *
 * @param buffer: The buffer that the requested byte must come from.
 *
 * @param index: The index of the requested byte.
 *
 * @return A pointer to the requested byte.
 */
int8u *emberGetLinkedBuffersPointer(EmberMessageBuffer buffer, int8u index);

/**
 * @description Gets a specified byte in a linked list of buffers.
 *
 * @param buffer: The buffer that the requested byte must come from.
 *
 * @param index: The index of the requested byte.
 *
 * @return A byte.
 */
int8u emberGetLinkedBuffersByte(EmberMessageBuffer buffer, int8u index);

/**
 * @description Sets the indexed byte in a linked list of buffers.
 *
 * @param buffer: The buffer holding the byte to be set.
 *
 * @param index: The index of the byte to set.
 
 * @param byte: The value to set the byte to.
 */
void emberSetLinkedBuffersByte(EmberMessageBuffer buffer, int8u index, int8u byte);

/**
 * @description Copies a chain of linked buffers.
 *
 * @param buffer: The first buffer in the chain to copy. 
 *
 * @return A newly created copy of the \c buffer chain.
 */
EmberMessageBuffer emberCopyLinkedBuffers(EmberMessageBuffer buffer);

/**
 * @description Creates a new, unshared copy of a specified buffer, if that
 * buffer is shared.  If it isn't shared, increments the reference count by
 * 1 so that the user of the returned buffer can release it in either case.
 *
 * @param buffer: The buffer to copy.
 *
 * @param isShared: A flag indicating whether the buffer is shared.
 *
 * @return A fresh copy of \c buffer if \c isShared is true, and \c buffer
 *  if \c isShared is not true.
 */
EmberMessageBuffer emberMakeUnsharedLinkedBuffer(EmberMessageBuffer buffer, boolean isShared);

//----------------------------------------------------------------
#ifndef DOXYGEN_SHOULD_SKIP_THIS
// This is obsolete.  Calls to it should be removed when convenient.
#define emInitializePacketBuffers()

// Function to return true if we have at least buffersNeeded free buffers
// available.
boolean emCheckAvailableBuffers( int8u buffersNeeded );

// For debugging.  This returns the number of free buffers.
int8u emFreePacketBufferCount(void);
#endif

#ifdef SIMULATOR
void printPacketBuffers(EmberMessageBuffer buffer);
#endif 

#endif  // __PACKET_BUFFER_H__

