/**
 * @file provisioning.h
 * @brief Definitions for the remote provisioning service on endpoint0
 *
 * A node can be interrogated and provisioned by sending datagrams to 
 * interface 0 of endpoint0. Each such datagram consists of a 1-byte command
 * optionally followed by 1 or more arguments. The possible requests and
 * their arguments are listed below. For each request received the node
 * responds with a datagram ACK. The general format of the responses is
 * the same as that of the requests.
 *
 * In these messages EmberInterfaceIds are encoded as follows:
 * @packetFormat
 * [ domain (1) ]
 * [ type (1) ]
 * [ specification (n) ]
 *
 * @param domain: The domain of the interface ID.
 * @param type: The type of the interface ID.
 * @param specification: 1 to 9 bytes of specification. The length of
 *  the specification depends on the domain, as described in the definition
 *  of EmberInterfaceId.
 *
 * In these messages an EmberBindingTableEntry is encoded as follows:
 * @packetFormat
 * [ type (1) ]
 * [ identifier (8) ]
 * [ local terminal (1) ]
 * [ remote terminal (1) ]
 *
 * The values of these packet fields are identical to the fields in an 
 * EmberBindingTableEntry.
 *
 * See the <i>Ember Application Development Guide</i> for further information 
 * about endpoint0 requests.
 *
 * <!--Author(s): James P. Parziale, jim@ember.com-->
 *
 * <!--Copyright 2004 by Ember Corporation. All rights reserved.         *80*-->
 */

 /** @description Defines requests that may be sent to endpoint0.
  */
enum {              
  /** @description Requests the number of interfaces implemented by each
   * endpoint.
   *       
   * @packetFormat
   * [ command (1) ]
   *
   * @param command: EMBER_ENDPOINT_REQUEST
   */
  EMBER_ENDPOINT_REQUEST            = 0,  
  /** @description Requests the endpoints that implement the given interfaces.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ interfaceId (n) ]
   * [...]
   *
   * @param command: EMBER_INTERFACE_REQUEST.
   * @param interfaceId: One or more EmberInterfaceIds, encoded as described
   * above.
   */
  EMBER_INTERFACE_REQUEST           = 1,  
  /** @description Requests the interfaces implemented by the given terminals.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ terminal (1) ]
   * [...]
   *
   * @param command: EMBER_ENDPOINT_INTERFACE_REQUEST
   * @param endpoint: One or more EmberTerminals, encoded as described above.
   */
  EMBER_ENDPOINT_INTERFACE_REQUEST  = 2,  
#ifndef DOXYGEN_SHOULD_SKIP_THIS
// These are not yet implemented.
  /** <interfaceID>. */
  EMBER_START_SEARCH_REQUEST        = 3,  
  /** <interfaceID>. */
  EMBER_STOP_SEARCH_REQUEST         = 4,  
  /** <no data besides type>. */
#endif //DOXYGEN_SHOULD_SKIP_THIS
  /** @description Requests the sizes of the nonvolatile and temporary portions
   * of the binding table.
   *       
   * @packetFormat
   * [ command (1) ]
   *
   * @param command: EMBER_BINDING_TABLE_SIZE_REQUEST.
   */
  EMBER_BINDING_TABLE_SIZE_REQUEST  = 5,  
  /** @description Requests entries in the binding table.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ index (1) ]
   * [ ... ]
   *
   * @param command: EMBER_BINDING_REQUEST.
   * @param index: One or more binding table indexes.
   */
  EMBER_BINDING_REQUEST             = 6,  
  /** @description Requests entries in the binding table.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ index (1) ]
   * [ entry (11)]
   * [ ... ]
   *
   * @param command: EMBER_SET_BINDING_REQUEST.
   * @param index: An index into the node's binding table.
   * @param entry: The EmberBindingTableEntry to be put at \c index.
   *  There may be one or more index + entry pairs in the request.
   */
  EMBER_SET_BINDING_REQUEST         = 7,  
  /** @description Requests the multicast IDs currently in the binding table.
   *       
   * @packetFormat
   * [ command (1) ]
   *
   * @param command: EMBER_MULTICAST_IDS_REQUEST.
   */
  EMBER_MULTICAST_IDS_REQUEST       = 8,  
};
typedef int8u EmberProvisioningRequestMessage;

/** @description Responses returned by endpoint0. */
enum {  
  /** @description Response for EMBER_ENDPOINT_REQUEST.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ counts (32) ]
   *
   * @param command: EMBER_ENDPOINT_RESPONSE
   * @param counts: The number of interfaces implemented by each of the
   * 32 endpoints. The first byte is the count for endpoint0, the
   * second is for endpoint 1, and so on.
   */
  EMBER_ENDPOINT_RESPONSE           =  0,
  /** @description Response for EMBER_INTERFACE_REQUEST.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ interfaceId (n) ]
   * [ count (1) ]
   * [ terminals (n) ]
   * [ ... ]
   *
   * @param command: EMBER_INTERFACE_RESPONSE.
   * @param interfaceId: The EmberInterfaceId implemented by the following
   *   terminals.
   * @param count: The number of terminals that implement the interface.
   * @param terminals: The terminals that implement the interface.
   *
   * There may be one or more sets of \c interfaceId, \c count, and \c terminals
   * in a response.  If a EMBER_INTERFACE_REQUEST includes more \c interfaceIds
   * than will fit in the response, the later \c interfaceIds will be ignored.
   */
  EMBER_INTERFACE_RESPONSE          =  1,  // (<interfaceID>
                                           //   <terminal count> <terminal>*)*
  /** @description Response for EMBER_ENDPOINT_INTERFACE_REQUEST.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ terminal (1) ]
   * [ interfaceId (n) ]
   * [...]
   *
   * @param command: EMBER_ENDPOINT_INTERFACE_RESPONSE.
   * @param terminal: The terminal whose interface was requested.
   * @param interfaceId: The EmberInterfaceId of the terminal.
   *
   * There may be one or more sets of \c terminal and \c interfaceId
   * in a response. If a EMBER_ENDPOINT_INTERFACE_REQUEST includes more
   * terminals than will fit in the response, the later terminals will be ignored.
   */
  EMBER_ENDPOINT_INTERFACE_RESPONSE =  2,  // (<terminal> <interfaceID>)* 

#ifndef DOXYGEN_SHOULD_SKIP_THIS
  EMBER_START_SEARCH_RESPONSE       =  3,  // <interfaceID>*
  EMBER_STOP_SEARCH_RESPONSE        =  4,  // <interfaceID>*
  EMBER_ENDPOINT_FOUND              =  5,  // <interfaceID>* <EUI64NodeId>
                                           //  <endpointMask> <distance>
  EMBER_SEARCH_COMPLETE             =  6,  // <interfaceID>*
#endif

  /** @description Response for EMBER_BINDING_TABLE_SIZE_REQUEST.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ nonvolatile (1) ]
   * [ temporary (1) ]
   *
   * @param command: EMBER_BINDING_TABLE_SIZE_RESPONSE.
   * @param nonvolatile: The size of the nonvolatile portion of the binding
   *  table.
   * @param temporary: The size of the temporary portion of the binding table.
   */
  EMBER_BINDING_TABLE_SIZE_RESPONSE =  7,  // <non-volatile count>
                                           //  <volatile count> <mask>*
  /** @description Response for EMBER_BINDING_REQUEST.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ index (1) ]
   * [ status (1) ]
   * [ entry (11)]
   * [ ... ]
   *
   * @param command: EMBER_SET_BINDING_RESPONSE.
   * @param index: The binding table index from the request.
   * @param status: The status returned by emberGetBinding(index, ...).
   * @param entry: The binding table entry itself.  This is present only
   *   if \c status is EMBER_SUCCESS.
   *
   * There may be one or more binding table entries in a response.
   * An EMBER_ENDPOINT_BINDING_REQUEST should not request more
   * binding table entries than will fit in a single response.
   */
  EMBER_BINDING_RESPONSE            =  8,  // (<index> <status>
                                           //  [<bindingEntry>])*
  /** @description Response for EMBER_SET_BINDING_REQUEST.
   * endpoint.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ index (1) ]
   * [ status (1) ]
   * [ entry (11)]
   * [ ... ]
   *
   * @param command: EMBER_SET_BINDING_RESPONSE.
   * @param index: The binding table index from the request.
   * @param status: The status returned by
   *   emberRemoteSetBindingHandler(index, ...)
   * @param entry: The binding table entry itself. This is present only
   *   if \c status is EMBER_SUCCESS.
   *
   * There may be one or more binding table entries in a response.
   * An EMBER_ENDPOINT_SET_BINDING_REQUEST should not request more
   * binding table entries than will fit in a single response.
   */
  EMBER_SET_BINDING_RESPONSE        =  9,  // (<index> <status> <bindingEntry>)*
  /** @description Response for EMBER_MULTICAST_IDS_REQUEST.
   *       
   * @packetFormat
   * [ command (1) ]
   * [ multicastId (2) ]
   * [ ... ]
   *
   * @param command: EMBER_MULTICAST_IDS_RESPONSE.
   * @param multicastId: A multicast group ID found in the node's binding table.
   *  The bytes are in the same order as in the identifiers in an 
   * EmberBindingTableEntry .
   */
  EMBER_MULTICAST_IDS_RESPONSE      = 10,  // <multicastID>
};
typedef int8u EmberProvisioningResponseMessage;

/**
 * @description Sets the length of EmberProvisioningResponseMessage
 * at a maximum of 64 bytes.
 */
#define EMBER_MAXIMUM_ENDPOINT_MESSAGE_LENGTH 64

