#!/usr/bin/python
# -*- coding: UTF-8 -*-

"""
A Marvin API to Read Each of the known Firmware Files;
determine their versions and checksums and return the results as JSON object

@author: "Nigel Winterbottom"
"""

import json

import api_marvin
from bottle import Bottle, response
from cipherdirectories import FIRMWARE_RF, FIRMWARE_PST, FIRMWARE_RH, FIRMWARE_ANA, FIRMWARE_PRT, FIRMWARE_NWPST
from cipherexception import MarvinApiError

app = Bottle()

#
# Map the FIRMWARE Binaries to the JSON keys
# This can be global I suppose, and it does remain constant
#
binaries = {
            "RFMOD" : {'filename':FIRMWARE_RF, 'mark':"rf    Issue"},
            "RF512" : {'filename':FIRMWARE_PST, 'mark':"RF512 Issue"},
            "RF513" : {'filename':FIRMWARE_RH, 'mark':"RF513 Issue"},
            "RF515" : {'filename':FIRMWARE_ANA, 'mark':"RF515 Issue"},
            "RF516" : {'filename':FIRMWARE_PRT, 'mark':"RF516 Issue"},
            "RF542" : {'filename':FIRMWARE_NWPST, 'mark':"RF542 Issue"},
            }


@app.route('/')
def firmwareinfo():
    """
    Serve a JSON response like this but as a object value of the 'Data' key in the Marvin API response:
    {
      "RF512" : {"Vn":"3.0.0", "ChkSum":42, "Size":23456}
      "RF513" : {"Vn":"3.0.0", "ChkSum":42, "Size":23456}
      "RFMOD" : {"Vn":"3.0.0", "ChkSum":42, "Size":23456}
    }
    """

    firmwareversions = {}
    returner = api_marvin.create_return_template()

    try:
        for (json_key, filedetail) in binaries.items():
            (version, filesize, checksum) = parse_firmware(filedetail)
            firmwareversions[json_key] = {"Vn":version, "ChkSum":checksum, "Size":filesize}
        # returner['Data'] = json.dumps(firmwareversions, skipkeys=False, sort_keys=True, ensure_ascii=True, check_circular=True, allow_nan=True, cls=None, indent=None, separators=None, encoding="utf-8", default=None)
        returner['Data'] = firmwareversions

    except Exception as exc:
        returner['SysError'] = True
        if isinstance(exc, MarvinApiError):
            returner['Warning'] = str(exc)
        else:
            returner['Warning'] = repr(exc)

    api_marvin.set_http_ok_response(response)
    return '(%s)' % json.dumps(returner, skipkeys=False, sort_keys=True, ensure_ascii=True, check_circular=True, allow_nan=True, cls=None, indent=None, separators=None, encoding="utf-8", default=None)


def scan_for_version(next_chunk, prev_chunk, mark):

    found_version = None

    # Join previous + next together in case mark spans across the join.
    cmp_str = prev_chunk + next_chunk
    mark_idx = cmp_str.find(mark)
    if mark_idx >= 0:
        end_idx = cmp_str[mark_idx:].find('\0')
        if end_idx >= 0:
            found_version = cmp_str[mark_idx + len(mark) + 1:mark_idx + end_idx]

    return found_version


def parse_firmware(filedetail):
    """ Return a tuple of (version, filesize, checksum)"""

    version = ""
    size = None
    checksum = None

    try:
        with open(filedetail['filename'], "rb") as f:
            checksum = 0;
            size = 0
            prev_chunk = ''
            while True:
                chunk = f.read(4096)
                if not chunk:
                    break
                size += len(chunk)
                checksum = sum(bytearray(chunk), checksum)
                potential_version = scan_for_version(chunk, prev_chunk, filedetail['mark'])
                if potential_version is not None:
                    version = potential_version
                prev_chunk = chunk

        if size < 0x7E00:
            checksum += (0x7E00 - size) * 0xFF;  # Pad with ff's till end
        checksum &= 0xFFFF

    except:
        pass



    return ((version, size, checksum))


