#!/usr/bin/python
# -*- coding: UTF-8 -*-

from bottle import Bottle, request, template, abort
import datetime

import gatewayId
import txsummary
import taskinfo
import txmodel
import alarmgrouplist
import periodsummary
from cipherconstants import TX_EPOCH
import txrdg
import ciphertypes
import math
from cipherexception import MarvinApiError
from report_helper import user_signed_in_check

app = Bottle()


# Expecting a query string like ?sno=CC10100060&PeriodNo=3&TaskNo=5&date=2013-11-28&u=4
@app.route('/')
def txtabular():

    # Read Request Variables
    date = None
    try:
        sno = request.query.get('sno')
        period_no = int(request.query.get('PeriodNo'))
        task_no = int(request.query.get('TaskNo'))
        date_str = request.query.get('date')
        if date_str:  # 2013-02-14 format expected
            date = datetime.date(int(date_str[0:4]), int(date_str[5:7]), int(date_str[8:10]))
    except ValueError as e:
        abort(400, e)
    param_u = request.query.get('u', '0')
    user_signed_in_check(param_u)

    # Read Stuff off disk
    try:
        gwid = gatewayId.rd_cu_id()
        gateway_vn = gatewayId.rd_versions()
        grouplist = alarmgrouplist.load_alarmgrouplist()
        summary = txsummary.load_summary(sno)
        task = taskinfo.load_task(sno, task_no)
        period_summary = periodsummary.load_period_summary(sno, task_no, period_no)
    except MarvinApiError as e:
        abort(404, e)
        # return template('fileerror', message=e)

    alarmgroup_number = txsummary.get_alarmgroup_no(summary)

    (ext1_scale, ext2_scale) = assign_scales(task)  # For RF515
    chname_intl = task.CuTask.chName[0].value
    chname_ext1 = task.CuTask.chName[1].value
    chname_ext2 = task.CuTask.chName[2].value
    chname_door = task.CuTask.chName[3].value
    latest_period = task.CuTask.periodNo
    print_lethality = task.CuTask.useLethCalc   #Distrust the lethality flag (and over-ride for the special cases)
    if task.CuTask.model == ciphertypes.TxModel.MDL_ANA or task.CuTask.model == ciphertypes.TxModel.MDL_RH :
		print_lethality = False

    # Is it old data
    historicData = (period_no != task.CuTask.periodNo or task_no != summary.taskNo)

    period_end_date = datetime.datetime.utcfromtimestamp(period_summary.stopTime + TX_EPOCH)
    if period_summary.startTime > period_summary.stopTime:
        period_end_date = None

    if date is None:
        if (historicData):
            date = period_end_date
        else:
            date = datetime.datetime.utcfromtimestamp(summary.last_rdg.TimeStamp + TX_EPOCH)

    # Get the actual readings
    txrdg_builder = txrdg.TxRdgBuilder()
    txrdgs = txrdg_builder.load_txrdg(sno, task_no, period_no, date)
    config = txrdg_builder.config

    return template('report_tabrdgs',
                    rqdate=date,
                    print_lethality=print_lethality,
                    historicData=historicData,
                    gateway_name=gwid.name,
                    gateway_sno=gwid.sno,
                    gateway_vn=gateway_vn,
                    start_group=task.CuTask.location,
                    current_group=grouplist.Name[alarmgroup_number].Name,
                    version=task.CuTask.issueNo,
                    task_descr=task.CuTask.description,
                    model=txmodel.TxModel.to_alt_str(task.CuTask.model),
                    latest_period=latest_period,
                    period_start_date=datetime.datetime.utcfromtimestamp(period_summary.startTime + TX_EPOCH),
                    period_end_date=period_end_date,
                    retrieved_task=task.CuTask.AutoTask,
                    latest_task=summary.taskNo,
                    task_create_date=datetime.datetime.utcfromtimestamp(task.CuTask.create_date),
                    datasigned=period_summary.signDone,
                    signer=period_summary.user,
                    signdate=datetime.datetime.utcfromtimestamp(period_summary.signDate),
                    signreason=period_summary.reason,
                    signauth=period_summary.authority,
                    config=config,
                    chname_intl=chname_intl,
                    chname_ext1=chname_ext1,
                    chname_ext2=chname_ext2,
                    chname_door=chname_door,
                    intl_scale=ciphertypes.ScaleT.to_alt_str(txrdg_builder.config.t_scale),
                    ext1_scale=ext1_scale,
                    ext2_scale=ext2_scale,
                    txrdgs=txrdgs,
                    rdg_formatters=assign_rdg_formatters(task),
                    is_rdgcode_graphable=is_rdgcode_graphable,
                    lethality_calculator=Accumulate_lethality(task),
                    )


class Accumulate_lethality(object):
    """Generator Object returning F-Values for EXT1 & EXT2 as a tuple
    Written using the standard Iterator pattern"""

    def __init__(self, task):
        self.lethT = task.CuTask.leth_T
        self.lethZ = task.CuTask.leth_z

        self.fval = [0.0, 0.0]
        self.prev_time = [None, None]
        self.prev_rate = [0.0, 0.0]

    def calc_next(self, txrdg, chan):
        curr_time = txrdg.rdg_time
        curr_temp = txrdg.rdgs[chan]
        if chan == 1 or chan == 2:

            idx = chan - 1

            if (self.prev_time[idx]) is None:
                time_step = 0
            else:
                time_step = (curr_time - self.prev_time[idx]).seconds / 60

            (curr_rate, curr_fval) = self._calc(self.prev_rate[idx], curr_temp, time_step)

            self.fval[idx] += curr_fval
            self.prev_rate[idx] = curr_rate
            self.prev_time[idx] = curr_time
            return (self.fval[idx])

        return 0.0

#         try:
#             ftd1 = datetime.timedelta(seconds=self.fval1)
#         except OverflowError:
#             ftd1 = datetime.timedelta.max
#
#         try:
#             ftd2 = datetime.timedelta(seconds=self.fval2)
#         except OverflowError:
#             ftd2 = datetime.timedelta.max
#        return ((ftd1, ftd2))

        # return ((datetime.timedelta(seconds=self.fval1), datetime.timedelta(seconds=self.fval2)))
        # return ((self.fval1, self.fval2))


    def _calc(self, prev_rate, curr_temp, time_step):
        """Calculate the current lethal rate given the current point's time and temperature,
        Calculate the F-Value using the trapezoid method"""
        curr_rate = math.pow(10, ((curr_temp - self.lethT) / self.lethZ))
        fval = time_step * (prev_rate + curr_rate) / 2
        return ((curr_rate, fval))



"""

    fval1 = 0
    fval2 = 0

    prev_time = None
    prev_temp1 = 0
    prev_temp2 = 0

    for txrdg in txrdgs:

        curr_time = txrdg.rdg_time
        curr_temp1 = txrdg.rdgs[1]
        curr_temp2 = txrdg.rdgs[2]

        if prev_time is None:
            fval1 = 0
            fval2 = 0
        else:
            x = math.pow(10, ((prev_temp1 - task.CuTask.leth_t) / task.CuTask.leth_z));
            y = math.pow(10, ((curr_temp1 - task.CuTask.leth_t) / task.CuTask.leth_z));
            fval1 += (x + y) / 2 / (curr_time - prev_time).seconds

            x = math.pow(10, ((prev_temp2 - task.CuTask.leth_t) / task.CuTask.leth_z));
            y = math.pow(10, ((curr_temp2 - task.CuTask.leth_t) / task.CuTask.leth_z));
            fval2 += (x + y) / 2 / (curr_time - prev_time).seconds

        prev_time = curr_time
        prev_temp1 = curr_temp1
        prev_temp2 = curr_temp2

        yield ((fval1, fval2))


    return"""


"""
    if (rdgIndex == 0) {
            lastLeathValue[channelIndex] = 0;
            lastTemp = 0;
            return NumberFormat.getFormat("0").format(lastLeathValue[channelIndex]);
        }

        lastTemp = dayRdgs.rdgs[channelIndex + 1][rdgIndex - 1];
        lastTime = new Date((long) (dayRdgs.startDate.getTime() + dayRdgs.minuteOffset[rdgIndex - 1])).getTime();

        currentTemp = dayRdgs.rdgs[channelIndex + 1][rdgIndex];
        currentTime = new Date((long) (dayRdgs.startDate.getTime() + dayRdgs.minuteOffset[rdgIndex])).getTime();

        x = Math.pow(10, ((lastTemp - task.cuTask.leth_T) / task.cuTask.leth_Z));
        y = Math.pow(10, ((currentTemp - task.cuTask.leth_T) / task.cuTask.leth_Z));

        lastLeathValue[channelIndex] += (x + y) / (2 * ((((currentTime - lastTime) / 1000) / 60)));

        return NumberFormat.getFormat("0").format(lastLeathValue[channelIndex]);
"""





def assign_rdg_formatters(task):
    """Assign 4 delegates to handle reading printout for each of the 4 channels.
    RF515 is a right bastard and requires a special case."""

    fmt_rdg_intl = _fmt_rdg_with_1_decimal  # This one's easy

    if task.CuTask.model == ciphertypes.TxModel.MDL_ANA:
        fmt_rdg_ext1 = _fmt_ana_rdg1
        fmt_rdg_ext2 = _fmt_ana_rdg2
    elif task.CuTask.model == ciphertypes.TxModel.MDL_PRT:
        fmt_rdg_ext1 = _fmt_rdg_with_2_decimal  # This one's easy
        fmt_rdg_ext2 = _fmt_rdg_with_2_decimal
    else:
        fmt_rdg_ext1 = _fmt_rdg_with_1_decimal
        fmt_rdg_ext2 = _fmt_rdg_with_1_decimal

    fmt_rdg_door = _fmt_rdg_with_2_decimal  # Door percentage
    fmt_rdg_door_mins = _fmt_rdg_with_no_decimal  # Door Mins

    return (fmt_rdg_intl, fmt_rdg_ext1, fmt_rdg_ext2, fmt_rdg_door, fmt_rdg_door_mins)


# NB: _fmt_rdg_with_1_decimal & _fmt_ana_rdg1 etc. operate as delegates
# and must share the same function signature.
def _fmt_rdg_with_no_decimal(rdg, _unused):
    return (_fmt_rdg_with_fmtstr(round(rdg, 0), "%0.0f"))

def _fmt_rdg_with_1_decimal(rdg, _unused):
    return (_fmt_rdg_with_fmtstr(rdg, "%0.1f"))

def _fmt_rdg_with_2_decimal(rdg, _unused):
    return (_fmt_rdg_with_fmtstr(rdg, "%0.2f"))

def _fmt_ana_rdg1(rdg, config):
    return _fmt_rdg_with_n_decimal(rdg, config.decimals[1])

def _fmt_ana_rdg2(rdg, config):
    return _fmt_rdg_with_n_decimal(rdg, config.decimals[2])

def _fmt_rdg_with_n_decimal(rdg, n):
    if not (0 <= n <= 4):
        n = 5
    fmtstr = "%%5.%if" % n
    return (_fmt_rdg_with_fmtstr(rdg, fmtstr))

# NB: This function will truncate the reading as opposed to rounding the value
#     Use the built in function round to mathamatically round anf format the number
def _fmt_rdg_with_fmtstr(rdg, formatstr):
    fmt_rdg = formatstr % rdg
    return fmt_rdg


def assign_scales(task):

    if task.CuTask.model == ciphertypes.TxModel.MDL_ANA:
        ext1_scale = task.CuTask.ana_task.EngScaling[1].MyScale
        ext2_scale = task.CuTask.ana_task.EngScaling[2].MyScale
    elif (task.CuTask.model == ciphertypes.TxModel.MDL_PRT or
          task.CuTask.model == ciphertypes.TxModel.MDL_PST or
          task.CuTask.model == ciphertypes.TxModel.MDL_NWPST):
        ext1_scale = ciphertypes.ScaleT.to_alt_str(task.CuTask.T_scale)
        ext2_scale = ext1_scale
    elif task.CuTask.model == ciphertypes.TxModel.MDL_RH:
        ext1_scale = ciphertypes.ScaleRH.to_alt_str(task.CuTask.RH_scale)
        ext2_scale = ciphertypes.ScaleT.to_alt_str(task.CuTask.T_scale)

    return (ext1_scale, ext2_scale)


#     if is_rdgcode_graphable(rdgcode):
#         fmt_rdg = formatstr % rdg
#     else:
#         fmt_rdg = ciphertypes.RdgCode.to_alt_str(rdgcode)
#     return fmt_rdg


def is_rdgcode_graphable(rdgcode):
    return (
            rdgcode != ciphertypes.RdgCode.RDG_SKIP and
            rdgcode != ciphertypes.RdgCode.RDG_FAULT and
            rdgcode != ciphertypes.RdgCode.RDG_OVR and
            rdgcode != ciphertypes.RdgCode.RDG_UNR
            )
