#! /usr/bin/python
# -*- coding: UTF-8 -*-

"""This Program attempts to duplicate the functionality of the original
Kantaris c code, rather than a wholesale rewrite.
Hopefully this code will become an excellent base upon which functionality mods
would become trivial

@author: Nigel Winterbottom
"""
import datetime
import struct

import taskinfo
import txmodel
from ciphertypes import ChanType, RdgStatus
import cipherdirectories
from viewmodels import TransmitterConfig
import ciphertypes


class TxRdg:
    """Nicely Object Oriented version of the C type rdg_rec_t"""
    def __init__(self):
        self.rdg_time = None
        self.rdg_status = [None] * 4
        self.rdgs = [None] * 5


class TxRdgBuilder:

    def __init__(self):
        self.task = None
        self.date = None
        self.config = TransmitterConfig()


    def load_txrdg(self, sno, task_no, period_no, date):
        """@param day: A datetime.date object
        Load TxRdg for 1 day from Disk - Return an Instance of TxRdg
        for the caller to do with as he pleases."""

        self.config.period_no = period_no
        self.date = date

        # Fetch the Task & update self.config if necessary
        if (sno != self.config.sno) or (task_no != self.config.task_no):
            self.config.sno = sno
            self.config.task_no = task_no
            self._load_task()

        # Build TXDATA filename
        filepath = cipherdirectories.TX_DATA_FILE_V2_PATTERN % (sno, task_no, period_no, date.strftime("%d_%b_%Y"))

        # Open File
        try:
            f = open (filepath, 'rb')
            rdgbytes = f.read(32768)  # Play safe: I don't want to read 1GB here. A proper file will always be < 32768
        except (IOError):
            return []

        return self.parse_daily_txrdgs(rdgbytes)


    def _load_task(self):
        task = taskinfo.load_task(self.config.sno, self.config.task_no)
        self.task = task

        self.config.model = task.CuTask.model

        self.config.decimals[0] = 1  # Internal always is 1 decimal place
        if task.CuTask.model == ciphertypes.TxModel.MDL_ANA:
            self.config.decimals[1] = task.CuTask.ana_task.EngScaling[ChanType.EXT1].Decimals
            self.config.decimals[2] = task.CuTask.ana_task.EngScaling[ChanType.EXT2].Decimals
        elif task.CuTask.model == ciphertypes.TxModel.MDL_PRT:
            self.config.decimals[1] = 2  # PRT is 2 decimal place
        else:
            self.config.decimals[1] = 1
            self.config.decimals[2] = 1
        self.config.decimals[3] = 2  # Door Percentage always is 2 decimal place

        self.config.t_scale = task.CuTask.T_scale
        self.config.rh_scale = task.CuTask.RH_scale

        self.config.bIntlEn = task.TxTask.chan_set[0].ChannelF.bChan_En
        self.config.bExt1En = task.TxTask.chan_set[1].ChannelF.bChan_En
        self.config.bExt2En = task.TxTask.chan_set[2].ChannelF.bChan_En
        self.config.bDoorEn = task.TxTask.chan_set[3].ChannelF.bChan_En
        self.config.door_avg_interval = task.TxTask.rate_set.DoorAveInterval


    def parse_daily_txrdgs(self, rdgbytes):
        """Make some sort of order out of Nick's binary chaos"""

        # Build a fully slotted array of bare TxRdg instances (remove empty ones later)
        txrdg_list = [None] * 1440  # Pre-Allocate list size
        for i in xrange(1440):
            txrdg = TxRdg()
            if self.config.bIntlEn: txrdg.rdg_status[0] = RdgStatus()
            if self.config.bExt1En: txrdg.rdg_status[1] = RdgStatus()
            if self.config.bExt2En: txrdg.rdg_status[2] = RdgStatus()
            if self.config.bDoorEn: txrdg.rdg_status[3] = RdgStatus()
            txrdg_list[i] = txrdg

        # Build a Binary Parser using the struct module (dependent on model, and number of enabled channels)
        txrdg_stuct = struct.Struct(self.calc_struct_format_str())

        offset = 0
        while (offset < len(rdgbytes)):
            elemental_txrdg = txrdg_stuct.unpack_from(rdgbytes, offset)
            self.parse_txrdg_from_elemental(txrdg_list, elemental_txrdg)
            offset += txrdg_stuct.size

        # Reduce the list to populated entries only
        return [txrdg for txrdg in txrdg_list if txrdg.rdg_time is not None]


    def calc_struct_format_str(self):
        """Build a binary format string to suit the struct module"""
        intlfmt = 'Bh' if self.config.bIntlEn else ''
#         if self.config.bExt1En:
#             if self.config.model == txmodel.TxModel.MDL_PRT:
#                 ext1fmt = 'Bi'
#             else:
#                 ext1fmt = 'Bh'
#         else:
#             ext1fmt = ''
        # This one liner is the ternary equivalent of the above if then else stuff
        ext1fmt = '' if not self.config.bExt1En else 'Bi' if self.config.model == txmodel.TxModel.MDL_PRT else 'Bh'
        ext2fmt = 'Bh' if self.config.bExt2En else ''
        doorfmt = 'Bh' if self.config.bDoorEn else ''

        return ''.join(('>h', intlfmt, ext1fmt, ext2fmt, doorfmt))


    def parse_txrdg_from_elemental(self, txrdg_list, raw_txrdg):

        minute_offset = raw_txrdg[0]
        if (0 <= minute_offset < 1440):

            txrdg = txrdg_list[minute_offset]
            txrdg.rdg_time = datetime.datetime.combine(self.date, datetime.time()) + datetime.timedelta(minutes=minute_offset)

            idx = 1
            if self.config.bIntlEn:
                self.update_rdg_status(RdgStatus(raw_txrdg[idx]), txrdg.rdg_status[0])
                idx += 1
                txrdg.rdgs[0] = raw_txrdg[idx] / 10.0
                idx += 1

            if self.config.bExt1En:
                self.update_rdg_status(RdgStatus(raw_txrdg[idx]), txrdg.rdg_status[1])
                idx += 1
                if self.config.model == txmodel.TxModel.MDL_ANA:
                    txrdg.rdgs[1] = raw_txrdg[idx] / pow(10, self.config.decimals[1])
                elif self.config.model == txmodel.TxModel.MDL_PRT:
                    txrdg.rdgs[1] = raw_txrdg[idx] / 100.0
                else:
                    txrdg.rdgs[1] = raw_txrdg[idx] / 10.0
                idx += 1

            if self.config.bExt2En:
                self.update_rdg_status(RdgStatus(raw_txrdg[idx]), txrdg.rdg_status[2])
                idx += 1
                if self.config.model == txmodel.TxModel.MDL_ANA:
                    txrdg.rdgs[2] = raw_txrdg[idx] / pow(10, self.config.decimals[2])
                else:
                    txrdg.rdgs[2] = raw_txrdg[idx] / 10.0
                idx += 1

            if self.config.bDoorEn:
                self.update_rdg_status(RdgStatus(raw_txrdg[idx]), txrdg.rdg_status[3])
                idx += 1
                mins = raw_txrdg[idx] / 120.0
                txrdg.rdgs[3] = self.calc_door_percentage(mins)
                txrdg.rdgs[4] = mins
                # idx += 1

        return txrdg


    def update_rdg_status(self, src_status, dest_status):
        dest_status.rdgcode = src_status.rdgcode
        if src_status.bPeakRdg:
            dest_status.bPeakRdg = True


    def calc_door_percentage(self, mins):
        pcnt = 100.0 * mins / self.config.door_avg_interval
        if (pcnt < 0.0):   pcnt = 0.0;
        if (pcnt > 100.0): pcnt = 100.0;
        return pcnt;


if __name__ == '__main__':
    builder = TxRdgBuilder()
    # txrdg_list = builder.load_txrdg("CA03110485", 1, 1, datetime.date(2012, 01, 11))
    txrdg_list = builder.load_txrdg("CE01130106", 4, 1, datetime.date(2013, 05, 02))  # From ALS-Coventry

    for txrdg in txrdg_list:
        print txrdg.rdg_time, txrdg.rdg_status[1].rdgcode, txrdg.rdg_status[1].bPeakRdg, txrdg.rdgs[1]


