#! /usr/bin/python
# -*- coding: UTF-8 -*-

"""This Program attempts to duplicate the functionality of the original
Kantaris c code, rather than a wholesale rewrite.
Hopefully this code will become an excellent base upon which functionality mods
would become trivial"""

# from __future__ import print_function
from ctypes import Structure, Union, sizeof, c_char, c_float, \
                   c_uint8, c_int16, c_uint16, c_int32, c_uint32
from cipherconstants import SNO_SIZE, STR_FILE_CORRUPT
from cipherdirectories import SUMMARY_FILE_IN_RAM
from cipherexception import MarvinApiError


class RdgStatusType (Structure):
    """Nigel's Reading Status bitfield"""
    _pack_ = 1
    _fields_ = [("StatusCode", c_uint8, 7), ("bPeakRdg", c_uint8, 1)]

class _ExtRdgType (Structure):
    """Internal Structure for Nigel's Merged Cipher Reading Record"""
    _pack_ = 1
    _fields_ = [("Ext1", c_int16), ("Ext2", c_int16)]

class MrgRdgType (Union):
    """Nigel's Merged Cipher Reading Record"""
    _pack_ = 1
    _fields_ = [("Ex", _ExtRdgType), ("Pt100_Rdg", c_int32)]

class RdgRecType (Structure):
    """Nigel's Cipher Reading Record"""
    _pack_ = 1
    _fields_ = [("RecordType", c_uint8),
                ("RecordF", c_uint8),
                ("TimeStamp", c_uint32),
                ("periodNo", c_uint16),
                ("RdgStatus", RdgStatusType * 4),
                ("MrgRdg", MrgRdgType),
                ("Door", c_uint16),
                ("Int", c_int16)]

class PrtCalType (Structure):
    """Nick's Cipher Prt Calibration Record"""
    _pack_ = 1
    _fields_ = [("T_Lo", c_float),
                ("T_Hi", c_float),
                ("Ohms_Lo", c_float),
                ("Ohms_Hi", c_float),
                ("offset", c_int16),
                ("scale", c_uint16)]

class AlarmStatType (Structure):
    """Nick's Cipher Prt Calibration Record"""
    _pack_ = 1
    _fields_ = [("LastAlarmStatus_int", c_char),
                ("LastAlarmStatus_ext1", c_char),
                ("LastAlarmStatus_ext2", c_char),
                ("LastAlarmStatus_door", c_char),
                ("LastDoorConOpenAlarmStatus", c_char)]

class TxSummaryStruct (Structure):
    """Equivalent of summaryFile_t"""
    _pack_ = 1
    _fields_ = [
            ("sno", c_char * SNO_SIZE),
            ("model", c_uint8),
            ("RfClass", c_uint8),
            ("version", c_char * 32),
            ("ModelStr", c_char * 12),
            ("description", c_char * 32),
            ("chName", (c_char * 24) * 4),
            ("location", c_uint8 * 64),
            ("LastGraphName", c_char * 32),
            ("Ymax", c_int32),
            ("Ymin", c_int32),
            ("taskNo", c_int32),
            ("periodNo", c_int32),
            ("last_rdg", RdgRecType),
            ("first_rdg", RdgRecType),
            ("lastRfTime", c_int32),
            ("PcTimeStamp", c_int32),
            ("ChMin", c_int32 * 6),
            ("ChMax", c_int32 * 6),
            ("ChSum", c_int32 * 6),
            ("NoOfRdgs", c_int32),
            ("RadioRate", c_int32),
            ("LastEventTime", c_int32),
            ("prt_cal", PrtCalType),
            ("alarm_stat", AlarmStatType),
            ("EmberEieioNo", c_char * 32),
            ("ChFaultFlag", c_char),
            ("prt_cal_date", c_int32),
            ("TaskReteiveMode", c_char),
            ("TaskNoInTx", c_int32),
            ("NotCompatible", c_uint8, 1),
            ("request_mesher", c_uint8, 1),
            ("unknownPeriod", c_uint8, 1),
            ("PreSetScale", c_uint8, 1),
            ("AuditOldDataFlag", c_uint8, 1),
            ("resend", c_uint8, 1),
            ("preScanFlag", c_uint8, 1),
            ("DataType2", c_uint8, 1),
            ("RF_firmware_blockNo", c_uint16),
            ("RF_firmware_blockSize", c_uint16),
            ("map_row", c_uint8),
            ("map_col", c_uint8),
            ("GwCmdVn", c_uint8),
            ("GwDataVn", c_uint8),
            ("nextXmlReading", c_uint32),
            ("enterDate", c_uint32),
            ("tempScale_Hi", c_int16),
            ("tempScale_Lo", c_int16),
            ("percentScale_Hi", c_int16),
            ("percentScale_Lo", c_int16),
            ("lastTxReset", c_uint32),
            ("Spare", c_char * (100 - 20 - 5 - 32 - 1 - 4 - 1 - 4 - 1 - 2 - 2 - 1 - 1 - 1 - 1 - 4 - 4 - 8 - 4)),
            ("DoorAveInterval", c_char),
            ("T_scale", c_char),
            ("Rh_scale", c_char),
            ("CnScale", (c_char * 9) * 6),
            ("ChDP", c_char * 6),
            ("batteryMv", c_int32),
            ("batteryStatus", c_char),
            ("SnoPos", c_char),
            ("noSignalStat", c_char),
            ("sw_1", c_char * 6),
            ("sw_2", c_char * 6),
            ("ANA_Ymax", c_int32),
            ("ANA_Ymin", c_int32),
            ("Spare1", c_char * 4),

            ("mesher", c_uint8, 1),
            ("newSno", c_uint8, 1),
            ("found", c_uint8, 1),
            ("remove", c_uint8, 1),
            ("ForcedRemove", c_uint8, 1),
            ("removed", c_uint8, 1),
            ("anyData", c_uint8, 1),
            ("anyTask", c_uint8, 1),

            ("logging", c_uint8, 1),
            ("dataTrue", c_uint8, 1),
            ("lowBat", c_uint8, 1),
            ("deadBat", c_uint8, 1),
            ("noSignal", c_uint8, 1),
            ("alarm", c_uint8, 1),
            ("lcdOff", c_uint8, 1),
            ("mainsPower", c_uint8, 1),

            ("CmdInQ", c_uint8, 1),
            ("TaskInQ", c_uint8, 1),
            ("FirmWareInQ", c_uint8, 1),
            ("ch_int", c_uint8, 1),
            ("ch_ext1", c_uint8, 1),
            ("ch_ext2", c_uint8, 1),
            ("ch_door", c_uint8, 1),
            ("re_add", c_uint8, 1),

            ("NewGraph", c_uint8, 1),
            ("fault", c_uint8, 1),
            ("newTimeCmdQ", c_uint8, 1),
            ("unknownTask", c_uint8, 1),
            ("taskRetrieve", c_uint8, 1),
            ("TxLoggingStatus", c_uint8, 1),
            ("fault_int", c_uint8, 1),
            ("fault_ext1", c_uint8, 1),

            ("fault_ext2", c_uint8, 1),
            ("fault_door", c_uint8, 1),
            ("edgeTrigAlarm", c_uint8, 1),
            ("spare2", c_uint8, 1),
            ("spare3", c_uint8, 1),
            ("spare4", c_uint8, 1),
            ("spare5", c_uint8, 1),
            ("spare6", c_uint8, 1),
           ]


def _rd_summary_offset(offset, summary):
    """Read from Summary File into structure passed"""
    try:
        with open(SUMMARY_FILE_IN_RAM, 'rb') as sumf:
            pos = sizeof(TxSummaryStruct) * (offset + 1)
            sumf.seek(-pos, 2)
            number_bytes_read = sumf.readinto(summary)
            if number_bytes_read == 0 or\
               number_bytes_read != sizeof(TxSummaryStruct):
                raise MarvinApiError(STR_FILE_CORRUPT + SUMMARY_FILE_IN_RAM)
    except EnvironmentError:
        raise MarvinApiError("File Open Error: " + SUMMARY_FILE_IN_RAM)


def read_summary(sfile, summary=None):
    """Generator Function to Sequentially Read from Summary File into structure
    """

    if summary is None:
        summary = TxSummaryStruct()
    number_bytes_read = sfile.readinto(summary)  # The 'none" entry
    if number_bytes_read != sizeof(TxSummaryStruct):
        raise MarvinApiError(STR_FILE_CORRUPT + sfile.name)

    while True:
        number_bytes_read = sfile.readinto(summary)
        if number_bytes_read == sizeof(TxSummaryStruct):
            yield summary
        elif number_bytes_read != 0:
            raise MarvinApiError(STR_FILE_CORRUPT + sfile)
        else:
            return


def load_summary(sno, summary=None):
    """Load the Specific Transmitter Summary from ramdisk into parameter summary"""
    try:
        with open(SUMMARY_FILE_IN_RAM, 'rb') as sumf:
            for summary in read_summary(sumf, summary):
                if summary.sno == sno:
                    break
        if summary.sno != sno:
            raise MarvinApiError("Transmitter Not Registered to Gateway: " + sno)

    except EnvironmentError:
        raise MarvinApiError("File Open Error: " + SUMMARY_FILE_IN_RAM)

    return summary


def badbattery_count1():
    """Count Number of Tx with low battery"""
    summary = TxSummaryStruct()
    badbat = 0
    slot = 0
    while True:
        _rd_summary_offset(slot, summary)
        if summary.sno == "none":
            break
        if ((summary.lowBat or summary.deadBat)
            and not (summary.remove or summary.removed)
            and not (summary.newSno or summary.re_add)
            and summary.found
        ):
            badbat += 1
        slot += 1
    return badbat


def badbattery_count():
    """Count Number of Tx with low battery"""
    badbat = 0
    try:
        with open(SUMMARY_FILE_IN_RAM, 'rb') as sumf:
            for summary in read_summary(sumf):
                if ((summary.lowBat or summary.deadBat) and
                    not (summary.remove or summary.removed) and
                    not (summary.newSno or summary.re_add) and
                    summary.found):
                    badbat += 1
    except EnvironmentError:
        raise MarvinApiError("File Open Error: " + SUMMARY_FILE_IN_RAM)

    return badbat


def get_alarmgroup_no(summary):
    for i in range(32):
        if summary.location[i]:
            return i



def test():
    """Test Structure Reading from file"""
    try:
        with open(SUMMARY_FILE_IN_RAM, 'rb') as sumf:
            for summary in read_summary(sumf):
                print ("Sno:{0}  lowBat:{1}  deadBat:{2}  Reject:{3}  ChNames: [{4}, {5}, {6}, {7}]".
                       format (summary.sno, summary.lowBat, summary.deadBat,
                               (summary.remove or summary.removed or
                                summary.newSno or summary.re_add or
                                not summary.found),
                               summary.chName[0].value, summary.chName[1].value, summary.chName[2].value, summary.chName[3].value,

                               )
                       )
    except EnvironmentError:
        raise MarvinApiError("File Open Error: " + SUMMARY_FILE_IN_RAM)


    #
    # summary = CipherSummaryStruct()
    #
    # slot = 0
    # while True:
    #    rd_summary_offset(slot, summary)
    #    if summary.sno == "none":
    #        break
    #    print ("Sno:{0}  lowBat:{1}  deadBat:{2}  Reject:{3}".
    #        format (summary.sno, summary.lowBat, summary.deadBat, (
    #            summary.remove
    #            or summary.removed
    #            or summary.newSno
    #            or summary.re_add
    #            or not summary.found)
    #               )
    #        )
    #    slot += 1


if __name__ == '__main__':
    # print (badbattery_count())
    test()
