#! /usr/bin/python
# -*- coding: UTF-8 -*-

"""This Program attempts to duplicate the functionality of the original
Kantaris c code, rather than a wholesale rewrite.
Hopefully this code will become an excellent base upon which functionality mods
would become trivial"""

from ctypes import Structure, c_char, c_uint8, c_int32, sizeof
from cipherdirectories import USERS_LIST_FILE
from cipherexception import MarvinFileError, MarvinApiError
from cipherconstants import STR_FILE_CORRUPT, STR_FILE_OPEN_ERROR
import time

SESSION_INACTIVITY_TIMEOUT = 60 * 60;  # 1 Hour inactivity

# class LocationStr (Structure):
#     """Nick's 16 char location name. NB Maxlen is 16 chars so no zero term"""
#     _fields_ = [("locationstr", c_char * 16)]


class UserAccountStruct (Structure):
    """Equivalent to Nick's userlist_t"""
    _pack_ = 1
    _fields_ = [
            ("userName", c_char * 33),
            ("userPassword", c_char * 33),
            ("expireDate", c_int32),
            ("userID", c_int32),
            ("location", (c_char * 16) * 32),
            ("restrictLocation", c_uint8 * 32),
            ("screenSize", c_char),
            ("auditNoAlarms", c_char),
            ("orderType", c_char),
            ("orderRevers", c_char),
            ("language", c_char),
            ("spare", c_char * (512 - 32 - 1 - 1 - 1 - 1 - 1)),
            ("lastIn", c_int32),

            ("removed", c_uint8, 1),
            ("renew", c_uint8, 1),
            ("admin", c_uint8, 1),
            ("allow_setTask", c_uint8, 1),
            ("allow_ackAlarms", c_uint8, 1),
            ("allow_addTx", c_uint8, 1),
            ("allow_startStop", c_uint8, 1),
            ("list_all_tx", c_uint8, 1),

            ("list_all_loc", c_uint8, 1),
            ("unack_alarms", c_uint8, 1),
            ("with_no_data", c_uint8, 1),
            ("with_data", c_uint8, 1),
            ("with_no_task", c_uint8, 1),
            ("with_task", c_uint8, 1),
            ("with_no_signal", c_uint8, 1),
            ("with_low_battery", c_uint8, 1),

            ("with_fault", c_uint8, 1),
            ("model_PST", c_uint8, 1),
            ("model_RH", c_uint8, 1),
            ("model_PRT", c_uint8, 1),
            ("units_removed", c_uint8, 1),
            ("previewOn", c_uint8, 1),
            ("autoRefreshOn", c_uint8, 1),
            ("at_days_d", c_uint8, 1),

            ("at_days_w", c_uint8, 1),
            ("at_days_m", c_uint8, 1),
            ("at_filter_t", c_uint8, 1),
            ("at_filter_a", c_uint8, 1),
            ("at_filter_u", c_uint8, 1),
            ("allow_sign", c_uint8, 1),
            ("backbone", c_uint8, 1),
            ("model_ANA", c_uint8, 1),

            ("model_NWPST", c_uint8, 1),
            ("spare_2", c_uint8, 1),
            ("spare_3", c_uint8, 1),
            ("spare_4", c_uint8, 1),
            ("spare_5", c_uint8, 1),
            ("spare_6", c_uint8, 1),
            ("spare_7", c_uint8, 1),
            ("spare_8", c_uint8, 1)]


class UserAccount (object):
    """Handle Cipher User Accounts"""
    def __init__(self):
        self.account = UserAccountStruct()

    def loaduseraccount(self, name=None, uid=None):
        """Load account info from persistent storage.
        @requires:At least one of name or uid must be given.
        @raise MarvinApiError: If either name or uid are invalid.
        @raise MarvinFileError: If UsersListFile cannot be read.
        """
        if (name is None and uid is None):
            raise MarvinApiError("At least one of name or uid expected")

        try:
            with open(USERS_LIST_FILE, 'rb') as ulf:
                while ((uid is not None and self.account.userID != uid)
                       or (name is not None and self.account.userName != name)):
                    number_bytes_read = ulf.readinto(self.account)
                    if number_bytes_read == 0:
                        raise MarvinApiError("User not found")
                    if number_bytes_read != sizeof(UserAccountStruct):
                        raise MarvinFileError(STR_FILE_CORRUPT + USERS_LIST_FILE)
        except EnvironmentError:
            raise MarvinFileError("File Open Error: " + USERS_LIST_FILE)

#     def loaduseraccount_id(self, userid):
#         """Load account info from persistent storage"""
#         if type(userid) is not int:
#             raise MarvinFileError("UserId must be +ve Integer")
#
#         try:
#             with open(USERS_LIST_FILE, 'rb') as ulf:
#                 ulf.seek(userid * sizeof(UserAccountStruct))
#                 number_bytes_read = ulf.readinto(self.account)
#                 if (self.account.userID != userid or
#                     number_bytes_read != sizeof(UserAccountStruct)):
#                     raise MarvinFileError(STR_FILE_CORRUPT + USERS_LIST_FILE)
#         except EnvironmentError:
#             raise MarvinFileError("File Open Error: " + USERS_LIST_FILE)

    def saveuseraccount(self):
        """Save to persistent storage"""
        try:
            with open(USERS_LIST_FILE, 'rb+') as ulf:
                account = UserAccountStruct()
                while account.userName != self.account.userName:
                    number_bytes_read = ulf.readinto(account)
                    if number_bytes_read == 0:
                        raise MarvinFileError("Username not found: " + self.account.username)
                    if number_bytes_read != sizeof(UserAccountStruct):
                        raise MarvinFileError(STR_FILE_CORRUPT + USERS_LIST_FILE)
                ulf.seek(-sizeof(UserAccountStruct), 1)
                ulf.write(self.account)
        except EnvironmentError:
            raise MarvinFileError("File Error: " + USERS_LIST_FILE)

    def signin_user(self, password):
        """Attempt to Sign In the User with the given password.
        @return: boolean Success or Fail
        """
        if password == self.account.userPassword:
            self.account.lastIn = int(time.time())
            self.saveuseraccount()
            return True
        return False


def iter_users():
    """Generate an Instance of UserAccount() for each user entry in USERS_LIST_FILE"""
    userid = 0
    while True:
        try:
            ulf = open(USERS_LIST_FILE, 'rb')
        except IOError:
            raise MarvinFileError(STR_FILE_OPEN_ERROR + USERS_LIST_FILE)

        account = UserAccount()
        ulf.seek(userid * sizeof(UserAccountStruct), 0)
        number_bytes_read = ulf.readinto(account.account)
        ulf.close()

        if number_bytes_read == sizeof(UserAccountStruct):
            yield account
            userid += 1
        elif number_bytes_read == 0:
            break
        else:
            raise MarvinFileError(STR_FILE_CORRUPT + USERS_LIST_FILE)


if __name__ == '__main__':
    user = UserAccount()
    user.loaduseraccount(uid=1)
    print user.account.userID
    print user.account.userName
